(** Bundled by bspack 08/25-14:49 *)
module 
Reason_config
= struct
#1 "reason_config.ml"
(**
 * Copyright (c) 2015-present, Facebook, Inc. All rights reserved.
 *)

let recoverable = ref false

let configure ~r = (
  recoverable := r
)

end
module 
Syntax_util
= struct
#1 "syntax_util.ml"
open Ast_mapper
open Asttypes
open Parsetree
open Longident

(** [is_prefixed prefix i str] checks if prefix is the prefix of str
  * starting from position i
  *)
let is_prefixed prefix str i =
  let len = String.length prefix in
  if i + len > String.length str then false else
  let rec loop j =
    if j >= len then true else
      if String.unsafe_get prefix j <> String.unsafe_get str (i + j) then false else loop (j + 1)
    in
  loop 0

(**
 * pick_while returns a tuple where first element is longest prefix (possibly empty) of the list of elements that satisfy p
 * and second element is the remainder of the list
 *)
let rec pick_while p = function
  | [] -> [], []
  | hd::tl when p hd ->
                  let (satisfied, not_satisfied) = pick_while p tl in
                  hd :: satisfied, not_satisfied
  | l -> ([], l)


let rec replace_string_ old_str new_str i str buffer =
  if i >= String.length str then
    ()
  else
    (* found match *)
    if is_prefixed old_str str i then
      (* split string *)
      let old_str_len = String.length old_str in
      Buffer.add_string buffer new_str;
      replace_string_ old_str new_str (i + old_str_len) str buffer
    else
      let start = String.sub str i 1 in
      Buffer.add_string buffer start;
      replace_string_ old_str new_str (i + 1) str buffer


(** [replace_string old_str new_str str] replaces old_str to new_str in str *)
let replace_string old_str new_str str =
  let buffer = Buffer.create (String.length old_str * 2) in
  replace_string_ old_str new_str 0 str buffer;
  Buffer.contents buffer


module StringMap = Map.Make (String)


(** Generate a suitable extension node for Merlin's consumption,
    for the purposes of reporting a syntax error - only used
    in recovery mode.
 *)
let syntax_error_extension_node loc message =
  let str = Location.mkloc "merlin.syntax-error" loc in
  let payload = PStr [{
    pstr_loc = Location.none;
    pstr_desc = Pstr_eval (
      {
        pexp_loc = Location.none;
        pexp_desc = Pexp_constant (Asttypes.Const_string (message, None));
        pexp_attributes = [];
      },
      []
    );
  }]
 in
 (str, payload)

let reason_to_ml_swapping_alist = [
  "===",  "==";
  "==",  "=";
  (* ===\/ and !==\/ are not representable in OCaml but
   * representable in Reason
   *)
  "\\!==", "!==";
  "\\===", "===";
  "!=", "<>";
  "!==", "!=";
]

let swap_txt map txt =
  if StringMap.mem txt map then
    StringMap.find txt map
  else
    txt

(** identifier_mapper maps all identifiers in an AST with a mapping function f
  *)
let identifier_mapper f =
{ default_mapper with
  expr = begin fun mapper expr ->
    let expr =
      match expr with
        | {pexp_desc=Pexp_ident ({txt} as id);
           pexp_loc;
           pexp_attributes} ->
             let swapped = match txt with
               | Lident s -> Lident (f s)
               | Ldot(longPrefix, s) -> Ldot(longPrefix, f s)
               | Lapply (y,s) -> Lapply (y, s)
             in
             {expr with pexp_desc=Pexp_ident ({id with txt=swapped})}
        | _ -> expr
    in
    default_mapper.expr mapper expr
  end;
  pat = begin fun mapper pat ->
    let pat =
      match pat with
        | {ppat_desc=Ppat_var ({txt} as id);
           ppat_loc;
           ppat_attributes} ->
             {pat with ppat_desc=Ppat_var ({id with txt=(f txt)})}
        | _ -> pat
    in
    default_mapper.pat mapper pat
  end;
}

(** unescape_stars_slashes_mapper unescapes all stars and slases in an AST
  *)
let unescape_stars_slashes_mapper =
  let unescape_stars_slashes str =
    let len = String.length str in
    if len < 2 then
      str
    else
      let ending = String.sub str 1 (len - 1) in
    String.sub str 0 1 ^
      replace_string "\\*" "*"
        (replace_string ("\\/") "/" ending)
  in
  identifier_mapper unescape_stars_slashes

(** escape_stars_slashes_mapper escapes all stars and slases in an AST
  *)
let escape_stars_slashes_mapper =
  let escape_stars_slashes str =
    let len = String.length str in
    if len < 2 then
      str
    else
      let ending = String.sub str 1 (len -1) in
      String.sub str 0 1 ^
        replace_string "*" "\\*"
          (replace_string "/" "\\/" ending)
  in
  identifier_mapper escape_stars_slashes

(**
 * swap_operator_mapper is a mapper that swaps two operators at parse/print time.
 * We need this since we want to transform operator such as "=" in Ocaml to "==" in Reason.
 * In this case, in the parser, everytime we see a token "==" in Reason, we transform it into "=";
 * Similarly, in the printer, everytime we see a token "=", we transform it into "==";
 *)
let swap_operator_mapper map = identifier_mapper (swap_txt map)

let reason_to_ml_swap_map = List.fold_left
                              (fun map (op1, op2) -> (StringMap.add op1 op2 map))
                              StringMap.empty
                              reason_to_ml_swapping_alist

let ml_to_reason_swap_map = List.fold_left
                              (fun map (op1, op2) -> (StringMap.add op2 op1 map))
                              StringMap.empty
                              reason_to_ml_swapping_alist

(* To be used in parser, transform a token into an ast node with different identifier
 *)
let reason_to_ml_swap_operator_mapper = swap_operator_mapper reason_to_ml_swap_map

(* To be used in printer, transform an ast node into a token with different identifier
 *)
let ml_to_reason_swap_operator_mapper = swap_operator_mapper ml_to_reason_swap_map

(* attribute_equals tests an attribute is txt
 *)
let attribute_equals to_compare = function
  | ({txt; _}, _) -> txt = to_compare

(* attribute_exists tests if an attribute exists in a list
 *)
let attribute_exists txt attributes = List.exists (attribute_equals txt) attributes

(* conflicted_attributes tests if both attribute1 and attribute2
 * exist
 *)
let attributes_conflicted attribute1 attribute2 attributes =
  attribute_exists attribute1 attributes &&
  attribute_exists attribute2 attributes

(* normalized_attributes removes attribute from a list of attributes
 *)
let normalized_attributes attribute attributes =
  List.filter (fun x -> not (attribute_equals attribute x)) attributes

(*
 * apply_mapper_chain family applies an ast_mapper_chain to an ast,
 * ordering from left to right.
 *)
let apply_mapper_chain_to_structure =
  List.fold_left (fun s mapper -> mapper.structure mapper s )
let apply_mapper_chain_to_signature =
  List.fold_left (fun s mapper -> mapper.signature mapper s )
let apply_mapper_chain_to_type =
  List.fold_left (fun s mapper -> mapper.typ mapper s )
let apply_mapper_chain_to_expr =
  List.fold_left (fun s mapper -> mapper.expr mapper s )
let apply_mapper_chain_to_pattern =
  List.fold_left (fun s mapper -> mapper.pat mapper s )

let apply_mapper_chain_to_toplevel_phrase toplevel_phrase chain =
  match toplevel_phrase with
  | Ptop_def x -> Ptop_def (apply_mapper_chain_to_structure x chain)
  | x -> x

let apply_mapper_chain_to_use_file use_file chain =
  List.map (fun x -> apply_mapper_chain_to_toplevel_phrase x chain) use_file

(* The following logic defines our own Error object
 * and register it with ocaml so it knows how to print it
 *)

type error = Syntax_error of string

exception Error of Location.t * error

let report_error ppf (Syntax_error err) =
  Format.(fprintf ppf "%s" err)

let () =
  Location.register_error_of_exn
    (function
     | Error (loc, err) ->
        Some (Location.error_of_printer loc report_error err)
     | _ ->
        None
     )

end
module Reason_parser : sig 
#1 "reason_parser.mli"

(* The type of tokens. *)

type token = 
  | WITH
  | WHILE
  | WHEN
  | VIRTUAL
  | VAL
  | UNDERSCORE
  | UIDENT of (string)
  | TYPE
  | TRY
  | TRUE
  | TO
  | TILDE
  | THEN
  | SWITCH
  | STRUCT
  | STRING of (string * string option)
  | STAR
  | SIG
  | SHARPOP of (string)
  | SHARP
  | SEMISEMI
  | SEMI
  | RPAREN
  | REC
  | RBRACKET
  | RBRACE
  | QUOTE
  | QUESTION
  | PRIVATE
  | PREFIXOP of (string)
  | PLUSEQ
  | PLUSDOT
  | PLUS
  | PERCENT
  | OR
  | OPTIONAL_NO_DEFAULT
  | OPEN
  | OF
  | OBJECT
  | NONREC
  | NEW
  | NATIVEINT of (nativeint)
  | MUTABLE
  | MODULE
  | MINUSGREATER
  | MINUSDOT
  | MINUS
  | METHOD
  | MATCH
  | LPAREN
  | LIDENT of (string)
  | LET
  | LESSMINUS
  | LESSGREATER
  | LESSDOTDOTGREATER
  | LESS
  | LBRACKETPERCENTPERCENT
  | LBRACKETPERCENT
  | LBRACKETLESS
  | LBRACKETGREATER
  | LBRACKETBAR
  | LBRACKETATATAT
  | LBRACKETATAT
  | LBRACKETAT
  | LBRACKET
  | LBRACELESS
  | LBRACE
  | LAZY
  | INT64 of (int64)
  | INT32 of (int32)
  | INT of (int)
  | INITIALIZER
  | INHERIT
  | INFIXOP4 of (string)
  | INFIXOP3 of (string)
  | INFIXOP2 of (string)
  | INFIXOP1 of (string)
  | INFIXOP0 of (string)
  | INCLUDE
  | IN
  | IF
  | GREATERRBRACKET
  | GREATERRBRACE
  | GREATER
  | FUNCTOR
  | FUNCTION
  | FUN
  | FOR
  | FLOAT of (string)
  | FALSE
  | EXTERNAL
  | EXPLICITLY_PASSED_OPTIONAL
  | EXCEPTION
  | EQUALGREATER
  | EQUAL
  | EOL
  | EOF
  | END
  | ELSE
  | DOWNTO
  | DOTDOTDOT
  | DOTDOT
  | DOT
  | DONE
  | DO
  | CONSTRAINT
  | COMMENT of (string * Location.t)
  | COMMA
  | COLONGREATER
  | COLONEQUAL
  | COLONCOLON
  | COLON
  | CLASS
  | CHAR of (char)
  | BEGIN
  | BARRBRACKET
  | BARBAR
  | BAR
  | BANG
  | BACKQUOTE
  | ASSERT
  | AS
  | AND
  | AMPERSAND
  | AMPERAMPER

(* This exception is raised by the monolithic API functions. *)

exception Error

(* The monolithic API. *)

val use_file: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.toplevel_phrase list)

val toplevel_phrase: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.toplevel_phrase)

val parse_pattern: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.pattern)

val parse_expression: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.expression)

val parse_core_type: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.core_type)

val interface: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.signature)

val implementation: (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (Parsetree.structure)

end = struct
#1 "reason_parser.ml"

module Basics = struct
  
  exception Error
  
  type token = 
    | WITH
    | WHILE
    | WHEN
    | VIRTUAL
    | VAL
    | UNDERSCORE
    | UIDENT of (
# 906 "reason_parser.mly"
       (string)
# 17 "reason_parser.ml"
  )
    | TYPE
    | TRY
    | TRUE
    | TO
    | TILDE
    | THEN
    | SWITCH
    | STRUCT
    | STRING of (
# 898 "reason_parser.mly"
       (string * string option)
# 30 "reason_parser.ml"
  )
    | STAR
    | SIG
    | SHARPOP of (
# 895 "reason_parser.mly"
       (string)
# 37 "reason_parser.ml"
  )
    | SHARP
    | SEMISEMI
    | SEMI
    | RPAREN
    | REC
    | RBRACKET
    | RBRACE
    | QUOTE
    | QUESTION
    | PRIVATE
    | PREFIXOP of (
# 882 "reason_parser.mly"
       (string)
# 52 "reason_parser.ml"
  )
    | PLUSEQ
    | PLUSDOT
    | PLUS
    | PERCENT
    | OR
    | OPTIONAL_NO_DEFAULT
    | OPEN
    | OF
    | OBJECT
    | NONREC
    | NEW
    | NATIVEINT of (
# 871 "reason_parser.mly"
       (nativeint)
# 68 "reason_parser.ml"
  )
    | MUTABLE
    | MODULE
    | MINUSGREATER
    | MINUSDOT
    | MINUS
    | METHOD
    | MATCH
    | LPAREN
    | LIDENT of (
# 857 "reason_parser.mly"
       (string)
# 81 "reason_parser.ml"
  )
    | LET
    | LESSMINUS
    | LESSGREATER
    | LESSDOTDOTGREATER
    | LESS
    | LBRACKETPERCENTPERCENT
    | LBRACKETPERCENT
    | LBRACKETLESS
    | LBRACKETGREATER
    | LBRACKETBAR
    | LBRACKETATATAT
    | LBRACKETATAT
    | LBRACKETAT
    | LBRACKET
    | LBRACELESS
    | LBRACE
    | LAZY
    | INT64 of (
# 841 "reason_parser.mly"
       (int64)
# 103 "reason_parser.ml"
  )
    | INT32 of (
# 840 "reason_parser.mly"
       (int32)
# 108 "reason_parser.ml"
  )
    | INT of (
# 839 "reason_parser.mly"
       (int)
# 113 "reason_parser.ml"
  )
    | INITIALIZER
    | INHERIT
    | INFIXOP4 of (
# 836 "reason_parser.mly"
       (string)
# 120 "reason_parser.ml"
  )
    | INFIXOP3 of (
# 835 "reason_parser.mly"
       (string)
# 125 "reason_parser.ml"
  )
    | INFIXOP2 of (
# 834 "reason_parser.mly"
       (string)
# 130 "reason_parser.ml"
  )
    | INFIXOP1 of (
# 833 "reason_parser.mly"
       (string)
# 135 "reason_parser.ml"
  )
    | INFIXOP0 of (
# 832 "reason_parser.mly"
       (string)
# 140 "reason_parser.ml"
  )
    | INCLUDE
    | IN
    | IF
    | GREATERRBRACKET
    | GREATERRBRACE
    | GREATER
    | FUNCTOR
    | FUNCTION
    | FUN
    | FOR
    | FLOAT of (
# 821 "reason_parser.mly"
       (string)
# 155 "reason_parser.ml"
  )
    | FALSE
    | EXTERNAL
    | EXPLICITLY_PASSED_OPTIONAL
    | EXCEPTION
    | EQUALGREATER
    | EQUAL
    | EOL
    | EOF
    | END
    | ELSE
    | DOWNTO
    | DOTDOTDOT
    | DOTDOT
    | DOT
    | DONE
    | DO
    | CONSTRAINT
    | COMMENT of (
# 913 "reason_parser.mly"
       (string * Location.t)
# 177 "reason_parser.ml"
  )
    | COMMA
    | COLONGREATER
    | COLONEQUAL
    | COLONCOLON
    | COLON
    | CLASS
    | CHAR of (
# 800 "reason_parser.mly"
       (char)
# 188 "reason_parser.ml"
  )
    | BEGIN
    | BARRBRACKET
    | BARBAR
    | BAR
    | BANG
    | BACKQUOTE
    | ASSERT
    | AS
    | AND
    | AMPERSAND
    | AMPERAMPER
  
end

include Basics

let _eRR =
  Basics.Error

type _menhir_env = {
  _menhir_lexer: Lexing.lexbuf -> token;
  _menhir_lexbuf: Lexing.lexbuf;
  _menhir_token: token;
  mutable _menhir_error: bool
}

and _menhir_state = 
  | MenhirState2760
  | MenhirState2757
  | MenhirState2753
  | MenhirState2748
  | MenhirState2745
  | MenhirState2743
  | MenhirState2742
  | MenhirState2739
  | MenhirState2738
  | MenhirState2737
  | MenhirState2736
  | MenhirState2733
  | MenhirState2732
  | MenhirState2731
  | MenhirState2730
  | MenhirState2717
  | MenhirState2710
  | MenhirState2709
  | MenhirState2707
  | MenhirState2706
  | MenhirState2705
  | MenhirState2704
  | MenhirState2701
  | MenhirState2700
  | MenhirState2698
  | MenhirState2697
  | MenhirState2690
  | MenhirState2688
  | MenhirState2687
  | MenhirState2683
  | MenhirState2682
  | MenhirState2681
  | MenhirState2680
  | MenhirState2679
  | MenhirState2677
  | MenhirState2673
  | MenhirState2669
  | MenhirState2666
  | MenhirState2665
  | MenhirState2664
  | MenhirState2663
  | MenhirState2662
  | MenhirState2660
  | MenhirState2657
  | MenhirState2655
  | MenhirState2654
  | MenhirState2652
  | MenhirState2651
  | MenhirState2649
  | MenhirState2647
  | MenhirState2645
  | MenhirState2644
  | MenhirState2642
  | MenhirState2640
  | MenhirState2639
  | MenhirState2637
  | MenhirState2636
  | MenhirState2635
  | MenhirState2632
  | MenhirState2631
  | MenhirState2630
  | MenhirState2629
  | MenhirState2627
  | MenhirState2626
  | MenhirState2620
  | MenhirState2619
  | MenhirState2613
  | MenhirState2612
  | MenhirState2610
  | MenhirState2609
  | MenhirState2608
  | MenhirState2606
  | MenhirState2605
  | MenhirState2604
  | MenhirState2602
  | MenhirState2601
  | MenhirState2600
  | MenhirState2597
  | MenhirState2596
  | MenhirState2595
  | MenhirState2593
  | MenhirState2592
  | MenhirState2591
  | MenhirState2590
  | MenhirState2589
  | MenhirState2588
  | MenhirState2587
  | MenhirState2586
  | MenhirState2584
  | MenhirState2583
  | MenhirState2582
  | MenhirState2579
  | MenhirState2578
  | MenhirState2576
  | MenhirState2574
  | MenhirState2570
  | MenhirState2565
  | MenhirState2563
  | MenhirState2557
  | MenhirState2556
  | MenhirState2555
  | MenhirState2554
  | MenhirState2552
  | MenhirState2551
  | MenhirState2548
  | MenhirState2540
  | MenhirState2539
  | MenhirState2538
  | MenhirState2537
  | MenhirState2536
  | MenhirState2533
  | MenhirState2528
  | MenhirState2527
  | MenhirState2523
  | MenhirState2522
  | MenhirState2521
  | MenhirState2520
  | MenhirState2516
  | MenhirState2515
  | MenhirState2511
  | MenhirState2510
  | MenhirState2507
  | MenhirState2504
  | MenhirState2503
  | MenhirState2501
  | MenhirState2500
  | MenhirState2498
  | MenhirState2497
  | MenhirState2494
  | MenhirState2493
  | MenhirState2491
  | MenhirState2490
  | MenhirState2487
  | MenhirState2486
  | MenhirState2484
  | MenhirState2483
  | MenhirState2481
  | MenhirState2480
  | MenhirState2479
  | MenhirState2478
  | MenhirState2476
  | MenhirState2475
  | MenhirState2473
  | MenhirState2472
  | MenhirState2470
  | MenhirState2466
  | MenhirState2465
  | MenhirState2464
  | MenhirState2463
  | MenhirState2461
  | MenhirState2459
  | MenhirState2457
  | MenhirState2456
  | MenhirState2449
  | MenhirState2442
  | MenhirState2441
  | MenhirState2439
  | MenhirState2434
  | MenhirState2433
  | MenhirState2432
  | MenhirState2428
  | MenhirState2426
  | MenhirState2423
  | MenhirState2422
  | MenhirState2421
  | MenhirState2418
  | MenhirState2417
  | MenhirState2416
  | MenhirState2415
  | MenhirState2411
  | MenhirState2407
  | MenhirState2406
  | MenhirState2405
  | MenhirState2396
  | MenhirState2395
  | MenhirState2393
  | MenhirState2391
  | MenhirState2389
  | MenhirState2387
  | MenhirState2382
  | MenhirState2379
  | MenhirState2377
  | MenhirState2376
  | MenhirState2375
  | MenhirState2374
  | MenhirState2368
  | MenhirState2366
  | MenhirState2364
  | MenhirState2362
  | MenhirState2360
  | MenhirState2358
  | MenhirState2354
  | MenhirState2351
  | MenhirState2350
  | MenhirState2347
  | MenhirState2346
  | MenhirState2343
  | MenhirState2339
  | MenhirState2337
  | MenhirState2326
  | MenhirState2324
  | MenhirState2323
  | MenhirState2320
  | MenhirState2315
  | MenhirState2314
  | MenhirState2312
  | MenhirState2310
  | MenhirState2309
  | MenhirState2307
  | MenhirState2306
  | MenhirState2304
  | MenhirState2303
  | MenhirState2302
  | MenhirState2301
  | MenhirState2296
  | MenhirState2295
  | MenhirState2294
  | MenhirState2293
  | MenhirState2292
  | MenhirState2291
  | MenhirState2289
  | MenhirState2288
  | MenhirState2281
  | MenhirState2278
  | MenhirState2277
  | MenhirState2275
  | MenhirState2274
  | MenhirState2272
  | MenhirState2271
  | MenhirState2270
  | MenhirState2269
  | MenhirState2268
  | MenhirState2259
  | MenhirState2258
  | MenhirState2257
  | MenhirState2255
  | MenhirState2253
  | MenhirState2252
  | MenhirState2248
  | MenhirState2247
  | MenhirState2238
  | MenhirState2235
  | MenhirState2234
  | MenhirState2233
  | MenhirState2232
  | MenhirState2231
  | MenhirState2230
  | MenhirState2229
  | MenhirState2224
  | MenhirState2223
  | MenhirState2222
  | MenhirState2221
  | MenhirState2220
  | MenhirState2219
  | MenhirState2218
  | MenhirState2217
  | MenhirState2214
  | MenhirState2213
  | MenhirState2211
  | MenhirState2210
  | MenhirState2208
  | MenhirState2202
  | MenhirState2198
  | MenhirState2197
  | MenhirState2192
  | MenhirState2191
  | MenhirState2188
  | MenhirState2187
  | MenhirState2183
  | MenhirState2182
  | MenhirState2179
  | MenhirState2177
  | MenhirState2173
  | MenhirState2172
  | MenhirState2169
  | MenhirState2168
  | MenhirState2165
  | MenhirState2161
  | MenhirState2158
  | MenhirState2155
  | MenhirState2153
  | MenhirState2149
  | MenhirState2147
  | MenhirState2144
  | MenhirState2143
  | MenhirState2142
  | MenhirState2140
  | MenhirState2139
  | MenhirState2138
  | MenhirState2137
  | MenhirState2134
  | MenhirState2132
  | MenhirState2129
  | MenhirState2128
  | MenhirState2127
  | MenhirState2126
  | MenhirState2125
  | MenhirState2123
  | MenhirState2122
  | MenhirState2118
  | MenhirState2117
  | MenhirState2116
  | MenhirState2114
  | MenhirState2113
  | MenhirState2109
  | MenhirState2103
  | MenhirState2096
  | MenhirState2093
  | MenhirState2092
  | MenhirState2088
  | MenhirState2087
  | MenhirState2086
  | MenhirState2083
  | MenhirState2082
  | MenhirState2078
  | MenhirState2077
  | MenhirState2076
  | MenhirState2075
  | MenhirState2074
  | MenhirState2068
  | MenhirState2065
  | MenhirState2063
  | MenhirState2062
  | MenhirState2061
  | MenhirState2058
  | MenhirState2057
  | MenhirState2056
  | MenhirState2055
  | MenhirState2054
  | MenhirState2052
  | MenhirState2051
  | MenhirState2049
  | MenhirState2048
  | MenhirState2046
  | MenhirState2044
  | MenhirState2043
  | MenhirState2041
  | MenhirState2039
  | MenhirState2037
  | MenhirState2036
  | MenhirState2033
  | MenhirState2032
  | MenhirState2027
  | MenhirState2025
  | MenhirState2016
  | MenhirState2006
  | MenhirState2005
  | MenhirState2001
  | MenhirState1999
  | MenhirState1998
  | MenhirState1997
  | MenhirState1995
  | MenhirState1994
  | MenhirState1991
  | MenhirState1990
  | MenhirState1989
  | MenhirState1988
  | MenhirState1987
  | MenhirState1985
  | MenhirState1984
  | MenhirState1983
  | MenhirState1982
  | MenhirState1980
  | MenhirState1979
  | MenhirState1978
  | MenhirState1977
  | MenhirState1975
  | MenhirState1974
  | MenhirState1973
  | MenhirState1972
  | MenhirState1971
  | MenhirState1969
  | MenhirState1965
  | MenhirState1962
  | MenhirState1958
  | MenhirState1955
  | MenhirState1952
  | MenhirState1950
  | MenhirState1949
  | MenhirState1948
  | MenhirState1947
  | MenhirState1945
  | MenhirState1944
  | MenhirState1943
  | MenhirState1939
  | MenhirState1935
  | MenhirState1932
  | MenhirState1929
  | MenhirState1924
  | MenhirState1923
  | MenhirState1922
  | MenhirState1921
  | MenhirState1920
  | MenhirState1919
  | MenhirState1917
  | MenhirState1912
  | MenhirState1910
  | MenhirState1909
  | MenhirState1906
  | MenhirState1904
  | MenhirState1902
  | MenhirState1900
  | MenhirState1898
  | MenhirState1896
  | MenhirState1895
  | MenhirState1894
  | MenhirState1893
  | MenhirState1891
  | MenhirState1888
  | MenhirState1887
  | MenhirState1886
  | MenhirState1885
  | MenhirState1882
  | MenhirState1881
  | MenhirState1880
  | MenhirState1877
  | MenhirState1876
  | MenhirState1875
  | MenhirState1872
  | MenhirState1871
  | MenhirState1870
  | MenhirState1867
  | MenhirState1864
  | MenhirState1861
  | MenhirState1859
  | MenhirState1858
  | MenhirState1857
  | MenhirState1856
  | MenhirState1855
  | MenhirState1854
  | MenhirState1853
  | MenhirState1846
  | MenhirState1842
  | MenhirState1834
  | MenhirState1832
  | MenhirState1831
  | MenhirState1827
  | MenhirState1826
  | MenhirState1825
  | MenhirState1824
  | MenhirState1823
  | MenhirState1820
  | MenhirState1819
  | MenhirState1816
  | MenhirState1815
  | MenhirState1814
  | MenhirState1809
  | MenhirState1808
  | MenhirState1807
  | MenhirState1806
  | MenhirState1805
  | MenhirState1801
  | MenhirState1800
  | MenhirState1799
  | MenhirState1795
  | MenhirState1794
  | MenhirState1790
  | MenhirState1788
  | MenhirState1787
  | MenhirState1786
  | MenhirState1781
  | MenhirState1780
  | MenhirState1779
  | MenhirState1774
  | MenhirState1773
  | MenhirState1772
  | MenhirState1771
  | MenhirState1770
  | MenhirState1768
  | MenhirState1767
  | MenhirState1766
  | MenhirState1763
  | MenhirState1757
  | MenhirState1753
  | MenhirState1749
  | MenhirState1745
  | MenhirState1744
  | MenhirState1742
  | MenhirState1739
  | MenhirState1738
  | MenhirState1737
  | MenhirState1735
  | MenhirState1734
  | MenhirState1731
  | MenhirState1729
  | MenhirState1728
  | MenhirState1726
  | MenhirState1725
  | MenhirState1723
  | MenhirState1721
  | MenhirState1720
  | MenhirState1719
  | MenhirState1718
  | MenhirState1714
  | MenhirState1713
  | MenhirState1712
  | MenhirState1711
  | MenhirState1710
  | MenhirState1709
  | MenhirState1708
  | MenhirState1707
  | MenhirState1704
  | MenhirState1703
  | MenhirState1702
  | MenhirState1700
  | MenhirState1698
  | MenhirState1695
  | MenhirState1693
  | MenhirState1692
  | MenhirState1691
  | MenhirState1690
  | MenhirState1689
  | MenhirState1688
  | MenhirState1686
  | MenhirState1685
  | MenhirState1684
  | MenhirState1683
  | MenhirState1680
  | MenhirState1678
  | MenhirState1676
  | MenhirState1674
  | MenhirState1671
  | MenhirState1669
  | MenhirState1668
  | MenhirState1667
  | MenhirState1666
  | MenhirState1664
  | MenhirState1663
  | MenhirState1662
  | MenhirState1658
  | MenhirState1657
  | MenhirState1650
  | MenhirState1648
  | MenhirState1647
  | MenhirState1643
  | MenhirState1642
  | MenhirState1641
  | MenhirState1640
  | MenhirState1639
  | MenhirState1637
  | MenhirState1635
  | MenhirState1634
  | MenhirState1633
  | MenhirState1632
  | MenhirState1631
  | MenhirState1630
  | MenhirState1627
  | MenhirState1626
  | MenhirState1623
  | MenhirState1617
  | MenhirState1615
  | MenhirState1610
  | MenhirState1609
  | MenhirState1606
  | MenhirState1603
  | MenhirState1601
  | MenhirState1599
  | MenhirState1595
  | MenhirState1594
  | MenhirState1593
  | MenhirState1592
  | MenhirState1588
  | MenhirState1587
  | MenhirState1586
  | MenhirState1581
  | MenhirState1579
  | MenhirState1576
  | MenhirState1575
  | MenhirState1573
  | MenhirState1566
  | MenhirState1563
  | MenhirState1561
  | MenhirState1559
  | MenhirState1558
  | MenhirState1556
  | MenhirState1554
  | MenhirState1553
  | MenhirState1552
  | MenhirState1551
  | MenhirState1545
  | MenhirState1544
  | MenhirState1542
  | MenhirState1539
  | MenhirState1538
  | MenhirState1536
  | MenhirState1534
  | MenhirState1532
  | MenhirState1530
  | MenhirState1528
  | MenhirState1526
  | MenhirState1525
  | MenhirState1524
  | MenhirState1522
  | MenhirState1520
  | MenhirState1519
  | MenhirState1518
  | MenhirState1516
  | MenhirState1514
  | MenhirState1513
  | MenhirState1512
  | MenhirState1509
  | MenhirState1507
  | MenhirState1506
  | MenhirState1505
  | MenhirState1503
  | MenhirState1501
  | MenhirState1500
  | MenhirState1498
  | MenhirState1496
  | MenhirState1494
  | MenhirState1493
  | MenhirState1492
  | MenhirState1491
  | MenhirState1490
  | MenhirState1489
  | MenhirState1487
  | MenhirState1485
  | MenhirState1483
  | MenhirState1482
  | MenhirState1480
  | MenhirState1478
  | MenhirState1477
  | MenhirState1476
  | MenhirState1475
  | MenhirState1474
  | MenhirState1472
  | MenhirState1471
  | MenhirState1468
  | MenhirState1467
  | MenhirState1466
  | MenhirState1464
  | MenhirState1462
  | MenhirState1460
  | MenhirState1459
  | MenhirState1457
  | MenhirState1445
  | MenhirState1444
  | MenhirState1443
  | MenhirState1442
  | MenhirState1441
  | MenhirState1438
  | MenhirState1436
  | MenhirState1435
  | MenhirState1434
  | MenhirState1433
  | MenhirState1432
  | MenhirState1430
  | MenhirState1429
  | MenhirState1428
  | MenhirState1427
  | MenhirState1425
  | MenhirState1422
  | MenhirState1421
  | MenhirState1420
  | MenhirState1419
  | MenhirState1418
  | MenhirState1414
  | MenhirState1411
  | MenhirState1409
  | MenhirState1408
  | MenhirState1407
  | MenhirState1406
  | MenhirState1405
  | MenhirState1402
  | MenhirState1401
  | MenhirState1400
  | MenhirState1398
  | MenhirState1397
  | MenhirState1393
  | MenhirState1392
  | MenhirState1390
  | MenhirState1389
  | MenhirState1387
  | MenhirState1386
  | MenhirState1384
  | MenhirState1383
  | MenhirState1380
  | MenhirState1379
  | MenhirState1377
  | MenhirState1376
  | MenhirState1375
  | MenhirState1374
  | MenhirState1372
  | MenhirState1369
  | MenhirState1367
  | MenhirState1364
  | MenhirState1361
  | MenhirState1359
  | MenhirState1358
  | MenhirState1357
  | MenhirState1352
  | MenhirState1351
  | MenhirState1347
  | MenhirState1346
  | MenhirState1345
  | MenhirState1344
  | MenhirState1338
  | MenhirState1337
  | MenhirState1332
  | MenhirState1330
  | MenhirState1328
  | MenhirState1324
  | MenhirState1320
  | MenhirState1319
  | MenhirState1316
  | MenhirState1313
  | MenhirState1312
  | MenhirState1310
  | MenhirState1309
  | MenhirState1307
  | MenhirState1306
  | MenhirState1303
  | MenhirState1302
  | MenhirState1300
  | MenhirState1299
  | MenhirState1296
  | MenhirState1295
  | MenhirState1293
  | MenhirState1292
  | MenhirState1290
  | MenhirState1289
  | MenhirState1288
  | MenhirState1287
  | MenhirState1285
  | MenhirState1284
  | MenhirState1282
  | MenhirState1280
  | MenhirState1278
  | MenhirState1277
  | MenhirState1275
  | MenhirState1272
  | MenhirState1271
  | MenhirState1269
  | MenhirState1268
  | MenhirState1267
  | MenhirState1264
  | MenhirState1260
  | MenhirState1256
  | MenhirState1254
  | MenhirState1253
  | MenhirState1252
  | MenhirState1251
  | MenhirState1250
  | MenhirState1249
  | MenhirState1248
  | MenhirState1247
  | MenhirState1246
  | MenhirState1245
  | MenhirState1244
  | MenhirState1239
  | MenhirState1238
  | MenhirState1237
  | MenhirState1236
  | MenhirState1235
  | MenhirState1234
  | MenhirState1232
  | MenhirState1227
  | MenhirState1225
  | MenhirState1224
  | MenhirState1221
  | MenhirState1219
  | MenhirState1217
  | MenhirState1215
  | MenhirState1213
  | MenhirState1211
  | MenhirState1208
  | MenhirState1207
  | MenhirState1206
  | MenhirState1205
  | MenhirState1204
  | MenhirState1199
  | MenhirState1198
  | MenhirState1197
  | MenhirState1196
  | MenhirState1195
  | MenhirState1194
  | MenhirState1193
  | MenhirState1192
  | MenhirState1189
  | MenhirState1188
  | MenhirState1185
  | MenhirState1184
  | MenhirState1183
  | MenhirState1181
  | MenhirState1176
  | MenhirState1175
  | MenhirState1174
  | MenhirState1171
  | MenhirState1167
  | MenhirState1166
  | MenhirState1161
  | MenhirState1159
  | MenhirState1158
  | MenhirState1157
  | MenhirState1156
  | MenhirState1155
  | MenhirState1154
  | MenhirState1149
  | MenhirState1147
  | MenhirState1144
  | MenhirState1142
  | MenhirState1139
  | MenhirState1138
  | MenhirState1137
  | MenhirState1135
  | MenhirState1134
  | MenhirState1131
  | MenhirState1124
  | MenhirState1123
  | MenhirState1120
  | MenhirState1117
  | MenhirState1115
  | MenhirState1113
  | MenhirState1109
  | MenhirState1108
  | MenhirState1103
  | MenhirState1102
  | MenhirState1099
  | MenhirState1098
  | MenhirState1092
  | MenhirState1089
  | MenhirState1087
  | MenhirState1086
  | MenhirState1084
  | MenhirState1083
  | MenhirState1081
  | MenhirState1078
  | MenhirState1076
  | MenhirState1074
  | MenhirState1071
  | MenhirState1068
  | MenhirState1066
  | MenhirState1065
  | MenhirState1063
  | MenhirState1061
  | MenhirState1060
  | MenhirState1059
  | MenhirState1057
  | MenhirState1056
  | MenhirState1054
  | MenhirState1053
  | MenhirState1052
  | MenhirState1051
  | MenhirState1048
  | MenhirState1047
  | MenhirState1046
  | MenhirState1045
  | MenhirState1044
  | MenhirState1043
  | MenhirState1041
  | MenhirState1038
  | MenhirState1037
  | MenhirState1036
  | MenhirState1035
  | MenhirState1034
  | MenhirState1033
  | MenhirState1032
  | MenhirState1027
  | MenhirState1025
  | MenhirState1024
  | MenhirState1023
  | MenhirState1022
  | MenhirState1021
  | MenhirState1020
  | MenhirState1019
  | MenhirState1018
  | MenhirState1016
  | MenhirState1015
  | MenhirState1014
  | MenhirState1013
  | MenhirState1012
  | MenhirState1011
  | MenhirState1009
  | MenhirState1008
  | MenhirState1007
  | MenhirState1006
  | MenhirState1003
  | MenhirState1002
  | MenhirState1001
  | MenhirState1000
  | MenhirState999
  | MenhirState998
  | MenhirState997
  | MenhirState996
  | MenhirState995
  | MenhirState994
  | MenhirState993
  | MenhirState992
  | MenhirState991
  | MenhirState990
  | MenhirState989
  | MenhirState988
  | MenhirState987
  | MenhirState986
  | MenhirState985
  | MenhirState984
  | MenhirState983
  | MenhirState982
  | MenhirState981
  | MenhirState980
  | MenhirState979
  | MenhirState978
  | MenhirState977
  | MenhirState976
  | MenhirState975
  | MenhirState974
  | MenhirState973
  | MenhirState972
  | MenhirState971
  | MenhirState970
  | MenhirState969
  | MenhirState968
  | MenhirState967
  | MenhirState966
  | MenhirState965
  | MenhirState964
  | MenhirState963
  | MenhirState961
  | MenhirState960
  | MenhirState955
  | MenhirState953
  | MenhirState951
  | MenhirState950
  | MenhirState949
  | MenhirState948
  | MenhirState947
  | MenhirState946
  | MenhirState940
  | MenhirState938
  | MenhirState936
  | MenhirState935
  | MenhirState933
  | MenhirState931
  | MenhirState930
  | MenhirState929
  | MenhirState926
  | MenhirState925
  | MenhirState924
  | MenhirState923
  | MenhirState922
  | MenhirState921
  | MenhirState920
  | MenhirState914
  | MenhirState913
  | MenhirState912
  | MenhirState911
  | MenhirState910
  | MenhirState909
  | MenhirState908
  | MenhirState907
  | MenhirState906
  | MenhirState903
  | MenhirState902
  | MenhirState899
  | MenhirState898
  | MenhirState897
  | MenhirState896
  | MenhirState895
  | MenhirState894
  | MenhirState893
  | MenhirState892
  | MenhirState891
  | MenhirState888
  | MenhirState887
  | MenhirState886
  | MenhirState885
  | MenhirState884
  | MenhirState883
  | MenhirState882
  | MenhirState881
  | MenhirState880
  | MenhirState879
  | MenhirState878
  | MenhirState877
  | MenhirState876
  | MenhirState875
  | MenhirState874
  | MenhirState873
  | MenhirState872
  | MenhirState871
  | MenhirState870
  | MenhirState869
  | MenhirState868
  | MenhirState867
  | MenhirState866
  | MenhirState865
  | MenhirState864
  | MenhirState863
  | MenhirState862
  | MenhirState861
  | MenhirState860
  | MenhirState859
  | MenhirState858
  | MenhirState857
  | MenhirState856
  | MenhirState855
  | MenhirState854
  | MenhirState853
  | MenhirState852
  | MenhirState851
  | MenhirState850
  | MenhirState848
  | MenhirState847
  | MenhirState842
  | MenhirState840
  | MenhirState838
  | MenhirState837
  | MenhirState836
  | MenhirState835
  | MenhirState834
  | MenhirState833
  | MenhirState827
  | MenhirState825
  | MenhirState823
  | MenhirState822
  | MenhirState820
  | MenhirState818
  | MenhirState817
  | MenhirState816
  | MenhirState815
  | MenhirState814
  | MenhirState813
  | MenhirState812
  | MenhirState811
  | MenhirState805
  | MenhirState801
  | MenhirState800
  | MenhirState799
  | MenhirState798
  | MenhirState797
  | MenhirState796
  | MenhirState794
  | MenhirState793
  | MenhirState791
  | MenhirState780
  | MenhirState775
  | MenhirState773
  | MenhirState772
  | MenhirState771
  | MenhirState770
  | MenhirState768
  | MenhirState767
  | MenhirState766
  | MenhirState765
  | MenhirState764
  | MenhirState763
  | MenhirState762
  | MenhirState761
  | MenhirState759
  | MenhirState758
  | MenhirState757
  | MenhirState756
  | MenhirState755
  | MenhirState754
  | MenhirState753
  | MenhirState749
  | MenhirState748
  | MenhirState747
  | MenhirState746
  | MenhirState743
  | MenhirState741
  | MenhirState740
  | MenhirState739
  | MenhirState738
  | MenhirState737
  | MenhirState736
  | MenhirState735
  | MenhirState734
  | MenhirState733
  | MenhirState730
  | MenhirState729
  | MenhirState722
  | MenhirState719
  | MenhirState714
  | MenhirState713
  | MenhirState712
  | MenhirState711
  | MenhirState705
  | MenhirState700
  | MenhirState699
  | MenhirState692
  | MenhirState688
  | MenhirState687
  | MenhirState686
  | MenhirState681
  | MenhirState678
  | MenhirState676
  | MenhirState668
  | MenhirState666
  | MenhirState665
  | MenhirState660
  | MenhirState658
  | MenhirState657
  | MenhirState655
  | MenhirState654
  | MenhirState650
  | MenhirState649
  | MenhirState646
  | MenhirState644
  | MenhirState642
  | MenhirState635
  | MenhirState634
  | MenhirState630
  | MenhirState619
  | MenhirState615
  | MenhirState614
  | MenhirState613
  | MenhirState612
  | MenhirState611
  | MenhirState610
  | MenhirState609
  | MenhirState605
  | MenhirState603
  | MenhirState602
  | MenhirState594
  | MenhirState593
  | MenhirState592
  | MenhirState590
  | MenhirState588
  | MenhirState587
  | MenhirState579
  | MenhirState577
  | MenhirState576
  | MenhirState574
  | MenhirState573
  | MenhirState564
  | MenhirState563
  | MenhirState554
  | MenhirState553
  | MenhirState550
  | MenhirState548
  | MenhirState545
  | MenhirState540
  | MenhirState533
  | MenhirState532
  | MenhirState531
  | MenhirState526
  | MenhirState525
  | MenhirState519
  | MenhirState518
  | MenhirState516
  | MenhirState514
  | MenhirState511
  | MenhirState505
  | MenhirState504
  | MenhirState503
  | MenhirState499
  | MenhirState495
  | MenhirState487
  | MenhirState486
  | MenhirState484
  | MenhirState481
  | MenhirState480
  | MenhirState477
  | MenhirState476
  | MenhirState475
  | MenhirState474
  | MenhirState473
  | MenhirState472
  | MenhirState471
  | MenhirState470
  | MenhirState469
  | MenhirState465
  | MenhirState464
  | MenhirState457
  | MenhirState456
  | MenhirState455
  | MenhirState454
  | MenhirState453
  | MenhirState452
  | MenhirState451
  | MenhirState449
  | MenhirState444
  | MenhirState443
  | MenhirState442
  | MenhirState441
  | MenhirState440
  | MenhirState439
  | MenhirState438
  | MenhirState437
  | MenhirState436
  | MenhirState434
  | MenhirState432
  | MenhirState426
  | MenhirState425
  | MenhirState423
  | MenhirState422
  | MenhirState420
  | MenhirState419
  | MenhirState418
  | MenhirState417
  | MenhirState416
  | MenhirState415
  | MenhirState413
  | MenhirState412
  | MenhirState411
  | MenhirState408
  | MenhirState405
  | MenhirState404
  | MenhirState403
  | MenhirState401
  | MenhirState400
  | MenhirState399
  | MenhirState397
  | MenhirState396
  | MenhirState395
  | MenhirState394
  | MenhirState392
  | MenhirState389
  | MenhirState337
  | MenhirState334
  | MenhirState333
  | MenhirState332
  | MenhirState329
  | MenhirState328
  | MenhirState327
  | MenhirState325
  | MenhirState324
  | MenhirState317
  | MenhirState313
  | MenhirState307
  | MenhirState306
  | MenhirState303
  | MenhirState301
  | MenhirState296
  | MenhirState295
  | MenhirState292
  | MenhirState291
  | MenhirState288
  | MenhirState287
  | MenhirState284
  | MenhirState277
  | MenhirState273
  | MenhirState272
  | MenhirState271
  | MenhirState270
  | MenhirState263
  | MenhirState255
  | MenhirState254
  | MenhirState251
  | MenhirState250
  | MenhirState249
  | MenhirState246
  | MenhirState245
  | MenhirState244
  | MenhirState243
  | MenhirState242
  | MenhirState240
  | MenhirState239
  | MenhirState234
  | MenhirState233
  | MenhirState216
  | MenhirState215
  | MenhirState212
  | MenhirState210
  | MenhirState196
  | MenhirState195
  | MenhirState193
  | MenhirState191
  | MenhirState190
  | MenhirState189
  | MenhirState187
  | MenhirState184
  | MenhirState181
  | MenhirState180
  | MenhirState178
  | MenhirState177
  | MenhirState175
  | MenhirState174
  | MenhirState173
  | MenhirState170
  | MenhirState164
  | MenhirState163
  | MenhirState162
  | MenhirState161
  | MenhirState159
  | MenhirState158
  | MenhirState156
  | MenhirState154
  | MenhirState153
  | MenhirState147
  | MenhirState143
  | MenhirState141
  | MenhirState138
  | MenhirState137
  | MenhirState135
  | MenhirState134
  | MenhirState130
  | MenhirState129
  | MenhirState128
  | MenhirState124
  | MenhirState122
  | MenhirState121
  | MenhirState120
  | MenhirState118
  | MenhirState115
  | MenhirState112
  | MenhirState111
  | MenhirState109
  | MenhirState108
  | MenhirState106
  | MenhirState105
  | MenhirState102
  | MenhirState98
  | MenhirState92
  | MenhirState90
  | MenhirState88
  | MenhirState87
  | MenhirState84
  | MenhirState83
  | MenhirState82
  | MenhirState81
  | MenhirState79
  | MenhirState78
  | MenhirState77
  | MenhirState76
  | MenhirState74
  | MenhirState73
  | MenhirState71
  | MenhirState19
  | MenhirState17
  | MenhirState15
  | MenhirState10
  | MenhirState9
  | MenhirState7
  | MenhirState6
  | MenhirState2
  | MenhirState1
  | MenhirState0

# 52 "reason_parser.mly"
  
open Location
open Asttypes
open Longident
open Parsetree
open Ast_helper
open Ast_mapper
open Syntax_util

(*
   TODO:
   - Remove all [open]s from the top of this file one by one and fix compilation
   failures that ensue by specifying the appropriate long identifiers. That
   will make the parser much easier to reason about.
   - Go back to trunk, do the same (remove [open]s, and fully specify long
   idents), to perform a clean diff.

*)

(**

   location.ml:
   ------------
   let mkloc txt loc = { txt ; loc }
   let rhs_loc n = {
     loc_start = Parsing.rhs_start_pos n;
     loc_end = Parsing.rhs_end_pos n;
     loc_ghost = false;
   }
   let symbol_rloc () = {
     loc_start = Parsing.symbol_start_pos ();
     loc_end = Parsing.symbol_end_pos ();
     loc_ghost = false;
   }

   let symbol_gloc () = {
     loc_start = Parsing.symbol_start_pos ();
     loc_end = Parsing.symbol_end_pos ();
     loc_ghost = true;
   }

   ast_helper.ml:
   ------------
   module Typ = struct
    val mk: ?loc:loc -> ?attrs:attrs -> core_type_desc -> core_type
    let mk ?(loc = !default_loc) ?(attrs = []) d =
       {ptyp_desc = d; ptyp_loc = loc; ptyp_attributes = attrs}
     ..
   end

   parse_tree.mli
   --------------
   and core_type = {
     ptyp_desc: core_type_desc;
     ptyp_loc: Location.t;
     ptyp_attributes: attributes; (* ... [@id1] [@id2] *)
   }

   and core_type_desc =
     | Ptyp_any
           (*  _ *)
     | Ptyp_var of string
           (* 'a *)
     | Ptyp_arrow of label * core_type * core_type
           (* T1 -> T2       (label = "")
              ~l:T1 -> T2    (label = "l")
              ?l:T1 -> T2    (label = "?l")
            *)
     | Ptyp_tuple of core_type list
           (* T1 * ... * Tn   (n >= 2) *)

   reason_parser.mly
   ---------------
   In general:

                                          syntax variant          {pblah_desc: core_blah_desc
                                                                   pblah_loc: {txt, loc}
                                                                   pblah_attributes: ... }
                                         /              \            /       \
   val mkblah: ~loc -> ~attributes ->     core_blah_desc     ->      core_blah
   let mkblah = Blah.mk

*)


let dummy_loc () = {
  loc_start = Lexing.dummy_pos;
  loc_end = Lexing.dummy_pos;
  loc_ghost = false;
}

let mklocation loc_start loc_end = {
  loc_start = loc_start;
  loc_end = loc_end;
  loc_ghost = false;
}

let with_txt a txt = {
    a with txt=txt;
}

let make_real_loc loc = {
    loc with loc_ghost = false
}

let make_ghost_loc loc = {
    loc with loc_ghost = true
}

let ghloc ?(loc=dummy_loc ()) d = { txt = d; loc = (make_ghost_loc loc) }

(**
  * turn an object into a real
  *)
let make_real_exp exp = {
    exp with pexp_loc = make_real_loc exp.pexp_loc
}
let make_real_pat pat = {
    pat with ppat_loc = make_real_loc pat.ppat_loc
}
let make_real_cf cf = {
    cf with pcf_loc = make_real_loc cf.pcf_loc
}

(**
  * turn a object into ghost
  *)
let make_ghost_cf cf = {
    cf with pcf_loc = make_ghost_loc cf.pcf_loc
}
let make_ghost_exp exp = {
    exp with pexp_loc = make_ghost_loc exp.pexp_loc
}

let make_ghost_pat pat = {
    pat with ppat_loc = make_ghost_loc pat.ppat_loc
}

(**
  * change the location state to be a ghost location or real location
  *)
let set_loc_state is_ghost loc =
    if is_ghost then make_ghost_loc loc else make_real_loc loc

let mktyp ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Typ.mk ~loc d

let mkpat ?(attrs=[]) ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Pat.mk ~loc ~attrs d

let mkexp ?(attrs=[]) ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Exp.mk ~loc ~attrs d

let mkmty ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Mty.mk ~loc d

let mksig ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Sig.mk ~loc d

let mkmod ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Mod.mk ~loc d

let mkstr ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Str.mk ~loc d

let mkclass ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Cl.mk ~loc d

let mkcty ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Cty.mk ~loc d

let mkctf ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Ctf.mk ~loc d

let mkcf ?(loc=dummy_loc()) ?(ghost=false) d =
    let loc = set_loc_state ghost loc in
    Cf.mk ~loc d

let mkoption d =
  let loc = {d.ptyp_loc with loc_ghost = true} in
  Typ.mk ~loc (Ptyp_constr(mkloc (Ldot (Lident "*predef*", "option")) loc,[d]))

let simple_ghost_text_attr ?(loc=dummy_loc ()) txt =
  let loc = set_loc_state true loc in
  [({txt; loc}, PStr [])]

let mkExplicitArityTuplePat ?(loc=dummy_loc ()) pat =
  (* Tell OCaml type system that what this tuple construction represents is
     not actually a tuple, and should represent several constructor
     arguments.  This allows the syntax the ability to distinguish between:

     X (10, 20)  -- One argument constructor
     X 10 20     -- Multi argument constructor
  *)
  mkpat
    ~loc
    ~attrs:(simple_ghost_text_attr ~loc "explicit_arity")
    pat

let mkExplicitArityTupleExp ?(loc=dummy_loc ()) exp =
  mkexp
    ~loc
    ~attrs:(simple_ghost_text_attr ~loc "explicit_arity")
    exp

let is_pattern_list_single_any = function
  | [{ppat_desc=Ppat_any; ppat_attributes=[]} as onlyItem] -> Some onlyItem
  | _ -> None

let set_structure_item_location x loc = {x with pstr_loc = loc};;

let mkoperator name =
  let loc = name.loc in
  Exp.mk ~loc (Pexp_ident(mkloc (Lident name.txt) loc))


(*
  Ghost expressions and patterns:
  expressions and patterns that do not appear explicitly in the
  source file they have the loc_ghost flag set to true.
  Then the profiler will not try to instrument them and the
  -annot option will not try to display their type.

  Every grammar rule that generates an element with a location must
  make at most one non-ghost element, the topmost one.

  How to tell whether your location must be ghost:
  A location corresponds to a range of characters in the source file.
  If the location contains a piece of code that is syntactically
  valid (according to the documentation), and corresponds to the
  AST node, then the location must be real; in all other cases,
  it must be ghost.

  jordwalke: Noticed that ghost expressions are often used when inserting
   additional AST nodes from a parse rule. Either an extra wrapping one, or an
   additional inner node. This is consistent with the above description, I
   believe.
*)


let ghunit ?(loc=dummy_loc ()) () =
  mkexp ~ghost:true ~loc (Pexp_construct (mknoloc (Lident "()"), None))

let mkinfix arg1 name_operator arg2 =
  mkexp (Pexp_apply(mkoperator name_operator, ["", arg1; "", arg2]))

let neg_float_string f =
  if String.length f > 0 && f.[0] = '-'
  then String.sub f 1 (String.length f - 1)
  else "-" ^ f

let mkuminus name arg =
  match name.txt, arg.pexp_desc with
  | "-", Pexp_constant(Const_int n) ->
      mkexp(Pexp_constant(Const_int(-n)))
  | "-", Pexp_constant(Const_int32 n) ->
      mkexp(Pexp_constant(Const_int32(Int32.neg n)))
  | "-", Pexp_constant(Const_int64 n) ->
      mkexp(Pexp_constant(Const_int64(Int64.neg n)))
  | "-", Pexp_constant(Const_nativeint n) ->
      mkexp(Pexp_constant(Const_nativeint(Nativeint.neg n)))
  | ("-" | "-."), Pexp_constant(Const_float f) ->
      mkexp(Pexp_constant(Const_float(neg_float_string f)))
  | _ ->
      let name = {name with txt=("~" ^ name.txt)} in
      mkexp(Pexp_apply(mkoperator name, ["", arg]))

let mkFunctorThatReturns functorArgs returns =
  List.fold_left (
    fun acc (n, t) -> mkmod (Pmod_functor(n, t, acc))
  ) returns functorArgs

let mkuplus name arg =
  let desc = arg.pexp_desc in
  match name.txt, desc with
  | "+", Pexp_constant(Const_int _)
  | "+", Pexp_constant(Const_int32 _)
  | "+", Pexp_constant(Const_int64 _)
  | "+", Pexp_constant(Const_nativeint _)
  | ("+" | "+."), Pexp_constant(Const_float _) -> mkexp desc
  | _ ->
      let name = {name with txt=("~" ^ name.txt)} in
      mkexp(Pexp_apply(mkoperator name, ["", arg]))

let mkexp_cons consloc args loc =
  mkexp ~loc (Pexp_construct(mkloc (Lident "::") consloc, Some args))

let ghexp_cons consloc args loc =
  mkexp ~ghost:true ~loc (Pexp_construct(mkloc (Lident "::") loc, Some args))

let mkpat_cons consloc args loc =
  mkpat ~loc (Ppat_construct(mkloc (Lident "::") loc, Some args))

let ghpat_cons consloc args loc =
  mkpat ~ghost:true ~loc (Ppat_construct(mkloc (Lident "::") loc, Some args))

let simple_pattern_list_to_tuple ?(loc=dummy_loc ()) lst =
  match lst with
    | [] -> assert false
    | _ -> mkpat ~loc (Ppat_tuple (List.rev lst))

let mktailexp_extension loc seq ext_opt =
  let rec handle_seq = function
    [] ->
      let base_case = match ext_opt with
        | Some ext ->
          ext
        | None ->
          let loc = make_ghost_loc loc in
          let nil = { txt = Lident "[]"; loc } in
          Exp.mk ~loc (Pexp_construct (nil, None)) in
      base_case
  | e1 :: el ->
      let exp_el = handle_seq el in
      let loc = mklocation e1.pexp_loc.loc_start exp_el.pexp_loc.loc_end in
      let arg = mkexp ~ghost:true ~loc (Pexp_tuple [e1; exp_el]) in
      ghexp_cons loc arg loc in
  handle_seq seq

let mktailpat_extension loc seq ext_opt =
  let rec handle_seq = function
    [] ->
      let base_case = match ext_opt with
        | Some ext ->
          ext
        | None ->
          let loc = make_ghost_loc loc in
          let nil = { txt = Lident "[]"; loc } in
          mkpat ~loc (Ppat_construct (nil, None)) in
      base_case
  | p1 :: pl ->
      let pat_pl = handle_seq pl in
      let loc = mklocation p1.ppat_loc.loc_start pat_pl.ppat_loc.loc_end in
      let arg = mkpat ~ghost:true ~loc (Ppat_tuple [p1; pat_pl]) in
      ghpat_cons loc arg loc in
  handle_seq seq

(* Applies attributes to the structure item, not the expression itself. Makes
 * structure item have same location as expression. *)
let mkstrexp e attrs =
  { pstr_desc = Pstr_eval (e, attrs); pstr_loc = e.pexp_loc }

let ghexp_constraint loc e (t1, t2) =
  match t1, t2 with
  | Some t, None -> mkexp ~ghost:true ~loc (Pexp_constraint(e, t))
  | _, Some t -> mkexp ~ghost:true ~loc (Pexp_coerce(e, t1, t))
  | None, None -> assert false

let array_function ?(loc=dummy_loc()) str name =
  ghloc ~loc (Ldot(Lident str, (if !Clflags.fast then "unsafe_" ^ name else name)))

let syntax_error_str loc msg =
  if !Reason_config.recoverable then
    Str.mk ~loc:loc (Pstr_extension (Syntax_util.syntax_error_extension_node loc msg, []))
  else
    raise(Syntaxerr.Error(Syntaxerr.Other loc))

let syntax_error () =
  raise Syntaxerr.Escape_error

let syntax_error_exp loc msg =
  if !Reason_config.recoverable then
    Exp.mk ~loc (Pexp_extension (Syntax_util.syntax_error_extension_node loc msg))
  else
    syntax_error ()

let unclosed opening closing =
  raise(Syntaxerr.Error(Syntaxerr.Unclosed(opening.loc, opening.txt,
                                           closing.loc, closing.txt)))

let unclosed_extension closing =
  Syntax_util.syntax_error_extension_node closing.loc ("Expecting \"" ^ closing.txt ^ "\"")

let unclosed_mod opening closing =
  if !Reason_config.recoverable then
    mkmod(Pmod_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_cl opening closing =
  if !Reason_config.recoverable then
    mkclass(Pcl_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_mty opening closing =
  if !Reason_config.recoverable then
    mkmty(Pmty_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_cty opening closing =
  if !Reason_config.recoverable then
    mkcty(Pcty_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_exp opening closing =
  if !Reason_config.recoverable then
    mkexp(Pexp_extension (unclosed_extension closing))
  else
    unclosed opening closing

let unclosed_pat opening closing =
  if !Reason_config.recoverable then
    mkpat(Ppat_extension (unclosed_extension closing))
  else
    unclosed opening closing

let expecting nonterm =
    raise Syntaxerr.(Error(Expecting(nonterm.loc, nonterm.txt)))

let expecting_pat nonterm =
  if !Reason_config.recoverable then
    mkpat(Ppat_extension (Syntax_util.syntax_error_extension_node nonterm.loc ("Expecting " ^ nonterm.txt)))
  else
    expecting nonterm

let not_expecting start_pos end_pos nonterm =
    raise Syntaxerr.(Error(Not_expecting(mklocation start_pos end_pos, nonterm)))

let bigarray_function ?(loc=dummy_loc()) str name =
  ghloc ~loc (Ldot(Ldot(Lident "Bigarray", str), name))

let bigarray_untuplify = function
    { pexp_desc = Pexp_tuple explist; pexp_loc = _ } -> explist
  | exp -> [exp]

let bigarray_get ?(loc=dummy_loc()) arr arg =
  let get = if !Clflags.fast then "unsafe_get" else "get" in
  match bigarray_untuplify arg with
    [c1] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array1" get)),
                       ["", arr; "", c1]))
  | [c1;c2] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array2" get)),
                       ["", arr; "", c1; "", c2]))
  | [c1;c2;c3] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array3" get)),
                       ["", arr; "", c1; "", c2; "", c3]))
  | coords ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Genarray" "get")),
                       ["", arr; "", mkexp ~ghost:true ~loc (Pexp_array coords)]))

let bigarray_set ?(loc=dummy_loc()) arr arg newval =
  let set = if !Clflags.fast then "unsafe_set" else "set" in
  match bigarray_untuplify arg with
    [c1] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array1" set)),
                       ["", arr; "", c1; "", newval]))
  | [c1;c2] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array2" set)),
                       ["", arr; "", c1; "", c2; "", newval]))
  | [c1;c2;c3] ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Array3" set)),
                       ["", arr; "", c1; "", c2; "", c3; "", newval]))
  | coords ->
      mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(bigarray_function ~loc "Genarray" "set")),
                       ["", arr;
                        "", mkexp ~ghost:true ~loc (Pexp_array coords);
                        "", newval]))

let lapply p1 p2 start_pos end_pos =
  if !Clflags.applicative_functors
  then Lapply(p1, p2)
  else raise (Syntaxerr.Error(Syntaxerr.Applicative_path (mklocation start_pos end_pos)))

let exp_of_label label =
  mkexp ~loc:label.loc (Pexp_ident {label with txt=Lident(Longident.last label.txt)})

let pat_of_label label =
  mkpat ~loc:label.loc (Ppat_var {label with txt=(Longident.last label.txt)})

let check_variable vl loc v =
  if List.mem v vl then
    raise Syntaxerr.(Error(Variable_in_scope(loc,v)))

let varify_constructors var_names t =
  let rec loop t =
    let desc =
      match t.ptyp_desc with
      | Ptyp_any -> Ptyp_any
      | Ptyp_var x ->
          check_variable var_names t.ptyp_loc x;
          Ptyp_var x
      | Ptyp_arrow (label,core_type,core_type') ->
          Ptyp_arrow(label, loop core_type, loop core_type')
      | Ptyp_tuple lst -> Ptyp_tuple (List.map loop lst)
      | Ptyp_constr( { txt = Lident s }, []) when List.mem s var_names ->
          Ptyp_var s
      | Ptyp_constr(longident, lst) ->
          Ptyp_constr(longident, List.map loop lst)
      | Ptyp_object (lst, o) ->
          Ptyp_object
            (List.map (fun (s, attrs, t) -> (s, attrs, loop t)) lst, o)
      | Ptyp_class (longident, lst) ->
          Ptyp_class (longident, List.map loop lst)
      | Ptyp_alias(core_type, string) ->
          check_variable var_names t.ptyp_loc string;
          Ptyp_alias(loop core_type, string)
      | Ptyp_variant(row_field_list, flag, lbl_lst_option) ->
          Ptyp_variant(List.map loop_row_field row_field_list,
                       flag, lbl_lst_option)
      | Ptyp_poly(string_lst, core_type) ->
          List.iter (check_variable var_names t.ptyp_loc) string_lst;
          Ptyp_poly(string_lst, loop core_type)
      | Ptyp_package(longident,lst) ->
          Ptyp_package(longident,List.map (fun (n,typ) -> (n,loop typ) ) lst)
      | Ptyp_extension (s, arg) ->
          Ptyp_extension (s, arg)
    in
    {t with ptyp_desc = desc}
  and loop_row_field  =
    function
      | Rtag(label,attrs,flag,lst) ->
          Rtag(label,attrs,flag,List.map loop lst)
      | Rinherit t ->
          Rinherit (loop t)
  in
  loop t

(**
  I believe that wrap_type_annotation will automatically generate the type
  arguments (type a) (type b) based on what was listed before the dot in a
  polymorphic type annotation that uses locally abstract types.
 *)
let wrap_type_annotation newtypes core_type body =
  let exp = mkexp(Pexp_constraint(body,core_type)) in
  let exp =
    List.fold_right (fun newtype exp -> mkexp (Pexp_newtype (newtype, exp)))
      newtypes exp
  in
  let typ = mktyp ~ghost:true (Ptyp_poly(newtypes,varify_constructors newtypes core_type)) in
  (exp, typ)


let struct_item_extension (ext_attrs, ext_id) structure_item =
  mkstr ~ghost:true (Pstr_extension ((ext_id, PStr [structure_item]), ext_attrs))

let extension_expression (ext_attrs, ext_id) item_expr =
  mkexp ~ghost:true ~attrs:ext_attrs (Pexp_extension (ext_id, PStr [mkstrexp item_expr []]))

(* There's no more need for these functions - this was for the following:
 *
 *     fun % ext [@foo] arg => arg;
 *
 *   Becoming
 *
 *     [%ext  (fun arg => arg) [@foo]]
 *
 *   Which we no longer support.
 *)
(* Applies the attributes to the body, then wraps entire thing in an extension
 * expression, whose payload consists of a single structure item that is body
 *)
(* let wrap_exp_attrs body (ext, attrs) = *)
(*   (* todo: keep exact location for the entire attribute *) *)
(*   let body = {body with pexp_attributes = attrs @ body.pexp_attributes} in *)
(*   match ext with *)
(*   | None -> body *)
(*   | Some id -> mkexp ~ghost:true (Pexp_extension (id, PStr [mkstrexp body []])) *)

(* Why not just mkexp with the right attributes in the first place? *)
(* let mkexp_attrs d attrs = *)
(*   wrap_exp_attrs (mkexp d) attrs *)

let mkcf_attrs ?(loc=dummy_loc()) d attrs =
  Cf.mk ~loc ~attrs d

let mkctf_attrs d attrs =
  Ctf.mk ~attrs d

let add_nonrec rf attrs =
  match rf.txt with
  | Recursive -> attrs
  | Nonrecursive ->
      let name = { txt = "nonrec"; loc = rf.loc } in
        (name, PStr []) :: attrs

type let_binding =
  { lb_pattern: pattern;
    lb_expression: expression;
    (* The meaning of lb_leading_attributes and lbs_extension are dependent on
     * the context of the let binding (module/expression etc) *)
    lb_attributes: attributes;
    (* lb_docs: docs Lazy.t; *)
    (* lb_text: text Lazy.t; *)
    lb_loc: Location.t; }

type let_bindings =
  { lbs_bindings: let_binding list;
    lbs_rec: rec_flag;
    lbs_extension: string Asttypes.loc option;
    (* In Reason, we use this field to represent
       extension attributes attached to the extension on a series of "let/and"
       bindings As in: let [@extAttrs ] [%id] [@attribute] x = ...; It only
       makes sense to have [lbs_attributes] when there is an [lbs_extension].
     *)
    lbs_attributes: attributes;
    lbs_loc: Location.t }

let mklb (p, e) attrs loc =
  { lb_pattern = p;
    lb_expression = e;
    (* Only some individual let bindings are allowed to have attributes
     * depending on the context *)
    lb_attributes = attrs;
    lb_loc = loc; }

let mklbs (extAttrs, extId) rf lb loc =
  { lbs_bindings = [lb];
    lbs_rec = rf;
    lbs_extension = extId ;
    lbs_attributes = extAttrs;
    lbs_loc = loc; }

let addlb lbs lb =
  { lbs with lbs_bindings = lb :: lbs.lbs_bindings }

let val_of_let_bindings lbs =
  let bindings =
    List.map
      (fun lb ->
         Vb.mk ~loc:lb.lb_loc ~attrs:lb.lb_attributes
           (* ~docs:(Lazy.force lb.lb_docs) *)
           (* ~text:(Lazy.force lb.lb_text) *)
           lb.lb_pattern lb.lb_expression)
      lbs.lbs_bindings
  in
  let str = mkstr(Pstr_value(lbs.lbs_rec, List.rev bindings)) in
  (* Note that for value bindings, when there's an extension, the
   * lbs_attributes are attributes on the extension *)
  match (lbs.lbs_extension) with
    | None -> str
    | Some ext_id -> struct_item_extension (lbs.lbs_attributes, ext_id) str

let expr_of_let_bindings lbs body =
  let bindings =
    List.map
      (fun lb ->
         (* Individual let bindings in an *expression* can't have item attributes. *)
         if lb.lb_attributes <> [] then
           raise Syntaxerr.(Error(Not_expecting(lb.lb_loc, "item attribute")));
         Vb.mk ~loc:lb.lb_loc lb.lb_pattern lb.lb_expression)
      lbs.lbs_bindings
  in
  (* The location of this expression unfortunately includes the entire rule,
   * which will include any preceeding extensions. *)
  let item_expr = mkexp (Pexp_let(lbs.lbs_rec, List.rev bindings, body)) in
  (* Note that for let expression bindings, when there's an extension, the
   * lbs_attributes are attributes on the entire [let ..in x] expression. *)
  match lbs.lbs_extension with
    | None -> item_expr
    | Some ext_id -> extension_expression (lbs.lbs_attributes, ext_id) item_expr

let class_of_let_bindings lbs body =
  let bindings =
    List.map
      (fun lb ->
         if lb.lb_attributes <> [] then
           raise Syntaxerr.(Error(Not_expecting(lb.lb_loc, "item attribute")));
         Vb.mk ~loc:lb.lb_loc lb.lb_pattern lb.lb_expression)
      lbs.lbs_bindings
  in
    if lbs.lbs_extension <> None then
      raise Syntaxerr.(Error(Not_expecting(lbs.lbs_loc, "extension")));
    if lbs.lbs_attributes <> [] then
      raise Syntaxerr.(Error(Not_expecting(lbs.lbs_loc, "attributes")));
    mkclass(Pcl_let (lbs.lbs_rec, List.rev bindings, body))

(*
 * arity_conflict_resolving_mapper is triggered when both "implicit_arity" "explicit_arity"
 * are in the attribtues. In that case we have to remove "explicit_arity"
 *
 * However, if we simply remove explicit_arity, we would end up with a
 * wrapping tuple which has only one component (inner tuple).
 * This is against the invariance where tuples must have 2+ components.
 * Therefore, in the case we have to remove explicit_arity, we also need to
 * unwrap the tuple to expose the inner tuple directly.
 *
 *)
let arity_conflict_resolving_mapper =
{ default_mapper with
  expr = begin fun mapper expr ->
    match expr with
      | {pexp_desc=Pexp_construct(lid, args);
         pexp_loc;
         pexp_attributes} when attributes_conflicted "implicit_arity" "explicit_arity" pexp_attributes ->
         let new_args =
           match args with
             | Some {pexp_desc = Pexp_tuple [sp]} -> Some sp
             | _ -> args in
         default_mapper.expr mapper
         {pexp_desc=Pexp_construct(lid, new_args); pexp_loc; pexp_attributes=
          normalized_attributes "explicit_arity" pexp_attributes}
      | x -> default_mapper.expr mapper x
  end;
  pat = begin fun mapper pattern ->
    match pattern with
      | {ppat_desc=Ppat_construct(lid, args);
         ppat_loc;
         ppat_attributes} when attributes_conflicted "implicit_arity" "explicit_arity" ppat_attributes ->
         let new_args =
           match args with
             | Some {ppat_desc = Ppat_tuple [sp]} -> Some sp
             | _ -> args in
         default_mapper.pat mapper
         {ppat_desc=Ppat_construct(lid, new_args); ppat_loc; ppat_attributes=
          normalized_attributes "explicit_arity" ppat_attributes}
      | x -> default_mapper.pat mapper x
  end;
}

let default_mapper_chain = [arity_conflict_resolving_mapper; reason_to_ml_swap_operator_mapper; unescape_stars_slashes_mapper]

let rec string_of_longident = function
    | Lident s -> s
    | Ldot(longPrefix, s) ->
        s
    | Lapply (y,s) -> string_of_longident s

let built_in_explicit_arity_constructors = ["Some"; "Assert_failure"; "Match_failure"]


# 2282 "reason_parser.ml"

let rec _menhir_error1727 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce70 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_leading_bar_match_cases_to_sequence_body)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2528 "reason_parser.mly"
      ( mkexp (Pexp_try(_2, List.rev _4)) )
# 2300 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce69 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_leading_bar_match_cases_to_sequence_body)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2526 "reason_parser.mly"
      ( mkexp (Pexp_match(_2, List.rev _4)) )
# 2315 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce672 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__poly_type_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__poly_type_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_poly_type = 
# 3741 "reason_parser.mly"
                                         (_1)
# 2325 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv8005) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_poly_type) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7987 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 2339 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1332
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1332
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1332) : 'freshtv7988)
    | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7989 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1581
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1581
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1581) : 'freshtv7990)
    | MenhirState1695 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7993 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7991 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_private_flag), _startpos__3_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_poly_type)) = _menhir_stack in
        let _5 = () in
        let _2 = () in
        let _endpos = _endpos__6_ in
        let _v : 'tv_method_ = let _4 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2380 "reason_parser.ml"
          
        in
        
# 2038 "reason_parser.mly"
      ( if _1 = Override then syntax_error ();
        _4, _3, Cfk_virtual _6 )
# 2387 "reason_parser.ml"
         in
        _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7992)) : 'freshtv7994)
    | MenhirState1700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7997 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7995 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_poly_type)) = _menhir_stack in
        let _5 = () in
        let _3 = () in
        let _2 = () in
        let _endpos = _endpos__6_ in
        let _v : 'tv_method_ = let _4 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2409 "reason_parser.ml"
          
        in
        
# 2035 "reason_parser.mly"
      ( if _1 = Override then syntax_error ();
        _4, Private, Cfk_virtual _6 )
# 2416 "reason_parser.ml"
         in
        _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7996)) : 'freshtv7998)
    | MenhirState1720 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv8001 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1727 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7999 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1728 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1728) : 'freshtv8000)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1727 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv8002)
    | MenhirState153 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv8003 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2459
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2459
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2459) : 'freshtv8004)
    | _ ->
        _menhir_fail ()) : 'freshtv8006)

and _menhir_reduce209 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_semi_terminated_seq_expr), _startpos_x0_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2541 "reason_parser.ml"
      
    in
    
# 2664 "reason_parser.mly"
      ( syntax_error_exp _2.loc "SyntaxError in block" )
# 2547 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_leading_bar_match_cases_to_sequence_body : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_leading_bar_match_cases_to_sequence_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7973 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7971 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2149 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce69 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7972)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2149) : 'freshtv7974)
    | MenhirState2346 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7977 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2347 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7975 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2347 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce70 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7976)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2347) : 'freshtv7978)
    | MenhirState2433 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2434 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7979 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2434 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce69 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7980)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2434) : 'freshtv7982)
    | MenhirState2551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7985 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2552 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2552 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce70 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7984)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2552) : 'freshtv7986)
    | _ ->
        _menhir_fail ()

and _menhir_goto_simple_pattern_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7967 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState564
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState564 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | BARRBRACKET | COLON | COLONCOLON | COMMA | LBRACKETAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce166 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState564)) : 'freshtv7968)
    | MenhirState649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7969 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState650
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState650 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | COLON | COLONCOLON | COMMA | EOF | EQUAL | EQUALGREATER | IN | LBRACKETAT | RBRACKET | RPAREN | SEMI | WHEN ->
              _menhir_reduce166 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState650)) : 'freshtv7970)
    | _ ->
        _menhir_fail ()

and _menhir_reduce559 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__poly_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__poly_type) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__poly_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1173 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 2761 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7965) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__poly_type_) = _v in
    ((match _menhir_s with
    | MenhirState153 | MenhirState1720 | MenhirState1328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7961) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__poly_type_) = _v in
        (_menhir_reduce672 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7962)
    | MenhirState1700 | MenhirState1695 | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7963) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__poly_type_) = _v in
        (_menhir_reduce672 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7964)
    | _ ->
        _menhir_fail ()) : 'freshtv7966)

and _menhir_goto_optional_type_parameter_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_optional_type_parameter_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2076 | MenhirState1973 | MenhirState1421 | MenhirState1367 | MenhirState1254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7957 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1364
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run1262 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run1258 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run1256 _menhir_env (Obj.magic _menhir_stack) MenhirState1364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run1255 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce641 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7958)
    | MenhirState90 | MenhirState1855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7959 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1864
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run100 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run96 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run92 _menhir_env (Obj.magic _menhir_stack) MenhirState1864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run91 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce641 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7960)
    | _ ->
        _menhir_fail ()

and _menhir_reduce338 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 2832 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters -> Lexing.position -> _menhir_state -> 'tv_class_fun_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_class_fun_binding) ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 2838 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)) = _menhir_stack in
    let _endpos = _endpos__4_ in
    let _v : 'tv_class_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2850 "reason_parser.ml"
      
    in
    
# 1811 "reason_parser.mly"
      (
       (_2, _4, _1, List.rev _3)
      )
# 2858 "reason_parser.ml"
     in
    _menhir_goto_class_declaration_details _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce28 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_binding) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_fun_binding = 
# 1861 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _2)) )
# 2870 "reason_parser.ml"
     in
    _menhir_goto__class_fun_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2328 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7955 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce209 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7956)

and _menhir_reduce208 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_semi_terminated_seq_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2662 "reason_parser.mly"
      ( _2 )
# 2893 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2241 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7953 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce209 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7954)

and _menhir_goto_leading_bar_match_case_to_sequence_body : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_leading_bar_match_case_to_sequence_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState2552 | MenhirState2434 | MenhirState2347 | MenhirState2149 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7947 * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7945 * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_leading_bar_match_case_to_sequence_body) : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_leading_bar_match_cases_to_sequence_body)) = _menhir_stack in
        let _v : 'tv_leading_bar_match_cases_to_sequence_body = 
# 3077 "reason_parser.mly"
                                                                                     ( _2 :: _1 )
# 2922 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_cases_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7946)) : 'freshtv7948)
    | MenhirState2551 | MenhirState2433 | MenhirState2346 | MenhirState766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7951) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7949) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_leading_bar_match_case_to_sequence_body) : 'tv_leading_bar_match_case_to_sequence_body) = _v in
        ((let _v : 'tv_leading_bar_match_cases_to_sequence_body = 
# 3076 "reason_parser.mly"
                                            ( [_1] )
# 2937 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_cases_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7950)) : 'freshtv7952)
    | _ ->
        _menhir_fail ()

and _menhir_reduce182 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 2946 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 2952 "reason_parser.ml"
    )), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_module_binding_body)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)), _endpos__6_), _endpos__7_, _, (_7 : 'tv_semi_terminated_seq_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__semi_terminated_seq_expr_row = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 2968 "reason_parser.ml"
      
    in
    
# 2399 "reason_parser.mly"
                                                                                                     (
      let item_attrs = _5 in
      mkexp ~attrs:item_attrs (Pexp_letmodule(_3, _4, _7))
    )
# 2977 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce184 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)), _endpos__3_, _), _endpos__4_, _, (_4 : 'tv_semi_terminated_seq_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__semi_terminated_seq_expr_row = 
# 2407 "reason_parser.mly"
                                                             (
      let item_attrs = _2 in
      mkexp ~attrs:item_attrs (Pexp_sequence(_1, _4))
    )
# 2993 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce179 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_semi_terminated_seq_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__semi_terminated_seq_expr = 
# 2378 "reason_parser.mly"
                                               (
      expr_of_let_bindings _1 _3
    )
# 3008 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce183 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_override_flag), _startpos__3_), _endpos_x0_, _, (x0 : 'tv_mod_longident), _startpos_x0_), _endpos__5_, _, (_5 : 'tv_post_item_attributes)), _endpos__6_), _endpos__7_, _, (_7 : 'tv_semi_terminated_seq_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__semi_terminated_seq_expr_row = let _4 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 3029 "reason_parser.ml"
      
    in
    
# 2403 "reason_parser.mly"
                                                                                                    (
      let item_attrs = _5 in
      mkexp ~attrs:item_attrs (Pexp_open(_3, _4, _7))
    )
# 3038 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce177 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_item_extension_sugar * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_semi_terminated_seq_expr_row) _startpos__2_ ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_item_extension_sugar), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__semi_terminated_seq_expr = 
# 2369 "reason_parser.mly"
                                                      (
      extension_expression _1 _2
    )
# 3052 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce178 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_semi_terminated_seq_expr_row) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__semi_terminated_seq_expr = 
# 2372 "reason_parser.mly"
                                 (
      _1
    )
# 3065 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce147 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_row_field)), _startpos__3_), _, (_4 : 'tv_row_field_list)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3961 "reason_parser.mly"
      ( mktyp(Ptyp_variant(_2 :: List.rev _4, Closed, None)) )
# 3080 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce148 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_opt_bar)), _, (_3 : 'tv_row_field_list)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3963 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Open, None)) )
# 3094 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce150 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_opt_bar)), _, (_3 : 'tv_row_field_list)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3967 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Closed, Some [])) )
# 3108 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce146 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, _startpos__2_), _, (_3 : 'tv_row_field_list)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3959 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Closed, None)) )
# 3123 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run170 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_row_field_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _startpos ->
    let _menhir_stack = (_menhir_stack, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState170 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState170

and _menhir_error727 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 3169 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce172 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_pattern), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3304 "reason_parser.mly"
      ( mkpat(Ppat_lazy _2) )
# 3184 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce728 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_pattern_list = 
# 3412 "reason_parser.mly"
    ( [_1] )
# 3195 "reason_parser.ml"
     in
    _menhir_goto_simple_pattern_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce727 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_list) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern_list)), _endpos__2_, _, (_2 : 'tv_simple_pattern), _startpos__2_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_simple_pattern_list = 
# 3410 "reason_parser.mly"
    ( _2::_1 )
# 3206 "reason_parser.ml"
     in
    _menhir_goto_simple_pattern_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce167 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_pattern), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3288 "reason_parser.mly"
    (
      mkpat (Ppat_variant(_1, Some _2))
    )
# 3220 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce163 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__pattern_without_or = 
# 3263 "reason_parser.mly"
      ( _1 )
# 3232 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_non_arrowed_simple_core_type_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7921 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState255
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND | AND | AS | BAR | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | GREATER | LBRACKETAT | LBRACKETATAT | QUESTION | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce134 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState255)) : 'freshtv7922)
    | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7923 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState317
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState317 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | BAR | BARRBRACKET | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce135 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState317)) : 'freshtv7924)
    | MenhirState328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7925 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState329
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState329 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | BAR | BARRBRACKET | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce134 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState329)) : 'freshtv7926)
    | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1345 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1289 | MenhirState1284 | MenhirState1269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7929 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1277
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7927 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1277 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1278) : 'freshtv7928)
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce450 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1277)) : 'freshtv7930)
    | MenhirState2521 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1490 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7933 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1500
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7931 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1500 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1501 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1501) : 'freshtv7932)
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce450 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1500)) : 'freshtv7934)
    | MenhirState1592 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7935 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1593
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce33 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1593)) : 'freshtv7936)
    | MenhirState1642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7937 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1643
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1643 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce16 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1643)) : 'freshtv7938)
    | MenhirState1799 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7939 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1800
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce33 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1800)) : 'freshtv7940)
    | MenhirState121 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7941 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2472
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2472 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND | AND | AS | BAR | COLONGREATER | COMMA | CONSTRAINT | EOF | EQUAL | EQUALGREATER | GREATER | LBRACKETAT | LBRACKETATAT | QUESTION | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce135 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2472)) : 'freshtv7942)
    | MenhirState2682 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7943 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2683
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2683 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce16 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2683)) : 'freshtv7944)
    | _ ->
        _menhir_fail ()

and _menhir_goto_non_arrowed_core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState666 | MenhirState272 | MenhirState270 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7857 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState263
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState263
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce57 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7858)
    | MenhirState324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7867 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 3744 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState325
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7859 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 3756 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState325 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2406 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2406) : 'freshtv7860)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState325
          | QUESTION ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7865 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 3801 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState325 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7861 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 3815 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKET ->
                      _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETGREATER ->
                      _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETLESS ->
                      _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESS ->
                      _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSDOTDOTGREATER ->
                      _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSGREATER ->
                      _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SHARP ->
                      _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState327 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState327) : 'freshtv7862)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7863 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 3859 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7864)) : 'freshtv7866)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState325)) : 'freshtv7868)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7869 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState332
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState332
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce57 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7870)
    | MenhirState1022 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7873 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1023
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7871 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1023 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1024 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1024) : 'freshtv7872)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1023
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1023)) : 'freshtv7874)
    | MenhirState1195 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7877 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1196
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7875 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1196 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1197 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1197) : 'freshtv7876)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1196
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1196)) : 'freshtv7878)
    | MenhirState1710 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7881 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1711
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7879 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1711 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1712) : 'freshtv7880)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1711
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1711)) : 'freshtv7882)
    | MenhirState1770 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7885 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4163 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1771
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7883 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4175 "reason_parser.ml"
              ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1771 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1769 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1767 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1772) : 'freshtv7884)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1771
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1771)) : 'freshtv7886)
    | MenhirState2669 | MenhirState1766 | MenhirState1781 | MenhirState1774 | MenhirState1772 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7889 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1773
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7887 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1773 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1769 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1767 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1774 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1774) : 'freshtv7888)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1773
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1773)) : 'freshtv7890)
    | MenhirState1779 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7893 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4281 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1780
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7891 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4293 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1780 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1769 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1767 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1781 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1781) : 'freshtv7892)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1780
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1780)) : 'freshtv7894)
    | MenhirState1805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7897 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4344 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1806
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7895 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4356 "reason_parser.ml"
              ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1806 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1804 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1807 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1807) : 'freshtv7896)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1806
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1806)) : 'freshtv7898)
    | MenhirState2698 | MenhirState2063 | MenhirState1827 | MenhirState1787 | MenhirState1816 | MenhirState1809 | MenhirState1807 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7901 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1808
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7899 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1808 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1804 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1809 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1809) : 'freshtv7900)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1808
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1808)) : 'freshtv7902)
    | MenhirState1814 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7905 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4462 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1815
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7903 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4474 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1815 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1804 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1816) : 'freshtv7904)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1815
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1815)) : 'freshtv7906)
    | MenhirState2220 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7909 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2221
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7907 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2221 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2222 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2222) : 'freshtv7908)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2221
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2221)) : 'freshtv7910)
    | MenhirState143 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7919 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4618 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2461
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7911 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4630 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2461 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2465 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2465) : 'freshtv7912)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2461
          | QUESTION ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7917 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4675 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2461 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7913 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4689 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKET ->
                      _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETGREATER ->
                      _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETLESS ->
                      _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESS ->
                      _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSDOTDOTGREATER ->
                      _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSGREATER ->
                      _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SHARP ->
                      _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2463 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2463) : 'freshtv7914)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv7915 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 4733 "reason_parser.ml"
                  ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7916)) : 'freshtv7918)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2461)) : 'freshtv7920)
    | _ ->
        _menhir_fail ()

and _menhir_reduce285 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_pattern_two_or_more_comma_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3372 "reason_parser.mly"
      ( mkpat(Ppat_tuple(List.rev _2)) )
# 4754 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run587 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_pattern_two_or_more_comma_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos ->
    let _menhir_stack = (_menhir_stack, _endpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState587

and _menhir_reduce657 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_or_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_or_pattern) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern = 
# 3258 "reason_parser.mly"
               ( _1 )
# 4822 "reason_parser.ml"
     in
    _menhir_goto_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2469 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_package_type_cstr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce382 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constrain -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_constrain) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constraints)), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constraints = 
# 3487 "reason_parser.mly"
                                                ( _3 :: _1 )
# 4839 "reason_parser.ml"
     in
    _menhir_goto_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto__poly_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__poly_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState153 | MenhirState1720 | MenhirState1328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7853) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__poly_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce559 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7854)
    | MenhirState1700 | MenhirState1695 | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7855) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__poly_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce559 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7856)
    | _ ->
        _menhir_fail ()

and _menhir_reduce137 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_core_type_comma_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_simple_core_type = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3932 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        match _2 with
        | [] -> raise (Syntaxerr.Error(Syntaxerr.Applicative_path loc))
        | one::[] -> one
        | moreThanOne -> mktyp(Ptyp_tuple(List.rev moreThanOne)) )
# 4882 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run270 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_core_type_comma_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos ->
    let _menhir_stack = (_menhir_stack, _endpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState270

and _menhir_error1908 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1687 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1223 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce207 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_type_constraint)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2657 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        ghexp_constraint loc _2 _3
      )
# 4953 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce420 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_type_constraint)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _v : 'tv_expr_optional_constraint = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3161 "reason_parser.mly"
                         (
      let loc = mklocation _symbolstartpos _endpos in
      ghexp_constraint loc _1 _2
    )
# 4969 "reason_parser.ml"
     in
    _menhir_goto_expr_optional_constraint _menhir_env _menhir_stack _menhir_s _v _startpos

and _menhir_reduce639 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list -> Lexing.position -> _menhir_state -> 'tv_optional_type_variable_with_variance -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_optional_type_variable_with_variance) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_optional_type_parameter_list)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_optional_type_parameter_list = 
# 3521 "reason_parser.mly"
                                                                         ( _2 :: _1 )
# 4980 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameter_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce638 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_optional_type_variable_with_variance -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_optional_type_variable_with_variance) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_optional_type_parameter_list = 
# 3520 "reason_parser.mly"
                                                                         ( [_1] )
# 4990 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameter_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce51 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_class_constructor_type)) = _menhir_stack in
    let x1 = () in
    let _3 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 5011 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 5023 "reason_parser.ml"
      
    in
    
# 1970 "reason_parser.mly"
      ( unclosed_cl (with_txt _1 "(") (with_txt _5 ")") )
# 5029 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce345 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_fun_binding_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_fun_binding_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_fun_binding = 
# 1858 "reason_parser.mly"
                                                        (_1)
# 5039 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7851) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_fun_binding) = _v in
    ((match _menhir_s with
    | MenhirState1823 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7843 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce28 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7844)
    | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7845 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 5059 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce338 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7846)
    | MenhirState2704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7847 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce28 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7848)
    | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7849 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 5077 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_binding) = _v in
        (_menhir_reduce338 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7850)
    | _ ->
        _menhir_fail ()) : 'freshtv7852)

and _menhir_reduce698 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__semi_terminated_seq_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__semi_terminated_seq_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_semi_terminated_seq_expr = 
# 2367 "reason_parser.mly"
                                                                        (_1)
# 5093 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7841) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_semi_terminated_seq_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7809 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce183 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7810)
    | MenhirState1167 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7811 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce179 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7812)
    | MenhirState1176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7813 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce184 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7814)
    | MenhirState2134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7815 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 5120 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce182 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7816)
    | MenhirState1155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7819 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7817 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_semi_terminated_seq_expr), _startpos__5_) = _menhir_stack in
        let _4 = () in
        let _2 = () in
        let _v : 'tv_leading_bar_match_case_to_sequence_body = 
# 3101 "reason_parser.mly"
      ( Exp.case _1 ~guard:_3 _5 )
# 5134 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_case_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7818)) : 'freshtv7820)
    | MenhirState2147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7823 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7821 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _startpos__2_), _endpos__3_, _, (_3 : 'tv_semi_terminated_seq_expr), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_leading_bar_match_case_to_sequence_body = 
# 3099 "reason_parser.mly"
      ( Exp.case _1 _3 )
# 5147 "reason_parser.ml"
         in
        _menhir_goto_leading_bar_match_case_to_sequence_body _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7822)) : 'freshtv7824)
    | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2241 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce208 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7826)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2241 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7828)
    | MenhirState2253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7829 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce179 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7830)
    | MenhirState2259 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7831 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce184 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7832)
    | MenhirState2320 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7833 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce183 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7834)
    | MenhirState2326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7835 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 5185 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce182 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7836)
    | MenhirState432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7839 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2328 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7837 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_semi_terminated_seq_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce208 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7838)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2328 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7840)
    | _ ->
        _menhir_fail ()) : 'freshtv7842)

and _menhir_reduce699 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__semi_terminated_seq_expr_row_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_semi_terminated_seq_expr_row = 
# 2387 "reason_parser.mly"
                                                                                (_1)
# 5216 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7807) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1176 | MenhirState1167 | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7799) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce178 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7800)
    | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7801 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce177 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7802)
    | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState2259 | MenhirState2253 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7803) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce178 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7804)
    | MenhirState2255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7805 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce177 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7806)
    | _ ->
        _menhir_fail ()) : 'freshtv7808)

and _menhir_reduce255 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_expr = 
# 1246 "reason_parser.mly"
      ( _2 )
# 5270 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_row_field_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_row_field_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState163 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7751 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7747 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce146 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7748)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7749 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7750)) : 'freshtv7752)
    | MenhirState292 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7759 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7753 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState295) : 'freshtv7754)
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7755 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce150 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7756)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7757 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7758)) : 'freshtv7760)
    | MenhirState303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7765 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7761 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce148 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7762)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7763 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7764)) : 'freshtv7766)
    | MenhirState307 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7771 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7767 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce146 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7768)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7769 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7770)) : 'freshtv7772)
    | MenhirState313 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7777 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7773 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce147 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7774)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7775 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7776)) : 'freshtv7778)
    | MenhirState2449 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7783 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7779 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce147 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7780)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv7781 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7782)) : 'freshtv7784)
    | MenhirState161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7789 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7785 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce148 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7786)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7787 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7788)) : 'freshtv7790)
    | MenhirState158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7797 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run170 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7791 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2456) : 'freshtv7792)
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7793 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce150 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv7794)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv7795 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7796)) : 'freshtv7798)
    | _ ->
        _menhir_fail ()

and _menhir_goto_str_extension_constructors : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_extension_constructors -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2538 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2539
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7739 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2539 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COLONCOLON ->
                  _menhir_run1524 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run1518 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run1505 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run1490 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2540) : 'freshtv7740)
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2539
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2539) : 'freshtv7742)
    | MenhirState2588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv7745 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2612
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7743 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2612 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COLONCOLON ->
                  _menhir_run2608 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2613 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run2604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2613 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run2599 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2613 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run2595 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2613 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run2589 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2613 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2613) : 'freshtv7744)
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2612
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2612) : 'freshtv7746)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constructor_declarations_leading_bar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1268 | MenhirState1344 | MenhirState1346 | MenhirState1287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7731 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1316
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run1298 _menhir_env (Obj.magic _menhir_stack) MenhirState1316 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce394 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1316)) : 'freshtv7732)
    | MenhirState1319 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7733 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1320
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run1298 _menhir_env (Obj.magic _menhir_stack) MenhirState1320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce396 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1320)) : 'freshtv7734)
    | MenhirState106 | MenhirState2520 | MenhirState2522 | MenhirState2478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7735 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2507
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run2489 _menhir_env (Obj.magic _menhir_stack) MenhirState2507 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce394 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2507)) : 'freshtv7736)
    | MenhirState2510 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7737 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2511
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run2489 _menhir_env (Obj.magic _menhir_stack) MenhirState2511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce396 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2511)) : 'freshtv7738)
    | _ ->
        _menhir_fail ()

and _menhir_reduce505 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_let_binding) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_let_bindings = 
# 2873 "reason_parser.mly"
                                                ( _1 )
# 5634 "reason_parser.ml"
     in
    _menhir_goto_let_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1886 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1886

and _menhir_run1183 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1183 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1183

and _menhir_reduce256 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_module_expr = 
# 1248 "reason_parser.mly"
      ( mkmod(Pmod_unpack _3) )
# 5763 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce726 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__simple_pattern_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__simple_pattern_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_pattern = 
# 3330 "reason_parser.mly"
                                                   (_1)
# 5774 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7729) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7695 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce163 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7696)
    | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7697 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce167 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7698)
    | MenhirState564 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7699 * Lexing.position * _menhir_state * 'tv_simple_pattern_list) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce727 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7700)
    | MenhirState563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7701 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce728 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7702)
    | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce172 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7704)
    | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState738 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7705 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce163 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7706)
    | MenhirState646 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7707 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce167 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7708)
    | MenhirState650 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7709 * Lexing.position * _menhir_state * 'tv_simple_pattern_list) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce727 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7710)
    | MenhirState649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7711 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce728 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7712)
    | MenhirState532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7713 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce172 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7714)
    | MenhirState486 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7723 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 5829 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error727 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7715 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 5841 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState729) : 'freshtv7716)
          | OPTIONAL_NO_DEFAULT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7719 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 5900 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7717 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 5907 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 5912 "reason_parser.ml"
              )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_pattern), _startpos__3_) = _menhir_stack in
              let _4 = () in
              let _2 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_labeled_simple_pattern = 
# 2487 "reason_parser.mly"
      ( ("?" ^ _1, None, _3) )
# 5920 "reason_parser.ml"
               in
              _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv7718)) : 'freshtv7720)
          | BACKQUOTE | CHAR _ | COLON | EQUALGREATER | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7721 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 5928 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 5933 "reason_parser.ml"
              )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_pattern), _startpos__3_) = _menhir_stack in
              let _2 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_labeled_simple_pattern = 
# 2484 "reason_parser.mly"
      ( (_1, None, _3) )
# 5940 "reason_parser.ml"
               in
              _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv7722)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error727 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7724)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1639 | MenhirState1637 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7727 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7725 * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _v : 'tv_labeled_simple_pattern = 
# 2492 "reason_parser.mly"
      ( ("", None, _1) )
# 5957 "reason_parser.ml"
         in
        _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv7726)) : 'freshtv7728)
    | _ ->
        _menhir_fail ()) : 'freshtv7730)

and _menhir_run1827 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1804 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run1788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1827

and _menhir_goto_direction_flag : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_direction_flag -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState741 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState746 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState746) : 'freshtv7692)
    | MenhirState911 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState912) : 'freshtv7694)
    | _ ->
        _menhir_fail ()

and _menhir_reduce292 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 6116 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 906 "reason_parser.mly"
       (string)
# 6122 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _6 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6139 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6151 "reason_parser.ml"
      
    in
    
# 3390 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _6 ")") )
# 6157 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce596 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_non_arrowed_simple_core_type) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_core_type = 
# 3916 "reason_parser.mly"
    ( _1 )
# 6168 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce600 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_non_arrowed_simple_core_type) _startpos__1_ ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_simple_core_type_list = 
# 4019 "reason_parser.mly"
                                     ( [_1] )
# 6178 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_simple_core_type_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce601 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_non_arrowed_simple_core_type) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_non_arrowed_simple_core_type_list = 
# 4020 "reason_parser.mly"
                                                                     ( _2 :: _1 )
# 6189 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_simple_core_type_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce595 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_non_arrowed_non_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_non_arrowed_non_simple_core_type) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_core_type = 
# 3914 "reason_parser.mly"
    ( _1 )
# 6200 "reason_parser.ml"
     in
    _menhir_goto_non_arrowed_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_pattern_two_or_more_comma_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_pattern_two_or_more_comma_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState495 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7683 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run587 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7679 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce285 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7680)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7681 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7682)) : 'freshtv7684)
    | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7689 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run587 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7685 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce285 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7686)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7687 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_pattern_two_or_more_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7688)) : 'freshtv7690)
    | _ ->
        _menhir_fail ()

and _menhir_reduce643 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__or_pattern_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__or_pattern_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_or_pattern = 
# 3080 "reason_parser.mly"
                                           (_1)
# 6262 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7677) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_or_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7673) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce657 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7674)
    | MenhirState2578 | MenhirState191 | MenhirState2314 | MenhirState738 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState676 | MenhirState587 | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7675) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce657 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7676)
    | _ ->
        _menhir_fail ()) : 'freshtv7678)

and _menhir_error2650 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2572 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run2521 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 6303 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2521 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2521
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2521

and _menhir_run2523 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run2480 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2523 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2523

and _menhir_goto_amper_type_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_amper_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv7671 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERSAND ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7669 * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2441 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2442) : 'freshtv7670)
    | LBRACKETAT ->
        _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2441
    | AMPERAMPER | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2441) : 'freshtv7672)

and _menhir_reduce648 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _v : 'tv_package_type_cstr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6427 "reason_parser.ml"
      
    in
    
# 3980 "reason_parser.mly"
                                                 ( (_2, _4) )
# 6433 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7667) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_package_type_cstr) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2411 | MenhirState239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7661 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2410 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7659 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run240 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2411) : 'freshtv7660)
          | RPAREN ->
              _menhir_reduce649 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2410 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7662)
    | MenhirState2470 | MenhirState134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7665 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2469 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7663 * _menhir_state * 'tv_package_type_cstr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2470) : 'freshtv7664)
          | COLONGREATER | RPAREN ->
              _menhir_reduce649 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2469 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7666)
    | _ ->
        _menhir_fail ()) : 'freshtv7668)

and _menhir_run2044 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 898 "reason_parser.mly"
       (string * string option)
# 6505 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | STRING _v ->
        _menhir_run2044 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2044 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | LBRACKETATAT | SEMI ->
        _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2044

and _menhir_error2042 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1918 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1905 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1899 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1860 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1724 : _menhir_env -> ((((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce838 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_core_type), _startpos__6_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_value = let _4 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6567 "reason_parser.ml"
      
    in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      _startpos__2_ in
    
# 2012 "reason_parser.mly"
                                                                (
      if _1 = Override
      then not_expecting _symbolstartpos _endpos "members marked virtual may not also be marked overridden"
      else _4, Mutable, Cfk_virtual _6
    )
# 6582 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1681 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce838 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce840 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_value = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6605 "reason_parser.ml"
      
    in
    
# 2020 "reason_parser.mly"
      ( _3, _2, Cfk_virtual _5 )
# 6611 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1672 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce840 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error1600 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_value_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_value_type -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1563
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1563
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1563) : 'freshtv7658)

and _menhir_error1546 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1483 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 898 "reason_parser.mly"
       (string * string option)
# 6647 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | STRING _v ->
        _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1483 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce677 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1483

and _menhir_error1481 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce850 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_optional_type_parameters)), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_with_constraint = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6683 "reason_parser.ml"
      
    in
    let _endpos__2_ = _endpos_x0_ in
    let _startpos__2_ = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3711 "reason_parser.mly"
    (
      let last = (
        match _2.txt with
            | Lident s -> s
            | _ -> not_expecting _startpos__2_ _endpos__2_ "Long type identifier"
      ) in
     let loc = mklocation _symbolstartpos _endpos in
      Pwith_typesubst
          (Type.mk {_2 with txt=last}
             ~params:_3
             ~manifest:_5
             ~loc)
    )
# 6705 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce379 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constrain = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2283 "reason_parser.mly"
                              (
    let loc = mklocation _symbolstartpos _endpos in
    (_1, _3, loc)
  )
# 6722 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7655) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_constrain) = _v in
    ((match _menhir_s with
    | MenhirState1359 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7651 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrain) = _v in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7652)
    | MenhirState1859 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7653 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrain) = _v in
        (_menhir_reduce382 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7654)
    | _ ->
        _menhir_fail ()) : 'freshtv7656)

and _menhir_error1360 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1345 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 6755 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1345 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1345
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1345

and _menhir_run1347 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run1289 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1347 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1347

and _menhir_reduce821 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_, _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_type_kind = 
# 3510 "reason_parser.mly"
      ( (Ptype_open, Public, Some _2) )
# 6822 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1343 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce175 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__poly_type = 
# 3744 "reason_parser.mly"
          ( _1 )
# 6839 "reason_parser.ml"
     in
    _menhir_goto__poly_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce176 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_typevar_list), _startpos__1_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__poly_type = 
# 3746 "reason_parser.mly"
          ( mktyp(Ptyp_poly(List.rev _1, _3)) )
# 6852 "reason_parser.ml"
     in
    _menhir_goto__poly_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce814 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_kind = 
# 3496 "reason_parser.mly"
      ( (Ptype_abstract, Private, Some _3) )
# 6865 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce451 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_simple_core_type_list)), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_generalized_constructor_arguments = 
# 3609 "reason_parser.mly"
                                                ( (List.rev _1,Some _3) )
# 6877 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce452 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_generalized_constructor_arguments = 
# 3611 "reason_parser.mly"
                                                ( ([],Some _2) )
# 6889 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1233 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1220 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1212 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce810 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_constraint_right_of_colon = 
# 3246 "reason_parser.mly"
                                          ( (Some _1, Some _3) )
# 6916 "reason_parser.ml"
     in
    _menhir_goto_type_constraint_right_of_colon _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce808 : _menhir_env -> ('ttv_tail * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_constraint = 
# 3251 "reason_parser.mly"
                                                ( (None, Some _2) )
# 6928 "reason_parser.ml"
     in
    _menhir_goto_type_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce287 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _startpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3377 "reason_parser.mly"
      ( mkpat(Ppat_constraint(_2, _4)) )
# 6943 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce288 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _3 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6964 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 6976 "reason_parser.ml"
      
    in
    
# 3379 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _5 ")") )
# 6982 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce162 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__pattern_optional_constraint = 
# 3418 "reason_parser.mly"
                                ( mkpat(Ppat_constraint(_1, _3)) )
# 6995 "reason_parser.ml"
     in
    _menhir_goto__pattern_optional_constraint _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_core_type_comma_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_core_type_comma_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2479 | MenhirState118 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7643 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7639 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce137 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7640)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7641 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7642)) : 'freshtv7644)
    | MenhirState128 | MenhirState1288 | MenhirState1271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7649 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7645 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce137 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7646)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7647 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_core_type_comma_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7648)) : 'freshtv7650)
    | _ ->
        _menhir_fail ()

and _menhir_error1916 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1231 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce270 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern_not_ident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern = 
# 3334 "reason_parser.mly"
                             ( _1 )
# 7068 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_simple_labeled_expr_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_labeled_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7631 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState827
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run821 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState827 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce65 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7632)
    | MenhirState933 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7633 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState940
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run934 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState940 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce65 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7634)
    | MenhirState1647 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7635 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1648
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run821 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1648 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce13 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1648)) : 'freshtv7636)
    | MenhirState2687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7637 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2688
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run934 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce13 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2688)) : 'freshtv7638)
    | _ ->
        _menhir_fail ()

and _menhir_reduce210 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_comma_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2666 "reason_parser.mly"
      ( mkexp(Pexp_tuple(List.rev _2)) )
# 7305 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1147 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1147

and _menhir_reduce214 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ _startpos__5_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_expr_comma_list)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7403 "reason_parser.ml"
      
    in
    
# 2680 "reason_parser.mly"
      ( mkexp(Pexp_open(Fresh, _1, mkexp(Pexp_tuple(List.rev _4)))) )
# 7409 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1076 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1076

and _menhir_goto_type_constraint : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2179 | MenhirState1078 | MenhirState1065 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7603 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        (_menhir_reduce420 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7604)
    | MenhirState1086 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7605 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        (_menhir_reduce420 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7606)
    | MenhirState1149 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7611 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : (('freshtv7607 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7608)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7609 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce207 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7610)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce420 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7612)
    | MenhirState1184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7615 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1223 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7613 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1224) : 'freshtv7614)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1223 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7616)
    | MenhirState1684 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7619 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1687 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7617 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1688) : 'freshtv7618)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1687 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7620)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7623 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1908 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7621 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1909) : 'freshtv7622)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1908 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7624)
    | MenhirState2382 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7629 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : (('freshtv7625 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7626)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7627 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce207 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7628)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce420 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7630)
    | _ ->
        _menhir_fail ()

and _menhir_reduce223 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_record_expr)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7834 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7846 "reason_parser.ml"
      
    in
    
# 2719 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "{") (with_txt _3 "}"))
# 7852 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce227 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _, (_4 : 'tv_record_expr)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7873 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7885 "reason_parser.ml"
      
    in
    
# 2731 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "{") (with_txt _5 "}") )
# 7891 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_type_variance : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_type_variance -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7601 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7597) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1617 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1617 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1617) : 'freshtv7598)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7599 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7600)) : 'freshtv7602)

and _menhir_reduce829 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_type_variance = 
# 3546 "reason_parser.mly"
                                                ( Covariant )
# 7934 "reason_parser.ml"
     in
    _menhir_goto_type_variance _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce830 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_type_variance = 
# 3547 "reason_parser.mly"
                                                ( Contravariant )
# 7945 "reason_parser.ml"
     in
    _menhir_goto_type_variance _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce264 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mty_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_module_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 7963 "reason_parser.ml"
      
    in
    
# 1546 "reason_parser.mly"
    ( mkmty(Pmty_ident _1) )
# 7969 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run240 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState240 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState240 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState240

and _menhir_run135 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState135 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState135 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState135

and _menhir_error133 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce151 : _menhir_env -> (((('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _ ->
    let (((((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_opt_bar)), _, (_3 : 'tv_row_field_list)), _endpos__4_, _startpos__4_), _, (_5 : 'tv_name_tag_list)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3969 "reason_parser.mly"
      ( mktyp(Ptyp_variant(List.rev _3, Closed, Some (List.rev _5))) )
# 8019 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce642 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__optional_type_variable_with_variance -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__optional_type_variable_with_variance) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_optional_type_variable_with_variance = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 3523 "reason_parser.mly"
                                                                                (
      let first = fst x in
      let second = snd x in
      ({
        first with ptyp_loc = {
        first.ptyp_loc with
        loc_start = _symbolstartpos;
        loc_end = _endpos}}, second )
)
# 8039 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7595) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_optional_type_variable_with_variance) = _v in
    ((match _menhir_s with
    | MenhirState1855 | MenhirState90 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7587) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7588)
    | MenhirState2076 | MenhirState1973 | MenhirState1421 | MenhirState1367 | MenhirState1254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7589) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce638 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7590)
    | MenhirState1364 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7591 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce639 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7592)
    | MenhirState1864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7593 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_optional_type_variable_with_variance) = _v in
        (_menhir_reduce639 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7594)
    | _ ->
        _menhir_fail ()) : 'freshtv7596)

and _menhir_reduce52 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_simple_expr = 
# 1972 "reason_parser.mly"
      ( _2 )
# 8088 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1767 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1544 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1543 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1767

and _menhir_run1769 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 8114 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7583 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 8126 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1779 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1779) : 'freshtv7584)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7585 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 8169 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1770) : 'freshtv7586)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce823 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce12 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_def) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_expr = 
# 1923 "reason_parser.mly"
      ( _2 )
# 8216 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce31 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_def) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_fun_def = 
# 1887 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _2)) )
# 8228 "reason_parser.ml"
     in
    _menhir_goto__class_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2699 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2670 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7581 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce51 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7582)

and _menhir_goto__non_arrowed_class_constructor_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_class_constructor_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7579) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__non_arrowed_class_constructor_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7577) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__non_arrowed_class_constructor_type) : 'tv__non_arrowed_class_constructor_type) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _v : 'tv_mark_position_cty__non_arrowed_class_constructor_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1158 "reason_parser.mly"
 ( {x with pcty_loc = {x.pcty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 8265 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7575) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7573) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7571) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) : 'tv_mark_position_cty__non_arrowed_class_constructor_type_) = _v in
    ((let _v : 'tv_non_arrowed_class_constructor_type = 
# 2205 "reason_parser.mly"
                                                                                           (_1)
# 8282 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7569) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_class_constructor_type) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1826 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7561 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7557 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1831 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1831 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1831 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1831 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1831 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1831 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1831 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1831) : 'freshtv7558)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7559 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7560)) : 'freshtv7562)
    | MenhirState2707 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7567 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7563 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2681 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2677 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2673 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2709 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2665 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2709 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2709) : 'freshtv7564)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7565 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7566)) : 'freshtv7568)
    | _ ->
        _menhir_fail ()) : 'freshtv7570)) : 'freshtv7572)) : 'freshtv7574)) : 'freshtv7576)) : 'freshtv7578)) : 'freshtv7580)

and _menhir_error1828 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1818 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce50 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _startpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_class_constructor_type)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__class_simple_expr = 
# 1968 "reason_parser.mly"
      ( mkclass(Pcl_constraint(_2, _4)) )
# 8392 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1783 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7555 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce51 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7556)

and _menhir_reduce9 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 8408 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 8414 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_constructor_type)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__class_constructor_type = 
# 2200 "reason_parser.mly"
      ( mkcty(Pcty_arrow(_1, _3, _5)) )
# 8423 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce8 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 8430 "reason_parser.ml"
) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 8436 "reason_parser.ml"
    )), _startpos__1_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_constructor_type)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__class_constructor_type = 
# 2198 "reason_parser.mly"
      ( mkcty(Pcty_arrow("?" ^ _1, mkoption _3, _5)) )
# 8445 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce10 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_core_type), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_constructor_type)) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_constructor_type = 
# 2202 "reason_parser.mly"
      ( mkcty(Pcty_arrow("", _1, _3)) )
# 8458 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce365 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 8465 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 857 "reason_parser.mly"
       (string)
# 8470 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_clty_longident = 
# 4164 "reason_parser.mly"
                                                ( Lident _1 )
# 8477 "reason_parser.ml"
     in
    _menhir_goto_clty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1545 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1545

and _menhir_reduce524 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_fun_binding) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_fun_binding_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1193 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 8526 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7553) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_fun_binding_) = _v in
    ((match _menhir_s with
    | MenhirState1823 | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7549) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_binding_) = _v in
        (_menhir_reduce345 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7550)
    | MenhirState2704 | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7551) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_binding_) = _v in
        (_menhir_reduce345 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv7552)
    | _ ->
        _menhir_fail ()) : 'freshtv7554)

and _menhir_run1289 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1289
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1289

and _menhir_run2480 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2480
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2480

and _menhir_reduce294 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_item_extension_sugar * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_structure_item_without_item_extension_sugar) _startpos__2_ ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_item_extension_sugar), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__structure_item = 
# 1389 "reason_parser.mly"
                                                                     (
    struct_item_extension _1 _2
  )
# 8647 "reason_parser.ml"
     in
    _menhir_goto__structure_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce295 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_structure_item_without_item_extension_sugar) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item = 
# 1392 "reason_parser.mly"
                                                (
    _1
  )
# 8660 "reason_parser.ml"
     in
    _menhir_goto__structure_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2759 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2723 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1845 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7547 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | RBRACE | RBRACKET ->
        _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState1846
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1846) : 'freshtv7548)

and _menhir_reduce499 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_curried_binding_return_typed) ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1213 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 8793 "reason_parser.ml"
      
    in
    
# 2915 "reason_parser.mly"
      ( (_1, _2) )
# 8799 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error86 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1972 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1972

and _menhir_run1980 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7545 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1980 in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 8837 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONEQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7541 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 8849 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run1983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1982) : 'freshtv7542)
        | DOT | EQUAL ->
            _menhir_reduce576 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7543 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 8872 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7544)) : 'freshtv7546)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1980

and _menhir_run1420 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1420 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1420 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1420

and _menhir_run1430 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7539 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1430 in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 8910 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONEQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7535 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 8922 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1432) : 'freshtv7536)
        | DOT | EQUAL ->
            _menhir_reduce576 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7537 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 8945 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7538)) : 'freshtv7540)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1430

and _menhir_error1437 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce262 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_type = 
# 1542 "reason_parser.mly"
      ( _2 )
# 8969 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1971 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        _menhir_run1980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1971

and _menhir_reduce253 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ _startpos__5_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_module_expr = 
# 1242 "reason_parser.mly"
      ( mkmod(Pmod_constraint(_2, _4)) )
# 8999 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run2005 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2005 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2005 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2005 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2005 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2005 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run234 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2005 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2005

and _menhir_goto__module_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7533) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__module_declaration) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7531) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__module_declaration) : 'tv__module_declaration) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__module_declaration_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1178 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 9046 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7529) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__module_declaration_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7527) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__module_declaration_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7525) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_mty__module_declaration_) : 'tv_mark_position_mty__module_declaration_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_module_declaration = 
# 1745 "reason_parser.mly"
                                                           (_1)
# 9067 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7523) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_declaration) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7515 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 9081 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7513 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 9087 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 9092 "reason_parser.ml"
        )), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_), _endpos__5_, _, _startpos__5_), _endpos__6_, _, (_6 : 'tv_module_declaration)) = _menhir_stack in
        let _5 = () in
        let _3 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__module_declaration = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 9108 "reason_parser.ml"
          
        in
        
# 1750 "reason_parser.mly"
      ( mkmty(Pmty_functor(_2, Some _4, _6)) )
# 9114 "reason_parser.ml"
         in
        _menhir_goto__module_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv7514)) : 'freshtv7516)
    | MenhirState1457 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7519 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7517 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_declaration)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__module_declaration = 
# 1752 "reason_parser.mly"
      ( mkmty(Pmty_functor(mkloc "*" (mklocation _startpos__1_ _endpos__1_), None, _3)) )
# 9130 "reason_parser.ml"
         in
        _menhir_goto__module_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv7518)) : 'freshtv7520)
    | MenhirState1411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7521 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 9138 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKETATAT ->
            _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1462
        | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
            _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1462) : 'freshtv7522)
    | _ ->
        _menhir_fail ()) : 'freshtv7524)) : 'freshtv7526)) : 'freshtv7528)) : 'freshtv7530)) : 'freshtv7532)) : 'freshtv7534)

and _menhir_run1419 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        _menhir_run1430 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1419

and _menhir_run1444 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1444 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1444 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1444 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1444 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1444 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1444 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1444

and _menhir_reduce131 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_simple_module_type) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_module_type = 
# 1533 "reason_parser.mly"
                       (_1)
# 9195 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce118 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_simple_module_expr) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__module_expr = 
# 1284 "reason_parser.mly"
      ( mkmod (Pmod_apply(_1, _2)) )
# 9207 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce116 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_simple_module_expr) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__module_expr = 
# 1274 "reason_parser.mly"
      ( _1 )
# 9218 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce534 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__semi_terminated_seq_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__semi_terminated_seq_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1168 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 9231 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7511) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1176 | MenhirState1167 | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7507) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce698 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7508)
    | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState2259 | MenhirState2253 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7509) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce698 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7510)
    | _ ->
        _menhir_fail ()) : 'freshtv7512)

and _menhir_reduce535 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__semi_terminated_seq_expr_row) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1168 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 9268 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7505) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7501) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce699 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7502)
    | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7503) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__semi_terminated_seq_expr_row_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce699 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7504)
    | _ ->
        _menhir_fail ()) : 'freshtv7506)

and _menhir_reduce603 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 9299 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 9305 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_binding_body)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_nonlocal_module_binding_details = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 9317 "reason_parser.ml"
      
    in
    
# 1509 "reason_parser.mly"
                                      (
      (_1, _2)
    )
# 9325 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7499) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_nonlocal_module_binding_details) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1871 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7487 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1872
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1872
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1872) : 'freshtv7488)
    | MenhirState2309 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7489 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2310
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2310
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2310) : 'freshtv7490)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7491 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2312
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2312
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2312) : 'freshtv7492)
    | MenhirState2639 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7493 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2640
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2640
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2640) : 'freshtv7494)
    | MenhirState2636 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7495 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2642
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2642
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2642) : 'freshtv7496)
    | MenhirState2732 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7497 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2733
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2733
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2733) : 'freshtv7498)
    | _ ->
        _menhir_fail ()) : 'freshtv7500)

and _menhir_reduce579 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body_functor -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_module_binding_body_functor) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_binding_body = 
# 1487 "reason_parser.mly"
                                ( _1 )
# 9415 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run2167 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce255 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run2168 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2168 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2168 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2168 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2168 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2168 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run234 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2168 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2168

and _menhir_run1964 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce255 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1965 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run234 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1965

and _menhir_run1948 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run1961 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7485 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1948 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1949 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1949) : 'freshtv7486)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1948

and _menhir_reduce245 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = 
# 2792 "reason_parser.mly"
      ( mkexp (Pexp_pack _3) )
# 9589 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run762 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run2164 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState762 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7483 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState762 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState763 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState763) : 'freshtv7484)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState762

and _menhir_error2305 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2141 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce126 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_module_type), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__module_type = 
# 1588 "reason_parser.mly"
                                                     (
    (* below_EQUALGREATER to prevent following shift reduce conflict:
     *  1158: shift/reduce conflict (shift 1285, reduce 75) on EQUALGREATER
     *  state 1158
     *    module_binding_body_functor : functor_args COLON non_arrowed_module_type . EQUALGREATER module_expr  (59)
     *    module_type : non_arrowed_module_type .  (75)
     *
     *    EQUALGREATER  shift 1285
     *    LBRACKETAT  reduce 75
     *    WITH  reduce 75
     *)
    _1
  )
# 9726 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run687 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState687

and _menhir_run272 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState272

and _menhir_reduce170 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _ _startpos__8_ ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_pattern_without_or), _startpos__5_), _endpos__6_, _), _endpos__7_, _, (_7 : 'tv_pattern_without_or), _startpos__7_) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__pattern_without_or = let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3296 "reason_parser.mly"
      (
         let loc_coloncolon = mklocation _startpos__2_ _endpos__2_ in
         let loc = mklocation _symbolstartpos _endpos in
         mkpat_cons loc_coloncolon (mkpat ~ghost:true ~loc (Ppat_tuple[_5;_7])) loc
      )
# 9824 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run655 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState655 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState655

and _menhir_run658 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState658 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState658

and _menhir_run574 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState574

and _menhir_run577 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState577 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7481) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState577 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | AMPERAMPER ->
            _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERSAND ->
            _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BARBAR ->
            _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONEQUAL ->
            _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP0 _v ->
            _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP1 _v ->
            _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP2 _v ->
            _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP3 _v ->
            _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP4 _v ->
            _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run581 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OR ->
            _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run580 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSEQ ->
            _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STAR ->
            _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState579) : 'freshtv7482)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState577

and _menhir_reduce233 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_expr_comma_seq_extension) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = 
# 2747 "reason_parser.mly"
      ( let seq, ext_opt = _2 in
        let loc = mklocation _startpos__2_ _endpos__2_ in
        make_real_exp (mktailexp_extension loc seq ext_opt) )
# 10041 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce234 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_expr_comma_seq_extension) _startpos__4_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 10061 "reason_parser.ml"
      
    in
    
# 2751 "reason_parser.mly"
      ( let seq, ext_opt = _4 in
        let loc = mklocation _startpos__4_ _endpos__4_ in
        let list_exp = make_real_exp (mktailexp_extension loc seq ext_opt) in
        let list_exp = { list_exp with pexp_loc = loc } in
        mkexp (Pexp_open (Fresh, _1, list_exp)) )
# 10071 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce418 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_expr_comma_seq_extension) _startpos__3_ ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_), _endpos__2_, _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_expr_comma_seq_extension = 
# 3151 "reason_parser.mly"
    ( let seq, ext = _3 in (_1::seq, ext) )
# 10084 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_seq_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_row_field : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_row_field -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState170 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7463 * _menhir_state * 'tv_row_field_list) * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7461 * _menhir_state * 'tv_row_field_list) * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_row_field_list)), _startpos__2_), _, (_3 : 'tv_row_field)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_row_field_list = 
# 3988 "reason_parser.mly"
                                                ( _3 :: _1 )
# 10102 "reason_parser.ml"
         in
        _menhir_goto_row_field_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7462)) : 'freshtv7464)
    | MenhirState158 | MenhirState161 | MenhirState2449 | MenhirState163 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7467 * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7465 * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_row_field)) = _menhir_stack in
        let _v : 'tv_row_field_list = 
# 3987 "reason_parser.mly"
                                                ( [_1] )
# 10114 "reason_parser.ml"
         in
        _menhir_goto_row_field_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7466)) : 'freshtv7468)
    | MenhirState306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7473 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7469 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState313 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState313) : 'freshtv7470)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7471 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7472)) : 'freshtv7474)
    | MenhirState162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7479 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7475 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2449) : 'freshtv7476)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7477 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7478)) : 'freshtv7480)
    | _ ->
        _menhir_fail ()

and _menhir_reduce783 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_extension_constructor_rebind)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_str_extension_constructors = 
# 3664 "reason_parser.mly"
                                                          ( [_1] )
# 10233 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce785 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_str_extension_constructors)), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_extension_constructor_rebind)) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_extension_constructors = 
# 3668 "reason_parser.mly"
      ( _3 :: _1 )
# 10245 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce782 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_extension_constructor_declaration)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_str_extension_constructors = 
# 3663 "reason_parser.mly"
                                                          ( [_1] )
# 10256 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce784 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_str_extension_constructors)), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_extension_constructor_declaration)) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_extension_constructors = 
# 3666 "reason_parser.mly"
      ( _3 :: _1 )
# 10268 "reason_parser.ml"
     in
    _menhir_goto_str_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_sig_extension_constructors : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_sig_extension_constructors -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((((('freshtv7459 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7457 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState1392 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run1389 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1386 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1382 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1379 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1393 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1393) : 'freshtv7458)
    | LBRACKETATAT ->
        _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1392
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1392) : 'freshtv7460)

and _menhir_reduce397 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_constructor_declaration_leading_bar) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_constructor_declarations_leading_bar = 
# 3556 "reason_parser.mly"
                                                                             ( [_1] )
# 10314 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce398 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_constructor_declaration_leading_bar) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declarations_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constructor_declarations_leading_bar = 
# 3557 "reason_parser.mly"
                                                                             ( _2 :: _1 )
# 10325 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_let_binding : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2673 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1632 | MenhirState1663 | MenhirState1161 | MenhirState1176 | MenhirState1167 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7453) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce505 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7454)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState475 | MenhirState2259 | MenhirState2253 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7455) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce505 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7456)
    | _ ->
        _menhir_fail ()

and _menhir_run2087 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | VIRTUAL ->
        _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState2087
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2087

and _menhir_reduce301 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_exception_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_str_exception_declaration) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1418 "reason_parser.mly"
                              (
      mkstr(Pstr_exception _1)
    )
# 10375 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run2075 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7451) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2075 in
        let (_v : (
# 857 "reason_parser.mly"
       (string)
# 10393 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MINUS ->
            _menhir_run1262 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1258 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run1256 _menhir_env (Obj.magic _menhir_stack) MenhirState2076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run1255 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2076 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
            _menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState2076) : 'freshtv7452)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2075

and _menhir_goto_let_bindings : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_bindings -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1176 | MenhirState1167 | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7437 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1166
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1166 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7435 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1166 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1156 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1167 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce637 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7436)
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState1166) : 'freshtv7438)
    | MenhirState2673 | MenhirState1663 | MenhirState1632 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7441 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1662
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1662 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7439 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1662 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AS ->
                  _menhir_run1657 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1633 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
                  _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1663
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1663) : 'freshtv7440)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1662)) : 'freshtv7442)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7443 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1885
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce296 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7444)
    | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState2259 | MenhirState2253 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7447 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2252
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2252 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7445 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2252 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2253 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce637 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7446)
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState2252) : 'freshtv7448)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7449 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2742
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run1183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2742 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce296 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7450)
    | _ ->
        _menhir_fail ()

and _menhir_run2154 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce256 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1951 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce256 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce205 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2650 "reason_parser.mly"
      ( _2 )
# 10701 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce78 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _ _startpos__8_ ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_), _endpos__6_, _), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__expr = let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2556 "reason_parser.mly"
      (
        let loc_colon = mklocation _startpos__2_ _endpos__2_ in
        let loc = mklocation _symbolstartpos _endpos in
        mkexp_cons loc_colon (mkexp ~ghost:true ~loc (Pexp_tuple[_5;_7])) loc
      )
# 10725 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce213 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ _startpos__5_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 10746 "reason_parser.ml"
      
    in
    
# 2672 "reason_parser.mly"
      ( mkexp(Pexp_open(Fresh, _1, _4)) )
# 10752 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1013 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1013 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1013

and _menhir_run948 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState948 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState948

and _menhir_run963 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState963 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState963

and _menhir_run969 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState969

and _menhir_run971 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState971 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState971

and _menhir_run965 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState965 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState965

and _menhir_run973 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState973

and _menhir_run975 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState975

and _menhir_run977 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState977

and _menhir_run979 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState979

and _menhir_run985 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState985 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState985

and _menhir_run987 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState987 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState987

and _menhir_run950 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 836 "reason_parser.mly"
       (string)
# 11683 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState950

and _menhir_run967 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 835 "reason_parser.mly"
       (string)
# 11764 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState967 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState967

and _menhir_run981 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 834 "reason_parser.mly"
       (string)
# 11845 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState981 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState981

and _menhir_run983 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 833 "reason_parser.mly"
       (string)
# 11926 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState983 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState983

and _menhir_run989 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 832 "reason_parser.mly"
       (string)
# 12007 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState989 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState989

and _menhir_run991 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState991 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState991

and _menhir_run999 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState999 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState999

and _menhir_run993 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState993 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState993

and _menhir_run995 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState995 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState995

and _menhir_run997 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState997 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState997

and _menhir_run893 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState893

and _menhir_run835 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState835 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState835

and _menhir_run850 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState850 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState850

and _menhir_run856 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState856

and _menhir_run858 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState858

and _menhir_run852 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState852 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState852

and _menhir_run860 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState860 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState860

and _menhir_run862 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState862 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState862

and _menhir_run864 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState864 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState864

and _menhir_run866 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState866 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState866

and _menhir_run872 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState872 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState872

and _menhir_run874 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState874 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState874

and _menhir_run837 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 836 "reason_parser.mly"
       (string)
# 13397 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState837 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState837

and _menhir_run854 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 835 "reason_parser.mly"
       (string)
# 13478 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState854 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState854

and _menhir_run868 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 834 "reason_parser.mly"
       (string)
# 13559 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState868 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState868

and _menhir_run870 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 833 "reason_parser.mly"
       (string)
# 13640 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState870 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState870

and _menhir_run876 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 832 "reason_parser.mly"
       (string)
# 13721 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState876

and _menhir_run878 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState878 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState878

and _menhir_run886 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState886 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState886

and _menhir_run880 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState880 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState880

and _menhir_run882 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState882 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState882

and _menhir_run884 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState884 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState884

and _menhir_reduce550 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_pattern) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__simple_pattern_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1203 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 14193 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7433) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__simple_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7429) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce726 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7430)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState486 | MenhirState738 | MenhirState1886 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1657 | MenhirState1639 | MenhirState1637 | MenhirState1634 | MenhirState1183 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7431) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce726 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7432)
    | _ ->
        _menhir_fail ()) : 'freshtv7434)

and _menhir_reduce201 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_val_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _menhir_s (x0 : 'tv_val_longident) _startpos_x0_ ->
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 14234 "reason_parser.ml"
      
    in
    
# 2638 "reason_parser.mly"
      ( mkexp (Pexp_ident _1) )
# 14240 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce71 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__expr = 
# 2530 "reason_parser.mly"
      ( syntax_error_exp (mklocation _startpos__4_ _endpos__4_) "Invalid try with")
# 14255 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_labeled_simple_pattern : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_labeled_simple_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1041 | MenhirState902 | MenhirState1027 | MenhirState906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7393 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1022 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run907 _menhir_env (Obj.magic _menhir_stack) MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1027) : 'freshtv7394)
    | MenhirState898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7395 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1022 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run907 _menhir_env (Obj.magic _menhir_stack) MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1041) : 'freshtv7396)
    | MenhirState1227 | MenhirState1188 | MenhirState1199 | MenhirState1192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7397 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1195 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1193 _menhir_env (Obj.magic _menhir_stack) MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1189 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1199 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1199) : 'freshtv7398)
    | MenhirState1184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7399 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1195 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1193 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1189 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1227 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1227) : 'freshtv7400)
    | MenhirState1639 | MenhirState1637 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7403 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7401 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState1639 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1640 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1640 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1640 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1640 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1640 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1640 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1640 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1640) : 'freshtv7402)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1639 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1639) : 'freshtv7404)
    | MenhirState1912 | MenhirState1891 | MenhirState1703 | MenhirState1714 | MenhirState1707 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7405 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1710 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1708 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1714) : 'freshtv7406)
    | MenhirState1823 | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7411 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7407) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1823 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1790 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1826 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1826 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1789 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1826 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1827 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1826 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1826 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1826) : 'freshtv7408)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7409) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState1823 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1824 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1824 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1824 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1824 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1824 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1824 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1824 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1824) : 'freshtv7410)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1823) : 'freshtv7412)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7413 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run1710 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run1708 _menhir_env (Obj.magic _menhir_stack) MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1912 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1912) : 'freshtv7414)
    | MenhirState2238 | MenhirState480 | MenhirState2224 | MenhirState484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7415 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run2220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run2218 _menhir_env (Obj.magic _menhir_stack) MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2224 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2224) : 'freshtv7416)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7417 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            _menhir_run2220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            _menhir_run2218 _menhir_env (Obj.magic _menhir_stack) MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2238) : 'freshtv7418)
    | MenhirState2679 | MenhirState2677 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7421 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7419 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2679 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2681 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2677 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2673 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2665 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2680) : 'freshtv7420)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2679 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2679) : 'freshtv7422)
    | MenhirState2704 | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7427 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7423) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2704 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1790 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1789 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1827 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2707) : 'freshtv7424)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7425) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2704 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2681 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2705 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2677 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2705 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2673 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2705 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2705 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2705 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2665 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2705 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2705 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2705) : 'freshtv7426)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2704) : 'freshtv7428)
    | _ ->
        _menhir_fail ()

and _menhir_run742 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7391) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_direction_flag = 
# 4198 "reason_parser.mly"
                                                ( Upto )
# 15090 "reason_parser.ml"
     in
    _menhir_goto_direction_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7392)

and _menhir_run745 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7389) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_direction_flag = 
# 4199 "reason_parser.mly"
                                                ( Downto )
# 15106 "reason_parser.ml"
     in
    _menhir_goto_direction_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7390)

and _menhir_run734 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 895 "reason_parser.mly"
       (string)
# 15113 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState734

and _menhir_run743 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState743

and _menhir_run736 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7383 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState736 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2213) : 'freshtv7384)
    | LBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7385 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState736 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2210) : 'freshtv7386)
    | LIDENT _v ->
        _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState736 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7387 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState736 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState737) : 'freshtv7388)
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState736 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState736

and _menhir_run470 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> Lexing.position -> _menhir_state -> (
# 895 "reason_parser.mly"
       (string)
# 15438 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState470 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState470

and _menhir_run749 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7381) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState749 in
        let (_v : (
# 857 "reason_parser.mly"
       (string)
# 15506 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce460 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7382)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState749

and _menhir_run472 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7375 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState472 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2277 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2277) : 'freshtv7376)
    | LBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7377 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState472 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2274 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2274) : 'freshtv7378)
    | LIDENT _v ->
        _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7379 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState472 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState473 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState473) : 'freshtv7380)
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState472 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState472

and _menhir_reduce17 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_expr = 
# 1957 "reason_parser.mly"
      ( mkclass(Pcl_extension _1) )
# 15779 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce37 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_instance_type = 
# 2227 "reason_parser.mly"
      ( mkcty(Pcty_extension _1) )
# 15790 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce267 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_module_type = 
# 1552 "reason_parser.mly"
      ( mkmty(Pmty_extension _1) )
# 15801 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce250 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_expr = 
# 2809 "reason_parser.mly"
      ( mkexp (Pexp_extension _1) )
# 15812 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce261 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_module_expr = 
# 1268 "reason_parser.mly"
      ( mkmod (Pmod_extension _1) )
# 15823 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce293 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3392 "reason_parser.mly"
      ( mkpat(Ppat_extension _1) )
# 15834 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce153 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3973 "reason_parser.mly"
      ( mktyp(Ptyp_extension _1) )
# 15845 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce14 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_expr = 
# 1934 "reason_parser.mly"
      ( Cl.attr _1 _2 )
# 15857 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce36 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_instance_type), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_instance_type = 
# 2225 "reason_parser.mly"
      ( Cty.attr _1 _2 )
# 15869 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce133 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_type), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_module_type = 
# 1537 "reason_parser.mly"
      ( Mty.attr _1 _2 )
# 15881 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce110 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2632 "reason_parser.mly"
      ( Exp.attr _1 _2 )
# 15893 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce124 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__module_expr = 
# 1296 "reason_parser.mly"
      ( Mod.attr _1 _2 )
# 15905 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce174 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3313 "reason_parser.mly"
      ( Pat.attr _1 _2 )
# 15917 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce136 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_core_type), _startpos__1_), _endpos__2_, _, (_2 : 'tv_attribute)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_non_simple_core_type = 
# 3926 "reason_parser.mly"
      ( Typ.attr _1 _2 )
# 15929 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_non_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce311 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_floating_attribute -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_floating_attribute) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1461 "reason_parser.mly"
      ( mkstr(Pstr_attribute _1) )
# 15940 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce273 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_signed_constant), _startpos__1_), _endpos__2_), _endpos__3_, _, (_3 : 'tv_signed_constant), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3344 "reason_parser.mly"
      ( mkpat(Ppat_interval (_1, _3)) )
# 15953 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run191 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState191 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState191

and _menhir_run2027 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2027 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2027

and _menhir_reduce333 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) * Lexing.position * _menhir_state * 'tv_attr_id -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_single_attr_id), _startpos__1_), _endpos__3_, _, (_3 : 'tv_attr_id)) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_attr_id = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 4303 "reason_parser.mly"
                               ( mkloc (_1 ^ "." ^ _3.txt) (mklocation _symbolstartpos _endpos) )
# 16061 "reason_parser.ml"
     in
    _menhir_goto_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2370 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2175 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2162 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2159 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2156 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce258 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _startpos__8_ ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _, (_5 : 'tv_package_type)), _, (_7 : 'tv_package_type)) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__simple_module_expr = let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1255 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod (Pmod_unpack(
               mkexp ~ghost:true ~loc (Pexp_coerce(_3, Some(mktyp ~ghost:true ~loc (Ptyp_package _5)),
                                      mktyp ~ghost:true ~loc (Ptyp_package _7))))) )
# 16109 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1959 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce257 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_module_expr = let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1250 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod (Pmod_unpack(
               mkexp ~ghost:true ~loc (Pexp_constraint(_3, mktyp ~ghost:true ~loc (Ptyp_package _5))))) )
# 16135 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1956 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce259 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_module_expr = let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1261 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod (Pmod_unpack(
               mkexp ~ghost:true ~loc (Pexp_coerce(_3, None, mktyp ~ghost:true ~loc (Ptyp_package _5))))) )
# 16161 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1953 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce248 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _startpos__8_ ->
    let (((((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_expr), _startpos__5_), _endpos__6_, _, _startpos__6_), _, (_7 : 'tv_package_type)) = _menhir_stack in
    let _8 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16189 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2801 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_open(Fresh, _1,
        mkexp ~loc (Pexp_constraint (mkexp ~ghost:true ~loc (Pexp_pack _5),
                                     mktyp ~ghost:true ~loc (Ptyp_package _7))))) )
# 16202 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error803 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce246 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2794 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp (Pexp_constraint (mkexp ~ghost:true ~loc (Pexp_pack _3),
                                mktyp ~ghost:true ~loc (Ptyp_package _5))) )
# 16228 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error777 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error500 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 16240 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv7373 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 16247 "reason_parser.ml"
    ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce292 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7374)

and _menhir_error247 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce291 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 16262 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 16268 "reason_parser.ml"
    )), _startpos_x0_), _endpos__4_, _startpos__4_), _, (_5 : 'tv_package_type)) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16285 "reason_parser.ml"
      
    in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3385 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkpat(Ppat_constraint(mkpat ~ghost:true ~loc (Ppat_unpack _3),
                              mktyp ~ghost:true ~loc (Ptyp_package _5))) )
# 16296 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error235 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 16303 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv7371 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 16310 "reason_parser.ml"
    ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce292 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7372)

and _menhir_reduce152 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _, (_3 : 'tv_package_type)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3971 "reason_parser.mly"
      ( mktyp(Ptyp_package _3) )
# 16328 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error131 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce599 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__non_arrowed_simple_core_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__non_arrowed_simple_core_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_simple_core_type = 
# 3929 "reason_parser.mly"
                                                                               (_1)
# 16344 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7369) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_simple_core_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState158 | MenhirState161 | MenhirState162 | MenhirState2449 | MenhirState163 | MenhirState306 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState170 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7355) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7353) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_non_arrowed_simple_core_type) : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _v : 'tv_row_field = 
# 3992 "reason_parser.mly"
                                                            ( Rinherit _1 )
# 16369 "reason_parser.ml"
         in
        _menhir_goto_row_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7354)) : 'freshtv7356)
    | MenhirState2683 | MenhirState2472 | MenhirState1800 | MenhirState1277 | MenhirState255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7357 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce601 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7358)
    | MenhirState2682 | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState121 | MenhirState1799 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1345 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1289 | MenhirState1284 | MenhirState1269 | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7359) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce600 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7360)
    | MenhirState2698 | MenhirState2669 | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState2465 | MenhirState2463 | MenhirState143 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1816 | MenhirState1814 | MenhirState1809 | MenhirState1807 | MenhirState1805 | MenhirState1766 | MenhirState1781 | MenhirState1779 | MenhirState1774 | MenhirState1772 | MenhirState1770 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState324 | MenhirState272 | MenhirState270 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7361) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce596 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7362)
    | MenhirState1643 | MenhirState1593 | MenhirState1500 | MenhirState329 | MenhirState317 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7363 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce601 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7364)
    | MenhirState2521 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState1642 | MenhirState1592 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1490 | MenhirState328 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7365) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce600 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7366)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7367) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce596 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7368)
    | _ ->
        _menhir_fail ()) : 'freshtv7370)

and _menhir_reduce598 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__non_arrowed_non_simple_core_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_non_simple_core_type = 
# 3919 "reason_parser.mly"
                                                                                       (_1)
# 16430 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7351) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_non_simple_core_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2698 | MenhirState2669 | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState2465 | MenhirState2463 | MenhirState143 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1816 | MenhirState1814 | MenhirState1809 | MenhirState1807 | MenhirState1805 | MenhirState1766 | MenhirState1781 | MenhirState1779 | MenhirState1774 | MenhirState1772 | MenhirState1770 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState324 | MenhirState272 | MenhirState270 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7347) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce595 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7348)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7349) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce595 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7350)
    | _ ->
        _menhir_fail ()) : 'freshtv7352)

and _menhir_reduce278 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_lbl_pattern_list)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16474 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 16486 "reason_parser.ml"
      
    in
    
# 3354 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "{") (with_txt _3 "}") )
# 16492 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__pattern_optional_constraint : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__pattern_optional_constraint -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7345) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__pattern_optional_constraint) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7343) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__pattern_optional_constraint) : 'tv__pattern_optional_constraint) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _v : 'tv_mark_position_pat__pattern_optional_constraint_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1203 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 16515 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7341) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__pattern_optional_constraint_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7339) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__pattern_optional_constraint_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7337) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_pat__pattern_optional_constraint_) : 'tv_mark_position_pat__pattern_optional_constraint_) = _v in
    ((let _v : 'tv_pattern_optional_constraint = 
# 3415 "reason_parser.mly"
                                                                             (_1)
# 16532 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7335) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_pattern_optional_constraint) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7323 * _menhir_state * 'tv_pattern_two_or_more_comma_list) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7321 * _menhir_state * 'tv_pattern_two_or_more_comma_list) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_two_or_more_comma_list)), _endpos__2_), _, (_3 : 'tv_pattern_optional_constraint)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_pattern_two_or_more_comma_list = 
# 3403 "reason_parser.mly"
    ( _3::_1 )
# 16550 "reason_parser.ml"
         in
        _menhir_goto_pattern_two_or_more_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7322)) : 'freshtv7324)
    | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7329 * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7325 * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState676 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState676) : 'freshtv7326)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv7327 * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7328)) : 'freshtv7330)
    | MenhirState676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7333 * _menhir_state * 'tv_pattern_optional_constraint) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7331 * _menhir_state * 'tv_pattern_optional_constraint) * Lexing.position) * _menhir_state * 'tv_pattern_optional_constraint) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_optional_constraint)), _endpos__2_), _, (_3 : 'tv_pattern_optional_constraint)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_pattern_two_or_more_comma_list = 
# 3405 "reason_parser.mly"
    ( [_3; _1] )
# 16634 "reason_parser.ml"
         in
        _menhir_goto_pattern_two_or_more_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7332)) : 'freshtv7334)
    | _ ->
        _menhir_fail ()) : 'freshtv7336)) : 'freshtv7338)) : 'freshtv7340)) : 'freshtv7342)) : 'freshtv7344)) : 'freshtv7346)

and _menhir_reduce547 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__or_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__or_pattern) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__or_pattern_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1203 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 16649 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7319) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__or_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7315) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__or_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce643 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7316)
    | MenhirState2578 | MenhirState191 | MenhirState2314 | MenhirState738 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState676 | MenhirState587 | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7317) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__or_pattern_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce643 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7318)
    | _ ->
        _menhir_fail ()) : 'freshtv7320)

and _menhir_reduce399 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__core_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__core_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_core_type = 
# 3847 "reason_parser.mly"
                                         (_1)
# 16684 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7313) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_core_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState270 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7101 * _menhir_state * 'tv_core_type_comma_list) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7099 * _menhir_state * 'tv_core_type_comma_list) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_core_type_comma_list)), _endpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_core_type_comma_list = 
# 4015 "reason_parser.mly"
                                                ( _3 :: _1 )
# 16704 "reason_parser.ml"
         in
        _menhir_goto_core_type_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7100)) : 'freshtv7102)
    | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState1288 | MenhirState1271 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7105 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7103 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
        let _v : 'tv_core_type_comma_list = 
# 4014 "reason_parser.mly"
                                                ( [_1] )
# 16716 "reason_parser.ml"
         in
        _menhir_goto_core_type_comma_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7104)) : 'freshtv7106)
    | MenhirState666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7107 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce162 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7108)
    | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7115 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7111 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7109 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
          ((let _menhir_env = _menhir_discard _menhir_env in
          _menhir_reduce288 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7110)) : 'freshtv7112)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7113 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce287 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7114)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce162 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7116)
    | MenhirState722 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7123 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7119 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ((('freshtv7117 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
          ((let _menhir_env = _menhir_discard _menhir_env in
          _menhir_reduce288 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7118)) : 'freshtv7120)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7121 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce287 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7122)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce162 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7124)
    | MenhirState1066 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7125 * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce808 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7126)
    | MenhirState1895 | MenhirState1206 | MenhirState1068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7131 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv7127 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7128)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7129 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1071 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1071) : 'freshtv7130)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce809 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7132)
    | MenhirState1071 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7133 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce810 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7134)
    | MenhirState1087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7135 * _menhir_state) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce808 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7136)
    | MenhirState1089 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7139 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1091 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7137 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1092) : 'freshtv7138)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce809 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv7140)
    | MenhirState1092 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7141 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce810 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7142)
    | MenhirState1211 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7145 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1212 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7143 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1213 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1213) : 'freshtv7144)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1212 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7146)
    | MenhirState1219 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7149 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1220 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7147 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1221 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1221) : 'freshtv7148)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1220 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7150)
    | MenhirState1232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7153 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1233 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7151 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1234) : 'freshtv7152)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1233 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7154)
    | MenhirState1275 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7155 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce452 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7156)
    | MenhirState1278 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7157 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce451 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7158)
    | MenhirState1287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7159 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce814 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7160)
    | MenhirState1330 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7161 * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce176 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7162)
    | MenhirState153 | MenhirState1720 | MenhirState1328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7163 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce175 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7164)
    | MenhirState1268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7171 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1343 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7169 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run1298 _menhir_env (Obj.magic _menhir_stack) MenhirState1344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONCOLON ->
                  _menhir_run1295 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOT ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7165 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1344 in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce821 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7166)
              | FALSE ->
                  _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1347 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7167 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1344 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | BAR ->
                      _menhir_run1298 _menhir_env (Obj.magic _menhir_stack) MenhirState1346 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | COLONCOLON ->
                      _menhir_run1295 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1346 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1346 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run1347 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1346 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run1284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1346 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run1345 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1346 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_reduce680 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1346) : 'freshtv7168)
              | TRUE ->
                  _menhir_run1284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1344 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run1345 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1344 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState1344
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1344) : 'freshtv7170)
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce813 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1343 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7172)
    | MenhirState1359 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7175 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1360 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7173 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1361 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1361) : 'freshtv7174)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1360 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7176)
    | MenhirState1361 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7177 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce379 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7178)
    | MenhirState1425 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7179 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce850 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7180)
    | MenhirState1427 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7181 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) MenhirState1428 : 'freshtv7182)
    | MenhirState1471 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7183 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1472
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1472
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1472) : 'freshtv7184)
    | MenhirState1480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7187 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1481 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7185 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | STRING _v ->
                  _menhir_run1483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1482 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1482) : 'freshtv7186)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1481 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7188)
    | MenhirState1498 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7189 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce452 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7190)
    | MenhirState1501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7191 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce451 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7192)
    | MenhirState1545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7197 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1546 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7195 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7193 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _endpos = _endpos__3_ in
              let _v : 'tv_class_self_type = 
# 2238 "reason_parser.mly"
      ( _2 )
# 17384 "reason_parser.ml"
               in
              _menhir_goto_class_self_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7194)) : 'freshtv7196)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1546 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7198)
    | MenhirState1556 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7201 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7199 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _endpos = _endpos__5_ in
        let _v : 'tv_value_type = 
# 2276 "reason_parser.mly"
      ( _3, _2, Virtual, _5 )
# 17403 "reason_parser.ml"
         in
        _menhir_goto_value_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7200)) : 'freshtv7202)
    | MenhirState1561 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7205 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7203 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_virtual_flag)), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _endpos = _endpos__5_ in
        let _v : 'tv_value_type = 
# 2278 "reason_parser.mly"
      ( _3, Mutable, _2, _5 )
# 17418 "reason_parser.ml"
         in
        _menhir_goto_value_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7204)) : 'freshtv7206)
    | MenhirState1566 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7209 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7207 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_label), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_value_type = 
# 2280 "reason_parser.mly"
      ( _1, Immutable, Concrete, _3 )
# 17432 "reason_parser.ml"
         in
        _menhir_goto_value_type _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7208)) : 'freshtv7210)
    | MenhirState1579 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7211 * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce176 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7212)
    | MenhirState1700 | MenhirState1695 | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7213 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce175 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7214)
    | MenhirState1744 | MenhirState1599 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7217 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1600 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7215 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1601) : 'freshtv7216)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1600 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7218)
    | MenhirState1601 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7227 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7225 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_constrain_field = 
# 2289 "reason_parser.mly"
                                           ( _1, _3 )
# 17505 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7223) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrain_field) = _v in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
        match _menhir_s with
        | MenhirState1599 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7219 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1603
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKETATAT ->
                  _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1603
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1603) : 'freshtv7220)
        | MenhirState1744 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv7221 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1745
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKETATAT ->
                  _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1745
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1745) : 'freshtv7222)
        | _ ->
            _menhir_fail ()) : 'freshtv7224)) : 'freshtv7226)) : 'freshtv7228)
    | MenhirState1671 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7233 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1672 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7231 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7229 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__6_ : Lexing.position) = _endpos in
              let (_startpos__6_ : Lexing.position) = _startpos in
              ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_) = _menhir_stack in
              let _6 = () in
              let _4 = () in
              let _1 = () in
              let _endpos = _endpos__6_ in
              let _v : 'tv_value = 
# 2022 "reason_parser.mly"
      ( not_expecting _startpos__6_ _endpos__6_ "not expecting equal - cannot specify value for virtual val" )
# 17566 "reason_parser.ml"
               in
              _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7230)) : 'freshtv7232)
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce840 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1672 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7234)
    | MenhirState1680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7239 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1681 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7237 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7235 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__7_ : Lexing.position) = _endpos in
              let (_startpos__7_ : Lexing.position) = _startpos in
              ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_label), _startpos__4_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_core_type), _startpos__6_) = _menhir_stack in
              let _7 = () in
              let _5 = () in
              let _3 = () in
              let _2 = () in
              let _endpos = _endpos__7_ in
              let _v : 'tv_value = 
# 2018 "reason_parser.mly"
      ( not_expecting _startpos__7_ _endpos__7_ "not expecting equal - cannot specify value for virtual val" )
# 17602 "reason_parser.ml"
               in
              _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7236)) : 'freshtv7238)
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce838 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1681 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7240)
    | MenhirState1723 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv7243 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1724 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((((('freshtv7241 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1725 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1725) : 'freshtv7242)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1724 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7244)
    | MenhirState1859 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7247 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1860 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7245 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1861) : 'freshtv7246)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1860 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7248)
    | MenhirState1861 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7249 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce379 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7250)
    | MenhirState1898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv7253 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1899 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((((('freshtv7251 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1900 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1900) : 'freshtv7252)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1899 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7254)
    | MenhirState1904 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7257 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1905 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv7255 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1906) : 'freshtv7256)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1905 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7258)
    | MenhirState1917 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7261 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1918 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv7259 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1919 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1919) : 'freshtv7260)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1918 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7262)
    | MenhirState1975 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7263 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce850 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7264)
    | MenhirState1977 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7265 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) MenhirState1978 : 'freshtv7266)
    | MenhirState2027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7269 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7267 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
        let _1 = () in
        let _v : 'tv_payload = 
# 4354 "reason_parser.mly"
                    ( PTyp _2 )
# 18048 "reason_parser.ml"
         in
        _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv7268)) : 'freshtv7270)
    | MenhirState2041 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7273 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2042 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7271 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | STRING _v ->
                  _menhir_run2044 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2043) : 'freshtv7272)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2042 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7274)
    | MenhirState242 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7275 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce648 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7276)
    | MenhirState2439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7279 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7277 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
        let _endpos = _endpos__1_ in
        let _v : 'tv_amper_type_list = 
# 4005 "reason_parser.mly"
                                                ( [_1] )
# 18092 "reason_parser.ml"
         in
        _menhir_goto_amper_type_list _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7278)) : 'freshtv7280)
    | MenhirState2442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7283 * Lexing.position * _menhir_state * 'tv_amper_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7281 * Lexing.position * _menhir_state * 'tv_amper_type_list) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_amper_type_list)), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_amper_type_list = 
# 4006 "reason_parser.mly"
                                                ( _3 :: _1 )
# 18106 "reason_parser.ml"
         in
        _menhir_goto_amper_type_list _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv7282)) : 'freshtv7284)
    | MenhirState141 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7285 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce648 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7286)
    | MenhirState2478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv7287 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce814 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7288)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7295 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2519 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7293 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run2489 _menhir_env (Obj.magic _menhir_stack) MenhirState2520 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONCOLON ->
                  _menhir_run2486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2520 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOT ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7289 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState2520 in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce821 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7290)
              | FALSE ->
                  _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2520 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run2523 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2520 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv7291 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState2520 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | BAR ->
                      _menhir_run2489 _menhir_env (Obj.magic _menhir_stack) MenhirState2522 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | COLONCOLON ->
                      _menhir_run2486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2522 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2522 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run2523 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2522 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run2475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2522 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run2521 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2522 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_reduce680 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2522) : 'freshtv7292)
              | TRUE ->
                  _menhir_run2475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2520 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run2521 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2520 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState2520
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2520) : 'freshtv7294)
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce813 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2519 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7296)
    | MenhirState2570 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7307 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2572 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7305 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7303 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1067 "reason_parser.mly"
      (Parsetree.core_type)
# 18211 "reason_parser.ml"
              ) = 
# 1139 "reason_parser.mly"
                  ( apply_mapper_chain_to_type _1 default_mapper_chain )
# 18215 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7301) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1067 "reason_parser.mly"
      (Parsetree.core_type)
# 18223 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7299) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1067 "reason_parser.mly"
      (Parsetree.core_type)
# 18231 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7297) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1067 "reason_parser.mly"
      (Parsetree.core_type)
# 18239 "reason_parser.ml"
              )) : (
# 1067 "reason_parser.mly"
      (Parsetree.core_type)
# 18243 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv7298)) : 'freshtv7300)) : 'freshtv7302)) : 'freshtv7304)) : 'freshtv7306)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2572 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7308)
    | MenhirState2649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7311 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2650 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7309 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | STRING _v ->
                  _menhir_run2044 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2651 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2651) : 'freshtv7310)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2650 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7312)
    | _ ->
        _menhir_fail ()) : 'freshtv7314)

and _menhir_reduce729 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__simple_pattern_not_ident_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__simple_pattern_not_ident_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_pattern_not_ident = 
# 3337 "reason_parser.mly"
                                                                       (_1)
# 18287 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7097) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_pattern_not_ident) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState564 | MenhirState563 | MenhirState550 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7085 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce270 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7086)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1657 | MenhirState1639 | MenhirState1637 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 | MenhirState908 | MenhirState767 | MenhirState738 | MenhirState486 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState650 | MenhirState649 | MenhirState646 | MenhirState642 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7087 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce270 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7088)
    | MenhirState1634 | MenhirState1183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7091 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1231 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7089 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1232) : 'freshtv7090)
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce270 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1231 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7092)
    | MenhirState2314 | MenhirState1886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7095 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1916 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7093 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1917 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1917) : 'freshtv7094)
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce270 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1916 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7096)
    | _ ->
        _menhir_fail ()) : 'freshtv7098)

and _menhir_reduce719 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_labeled_simple_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_labeled_simple_expr) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_labeled_expr_list = 
# 2837 "reason_parser.mly"
      ( [_1] )
# 18420 "reason_parser.ml"
     in
    _menhir_goto_simple_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce720 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list -> Lexing.position -> _menhir_state -> 'tv_labeled_simple_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_labeled_simple_expr) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_labeled_expr_list)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_simple_labeled_expr_list = 
# 2839 "reason_parser.mly"
      ( _2 :: _1 )
# 18431 "reason_parser.ml"
     in
    _menhir_goto_simple_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce469 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_label_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_label_expr) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_labeled_simple_expr = 
# 2846 "reason_parser.mly"
      ( _1 )
# 18441 "reason_parser.ml"
     in
    _menhir_goto_labeled_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce72 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_simple_non_labeled_expr_list_as_tuple -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_simple_non_labeled_expr_list_as_tuple) ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 18459 "reason_parser.ml"
      
    in
    
# 2532 "reason_parser.mly"
    (
      if List.mem (string_of_longident _1.txt)
         built_in_explicit_arity_constructors then
        (* unboxing the inner tupple *)
        match _2 with
          | {pexp_desc=Pexp_tuple [inner]; pexp_loc; pexp_attributes} -> mkexp (Pexp_construct(_1, Some inner))
          | _ -> assert false
      else
        mkExplicitArityTupleExp (Pexp_construct(_1, Some _2))
    )
# 18474 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1084 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_comma_seq -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1084 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1084

and _menhir_goto_expr_comma_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_expr_comma_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState798 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7071 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1074
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1076 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1074
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7069 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1074 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce214 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7070)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1074)) : 'freshtv7072)
    | MenhirState15 | MenhirState771 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7075 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1144
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1144
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7073 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1144 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce210 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7074)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1144)) : 'freshtv7076)
    | MenhirState755 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7079 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2177
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1076 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2177
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv7077 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2177 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce214 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7078)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2177)) : 'freshtv7080)
    | MenhirState184 | MenhirState408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7083 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2379
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2379
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv7081 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2379 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce210 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv7082)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2379)) : 'freshtv7084)
    | _ ->
        _menhir_fail ()

and _menhir_reduce807 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_type_constraint_right_of_colon -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_type_constraint_right_of_colon) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_constraint = 
# 3250 "reason_parser.mly"
                                                ( _2 )
# 18660 "reason_parser.ml"
     in
    _menhir_goto_type_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2331 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7067 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce223 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7068)

and _menhir_reduce222 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_record_expr)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2717 "reason_parser.mly"
      ( let (exten, fields) = _2 in mkexp (Pexp_record(fields, exten)) )
# 18683 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2244 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7065 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce223 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7066)

and _menhir_error2204 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7063 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce227 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7064)

and _menhir_reduce226 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_record_expr)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 18722 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2726 "reason_parser.mly"
      ( let (exten, fields) = _4 in
        let loc = mklocation _symbolstartpos _endpos in
        let rec_exp = mkexp ~loc (Pexp_record (fields, exten)) in
        mkexp(Pexp_open(Fresh, _1, rec_exp)) )
# 18734 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1125 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv7061 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce227 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv7062)

and _menhir_run2665 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2665 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2665 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2665

and _menhir_run2673 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1657 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1633 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2673 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState2673
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2673

and _menhir_run2677 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2677 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2677

and _menhir_run2681 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2681

and _menhir_run1789 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 18876 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce365 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run1790 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1545 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1790 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | SEMI | VAL ->
        _menhir_reduce355 _menhir_env (Obj.magic _menhir_stack) MenhirState1790
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1790

and _menhir_run1628 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run202 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run201 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run200 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run199 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_reduce829 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7059 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7060)

and _menhir_run1629 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run207 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run206 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run205 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_reduce830 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7057 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv7058)

and _menhir_run1788 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1790 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1788 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1789 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1788 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1788

and _menhir_run1804 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 18971 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7053 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 18983 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1814) : 'freshtv7054)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7055 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 19026 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1805) : 'freshtv7056)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce823 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce828 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_type_variance = 
# 3545 "reason_parser.mly"
                                                ( Invariant )
# 19069 "reason_parser.ml"
     in
    _menhir_goto_type_variance _menhir_env _menhir_stack _menhir_s _v

and _menhir_run1540 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce829 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run1541 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce830 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run1217 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_typevar_list * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _startpos ->
    let _menhir_stack = (_menhir_stack, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1217

and _menhir_goto_mty_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mty_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2391 | MenhirState2389 | MenhirState2368 | MenhirState2362 | MenhirState2360 | MenhirState2173 | MenhirState2158 | MenhirState2161 | MenhirState2155 | MenhirState1955 | MenhirState1958 | MenhirState1952 | MenhirState801 | MenhirState775 | MenhirState246 | MenhirState129 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7043 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error133 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7041 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run135 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState134) : 'freshtv7042)
          | COLONGREATER | RPAREN ->
              _menhir_reduce646 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error133 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7044)
    | MenhirState499 | MenhirState233 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7047 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error238 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7045 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | TYPE ->
                  _menhir_run240 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState239 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState239) : 'freshtv7046)
          | RPAREN ->
              _menhir_reduce646 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error238 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv7048)
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2125 | MenhirState2122 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1444 | MenhirState1442 | MenhirState1414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7049 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce264 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7050)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState2005 | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7051 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce264 _menhir_env (Obj.magic _menhir_stack) : 'freshtv7052)
    | _ ->
        _menhir_fail ()

and _menhir_goto_name_tag_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_name_tag_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7035 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (((('freshtv7033 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState296 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce151 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7034)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState296) : 'freshtv7036)
    | MenhirState2456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7039 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (((('freshtv7037 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2457 in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce151 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv7038)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2457) : 'freshtv7040)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_ampersand : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_opt_ampersand -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv7031 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2439) : 'freshtv7032)

and _menhir_goto__optional_type_variable_with_variance : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__optional_type_variable_with_variance -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2076 | MenhirState1973 | MenhirState1421 | MenhirState1367 | MenhirState1254 | MenhirState1364 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7027) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__optional_type_variable_with_variance) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7028)
    | MenhirState90 | MenhirState1855 | MenhirState1864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7029) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__optional_type_variable_with_variance) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce642 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv7030)
    | _ ->
        _menhir_fail ()

and _menhir_reduce349 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cty__class_instance_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cty__class_instance_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_instance_type = 
# 2212 "reason_parser.mly"
                                                             (_1)
# 19286 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7025) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_instance_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7011 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1594
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1594
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1594
          | RBRACE | SEMI ->
              _menhir_reduce40 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1594)) : 'freshtv7012)
    | MenhirState1542 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7013 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 19319 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1615
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1615
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce362 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1615)) : 'freshtv7014)
    | MenhirState1767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7015 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1768
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1768
          | RPAREN ->
              _menhir_reduce7 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1768)) : 'freshtv7016)
    | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7017 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1801
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1801
          | AND | EQUAL | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce7 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1801)) : 'freshtv7018)
    | MenhirState2707 | MenhirState1826 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv7021 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1834
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1834
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv7019 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_instance_type), _startpos__1_) = _menhir_stack in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__1_ in
              let _v : 'tv__non_arrowed_class_constructor_type = 
# 2208 "reason_parser.mly"
      ( _1 )
# 19385 "reason_parser.ml"
               in
              _menhir_goto__non_arrowed_class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv7020)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1834)) : 'freshtv7022)
    | MenhirState2056 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7023 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 19397 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2057
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2057
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce362 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2057)) : 'freshtv7024)
    | _ ->
        _menhir_fail ()) : 'freshtv7026)

and _menhir_reduce342 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_expr = 
# 1918 "reason_parser.mly"
                                          (_1)
# 19422 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7009) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6965 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1650
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1650
          | AND | AS | COLON | EOF | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce30 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1650)) : 'freshtv6966)
    | MenhirState1738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6975 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1739
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6973) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1739 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6969 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 19473 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6967 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos__2_ : Lexing.position) = _endpos in
                  let ((_2 : (
# 857 "reason_parser.mly"
       (string)
# 19483 "reason_parser.ml"
                  )) : (
# 857 "reason_parser.mly"
       (string)
# 19487 "reason_parser.ml"
                  )) = _v in
                  let (_startpos__2_ : Lexing.position) = _startpos in
                  ((let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
                  let _1 = () in
                  let _endpos = _endpos__2_ in
                  let _v : 'tv_parent_binder = 
# 2005 "reason_parser.mly"
          ( Some _2 )
# 19496 "reason_parser.ml"
                   in
                  _menhir_goto_parent_binder _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv6968)) : 'freshtv6970)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6971 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6972)) : 'freshtv6974)
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1739
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce652 _menhir_env (Obj.magic _menhir_stack) MenhirState1739
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1739)) : 'freshtv6976)
    | MenhirState2673 | MenhirState1632 | MenhirState1663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6977 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1757
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1757
          | RBRACE ->
              _menhir_reduce19 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1757)) : 'freshtv6978)
    | MenhirState1631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1763
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6979 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1763 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1769 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1767 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1766) : 'freshtv6980)
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1763
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1763 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce52 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6982)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1763)) : 'freshtv6984)
    | MenhirState1630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6985 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 19599 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1786
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1786
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce340 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1786)) : 'freshtv6986)
    | MenhirState1819 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6987 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1820
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1820
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce54 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1820)) : 'freshtv6988)
    | MenhirState1824 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6989 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1825
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1825
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce347 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1825)) : 'freshtv6990)
    | MenhirState1831 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6991 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1832
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1832
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce348 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1832)) : 'freshtv6992)
    | MenhirState2665 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2666
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6993 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2666 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run1769 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run1767 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2669 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2669) : 'freshtv6994)
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2666
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2666 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce52 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6996)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2666)) : 'freshtv6998)
    | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6999 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2690
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2690
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce30 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2690)) : 'freshtv7000)
    | MenhirState2664 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv7001 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 19747 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2697
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2697
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce340 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2697)) : 'freshtv7002)
    | MenhirState2700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7003 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2701
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2701
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce54 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2701)) : 'freshtv7004)
    | MenhirState2705 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv7005 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2706
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2706
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce347 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2706)) : 'freshtv7006)
    | MenhirState2709 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv7007 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2710
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2710
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce348 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2710)) : 'freshtv7008)
    | _ ->
        _menhir_fail ()) : 'freshtv7010)

and _menhir_reduce346 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_fun_def_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_fun_def_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_fun_def = 
# 1882 "reason_parser.mly"
                                               (_1)
# 19819 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6963) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_fun_def) = _v in
    ((match _menhir_s with
    | MenhirState1639 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6955 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce31 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6956)
    | MenhirState1637 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6957 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce12 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6958)
    | MenhirState2679 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6959 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce31 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6960)
    | MenhirState2677 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6961 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_def) = _v in
        (_menhir_reduce12 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6962)
    | _ ->
        _menhir_fail ()) : 'freshtv6964)

and _menhir_reduce225 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_self_pattern_and_structure), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19874 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19886 "reason_parser.ml"
      
    in
    
# 2724 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "{") (with_txt _3 "}") )
# 19892 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce49 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_expr_lets_and_rest)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19912 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 19924 "reason_parser.ml"
      
    in
    
# 1966 "reason_parser.mly"
    ( unclosed_cl (with_txt _1 "{") (with_txt _3 "}") )
# 19930 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce337 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cty__class_constructor_type_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cty__class_constructor_type_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_constructor_type = 
# 2193 "reason_parser.mly"
                                                                   (_1)
# 19940 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6953) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_constructor_type) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1774 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6909 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce10 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6910)
    | MenhirState1772 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6911 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 19958 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce8 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6912)
    | MenhirState1781 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6913 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 19966 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce9 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6914)
    | MenhirState1766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6917 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1783 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6915 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce50 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv6916)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1783 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6918)
    | MenhirState1809 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6919 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce10 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6920)
    | MenhirState1807 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6921 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 19997 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce8 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6922)
    | MenhirState1816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6923 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 20005 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        (_menhir_reduce9 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6924)
    | MenhirState1787 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6927 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1818 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6925 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CLASS ->
                  _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1819 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1819 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1819 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1819 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1819 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1819 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1819 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1819) : 'freshtv6926)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1818 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6928)
    | MenhirState1827 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6933 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1828 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6931 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6929 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_constructor_type)) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__3_ in
              let _v : 'tv__non_arrowed_class_constructor_type = 
# 2209 "reason_parser.mly"
                                         ( _2)
# 20073 "reason_parser.ml"
               in
              _menhir_goto__non_arrowed_class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6930)) : 'freshtv6932)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1828 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6934)
    | MenhirState2063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6943 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 20085 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6941 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 20091 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 20096 "reason_parser.ml"
        )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_constructor_type)) = _menhir_stack in
        let _4 = () in
        let _endpos = _endpos__5_ in
        let _v : 'tv_class_description_details = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 20109 "reason_parser.ml"
          
        in
        
# 2309 "reason_parser.mly"
                                                                                   (
    (_2, _5, _1, List.rev _3)
  )
# 20117 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6939) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_description_details) = _v in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
        match _menhir_s with
        | MenhirState2051 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6935 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKETATAT ->
                _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2065
            | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2065) : 'freshtv6936)
        | MenhirState2092 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6937 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKETATAT ->
                _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2093
            | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2093) : 'freshtv6938)
        | _ ->
            _menhir_fail ()) : 'freshtv6940)) : 'freshtv6942)) : 'freshtv6944)
    | MenhirState2669 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6947 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2670 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6945 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce50 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv6946)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2670 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6948)
    | MenhirState2698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6951 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2699 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6949 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CLASS ->
                  _menhir_run2681 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run2677 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run2673 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2700 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run2665 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2700 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2700) : 'freshtv6950)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2699 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6952)
    | _ ->
        _menhir_fail ()) : 'freshtv6954)

and _menhir_goto_private_virtual_flags : _menhir_env -> 'ttv_tail -> 'tv_private_virtual_flags -> 'ttv_return =
  fun _menhir_env _menhir_stack _v ->
    let _menhir_stack = (_menhir_stack, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv6907 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1573 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1573) : 'freshtv6908)

and _menhir_run1543 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 20227 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce365 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run1544 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1545 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1544 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | SEMI | VAL ->
        _menhir_reduce355 _menhir_env (Obj.magic _menhir_stack) MenhirState1544
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1544

and _menhir_reduce339 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 20251 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters -> Lexing.position -> _menhir_state -> 'tv_constrained_class_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_constrained_class_declaration) ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 20257 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)) = _menhir_stack in
    let _endpos = _endpos__4_ in
    let _v : 'tv_class_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 20269 "reason_parser.ml"
      
    in
    
# 1820 "reason_parser.mly"
    (
       (_2, _4, _1, List.rev _3)
    )
# 20277 "reason_parser.ml"
     in
    _menhir_goto_class_declaration_details _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto__class_fun_binding : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_binding -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1627 | MenhirState1823 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6903) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce524 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6904)
    | MenhirState2663 | MenhirState2704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6905) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_binding) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce524 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6906)
    | _ ->
        _menhir_fail ()

and _menhir_goto_with_type_binder : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_with_type_binder -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6899 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1427 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1427) : 'freshtv6900)
    | MenhirState1974 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6901 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1977 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1977) : 'freshtv6902)
    | _ ->
        _menhir_fail ()

and _menhir_run1269 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 20385 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6897 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 20415 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1269 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1271) : 'freshtv6898)
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1269
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1269

and _menhir_run1284 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1284 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1284
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1284

and _menhir_run1288 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run1289 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1288

and _menhir_run1292 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1292
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1292

and _menhir_run1295 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1295
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1295

and _menhir_run1298 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6881 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1312 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1312
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1312) : 'freshtv6882)
    | FALSE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6883 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1309 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1309
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1309) : 'freshtv6884)
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6889 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6885 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OF ->
                _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
                _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1306
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1306) : 'freshtv6886)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6887 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6888)) : 'freshtv6890)
    | TRUE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6891 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1302
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1302) : 'freshtv6892)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6893 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 20854 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1299 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1299
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1299) : 'freshtv6894)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6895 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6896)

and _menhir_run108 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 20908 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6879 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 20938 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState108 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState118) : 'freshtv6880)
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState108 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState108
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState108

and _menhir_run2475 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2475
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2475

and _menhir_run2479 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run2480 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2479 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2479

and _menhir_run2483 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2483 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2483
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2483

and _menhir_reduce817 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_kind = 
# 3502 "reason_parser.mly"
      ( (Ptype_open, Public, None) )
# 21134 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run2486 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2486
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2486

and _menhir_run2489 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6863 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2503 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2503
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2503) : 'freshtv6864)
    | FALSE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6865 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2500 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2500
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2500) : 'freshtv6866)
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6871 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6867 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OF ->
                _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2497 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
                _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2497
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2497) : 'freshtv6868)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6869 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6870)) : 'freshtv6872)
    | TRUE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6873 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2493 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2493
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2493) : 'freshtv6874)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6875 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 21390 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AND | BAR | CONSTRAINT | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2490
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2490) : 'freshtv6876)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6877 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6878)

and _menhir_reduce794 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_str__structure_item_without_item_extension_sugar_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_structure_item_without_item_extension_sugar = 
# 1400 "reason_parser.mly"
                                                                  (_1)
# 21448 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6861) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1939 | MenhirState1846 | MenhirState1477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6853) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce295 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6854)
    | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6855 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce294 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6856)
    | MenhirState2753 | MenhirState2760 | MenhirState2757 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6857) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce295 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6858)
    | MenhirState2743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6859 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce294 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6860)
    | _ ->
        _menhir_fail ()) : 'freshtv6862)

and _menhir_reduce793 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_str__structure_item_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_str__structure_item_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_structure_item = 
# 1387 "reason_parser.mly"
                                                   (_1)
# 21499 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6851) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_structure_item) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1939 | MenhirState1846 | MenhirState1477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6835 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1845 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6831 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETATATAT ->
                  _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENTPERCENT ->
                  _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EOF | RBRACE | RBRACKET ->
                  _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState1939
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1939) : 'freshtv6832)
          | EOF | RBRACE | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6833 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_) = _menhir_stack in
              let _v : 'tv_structure = 
# 1376 "reason_parser.mly"
                   ( [_1] )
# 21626 "reason_parser.ml"
               in
              _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6834)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1845 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6836)
    | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6841 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2723 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6839 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6837 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__2_ : Lexing.position) = _endpos in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : 'tv__toplevel_phrase = 
# 1124 "reason_parser.mly"
                                        ( Ptop_def [_1])
# 21653 "reason_parser.ml"
               in
              _menhir_goto__toplevel_phrase _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6838)) : 'freshtv6840)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2723 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6842)
    | MenhirState2753 | MenhirState2760 | MenhirState2757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6849 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2759 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6845 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6843 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : 'tv__use_file = 
# 1134 "reason_parser.mly"
                                                   ( [Ptop_def[_1]] )
# 21678 "reason_parser.ml"
               in
              _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6844)) : 'freshtv6846)
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6847 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run2660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EOF ->
                  _menhir_run2754 _menhir_env (Obj.magic _menhir_stack) MenhirState2760
              | EXCEPTION ->
                  _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run2647 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run2644 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETATATAT ->
                  _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENTPERCENT ->
                  _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run2635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run2628 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run2619 _menhir_env (Obj.magic _menhir_stack) MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2760 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2760) : 'freshtv6848)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2759 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6850)
    | _ ->
        _menhir_fail ()) : 'freshtv6852)

and _menhir_reduce406 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__curried_binding_return_typed_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_curried_binding_return_typed = 
# 3007 "reason_parser.mly"
                                                                               (_1)
# 21798 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6829) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_curried_binding_return_typed) = _v in
    ((match _menhir_s with
    | MenhirState1184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6825 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed) = _v in
        (_menhir_reduce499 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6826)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6827 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed) = _v in
        (_menhir_reduce499 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6828)
    | _ ->
        _menhir_fail ()) : 'freshtv6830)

and _menhir_goto_clty_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_clty_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1767 | MenhirState1542 | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6821 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1592
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce32 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1592)) : 'freshtv6822)
    | MenhirState2707 | MenhirState2056 | MenhirState1826 | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6823 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1799
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce32 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1799)) : 'freshtv6824)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mod_ext2 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mod_ext2 -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2707 | MenhirState2698 | MenhirState2682 | MenhirState2683 | MenhirState2669 | MenhirState2649 | MenhirState2631 | MenhirState2584 | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState2570 | MenhirState76 | MenhirState2521 | MenhirState106 | MenhirState2478 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2479 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState118 | MenhirState121 | MenhirState2472 | MenhirState128 | MenhirState129 | MenhirState141 | MenhirState143 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState158 | MenhirState161 | MenhirState162 | MenhirState2449 | MenhirState163 | MenhirState170 | MenhirState2442 | MenhirState2439 | MenhirState233 | MenhirState242 | MenhirState324 | MenhirState2406 | MenhirState327 | MenhirState401 | MenhirState2391 | MenhirState2389 | MenhirState2368 | MenhirState2362 | MenhirState2360 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2220 | MenhirState2173 | MenhirState2168 | MenhirState2158 | MenhirState2161 | MenhirState2155 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState1248 | MenhirState2113 | MenhirState2116 | MenhirState1250 | MenhirState1407 | MenhirState2063 | MenhirState2056 | MenhirState2041 | MenhirState2036 | MenhirState2027 | MenhirState1965 | MenhirState2005 | MenhirState1994 | MenhirState1990 | MenhirState1984 | MenhirState1977 | MenhirState1975 | MenhirState1955 | MenhirState1958 | MenhirState1952 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1861 | MenhirState1859 | MenhirState1826 | MenhirState1827 | MenhirState1787 | MenhirState1814 | MenhirState1816 | MenhirState1805 | MenhirState1807 | MenhirState1809 | MenhirState1799 | MenhirState1800 | MenhirState1788 | MenhirState1766 | MenhirState1779 | MenhirState1781 | MenhirState1770 | MenhirState1772 | MenhirState1774 | MenhirState1767 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1642 | MenhirState1643 | MenhirState1542 | MenhirState1601 | MenhirState1599 | MenhirState1592 | MenhirState1593 | MenhirState1586 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1545 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1490 | MenhirState1500 | MenhirState1501 | MenhirState1498 | MenhirState1480 | MenhirState1471 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 | MenhirState1435 | MenhirState1432 | MenhirState1427 | MenhirState1425 | MenhirState1253 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1361 | MenhirState1359 | MenhirState1345 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1288 | MenhirState1289 | MenhirState1284 | MenhirState1269 | MenhirState1277 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState801 | MenhirState775 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState666 | MenhirState499 | MenhirState328 | MenhirState329 | MenhirState244 | MenhirState317 | MenhirState306 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState272 | MenhirState270 | MenhirState245 | MenhirState254 | MenhirState255 | MenhirState246 | MenhirState212 | MenhirState193 | MenhirState78 | MenhirState87 | MenhirState83 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6815 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error86 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6813 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState87 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState87) : 'freshtv6814)
          | AND | DOT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce575 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error86 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6816)
    | MenhirState1998 | MenhirState1982 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6819 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1993 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6817 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1994) : 'freshtv6818)
          | AND | DOT | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce575 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1993 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6820)
    | _ ->
        _menhir_fail ()

and _menhir_goto_with_constraints : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6807 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1437 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6805 * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | MODULE ->
                  _menhir_run1430 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run1420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1438 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1438) : 'freshtv6806)
          | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce125 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1437 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6808)
    | MenhirState1971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6811 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2000 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6809 * Lexing.position * _menhir_state * 'tv_with_constraints) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | MODULE ->
                  _menhir_run1980 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run1972 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2001) : 'freshtv6810)
          | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce125 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2000 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6812)
    | _ ->
        _menhir_fail ()

and _menhir_reduce586 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mty__module_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mty__module_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_type = 
# 1568 "reason_parser.mly"
                                             (_1)
# 22051 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6803) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6723 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 22066 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1418
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState1418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1418
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6721 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 22082 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1418 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COLON ->
                  _menhir_run1442 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1441 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1441 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1441) : 'freshtv6722)
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1418)) : 'freshtv6724)
    | MenhirState1442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6727 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1443
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState1443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1443
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6725 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__2_ in
              let _v : 'tv__module_declaration = 
# 1748 "reason_parser.mly"
      ( _2 )
# 22129 "reason_parser.ml"
               in
              _menhir_goto__module_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6726)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1443)) : 'freshtv6728)
    | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6729 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1445
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState1445 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1445
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1445 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce128 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1445)) : 'freshtv6730)
    | MenhirState1466 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6739 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 22161 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1467
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState1467 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1467
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1467 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | LBRACKETATAT | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6737 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 22179 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 22184 "reason_parser.ml"
              )), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_type), _startpos__3_) = _menhir_stack in
              let _2 = () in
              let _endpos = _endpos__3_ in
              let _v : 'tv_module_rec_declaration_details = let _1 =
                let _endpos_x_ = _endpos_x0_ in
                let _startpos_x_ = _startpos_x0_ in
                let x = x0 in
                let _endpos = _endpos_x_ in
                let _symbolstartpos = _startpos_x_ in
                
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22197 "reason_parser.ml"
                
              in
              
# 1756 "reason_parser.mly"
                                     ( (_1, _3) )
# 22203 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6735) = _menhir_stack in
              let (_endpos : Lexing.position) = _endpos in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : 'tv_module_rec_declaration_details) = _v in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
              match _menhir_s with
              | MenhirState1464 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv6731 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKETATAT ->
                      _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1468
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1468) : 'freshtv6732)
              | MenhirState2082 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6733 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACKETATAT ->
                      _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2083
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2083) : 'freshtv6734)
              | _ ->
                  _menhir_fail ()) : 'freshtv6736)) : 'freshtv6738)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1467)) : 'freshtv6740)
    | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6743 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1969
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run2005 _menhir_env (Obj.magic _menhir_stack) MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1969
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1969 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce253 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6742)
          | WITH ->
              _menhir_run1971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1969 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1969)) : 'freshtv6744)
    | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6745 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2006
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run2005 _menhir_env (Obj.magic _menhir_stack) MenhirState2006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2006
          | WITH ->
              _menhir_run1971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce128 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2006)) : 'freshtv6746)
    | MenhirState2036 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6747 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2037
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2037
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2037
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2037) : 'freshtv6748)
    | MenhirState1407 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6749 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2103
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2103
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2103
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2103) : 'freshtv6750)
    | MenhirState1250 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6753 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2109
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run2005 _menhir_env (Obj.magic _menhir_stack) MenhirState2109 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2109
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6751 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2109 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce262 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6752)
          | WITH ->
              _menhir_run1971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2109 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2109)) : 'freshtv6754)
    | MenhirState2113 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6761 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2114
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2114 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2114
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6759 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2114 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6755 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACE ->
                      _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MODULE ->
                      _menhir_run402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run234 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2116 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2116) : 'freshtv6756)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6757 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6758)) : 'freshtv6760)
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2114 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2114)) : 'freshtv6762)
    | MenhirState2116 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6763 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2117
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run2005 _menhir_env (Obj.magic _menhir_stack) MenhirState2117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2117
          | WITH ->
              _menhir_run1971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2117 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce127 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2117)) : 'freshtv6764)
    | MenhirState1248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6767 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2118
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run2005 _menhir_env (Obj.magic _menhir_stack) MenhirState2118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2118
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6765 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2118 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce262 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6766)
          | WITH ->
              _menhir_run1971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2118 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2118)) : 'freshtv6768)
    | MenhirState2122 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6775 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2123
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2123
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6773 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2123 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUALGREATER ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6769 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LBRACE ->
                      _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2125 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2125 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2125 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2125 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MODULE ->
                      _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2125 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2125 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2125) : 'freshtv6770)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6771 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6772)) : 'freshtv6774)
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2123)) : 'freshtv6776)
    | MenhirState2125 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6777 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2126
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2126 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2126
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2126 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | LBRACKETATAT | RPAREN | SEMI ->
              _menhir_reduce127 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2126)) : 'freshtv6778)
    | MenhirState1247 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6781 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2127
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6779 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2127 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2128) : 'freshtv6780)
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2127
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2127 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2127)) : 'freshtv6782)
    | MenhirState2304 | MenhirState2140 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6783 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2144
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2144
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2144 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2144)) : 'freshtv6784)
    | MenhirState2168 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6787 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2169
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run2005 _menhir_env (Obj.magic _menhir_stack) MenhirState2169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2169
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6785 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2169 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce253 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6786)
          | WITH ->
              _menhir_run1971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2169 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2169)) : 'freshtv6788)
    | MenhirState449 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6793 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2289
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2289
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6791 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2289 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__5_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let (_startpos__5_ : Lexing.position) = _startpos in
              ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_functor_arg_name), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_) = _menhir_stack in
              let _5 = () in
              let _3 = () in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_functor_arg = let _2 =
                let _endpos_x_ = _endpos_x0_ in
                let _startpos_x_ = _startpos_x0_ in
                let x = x0 in
                let _endpos = _endpos_x_ in
                let _symbolstartpos = _startpos_x_ in
                
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22654 "reason_parser.ml"
                
              in
              
# 1220 "reason_parser.mly"
      ( _2, Some _4 )
# 22660 "reason_parser.ml"
               in
              _menhir_goto_functor_arg _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv6790)) : 'freshtv6792)
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2289 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2289)) : 'freshtv6794)
    | MenhirState2293 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6797 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2294
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6795 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2294 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2295 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2295 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2295) : 'freshtv6796)
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2294 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2294
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2294 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2294)) : 'freshtv6798)
    | MenhirState401 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6799 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2396
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run2005 _menhir_env (Obj.magic _menhir_stack) MenhirState2396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2396
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2396
          | WITH ->
              _menhir_run1971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2396) : 'freshtv6800)
    | MenhirState2631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6801 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2632
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run1444 _menhir_env (Obj.magic _menhir_stack) MenhirState2632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2632
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2632
          | WITH ->
              _menhir_run1419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2632) : 'freshtv6802)
    | _ ->
        _menhir_fail ()) : 'freshtv6804)

and _menhir_reduce722 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mty__simple_module_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mty__simple_module_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_module_type = 
# 1539 "reason_parser.mly"
                                                           (_1)
# 22757 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6719) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_module_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2125 | MenhirState2122 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1444 | MenhirState1442 | MenhirState1414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6715) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce131 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6716)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState2005 | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6717) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce131 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6718)
    | _ ->
        _menhir_fail ()) : 'freshtv6720)

and _menhir_reduce721 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mod__simple_module_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mod__simple_module_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_module_expr = 
# 1235 "reason_parser.mly"
                                                           (_1)
# 22792 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6713) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_module_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2644 | MenhirState411 | MenhirState451 | MenhirState756 | MenhirState2142 | MenhirState2138 | MenhirState2128 | MenhirState1245 | MenhirState799 | MenhirState772 | MenhirState759 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6705) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce116 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6706)
    | MenhirState2645 | MenhirState2366 | MenhirState2288 | MenhirState2172 | MenhirState761 | MenhirState2143 | MenhirState2139 | MenhirState2129 | MenhirState1246 | MenhirState800 | MenhirState773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6707 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce118 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6708)
    | MenhirState403 | MenhirState404 | MenhirState412 | MenhirState2306 | MenhirState2302 | MenhirState2295 | MenhirState2291 | MenhirState762 | MenhirState1476 | MenhirState1948 | MenhirState1945 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6709) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce116 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6710)
    | MenhirState2395 | MenhirState2393 | MenhirState2364 | MenhirState2307 | MenhirState2303 | MenhirState2296 | MenhirState2292 | MenhirState2165 | MenhirState2025 | MenhirState1947 | MenhirState1962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6711 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce118 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6712)
    | _ ->
        _menhir_fail ()) : 'freshtv6714)

and _menhir_goto_leading_bar_match_cases : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_cases -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1038
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState1038 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce68 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6702)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2235
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce68 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6704)
    | _ ->
        _menhir_fail ()

and _menhir_reduce35 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_class_sig_body)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_instance_type = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22883 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22895 "reason_parser.ml"
      
    in
    
# 2221 "reason_parser.mly"
      ( unclosed_cty (with_txt _1 "{") (with_txt _3 "}") )
# 22901 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__semi_terminated_seq_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1161 | MenhirState1167 | MenhirState1176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6697) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce534 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6698)
    | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState475 | MenhirState2253 | MenhirState2259 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6699) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce534 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6700)
    | _ ->
        _menhir_fail ()

and _menhir_goto__semi_terminated_seq_expr_row : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__semi_terminated_seq_expr_row -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1161 | MenhirState1167 | MenhirState1171 | MenhirState1176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6693) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce535 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6694)
    | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState475 | MenhirState2253 | MenhirState2255 | MenhirState2259 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6695) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__semi_terminated_seq_expr_row) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce535 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6696)
    | _ ->
        _menhir_fail ()

and _menhir_reduce280 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list_extension), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22965 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 22977 "reason_parser.ml"
      
    in
    
# 3361 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "[") (with_txt _4 "]") )
# 22983 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce283 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 23003 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 23015 "reason_parser.ml"
      
    in
    
# 3368 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "[|") (with_txt _4 "|]") )
# 23021 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_module_binding_body : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 23034 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2132
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2132
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2132) : 'freshtv6686)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6687 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 23050 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce603 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6688)
    | MenhirState2323 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 23058 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2324
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2324
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2324) : 'freshtv6690)
    | MenhirState2637 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6691 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 23074 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce603 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6692)
    | _ ->
        _menhir_fail ()

and _menhir_reduce582 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mod__module_binding_body_functor_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mod__module_binding_body_functor_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_binding_body_functor = 
# 1474 "reason_parser.mly"
                                                                             (_1)
# 23086 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6683) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_binding_body_functor) = _v in
    ((match _menhir_s with
    | MenhirState2637 | MenhirState2323 | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6679) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_functor) = _v in
        (_menhir_reduce579 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6680)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6681) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_functor) = _v in
        (_menhir_reduce579 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6682)
    | _ ->
        _menhir_fail ()) : 'freshtv6684)

and _menhir_run426 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_field_expr_list -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run421 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState426 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATERRBRACE ->
        _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState426

and _menhir_reduce584 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mod__module_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mod__module_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_expr = 
# 1271 "reason_parser.mly"
                                             (_1)
# 23133 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6677) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_module_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState759 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6621 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState761
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState761 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState761
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState761 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState761 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState761 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | COLON | EOF | EQUAL | EQUALGREATER | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce117 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState761)) : 'freshtv6622)
    | MenhirState772 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6627 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState773
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6623 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState773 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState775 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState775 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState775) : 'freshtv6624)
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState773
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState773 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6625 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState773 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce245 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6626)
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState773 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState773)) : 'freshtv6628)
    | MenhirState799 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6631 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState800
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6629 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState800 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState801 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState801 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState801) : 'freshtv6630)
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState800
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState800 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState800)) : 'freshtv6632)
    | MenhirState1245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6633 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1246
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1246
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1246 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce580 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1246)) : 'freshtv6634)
    | MenhirState1945 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6635 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1947
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1947
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1947 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | COLON | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce117 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1947)) : 'freshtv6636)
    | MenhirState1948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6637 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1962
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1962
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1962 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1962 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1962)) : 'freshtv6638)
    | MenhirState1476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6639 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2025
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2025
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2025
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2025 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2025) : 'freshtv6640)
    | MenhirState2128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6641 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2129
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2129
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2129 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce581 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2129)) : 'freshtv6642)
    | MenhirState2138 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6643 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2139
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2139
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2139 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce111 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2139)) : 'freshtv6644)
    | MenhirState2142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6645 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2143
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2143
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2143 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce112 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2143)) : 'freshtv6646)
    | MenhirState762 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6647 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2165
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run2168 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2165 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2165 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2165
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2165 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2165 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2167 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2165 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2165 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2165)) : 'freshtv6648)
    | MenhirState756 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6651 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2172
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6649 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2172 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2173 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2173) : 'freshtv6650)
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2172 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2172
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2172 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2172 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2172 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2172)) : 'freshtv6652)
    | MenhirState451 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6653 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2288
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2288
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2288 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2288 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce132 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2288)) : 'freshtv6654)
    | MenhirState2291 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6655 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2292
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2292
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2292 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce580 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2292)) : 'freshtv6656)
    | MenhirState2295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2296
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2296
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2296 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2296 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce581 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2296)) : 'freshtv6658)
    | MenhirState2302 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6659 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2303
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2303
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2303 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce111 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2303)) : 'freshtv6660)
    | MenhirState2306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6661 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2307
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2307
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2307 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce112 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2307)) : 'freshtv6662)
    | MenhirState412 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6663 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2364
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run2168 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2364
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2167 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2364 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2364 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2364)) : 'freshtv6664)
    | MenhirState411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6669 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2366
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6665 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2366 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2368 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2368) : 'freshtv6666)
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2366 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2366
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2366 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2366 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6667 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2366 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce245 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6668)
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2366 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2366)) : 'freshtv6670)
    | MenhirState404 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6671 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2393
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run1965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2393
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1964 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2393 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2393 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2393)) : 'freshtv6672)
    | MenhirState403 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6673 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2395
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2395
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2395 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2395 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce132 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2395)) : 'freshtv6674)
    | MenhirState2644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6675 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2645
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACE ->
              _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2645 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2645
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2645
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2645 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run762 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2645 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2645 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2645) : 'freshtv6676)
    | _ ->
        _menhir_fail ()) : 'freshtv6678)

and _menhir_reduce597 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_mty__non_arrowed_module_type_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_mty__non_arrowed_module_type_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_non_arrowed_module_type = 
# 1531 "reason_parser.mly"
                                                                     (_1)
# 23765 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6619) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_arrowed_module_type) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2631 | MenhirState2293 | MenhirState449 | MenhirState1247 | MenhirState2125 | MenhirState2122 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1444 | MenhirState1442 | MenhirState1414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6607 * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce126 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6608)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState2005 | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6609 * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce126 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6610)
    | MenhirState2140 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6613 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2141 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6611 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2142) : 'freshtv6612)
          | LBRACKETAT | WITH ->
              _menhir_reduce126 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2141 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6614)
    | MenhirState2304 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6617 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2305 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6615 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) = Obj.magic _menhir_stack in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | FUN ->
                  _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2306) : 'freshtv6616)
          | LBRACKETAT | WITH ->
              _menhir_reduce126 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2305 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv6618)
    | _ ->
        _menhir_fail ()) : 'freshtv6620)

and _menhir_reduce400 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_typ__core_type2_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_typ__core_type2_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_core_type2 = 
# 3874 "reason_parser.mly"
                                           (_1)
# 23865 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6605) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_core_type2) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState666 | MenhirState245 | MenhirState270 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6583 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState271
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6581 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState271 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | QUOTE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6577 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState277 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState277) : 'freshtv6578)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6579 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6580)) : 'freshtv6582)
          | EQUALGREATER ->
              _menhir_run272 _menhir_env (Obj.magic _menhir_stack) MenhirState271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce55 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6584)
    | MenhirState272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6585 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState273
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run272 _menhir_env (Obj.magic _menhir_stack) MenhirState273 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce60 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6586)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6593 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState686
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6591 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState686 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | QUOTE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6587 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState692 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState692 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState692) : 'freshtv6588)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv6589 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6590)) : 'freshtv6592)
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce55 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6594)
    | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6595 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState688
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState688 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce60 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6596)
    | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv6597 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 23994 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2405
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState2405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce58 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6598)
    | MenhirState2406 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6599 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 24010 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2407
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run687 _menhir_env (Obj.magic _menhir_stack) MenhirState2407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce59 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6600)
    | MenhirState2463 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv6601 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 24026 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2464
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run272 _menhir_env (Obj.magic _menhir_stack) MenhirState2464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce58 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6602)
    | MenhirState2465 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6603 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 24042 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2466
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUALGREATER ->
              _menhir_run272 _menhir_env (Obj.magic _menhir_stack) MenhirState2466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce59 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6604)
    | _ ->
        _menhir_fail ()) : 'freshtv6606)

and _menhir_reduce650 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position -> _menhir_state -> 'tv_package_type_cstrs -> 'ttv_return =
  fun _menhir_env _menhir_stack _ (_3 : 'tv_package_type_cstrs) ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_package_type_cstr)), _endpos__2_, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_package_type_cstrs = 
# 3984 "reason_parser.mly"
                                             ( _1::_3 )
# 24063 "reason_parser.ml"
     in
    _menhir_goto_package_type_cstrs _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce647 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position -> _menhir_state -> 'tv_package_type_cstrs -> 'ttv_return =
  fun _menhir_env _menhir_stack _ (_3 : 'tv_package_type_cstrs) ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mty_longident), _startpos_x0_), _endpos__2_, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_package_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24080 "reason_parser.ml"
      
    in
    
# 3977 "reason_parser.mly"
                                                  ( (_1, _3) )
# 24086 "reason_parser.ml"
     in
    _menhir_goto_package_type _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce667 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_pat__pattern_without_or_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_pat__pattern_without_or_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern_without_or = 
# 3260 "reason_parser.mly"
                                                           (_1)
# 24097 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6575) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_pattern_without_or) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6547 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState548
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState548
          | AS | BAR | BARRBRACKET | COLON | COLONCOLON | COMMA | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce173 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState548)) : 'freshtv6548)
    | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6549 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState573
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState573 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState573
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce656 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6550)
    | MenhirState574 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6551 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState576
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState576
          | AS | BAR | BARRBRACKET | COLON | COMMA | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce168 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState576)) : 'freshtv6552)
    | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6553 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState644
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState644
          | AS | BAR | COLON | COLONCOLON | COMMA | EOF | EQUAL | EQUALGREATER | IN | RBRACKET | RPAREN | SEMI | WHEN ->
              _menhir_reduce173 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState644)) : 'freshtv6554)
    | MenhirState2578 | MenhirState191 | MenhirState2314 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState738 | MenhirState676 | MenhirState668 | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6555 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState654
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run658 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState654
          | AMPERAMPER | AMPERSAND | AND | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce656 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6556)
    | MenhirState655 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6557 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState657
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState657
          | AS | BAR | COLON | COMMA | EOF | EQUAL | EQUALGREATER | IN | RBRACKET | RPAREN | SEMI | WHEN ->
              _menhir_reduce168 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState657)) : 'freshtv6558)
    | MenhirState711 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6561 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState712
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run658 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState712 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6559 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState712 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState713) : 'freshtv6560)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState712
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState712)) : 'freshtv6562)
    | MenhirState713 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6565 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState714
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState714 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState714
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6563 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState714 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce170 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6564)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState714)) : 'freshtv6566)
    | MenhirState2415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6569 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2416
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run658 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run655 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6567 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2416 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2417) : 'freshtv6568)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2416
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2416)) : 'freshtv6570)
    | MenhirState2417 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6573 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2418
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AS ->
              _menhir_run577 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONCOLON ->
              _menhir_run574 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2418
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6571 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2418 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce170 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6572)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2418)) : 'freshtv6574)
    | _ ->
        _menhir_fail ()) : 'freshtv6576)

and _menhir_reduce229 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_expr_comma_seq)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24432 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24444 "reason_parser.ml"
      
    in
    
# 2735 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "[|") (with_txt _4 "|]") )
# 24450 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce242 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _, (_4 : 'tv_field_expr_list)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _6 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24471 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24483 "reason_parser.ml"
      
    in
    
# 2786 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "{<") (with_txt _6 ">}") )
# 24489 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_expr_comma_seq_extension : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_expr_comma_seq_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1103 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6535 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce418 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6536)
    | MenhirState1098 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6537 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce234 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6538)
    | MenhirState2192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6539 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce418 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6540)
    | MenhirState2187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6541 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce234 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6542)
    | MenhirState418 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6543 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce233 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6544)
    | MenhirState416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6545 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_expr_comma_seq_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce233 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6546)
    | _ ->
        _menhir_fail ()

and _menhir_reduce232 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _, (_4 : 'tv_expr_comma_seq)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _6 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24564 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24576 "reason_parser.ml"
      
    in
    
# 2745 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "[|") (with_txt _6 "|]") )
# 24582 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce239 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_field_expr_list)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24602 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 24614 "reason_parser.ml"
      
    in
    
# 2776 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "{<") (with_txt _4 ">}" ) )
# 24620 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce145 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_tag_field)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3953 "reason_parser.mly"
      ( mktyp(Ptyp_variant([_2], Closed, None)) )
# 24634 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce691 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_tag_field -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_tag_field)) = _menhir_stack in
    let _v : 'tv_row_field = 
# 3991 "reason_parser.mly"
                                                ( _1 )
# 24644 "reason_parser.ml"
     in
    _menhir_goto_row_field _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_extension_constructor_rebind : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension_constructor_rebind -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1489 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6523 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1530
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1530
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1530) : 'freshtv6524)
    | MenhirState2540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6525 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce785 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6526)
    | MenhirState2538 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6527 * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce783 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6528)
    | MenhirState2613 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6529 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce785 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6530)
    | MenhirState2588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6531 * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce783 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6532)
    | MenhirState2654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6533 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2655
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2655
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2655) : 'freshtv6534)
    | _ ->
        _menhir_fail ()

and _menhir_goto_extension_constructor_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_extension_constructor_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1393 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6503 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6501 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_sig_extension_constructors)), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_extension_constructor_declaration)) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_sig_extension_constructors = 
# 3673 "reason_parser.mly"
      ( _3 :: _1 )
# 24710 "reason_parser.ml"
         in
        _menhir_goto_sig_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv6502)) : 'freshtv6504)
    | MenhirState1375 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6507 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6505 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_extension_constructor_declaration)) = _menhir_stack in
        let _endpos = _endpos__1_ in
        let _v : 'tv_sig_extension_constructors = 
# 3671 "reason_parser.mly"
                                                          ( [_1] )
# 24723 "reason_parser.ml"
         in
        _menhir_goto_sig_extension_constructors _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv6506)) : 'freshtv6508)
    | MenhirState1489 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6509 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1532
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1532
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1532) : 'freshtv6510)
    | MenhirState2048 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6511 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2049
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2049
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2049) : 'freshtv6512)
    | MenhirState2540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6513 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6514)
    | MenhirState2538 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6515 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce782 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6516)
    | MenhirState2613 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6517 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce784 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6518)
    | MenhirState2588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6519 * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce782 _menhir_env (Obj.magic _menhir_stack) : 'freshtv6520)
    | MenhirState2654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6521 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2657
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2657
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2657) : 'freshtv6522)
    | _ ->
        _menhir_fail ()

and _menhir_run1338 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_label_declarations -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MUTABLE ->
        _menhir_run1325 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1338 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)
    | LIDENT _ ->
        _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState1338
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1338

and _menhir_goto_constructor_declaration_leading_bar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1320 | MenhirState1316 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6493 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce398 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6494)
    | MenhirState1268 | MenhirState1344 | MenhirState1346 | MenhirState1287 | MenhirState1319 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6495) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce397 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6496)
    | MenhirState2511 | MenhirState2507 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6497 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce398 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6498)
    | MenhirState106 | MenhirState2520 | MenhirState2522 | MenhirState2478 | MenhirState2510 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6499) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declaration_leading_bar) = _v in
        (_menhir_reduce397 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv6500)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constructor_declaration_no_leading_bar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declaration_no_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1268 | MenhirState1344 | MenhirState1346 | MenhirState1287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6489 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1319
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run1298 _menhir_env (Obj.magic _menhir_stack) MenhirState1319 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce395 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1319)) : 'freshtv6490)
    | MenhirState106 | MenhirState2520 | MenhirState2522 | MenhirState2478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6491 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2510
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run2489 _menhir_env (Obj.magic _menhir_stack) MenhirState2510 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce395 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2510)) : 'freshtv6492)
    | _ ->
        _menhir_fail ()

and _menhir_reduce300 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_type_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_str_type_extension) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1415 "reason_parser.mly"
                       (
      mkstr(Pstr_typext _1)
    )
# 24880 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_nonlocal_module_bindings : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_nonlocal_module_bindings -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6483 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1870
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6481) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1870 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1871) : 'freshtv6482)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce303 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1870)) : 'freshtv6484)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6487 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2731
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6485) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2731 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run2637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2732 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2732) : 'freshtv6486)
          | EOF | SEMI ->
              _menhir_reduce303 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2731)) : 'freshtv6488)
    | _ ->
        _menhir_fail ()

and _menhir_reduce510 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_class_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_class_declaration) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_class_declarations = 
# 1787 "reason_parser.mly"
                                                  (
    _2::_1
  )
# 24961 "reason_parser.ml"
     in
    _menhir_goto_many_class_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce514 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_class_type_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_class_type_declaration) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_class_type_declarations = 
# 2320 "reason_parser.mly"
                                                            (
    _2::_1
  )
# 24975 "reason_parser.ml"
     in
    _menhir_goto_many_class_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce518 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_nonlocal_module_bindings -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_nonlocal_module_bindings) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_nonlocal_module_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_nonlocal_module_bindings = 
# 1495 "reason_parser.mly"
                                                               (
    _2::_1
  )
# 24989 "reason_parser.ml"
     in
    _menhir_goto_many_nonlocal_module_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce520 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_type_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_type_declaration) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_many_type_declarations = 
# 3467 "reason_parser.mly"
                                                ( _2 :: _1 )
# 25001 "reason_parser.ml"
     in
    _menhir_goto_many_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_class_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_class_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6475 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1880
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6473) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1880 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | VIRTUAL ->
                  _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _ ->
                  _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState1881
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1881) : 'freshtv6474)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce307 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1880)) : 'freshtv6476)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6479 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2737
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6477) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2737 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | VIRTUAL ->
                  _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _ ->
                  _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState2738
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2738) : 'freshtv6478)
          | EOF | SEMI ->
              _menhir_reduce307 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2737)) : 'freshtv6480)
    | _ ->
        _menhir_fail ()

and _menhir_reduce696 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_field -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_field)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_semi_terminated_class_fields = 
# 1999 "reason_parser.mly"
                          ( [_1] )
# 25083 "reason_parser.ml"
     in
    _menhir_goto_semi_terminated_class_fields _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1752 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_field -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce696 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce497 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_item_extension_sugar * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_let_binding_impl -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_let_binding_impl) _startpos__2_ ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_item_extension_sugar), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_let_binding = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2899 "reason_parser.mly"
                                          (
      let (rec_flag, body, item_attrs) = _2 in
      let (ext_attrs, ext_id) = _1 in
      let loc = mklocation _symbolstartpos _endpos in
      mklbs (ext_attrs, Some ext_id) rec_flag (mklb body item_attrs loc) loc
    )
# 25106 "reason_parser.ml"
     in
    _menhir_goto_let_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce496 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding_impl -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_let_binding_impl) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_let_binding = let _endpos = _endpos__1_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2889 "reason_parser.mly"
                     (
    let (rec_flag, body, item_attrs) = _1 in
    let loc = mklocation _symbolstartpos _endpos in
    mklbs ([], None) rec_flag (mklb body item_attrs loc) loc
  )
# 25123 "reason_parser.ml"
     in
    _menhir_goto_let_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_class_type_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_class_type_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6465 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1875
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6463) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1875 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | VIRTUAL ->
                  _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1876 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _ ->
                  _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState1876
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1876) : 'freshtv6464)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce308 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1875)) : 'freshtv6466)
    | MenhirState2563 | MenhirState2565 | MenhirState1251 | MenhirState1408 | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6469 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2086
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2087 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6467 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_type_declarations), _startpos__1_) = _menhir_stack in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__1_ in
              let _v : 'tv__signature_item = 
# 1726 "reason_parser.mly"
                                 (
      mksig(Psig_class_type (List.rev _1))
    )
# 25184 "reason_parser.ml"
               in
              _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6468)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2086)) : 'freshtv6470)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6471 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2736
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2087 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2736 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | SEMI ->
              _menhir_reduce308 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2736)) : 'freshtv6472)
    | _ ->
        _menhir_fail ()

and _menhir_goto_str_exception_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_str_exception_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6459) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce301 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6460)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6461) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce301 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6462)
    | _ ->
        _menhir_fail ()

and _menhir_reduce306 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_open_statement -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_open_statement) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1440 "reason_parser.mly"
                   (
      mkstr(Pstr_open _1)
    )
# 25241 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_type_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_type_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6451 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1853
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6449) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1853 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv6447) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1854 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 25275 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | MINUS ->
                      _menhir_run100 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run96 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run92 _menhir_env (Obj.magic _menhir_stack) MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run91 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1855 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState1855) : 'freshtv6448)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1854) : 'freshtv6450)
          | EOF | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce299 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1853)) : 'freshtv6452)
    | MenhirState2563 | MenhirState2565 | MenhirState1251 | MenhirState1408 | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6455 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2074
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2075 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2074 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6453 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_type_declarations), _startpos__1_) = _menhir_stack in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__1_ in
              let _v : 'tv__signature_item = 
# 1679 "reason_parser.mly"
                           (
      mksig(Psig_type (List.rev _1))
    )
# 25323 "reason_parser.ml"
               in
              _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv6454)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2074)) : 'freshtv6456)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6457 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2730
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AND ->
              _menhir_run2075 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | SEMI ->
              _menhir_reduce299 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2730)) : 'freshtv6458)
    | _ ->
        _menhir_fail ()

and _menhir_reduce506 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_and_let_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_and_let_binding) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_let_bindings = 
# 2874 "reason_parser.mly"
                                                ( addlb _1 _2 )
# 25357 "reason_parser.ml"
     in
    _menhir_goto_let_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce411 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_expr = 
# 2501 "reason_parser.mly"
                               (_1)
# 25368 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6445) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState833 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6095 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState834
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState834
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState834 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce106 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6096)
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6097 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState836
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState836 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState836
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce88 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6098)
    | MenhirState837 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6099 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 836 "reason_parser.mly"
       (string)
# 25451 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState838
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState838 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState838
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce83 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6100)
    | MenhirState847 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6101 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState848
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState848
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce101 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6102)
    | MenhirState850 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6103 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState851
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState851
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState851 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce96 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6104)
    | MenhirState852 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6105 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState853
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState853 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState853
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce97 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6106)
    | MenhirState854 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6107 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 835 "reason_parser.mly"
       (string)
# 25515 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState855
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState855 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState855
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce82 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6108)
    | MenhirState856 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6109 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState857
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState857
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState857 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce85 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6110)
    | MenhirState858 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6111 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState859
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState859
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce84 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6112)
    | MenhirState860 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6113 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState861
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState861
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState861 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce91 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6114)
    | MenhirState862 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6115 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState863
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState863
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState863 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce87 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6116)
    | MenhirState864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6117 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState865
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState865
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState865 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce86 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6118)
    | MenhirState866 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6119 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState867
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState867
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState867 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce98 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6120)
    | MenhirState868 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6121 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 834 "reason_parser.mly"
       (string)
# 25699 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState869
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState869
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState869 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce81 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6122)
    | MenhirState870 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6123 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 833 "reason_parser.mly"
       (string)
# 25723 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState871
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState871
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState871 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce80 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6124)
    | MenhirState872 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6125 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState873
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState873
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState873 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce99 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6126)
    | MenhirState874 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6127 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState875
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState875
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState875 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce89 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6128)
    | MenhirState876 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6129 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 832 "reason_parser.mly"
       (string)
# 25833 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState877
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState877
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState877 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce79 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6130)
    | MenhirState878 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6131 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState879
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState879
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState879 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce90 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6132)
    | MenhirState880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6133 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState881
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState881
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState881 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce92 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6134)
    | MenhirState882 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6135 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState883
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState883
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState883 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce93 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6136)
    | MenhirState884 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6137 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState885
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState885
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState885 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce94 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6138)
    | MenhirState886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6139 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState887
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState887
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce95 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6140)
    | MenhirState816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6145 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState888
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState888
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6143 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState888 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6141 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState891) : 'freshtv6142)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6144)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState888)) : 'freshtv6146)
    | MenhirState891 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6147 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState892
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState892
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState892 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce103 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6148)
    | MenhirState925 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6153 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState926
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState926
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6151 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState926 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6149 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState929) : 'freshtv6150)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6152)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState926)) : 'freshtv6154)
    | MenhirState946 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6155 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState947
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState947
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState947 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce106 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6156)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6157 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState949
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState949 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState949
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce88 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6158)
    | MenhirState950 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6159 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 836 "reason_parser.mly"
       (string)
# 26535 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState951
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState951 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState951
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce83 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6160)
    | MenhirState960 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6161 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState961
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState961
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce101 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6162)
    | MenhirState963 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6163 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState964
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState964
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState964 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce96 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6164)
    | MenhirState965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6165 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState966
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState966 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState966
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce97 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6166)
    | MenhirState967 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6167 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 835 "reason_parser.mly"
       (string)
# 26599 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState968
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState968 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState968
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce82 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6168)
    | MenhirState969 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6169 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState970
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState970
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState970 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce85 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6170)
    | MenhirState971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6171 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState972
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState972
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState972 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce84 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6172)
    | MenhirState973 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6173 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState974
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState974
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce91 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6174)
    | MenhirState975 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6175 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState976
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState976
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState976 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce87 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6176)
    | MenhirState977 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6177 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState978
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState978
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState978 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce86 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6178)
    | MenhirState979 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6179 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState980
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState980
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState980 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce98 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6180)
    | MenhirState981 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6181 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 834 "reason_parser.mly"
       (string)
# 26783 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState982
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState982
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState982 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce81 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6182)
    | MenhirState983 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6183 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 833 "reason_parser.mly"
       (string)
# 26807 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState984
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState984
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState984 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce80 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6184)
    | MenhirState985 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6185 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState986
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState986
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState986 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce99 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6186)
    | MenhirState987 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6187 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState988
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState988
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState988 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce89 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6188)
    | MenhirState989 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6189 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 832 "reason_parser.mly"
       (string)
# 26917 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState990
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState990
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState990 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce79 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6190)
    | MenhirState991 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6191 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState992
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState992
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState992 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce90 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6192)
    | MenhirState993 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6193 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState994
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState994
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState994 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce92 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6194)
    | MenhirState995 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6195 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState996
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState996
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState996 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce93 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6196)
    | MenhirState997 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6197 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState998
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState998
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState998 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce94 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6198)
    | MenhirState999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6199 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1000
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1000
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1000 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce95 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6200)
    | MenhirState929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6201 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1001
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1001
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1001 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce103 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6202)
    | MenhirState1002 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6207 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1003
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1003
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6205 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1003 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6203 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1006 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1006) : 'freshtv6204)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6206)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1003)) : 'freshtv6208)
    | MenhirState1006 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6209 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1007
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1007
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1007 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce104 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6210)
    | MenhirState1008 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6215 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1009
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1009
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6213 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1009 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6211 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1011 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1011) : 'freshtv6212)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6214)
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1009 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1009)) : 'freshtv6216)
    | MenhirState1011 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6217 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1012
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1012
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1012 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce105 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6218)
    | MenhirState1013 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6221 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1014
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6219 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1014 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1015 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1015) : 'freshtv6220)
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1014
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1014 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1014)) : 'freshtv6222)
    | MenhirState1015 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6223 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1016
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1016
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1016 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce109 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6224)
    | MenhirState1018 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6225 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1019
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1019
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1019 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce102 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6226)
    | MenhirState922 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6227 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1020
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1020
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce100 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6228)
    | MenhirState907 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6229 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1021
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1021
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1021 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce438 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1021)) : 'freshtv6230)
    | MenhirState1024 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6231 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1025
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1025
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1025 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce439 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1025)) : 'freshtv6232)
    | MenhirState1032 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6235 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1033
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6233 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1033 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1034 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1034) : 'freshtv6234)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1033
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1033)) : 'freshtv6236)
    | MenhirState1034 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6237 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1035
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1035
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1035 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce488 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1035)) : 'freshtv6238)
    | MenhirState1036 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6239 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1037
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1037
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1037 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | COLON | COLONGREATER | COMMA | EOF | EQUALGREATER | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce487 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1037)) : 'freshtv6240)
    | MenhirState897 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6241 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1043
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1043
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1043 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce74 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6242)
    | MenhirState893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6245 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1044
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6243 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1044 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1045 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1045) : 'freshtv6244)
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1044
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1044 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1044)) : 'freshtv6246)
    | MenhirState1045 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6247 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1046
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1046
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1046 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce109 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6248)
    | MenhirState1047 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6253 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1048
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1048
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6251 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1048 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6249 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1051) : 'freshtv6250)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6252)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1048)) : 'freshtv6254)
    | MenhirState1051 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6255 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1052
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1052
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce104 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6256)
    | MenhirState1053 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6261 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1054
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1054
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6259 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1054 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | EQUAL ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (((('freshtv6257 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1056) : 'freshtv6258)
              | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                  _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6260)
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1054 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1054)) : 'freshtv6262)
    | MenhirState1056 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6263 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1057
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1057
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1057 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce105 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6264)
    | MenhirState1059 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6265 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1060
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1060
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1060 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce102 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6266)
    | MenhirState813 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6267 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1061
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1061
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce100 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6268)
    | MenhirState416 | MenhirState418 | MenhirState2187 | MenhirState2192 | MenhirState2188 | MenhirState1098 | MenhirState1103 | MenhirState1099 | MenhirState1076 | MenhirState1063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6269 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1065
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1068 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState1065
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1065
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1065 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6270)
    | MenhirState798 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6273 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1078
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1068 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState1078
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1078
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6271 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1078 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce213 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6272)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1078 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6274)
    | MenhirState178 | MenhirState415 | MenhirState2182 | MenhirState1147 | MenhirState1142 | MenhirState1115 | MenhirState1081 | MenhirState1084 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6279 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1086
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6275) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1086 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1089 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1089) : 'freshtv6276)
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6277) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1086 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1087 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1087) : 'freshtv6278)
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1086
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1086 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6280)
    | MenhirState1123 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6281 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1124
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1124
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1124 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA | RBRACE ->
              _menhir_reduce474 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1124)) : 'freshtv6282)
    | MenhirState1134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6283 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1135
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1135
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1135 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              _menhir_reduce602 _menhir_env (Obj.magic _menhir_stack)
          | COMMA ->
              _menhir_reduce474 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1135)) : 'freshtv6284)
    | MenhirState791 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6287 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1137
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6285 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1137 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1138) : 'freshtv6286)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1137
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1137)) : 'freshtv6288)
    | MenhirState1138 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6291 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1139
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1139
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6289 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1139 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce78 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6290)
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1139 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1139)) : 'freshtv6292)
    | MenhirState15 | MenhirState771 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6295 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1149
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1068 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState1149
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1149
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6293 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1149 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce205 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6294)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1149 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6296)
    | MenhirState770 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6299 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1154
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6297 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1154 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1156 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1155 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1155) : 'freshtv6298)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1154
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1154 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1154)) : 'freshtv6300)
    | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1161 | MenhirState1167 | MenhirState1176 | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6301 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1174
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1174
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1174
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1174) : 'freshtv6302)
    | MenhirState1193 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6303 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1194
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1194
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1194 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce403 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6304)
    | MenhirState1197 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6305 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1198
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1198
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1198 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | EOF | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce408 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1198)) : 'freshtv6306)
    | MenhirState1204 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6307 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1205
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1205
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1205 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce61 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6308)
    | MenhirState1221 | MenhirState1213 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6309 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1215
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1215
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1215 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | EOF | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce538 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1215)) : 'freshtv6310)
    | MenhirState1224 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6311 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1225
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1225
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1225 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce498 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6312)
    | MenhirState1234 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6313 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1235
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1235
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1235 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce503 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6314)
    | MenhirState1237 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6315 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1238
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1238
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1238 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce502 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6316)
    | MenhirState1685 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6317 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1686
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1686
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1686 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce842 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1686)) : 'freshtv6318)
    | MenhirState1688 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv6319 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1689
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1689
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1689 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce843 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1689)) : 'freshtv6320)
    | MenhirState1708 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6321 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1709
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1709
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1709 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce403 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6322)
    | MenhirState1712 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6323 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1713
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1713
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1713 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce408 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1713)) : 'freshtv6324)
    | MenhirState1718 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6325 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1719
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1719
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce568 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1719)) : 'freshtv6326)
    | MenhirState1725 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((((('freshtv6327 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1726
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1726
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1726 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce569 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1726)) : 'freshtv6328)
    | MenhirState1728 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6329 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1729
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1729
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1729 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATAT | RBRACE | SEMI ->
              _menhir_reduce567 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1729)) : 'freshtv6330)
    | MenhirState1734 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6331 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1735
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1735
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1735
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1735 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1735) : 'freshtv6332)
    | MenhirState1893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6333 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1894
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1894
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce61 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6334)
    | MenhirState1906 | MenhirState1900 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6335 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1902
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1902
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce538 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1902)) : 'freshtv6336)
    | MenhirState1909 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6337 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1910
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1910
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce498 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6338)
    | MenhirState1919 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6339 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1920
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1920
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce503 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6340)
    | MenhirState1922 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6341 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1923
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1923
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce502 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6342)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6343 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1935
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1935
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1935
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1935) : 'freshtv6344)
    | MenhirState1949 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6349 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1950
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6345 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1950 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1955) : 'freshtv6346)
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6347 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1950 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1952 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1952 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1952) : 'freshtv6348)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1950
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1951 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1950 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1950)) : 'freshtv6350)
    | MenhirState763 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6355 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2153
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6351 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2153 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2158 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2158 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2158) : 'freshtv6352)
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6353 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2153 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2155 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2155) : 'freshtv6354)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2153
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2154 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2153)) : 'freshtv6356)
    | MenhirState755 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6359 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2179
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1068 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState2179
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2179
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6357 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2179 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce213 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6358)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2179 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6360)
    | MenhirState737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6363 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2208
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2208
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6361 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2208 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6362)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2208)) : 'freshtv6364)
    | MenhirState2210 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6367 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2211
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2211
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6365 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2211 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6366)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2211)) : 'freshtv6368)
    | MenhirState2213 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6371 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2214
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2214
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6369 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2214 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6370)
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2214 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2214)) : 'freshtv6372)
    | MenhirState2218 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6373 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2219
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2219
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce438 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2219)) : 'freshtv6374)
    | MenhirState2222 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6375 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2223
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2223
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2223 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce439 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2223)) : 'freshtv6376)
    | MenhirState2229 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6379 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2230
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6377 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2230 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2231 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2231) : 'freshtv6378)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2230
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2230 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2230)) : 'freshtv6380)
    | MenhirState2231 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6381 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2232
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2232
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2232 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce488 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2232)) : 'freshtv6382)
    | MenhirState2233 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6383 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2234
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2234
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | BAR | BARRBRACKET | COLON | COLONGREATER | COMMA | EOF | GREATERRBRACE | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce487 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2234)) : 'freshtv6384)
    | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState475 | MenhirState2253 | MenhirState2259 | MenhirState2255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6385 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2257
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2257
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2257
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2257 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2257) : 'freshtv6386)
    | MenhirState2270 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6387 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2271
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2271
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2271 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce74 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6388)
    | MenhirState473 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6391 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2272
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2272
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6389 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2272 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce217 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6390)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2272)) : 'freshtv6392)
    | MenhirState2274 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6395 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2275
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2275
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6393 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2275 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce219 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6394)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2275)) : 'freshtv6396)
    | MenhirState2277 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6399 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2278
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2278
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6397 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2278 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce221 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6398)
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2278 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2278)) : 'freshtv6400)
    | MenhirState422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6401 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 32225 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2337
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2337
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA | GREATERRBRACE ->
              _menhir_reduce433 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2337)) : 'freshtv6402)
    | MenhirState413 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6407 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2358
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6403 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2358 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2362 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2362 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2362) : 'freshtv6404)
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6405 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2358 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2360 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2360 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2360) : 'freshtv6406)
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2358
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run2154 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2358)) : 'freshtv6408)
    | MenhirState2374 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2375
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv6409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2375 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2376) : 'freshtv6410)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2375
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2375)) : 'freshtv6412)
    | MenhirState2376 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv6415 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2377
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2377
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv6413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2377 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce78 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6414)
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2377 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2377)) : 'freshtv6416)
    | MenhirState184 | MenhirState408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2382
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run1068 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState2382
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2382
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6417 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2382 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce205 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv6418)
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2382 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce419 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6420)
    | MenhirState405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2387
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run884 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run882 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run880 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6421 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2387 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2391 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2391 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2391) : 'freshtv6422)
          | COLONEQUAL ->
              _menhir_run886 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv6423 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2387 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2389 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2389 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2389) : 'freshtv6424)
          | GREATER ->
              _menhir_run878 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run876 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run870 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run868 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run854 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run837 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2387
          | LESS ->
              _menhir_run874 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run872 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run866 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run864 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run862 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run860 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run852 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run858 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run856 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run850 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run893 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RPAREN ->
              _menhir_run1951 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run835 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2387 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2387)) : 'freshtv6426)
    | MenhirState2422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6429 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2423
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2423
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2423 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv6427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _v : 'tv_payload = 
# 4356 "reason_parser.mly"
                               ( PPat (_2, Some _4) )
# 32806 "reason_parser.ml"
               in
              _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6428)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2423)) : 'freshtv6430)
    | MenhirState2574 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6441 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2576
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6439 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2576 in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6437 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_ : _menhir_state) = _menhir_s in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1069 "reason_parser.mly"
      (Parsetree.expression)
# 32841 "reason_parser.ml"
              ) = 
# 1142 "reason_parser.mly"
             ( apply_mapper_chain_to_expr _1 default_mapper_chain)
# 32845 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6435) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1069 "reason_parser.mly"
      (Parsetree.expression)
# 32853 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6433) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1069 "reason_parser.mly"
      (Parsetree.expression)
# 32861 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv6431) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1069 "reason_parser.mly"
      (Parsetree.expression)
# 32869 "reason_parser.ml"
              )) : (
# 1069 "reason_parser.mly"
      (Parsetree.expression)
# 32873 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv6432)) : 'freshtv6434)) : 'freshtv6436)) : 'freshtv6438)) : 'freshtv6440)
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2576
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2576 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2576)) : 'freshtv6442)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6443 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2748
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERAMPER ->
              _menhir_run997 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERSAND ->
              _menhir_run995 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARBAR ->
              _menhir_run993 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONEQUAL ->
              _menhir_run999 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | GREATER ->
              _menhir_run991 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP0 _v ->
              _menhir_run989 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP1 _v ->
              _menhir_run983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP2 _v ->
              _menhir_run981 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP3 _v ->
              _menhir_run967 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INFIXOP4 _v ->
              _menhir_run950 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2748
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2748
          | LESS ->
              _menhir_run987 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run985 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run979 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run977 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run975 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OR ->
              _menhir_run973 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run965 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run971 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run969 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSEQ ->
              _menhir_run963 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run1013 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STAR ->
              _menhir_run948 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | PREFIXOP _ | PRIVATE | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2748) : 'freshtv6444)
    | _ ->
        _menhir_fail ()) : 'freshtv6446)

and _menhir_reduce361 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__class_simple_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__class_simple_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_simple_expr = 
# 1960 "reason_parser.mly"
                                                        (_1)
# 32986 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6093) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_simple_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1632 | MenhirState1738 | MenhirState1663 | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6089 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1647
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run821 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce11 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1647)) : 'freshtv6090)
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6091 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2687
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run934 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2687 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce11 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2687)) : 'freshtv6092)
    | _ ->
        _menhir_fail ()) : 'freshtv6094)

and _menhir_reduce577 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 33114 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 906 "reason_parser.mly"
       (string)
# 33119 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_mod_longident = 
# 4143 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 33128 "reason_parser.ml"
     in
    _menhir_goto_mod_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce155 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3534 "reason_parser.mly"
                                                ( mktyp (Ptyp_any), Invariant )
# 33140 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce157 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3536 "reason_parser.mly"
                                                ( mktyp (Ptyp_any), Covariant )
# 33154 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce159 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3538 "reason_parser.mly"
                                                ( mktyp Ptyp_any, Contravariant )
# 33168 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1633 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | REC ->
        _menhir_run435 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1633 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce686 _menhir_env (Obj.magic _menhir_stack) MenhirState1633
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1633

and _menhir_goto_private_flag : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_private_flag -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6059 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6055 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1325 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState1324
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1324) : 'freshtv6056)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6057 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6058)) : 'freshtv6060)
    | MenhirState1344 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6065 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6061 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1325 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1351 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState1351
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1351) : 'freshtv6062)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6063 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6064)) : 'freshtv6066)
    | MenhirState1372 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6067 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run157 _menhir_env (Obj.magic _menhir_stack) MenhirState1374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState1374
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1374) : 'freshtv6068)
    | MenhirState1692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6069 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1693 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1693) : 'freshtv6070)
    | MenhirState1691 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6071 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1702 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1702) : 'freshtv6072)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6077 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6073 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1325 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2515 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState2515
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2515) : 'freshtv6074)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv6075 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6076)) : 'freshtv6078)
    | MenhirState2520 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv6083 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6079 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MUTABLE ->
                _menhir_run1325 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2527 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState2527
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2527) : 'freshtv6080)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv6081 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6082)) : 'freshtv6084)
    | MenhirState2536 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6085 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run157 _menhir_env (Obj.magic _menhir_stack) MenhirState2537 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2537
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2537) : 'freshtv6086)
    | MenhirState2586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv6087 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run157 _menhir_env (Obj.magic _menhir_stack) MenhirState2587 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
            _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState2587
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2587) : 'freshtv6088)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mutable_flag : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mutable_flag -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2527 | MenhirState2515 | MenhirState1351 | MenhirState1338 | MenhirState1324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6047 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv6043 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 33397 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv6039 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 33409 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | LBRACKET ->
                    _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETGREATER ->
                    _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETLESS ->
                    _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LESS ->
                    _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LESSDOTDOTGREATER ->
                    _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LESSGREATER ->
                    _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | QUOTE ->
                    _menhir_run154 _menhir_env (Obj.magic _menhir_stack) MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1328) : 'freshtv6040)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv6041 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 33454 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6042)) : 'freshtv6044)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv6045 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv6046)) : 'freshtv6048)
    | MenhirState1553 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6049 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1554 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1554) : 'freshtv6050)
    | MenhirState1668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6051 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1669 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1669) : 'freshtv6052)
    | MenhirState1676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6053 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1683 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1683) : 'freshtv6054)
    | _ ->
        _menhir_fail ()

and _menhir_reduce47 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__class_simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 33518 "reason_parser.ml"
      
    in
    
# 1963 "reason_parser.mly"
    ( mkclass(Pcl_constr(_1, [])) )
# 33524 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce237 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 33543 "reason_parser.ml"
      
    in
    
# 2772 "reason_parser.mly"
      ( mkexp (Pexp_new _2) )
# 33549 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__simple_pattern : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6035) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce550 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6036)
    | MenhirState2663 | MenhirState2704 | MenhirState2677 | MenhirState2679 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState476 | MenhirState2238 | MenhirState480 | MenhirState484 | MenhirState2224 | MenhirState486 | MenhirState738 | MenhirState1886 | MenhirState1887 | MenhirState1912 | MenhirState1891 | MenhirState1627 | MenhirState1823 | MenhirState1703 | MenhirState1707 | MenhirState1714 | MenhirState1657 | MenhirState1637 | MenhirState1639 | MenhirState1634 | MenhirState1183 | MenhirState1184 | MenhirState1227 | MenhirState1188 | MenhirState1192 | MenhirState1199 | MenhirState898 | MenhirState1041 | MenhirState902 | MenhirState906 | MenhirState1027 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6037) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce550 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6038)
    | _ ->
        _menhir_fail ()

and _menhir_goto_val_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_val_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState2555 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState180 | MenhirState184 | MenhirState187 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState457 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState2268 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1477 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1648 | MenhirState1647 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState842 | MenhirState840 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState827 | MenhirState820 | MenhirState825 | MenhirState822 | MenhirState817 | MenhirState816 | MenhirState813 | MenhirState798 | MenhirState811 | MenhirState805 | MenhirState747 | MenhirState470 | MenhirState464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6027) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_val_longident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce201 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6028)
    | MenhirState2760 | MenhirState2757 | MenhirState2753 | MenhirState2743 | MenhirState2582 | MenhirState2688 | MenhirState2687 | MenhirState2574 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState2350 | MenhirState417 | MenhirState418 | MenhirState419 | MenhirState2277 | MenhirState474 | MenhirState2229 | MenhirState729 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState955 | MenhirState953 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState940 | MenhirState933 | MenhirState938 | MenhirState935 | MenhirState930 | MenhirState929 | MenhirState922 | MenhirState907 | MenhirState920 | MenhirState913 | MenhirState912 | MenhirState910 | MenhirState895 | MenhirState894 | MenhirState793 | MenhirState764 | MenhirState746 | MenhirState740 | MenhirState734 | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv6029) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_val_longident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce201 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv6030)
    | MenhirState2620 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6033 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_val_longident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6031 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__3_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_val_longident) : 'tv_val_longident) = _v in
        let (_startpos__3_ : Lexing.position) = _startpos in
        ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
        let _1 = () in
        let _v : 'tv_toplevel_directive = 
# 4178 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_ident _3) )
# 33612 "reason_parser.ml"
         in
        _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv6032)) : 'freshtv6034)
    | _ ->
        _menhir_fail ()

and _menhir_error388 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce718 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__simple_expr_ -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__simple_expr_) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_expr = 
# 2635 "reason_parser.mly"
                                             (_1)
# 33630 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv6025) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_expr) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState805 | MenhirState464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5913 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState469
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState469
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState469 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState469 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce236 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5914)
    | MenhirState470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5915 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 895 "reason_parser.mly"
       (string)
# 33661 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState471
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState471
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce244 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5916)
    | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5917 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState733
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState733
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState733 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState733 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce236 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5918)
    | MenhirState734 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5919 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 895 "reason_parser.mly"
       (string)
# 33693 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState735
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState735
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce244 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5920)
    | MenhirState740 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5921 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState741
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState741
          | DOWNTO ->
              _menhir_run745 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState741 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState741 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState741 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TO ->
              _menhir_run742 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState741 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState741)) : 'freshtv5922)
    | MenhirState746 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5923 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState747
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState747
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState747 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState747)) : 'freshtv5924)
    | MenhirState747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5925 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState748
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState748
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState748 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState748 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce77 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5926)
    | MenhirState764 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5929 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState765
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState765
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5927 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState765 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState766 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState766) : 'freshtv5928)
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState765 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState765 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState765)) : 'freshtv5930)
    | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5931 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState794
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState794
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState794 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState794 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce108 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5932)
    | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5933 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState812
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState812
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState812 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState812 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce107 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5934)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState816 | MenhirState813 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5943 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState814
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5941 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState814 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5935 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState815 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1053 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1053) : 'freshtv5936)
              | LBRACKET ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5937 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState815 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1047 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1047) : 'freshtv5938)
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5939 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState815 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState816 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState816) : 'freshtv5940)
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState815 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState815) : 'freshtv5942)
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState814 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState814 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce495 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5944)
    | MenhirState817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5945 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState818
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState818
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState818 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState818 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce73 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5946)
    | MenhirState1647 | MenhirState1648 | MenhirState840 | MenhirState842 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5947 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState823
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState823
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState823 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState823 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce495 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5948)
    | MenhirState894 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5949 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState895
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState895
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState895)) : 'freshtv5950)
    | MenhirState895 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5953 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState896
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState896
          | ELSE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5951 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState896 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState897 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState897) : 'freshtv5952)
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState896 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState896 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce75 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5954)
    | MenhirState910 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5955 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState911
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState911
          | DOWNTO ->
              _menhir_run745 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState911 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TO ->
              _menhir_run742 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState911 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState911)) : 'freshtv5956)
    | MenhirState912 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5957 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState913
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState913
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState913 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState913)) : 'freshtv5958)
    | MenhirState913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5959 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState914
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState914
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState914 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState914 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce77 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5960)
    | MenhirState920 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5961 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState921
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState921
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState921 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce107 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5962)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 | MenhirState2574 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState929 | MenhirState922 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5971 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState923
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5969 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState923 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5963 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState924 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1008 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1008) : 'freshtv5964)
              | LBRACKET ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5965 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState924 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1002 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1002) : 'freshtv5966)
              | LIDENT _v ->
                  _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv5967 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState924 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState925 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState925) : 'freshtv5968)
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState924 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState924) : 'freshtv5970)
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState923 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState923 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce495 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5972)
    | MenhirState930 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5973 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState931
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState931
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState931 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState931 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce73 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5974)
    | MenhirState2687 | MenhirState2688 | MenhirState953 | MenhirState955 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5975 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState936
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState936
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState936 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState936 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce495 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5976)
    | MenhirState729 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5979 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 34747 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2217
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2217
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2217 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2217 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE | CHAR _ | COLON | EQUALGREATER | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5977 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 34765 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 34770 "reason_parser.ml"
              )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_pattern), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_simple_expr), _startpos__5_) = _menhir_stack in
              let _4 = () in
              let _2 = () in
              let _startpos = _startpos__1_ in
              let _v : 'tv_labeled_simple_pattern = 
# 2490 "reason_parser.mly"
      ( ("?" ^ _1, Some _5, _3) )
# 34778 "reason_parser.ml"
               in
              _menhir_goto_labeled_simple_pattern _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv5978)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2217)) : 'freshtv5980)
    | MenhirState474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2268
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2268
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2268)) : 'freshtv5982)
    | MenhirState2268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5985 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2269
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2269
          | ELSE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2269 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2270 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2270) : 'freshtv5984)
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState2269 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2269 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce75 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5986)
    | MenhirState457 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5987 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2281
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2281
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState2281 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2281 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce108 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5988)
    | MenhirState419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2343
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2343
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5989 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2343 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2346 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2346) : 'freshtv5990)
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2343 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2343 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2343 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv5993 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv5991 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce71 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5992)) : 'freshtv5994)) : 'freshtv5996)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2343)) : 'freshtv5998)
    | MenhirState417 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2350
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2350
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2350 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2350)) : 'freshtv6000)
    | MenhirState2350 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6001 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2351
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2351
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2351 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2351 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce76 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6002)
    | MenhirState180 | MenhirState187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6003 * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 35093 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2426
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2426
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState2426 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2426 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce235 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6004)
    | MenhirState177 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6007 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2432
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2432
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6005 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2432 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2433 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2433) : 'freshtv6006)
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2432)) : 'freshtv6008)
    | MenhirState17 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6017 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2548
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2548
          | LBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6009 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2548 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BAR ->
                  _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState2551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2551) : 'freshtv6010)
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2548 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2548 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WITH ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv6015 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2548 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv6013 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  ((let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : (('freshtv6011 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce71 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv6012)) : 'freshtv6014)) : 'freshtv6016)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2548)) : 'freshtv6018)
    | MenhirState6 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6019 * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 35202 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2554
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2554
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2554 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2554 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce235 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6020)
    | MenhirState2 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv6021 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2555
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | DOT ->
              _menhir_run736 _menhir_env (Obj.magic _menhir_stack) MenhirState2555
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run743 _menhir_env (Obj.magic _menhir_stack) MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run734 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2555 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2555)) : 'freshtv6022)
    | MenhirState2555 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv6023 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2556
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run472 _menhir_env (Obj.magic _menhir_stack) MenhirState2556
          | SHARP ->
              _menhir_run749 _menhir_env (Obj.magic _menhir_stack) MenhirState2556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARPOP _v ->
              _menhir_run470 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2556 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce76 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv6024)
    | _ ->
        _menhir_fail ()) : 'freshtv6026)

and _menhir_run1325 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce590 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error1133 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error241 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error140 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_functor_args : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_functor_args -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5895 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5893 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState758 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState759 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState759) : 'freshtv5894)
        | LPAREN ->
            _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState758 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState758) : 'freshtv5896)
    | MenhirState1943 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5899 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5897 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState1944 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1945 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1945 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1945) : 'freshtv5898)
        | LPAREN ->
            _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1944 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1944) : 'freshtv5900)
    | MenhirState2637 | MenhirState2323 | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5905 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5901 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2137 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2140 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2140 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2140 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2140) : 'freshtv5902)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5903 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2137 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2138 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2138) : 'freshtv5904)
        | LPAREN ->
            _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2137 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2137) : 'freshtv5906)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5911 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5907 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2301 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2304 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2304 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2304 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2304 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2304 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2304 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2304) : 'freshtv5908)
        | EQUALGREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5909 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = MenhirState2301 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FUN ->
                _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2302 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2302 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2302) : 'freshtv5910)
        | LPAREN ->
            _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2301 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2301) : 'freshtv5912)
    | _ ->
        _menhir_fail ()

and _menhir_reduce421 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension = 
# 4345 "reason_parser.mly"
                                           ( (_2, _3) )
# 35526 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5891) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_extension) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2698 | MenhirState2682 | MenhirState2683 | MenhirState2669 | MenhirState2649 | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState121 | MenhirState2472 | MenhirState128 | MenhirState141 | MenhirState143 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState158 | MenhirState161 | MenhirState162 | MenhirState2449 | MenhirState163 | MenhirState170 | MenhirState2442 | MenhirState2439 | MenhirState324 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1814 | MenhirState1816 | MenhirState1805 | MenhirState1807 | MenhirState1809 | MenhirState1799 | MenhirState1800 | MenhirState1766 | MenhirState1779 | MenhirState1781 | MenhirState1770 | MenhirState1772 | MenhirState1774 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1361 | MenhirState1359 | MenhirState1345 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1288 | MenhirState1289 | MenhirState1284 | MenhirState1269 | MenhirState1277 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState306 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState272 | MenhirState270 | MenhirState245 | MenhirState254 | MenhirState255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5863) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce153 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5864)
    | MenhirState2521 | MenhirState106 | MenhirState2478 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1642 | MenhirState1643 | MenhirState1601 | MenhirState1592 | MenhirState1593 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1490 | MenhirState1500 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState328 | MenhirState329 | MenhirState244 | MenhirState317 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5865) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce153 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5866)
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5867) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce293 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5868)
    | MenhirState2663 | MenhirState2704 | MenhirState2677 | MenhirState2679 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState476 | MenhirState2238 | MenhirState480 | MenhirState484 | MenhirState2224 | MenhirState486 | MenhirState738 | MenhirState1886 | MenhirState1887 | MenhirState1912 | MenhirState1891 | MenhirState1627 | MenhirState1823 | MenhirState1703 | MenhirState1707 | MenhirState1714 | MenhirState1657 | MenhirState1637 | MenhirState1639 | MenhirState1634 | MenhirState1183 | MenhirState1184 | MenhirState1227 | MenhirState1188 | MenhirState1192 | MenhirState1199 | MenhirState898 | MenhirState1041 | MenhirState902 | MenhirState906 | MenhirState1027 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5869) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce293 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5870)
    | MenhirState2644 | MenhirState2645 | MenhirState411 | MenhirState2366 | MenhirState451 | MenhirState2288 | MenhirState756 | MenhirState2172 | MenhirState759 | MenhirState761 | MenhirState2142 | MenhirState2143 | MenhirState2138 | MenhirState2139 | MenhirState2128 | MenhirState2129 | MenhirState1245 | MenhirState1246 | MenhirState799 | MenhirState800 | MenhirState772 | MenhirState773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5871) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce261 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5872)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState2555 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState180 | MenhirState184 | MenhirState187 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState457 | MenhirState464 | MenhirState470 | MenhirState2274 | MenhirState473 | MenhirState2268 | MenhirState2270 | MenhirState475 | MenhirState2253 | MenhirState2259 | MenhirState2255 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState747 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1647 | MenhirState1648 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState816 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState840 | MenhirState842 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 | MenhirState817 | MenhirState811 | MenhirState805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5873) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce250 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5874)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 | MenhirState2687 | MenhirState2688 | MenhirState2574 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState417 | MenhirState2350 | MenhirState418 | MenhirState419 | MenhirState2277 | MenhirState474 | MenhirState2229 | MenhirState729 | MenhirState730 | MenhirState734 | MenhirState2213 | MenhirState740 | MenhirState746 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState764 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1167 | MenhirState1176 | MenhirState1171 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState793 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState894 | MenhirState895 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState929 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState953 | MenhirState955 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 | MenhirState930 | MenhirState920 | MenhirState910 | MenhirState912 | MenhirState913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5875) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce250 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5876)
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5877) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce267 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5878)
    | MenhirState1767 | MenhirState1542 | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5879) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce37 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5880)
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1632 | MenhirState1738 | MenhirState1663 | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5881) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce17 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5882)
    | MenhirState2707 | MenhirState2056 | MenhirState1826 | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5883) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce37 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5884)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState1250 | MenhirState2116 | MenhirState1965 | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5885) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce267 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5886)
    | MenhirState403 | MenhirState2395 | MenhirState404 | MenhirState2393 | MenhirState412 | MenhirState2364 | MenhirState2306 | MenhirState2307 | MenhirState2302 | MenhirState2303 | MenhirState2295 | MenhirState2296 | MenhirState2291 | MenhirState2292 | MenhirState762 | MenhirState2165 | MenhirState1476 | MenhirState2025 | MenhirState1945 | MenhirState1947 | MenhirState1948 | MenhirState1962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5887) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce261 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5888)
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5889) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce17 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5890)
    | _ ->
        _menhir_fail ()) : 'freshtv5892)

and _menhir_reduce334 : _menhir_env -> (('ttv_tail * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_attribute = 
# 4306 "reason_parser.mly"
                                      ( (_2, _3) )
# 35659 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5861) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_attribute) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2461 | MenhirState325 | MenhirState2221 | MenhirState1815 | MenhirState1806 | MenhirState1808 | MenhirState1780 | MenhirState1771 | MenhirState1773 | MenhirState1711 | MenhirState1196 | MenhirState1023 | MenhirState263 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5829 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce136 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5830)
    | MenhirState2418 | MenhirState714 | MenhirState573 | MenhirState576 | MenhirState548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5831 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce174 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5832)
    | MenhirState2416 | MenhirState712 | MenhirState654 | MenhirState657 | MenhirState644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5833 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce174 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5834)
    | MenhirState2645 | MenhirState2366 | MenhirState2288 | MenhirState2172 | MenhirState761 | MenhirState2143 | MenhirState2139 | MenhirState2129 | MenhirState1246 | MenhirState800 | MenhirState773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5835 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce124 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5836)
    | MenhirState2387 | MenhirState2382 | MenhirState2358 | MenhirState2337 | MenhirState2275 | MenhirState2272 | MenhirState2271 | MenhirState2257 | MenhirState2234 | MenhirState2232 | MenhirState2223 | MenhirState2219 | MenhirState2211 | MenhirState2208 | MenhirState2179 | MenhirState1935 | MenhirState1923 | MenhirState1920 | MenhirState1910 | MenhirState1902 | MenhirState1894 | MenhirState1735 | MenhirState1729 | MenhirState1726 | MenhirState1719 | MenhirState1713 | MenhirState1709 | MenhirState1689 | MenhirState1686 | MenhirState1149 | MenhirState1135 | MenhirState1124 | MenhirState1086 | MenhirState1078 | MenhirState1061 | MenhirState1060 | MenhirState1057 | MenhirState1048 | MenhirState1052 | MenhirState888 | MenhirState1046 | MenhirState1003 | MenhirState926 | MenhirState892 | MenhirState834 | MenhirState887 | MenhirState861 | MenhirState881 | MenhirState883 | MenhirState885 | MenhirState879 | MenhirState877 | MenhirState875 | MenhirState873 | MenhirState867 | MenhirState871 | MenhirState869 | MenhirState865 | MenhirState863 | MenhirState859 | MenhirState857 | MenhirState851 | MenhirState855 | MenhirState853 | MenhirState836 | MenhirState848 | MenhirState838 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5837 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce110 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5838)
    | MenhirState2748 | MenhirState2576 | MenhirState2423 | MenhirState2375 | MenhirState2377 | MenhirState2278 | MenhirState2230 | MenhirState2214 | MenhirState2153 | MenhirState1154 | MenhirState1950 | MenhirState1238 | MenhirState1235 | MenhirState1225 | MenhirState1215 | MenhirState1205 | MenhirState1198 | MenhirState1194 | MenhirState1174 | MenhirState1137 | MenhirState1139 | MenhirState1065 | MenhirState1054 | MenhirState1044 | MenhirState1043 | MenhirState1037 | MenhirState1033 | MenhirState1035 | MenhirState1025 | MenhirState1021 | MenhirState1020 | MenhirState1019 | MenhirState1009 | MenhirState1014 | MenhirState1016 | MenhirState1012 | MenhirState1007 | MenhirState1001 | MenhirState947 | MenhirState1000 | MenhirState974 | MenhirState994 | MenhirState996 | MenhirState998 | MenhirState992 | MenhirState990 | MenhirState988 | MenhirState986 | MenhirState980 | MenhirState984 | MenhirState982 | MenhirState978 | MenhirState976 | MenhirState972 | MenhirState970 | MenhirState964 | MenhirState968 | MenhirState966 | MenhirState949 | MenhirState961 | MenhirState951 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5839 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce110 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5840)
    | MenhirState2610 | MenhirState2606 | MenhirState2602 | MenhirState2597 | MenhirState2593 | MenhirState2459 | MenhirState173 | MenhirState2441 | MenhirState1390 | MenhirState1387 | MenhirState1384 | MenhirState1380 | MenhirState1377 | MenhirState1332 | MenhirState1313 | MenhirState1310 | MenhirState1307 | MenhirState1303 | MenhirState1300 | MenhirState1296 | MenhirState1293 | MenhirState1290 | MenhirState1285 | MenhirState1282 | MenhirState1280 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5841 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1282
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1282
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1282) : 'freshtv5842)
    | MenhirState2632 | MenhirState2294 | MenhirState2289 | MenhirState2144 | MenhirState2127 | MenhirState2123 | MenhirState2126 | MenhirState2114 | MenhirState2103 | MenhirState2037 | MenhirState1467 | MenhirState1418 | MenhirState1443 | MenhirState1445 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5843 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce133 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5844)
    | MenhirState2504 | MenhirState2501 | MenhirState2498 | MenhirState2494 | MenhirState2491 | MenhirState2487 | MenhirState2484 | MenhirState2481 | MenhirState2476 | MenhirState2473 | MenhirState1528 | MenhirState1526 | MenhirState1522 | MenhirState1520 | MenhirState1516 | MenhirState1514 | MenhirState1509 | MenhirState1507 | MenhirState1503 | MenhirState1496 | MenhirState1494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5845 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1496
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1496
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1496) : 'freshtv5846)
    | MenhirState1768 | MenhirState1615 | MenhirState1594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5847 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce36 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5848)
    | MenhirState2666 | MenhirState1832 | MenhirState1825 | MenhirState1820 | MenhirState1786 | MenhirState1763 | MenhirState1757 | MenhirState1739 | MenhirState1650 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5849 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce14 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5850)
    | MenhirState2057 | MenhirState1834 | MenhirState1801 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5851 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce36 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5852)
    | MenhirState2396 | MenhirState2169 | MenhirState2118 | MenhirState2117 | MenhirState2109 | MenhirState1969 | MenhirState2006 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5853 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce133 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5854)
    | MenhirState2395 | MenhirState2393 | MenhirState2364 | MenhirState2307 | MenhirState2303 | MenhirState2296 | MenhirState2292 | MenhirState2165 | MenhirState2025 | MenhirState1947 | MenhirState1962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5855 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce124 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5856)
    | MenhirState332 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5857 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce136 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5858)
    | MenhirState2710 | MenhirState2706 | MenhirState2701 | MenhirState2697 | MenhirState2690 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5859 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce14 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5860)
    | _ ->
        _menhir_fail ()) : 'freshtv5862)

and _menhir_reduce457 : _menhir_env -> (('ttv_tail * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_item_attribute = 
# 4309 "reason_parser.mly"
                                        ( (_2, _3) )
# 35760 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5827) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_item_attribute) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2748 | MenhirState2745 | MenhirState2739 | MenhirState2733 | MenhirState2717 | MenhirState2657 | MenhirState2655 | MenhirState2652 | MenhirState2645 | MenhirState2642 | MenhirState2640 | MenhirState2630 | MenhirState2632 | MenhirState2612 | MenhirState2324 | MenhirState438 | MenhirState2132 | MenhirState1406 | MenhirState2103 | MenhirState2096 | MenhirState2093 | MenhirState2088 | MenhirState2083 | MenhirState2078 | MenhirState2065 | MenhirState2058 | MenhirState2049 | MenhirState2046 | MenhirState2037 | MenhirState1635 | MenhirState1472 | MenhirState1468 | MenhirState1462 | MenhirState1460 | MenhirState1402 | MenhirState1392 | MenhirState1369 | MenhirState1159 | MenhirState1239 | MenhirState1181 | MenhirState1174 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5821 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1181
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1181
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1181) : 'freshtv5822)
    | MenhirState2539 | MenhirState2533 | MenhirState395 | MenhirState400 | MenhirState2396 | MenhirState2315 | MenhirState2312 | MenhirState2310 | MenhirState2257 | MenhirState2025 | MenhirState1935 | MenhirState1932 | MenhirState1924 | MenhirState1882 | MenhirState1877 | MenhirState1872 | MenhirState1867 | MenhirState1842 | MenhirState1749 | MenhirState1745 | MenhirState1742 | MenhirState1735 | MenhirState1731 | MenhirState1674 | MenhirState1623 | MenhirState1606 | MenhirState1603 | MenhirState1595 | MenhirState1581 | MenhirState1563 | MenhirState1532 | MenhirState1530 | MenhirState1487 | MenhirState1485 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5823 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1487
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1487
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1487) : 'freshtv5824)
    | MenhirState1594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1595
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1595
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1595) : 'freshtv5826)
    | _ ->
        _menhir_fail ()) : 'freshtv5828)

and _menhir_reduce437 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_floating_attribute = 
# 4313 "reason_parser.mly"
                                            ((_2, _3))
# 35818 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5819) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_floating_attribute) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState1610 | MenhirState1551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5805) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5803) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_floating_attribute) : 'tv_floating_attribute) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__class_sig_field = 
# 2272 "reason_parser.mly"
      ( mkctf(Pctf_attribute _1) )
# 35845 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5804)) : 'freshtv5806)
    | MenhirState1753 | MenhirState1666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5809) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5807) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_floating_attribute) : 'tv_floating_attribute) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__class_field = 
# 1992 "reason_parser.mly"
      ( mkcf (Pcf_attribute _1) )
# 35866 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5808)) : 'freshtv5810)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5811) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce311 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5812)
    | MenhirState2563 | MenhirState2565 | MenhirState1251 | MenhirState1408 | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5815) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5813) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_floating_attribute) : 'tv_floating_attribute) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1732 "reason_parser.mly"
                       (
      mksig(Psig_attribute _1)
    )
# 35897 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5814)) : 'freshtv5816)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5817) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_floating_attribute) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce311 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5818)
    | _ ->
        _menhir_fail ()) : 'freshtv5820)

and _menhir_reduce458 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)), _, (_3 : 'tv_payload)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_item_extension = 
# 4350 "reason_parser.mly"
                                                  ( (_2, _3) )
# 35921 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5801) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_item_extension) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1610 | MenhirState1551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5791 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1606
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1606
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1606) : 'freshtv5792)
    | MenhirState1753 | MenhirState1666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5793 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1749
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1749
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1749) : 'freshtv5794)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5795 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1932
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1932
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1932) : 'freshtv5796)
    | MenhirState2563 | MenhirState2565 | MenhirState1251 | MenhirState1408 | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5797 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2096
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2096
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2096) : 'freshtv5798)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5799 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2745
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2745
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2745) : 'freshtv5800)
    | _ ->
        _menhir_fail ()) : 'freshtv5802)

and _menhir_reduce266 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_signature)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_module_type = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 36010 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 36022 "reason_parser.ml"
      
    in
    
# 1550 "reason_parser.mly"
      ( unclosed_mty (with_txt _1 "{") (with_txt _3 "}"))
# 36028 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_functor_arg_name : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_functor_arg_name -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5777 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5773 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState449 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState449 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState449 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState449) : 'freshtv5774)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5775 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5776)) : 'freshtv5778)
    | MenhirState1250 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5783 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5779 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2113 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2113 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2113 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2113 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2113 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2113 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2113) : 'freshtv5780)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5781 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5782)) : 'freshtv5784)
    | MenhirState1248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5785 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2122 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2122) : 'freshtv5786)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5787 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5788)) : 'freshtv5790)
    | _ ->
        _menhir_fail ()

and _menhir_reduce142 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_meth_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3947 "reason_parser.mly"
      ( let (f, c) = _2 in mktyp(Ptyp_object (f, c)) )
# 36163 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_signed_constant : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_signed_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState563 | MenhirState564 | MenhirState550 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5761 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error544 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOTDOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5759 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState545 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState545) : 'freshtv5760)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce272 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5762)
    | MenhirState545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5763 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce273 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5764)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState486 | MenhirState1886 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1657 | MenhirState1639 | MenhirState1637 | MenhirState1634 | MenhirState1183 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 | MenhirState908 | MenhirState767 | MenhirState738 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState642 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5769 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv5765 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5766)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOTDOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5767 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState630) : 'freshtv5768)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce272 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5770)
    | MenhirState630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5771 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce273 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5772)
    | _ ->
        _menhir_fail ()

and _menhir_goto_attr_id : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_attr_id -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState71 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5719 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_reduce333 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5720)
    | MenhirState19 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5721 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState73
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState73
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState73)) : 'freshtv5722)
    | MenhirState174 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5723 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState175
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState175
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState175)) : 'freshtv5724)
    | MenhirState189 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5725 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState190
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState190
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState190)) : 'freshtv5726)
    | MenhirState333 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5727 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState334
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState334
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState334)) : 'freshtv5728)
    | MenhirState389 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5729 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_reduce333 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5730)
    | MenhirState337 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5745 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5743 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_attr_id)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _v : 'tv_item_extension_sugar = 
# 4339 "reason_parser.mly"
                    (
      ([], _2)
    )
# 36714 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5741) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
        match _menhir_s with
        | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1161 | MenhirState1176 | MenhirState1167 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5731 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run1156 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1171 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1171) : 'freshtv5732)
        | MenhirState2673 | MenhirState1632 | MenhirState1663 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5733 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LET ->
                _menhir_run1633 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1664 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1664) : 'freshtv5734)
        | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5735 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CLASS ->
                _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1929 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1929) : 'freshtv5736)
        | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState475 | MenhirState2259 | MenhirState2253 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5737 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2255 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2255) : 'freshtv5738)
        | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5739 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
            ((assert (not _menhir_env._menhir_error);
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CLASS ->
                _menhir_run2660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2647 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2644 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run2635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run2628 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2743 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2743) : 'freshtv5740)
        | _ ->
            _menhir_fail ()) : 'freshtv5742)) : 'freshtv5744)) : 'freshtv5746)
    | MenhirState396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5747 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState397
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState397
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState397)) : 'freshtv5748)
    | MenhirState439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5749 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState440
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState440
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState440)) : 'freshtv5750)
    | MenhirState453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5751 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState454
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState454
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState454)) : 'freshtv5752)
    | MenhirState455 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5753 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState456
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState456
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState456)) : 'freshtv5754)
    | MenhirState1474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5755 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1475
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState1475
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1475)) : 'freshtv5756)
    | MenhirState2032 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5757 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2033
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | ASSERT ->
              _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CLASS ->
              _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              _menhir_run2027 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXCEPTION ->
              _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EXTERNAL ->
              _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FOR ->
              _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FUN ->
              _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IF ->
              _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INCLUDE ->
              _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LAZY ->
              _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LET ->
              _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUSDOT ->
              _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MODULE ->
              _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | OPEN ->
              _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PERCENT ->
              _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUSDOT ->
              _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUESTION ->
              _menhir_run191 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SWITCH ->
              _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRY ->
              _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TYPE ->
              _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHILE ->
              _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET ->
              _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState2033
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2033)) : 'freshtv5758)
    | _ ->
        _menhir_fail ()

and _menhir_goto_package_type : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_package_type -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState129 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5661 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error131 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5659 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce152 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5660)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error131 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5662)
    | MenhirState233 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5665 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 37752 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error235 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5663 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 37764 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce291 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5664)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error235 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5666)
    | MenhirState246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5669 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error247 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv5667 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce152 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5668)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error247 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5670)
    | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5673 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 37798 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error500 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5671 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 37810 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce291 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5672)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error500 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5674)
    | MenhirState775 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5677 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error777 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5675 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce246 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5676)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error777 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5678)
    | MenhirState801 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5681 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error803 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5679 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce248 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5680)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error803 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5682)
    | MenhirState2389 | MenhirState1952 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1953 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5683 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce259 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5684)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1953 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5686)
    | MenhirState2391 | MenhirState1955 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1956 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5687 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1958 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1958) : 'freshtv5688)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce257 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5690)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1956 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5692)
    | MenhirState1958 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5695 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1959 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce258 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5694)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1959 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5696)
    | MenhirState2360 | MenhirState2155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5699 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2156 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5697 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce259 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5698)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2156 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5700)
    | MenhirState2362 | MenhirState2158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5705 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2159 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2161) : 'freshtv5702)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce257 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5704)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2159 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5706)
    | MenhirState2161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2162 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5707 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce258 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5708)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2162 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5710)
    | MenhirState2173 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv5713 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2175 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv5711 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce248 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5712)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2175 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5714)
    | MenhirState2368 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv5717 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2370 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv5715 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce246 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5716)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2370 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5718)
    | _ ->
        _menhir_fail ()

and _menhir_reduce276 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_type_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern_not_ident = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 38056 "reason_parser.ml"
      
    in
    
# 3350 "reason_parser.mly"
      ( mkpat(Ppat_type (_2)) )
# 38062 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce558 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__non_arrowed_simple_core_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__non_arrowed_simple_core_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1173 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 38075 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5657) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__non_arrowed_simple_core_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2698 | MenhirState2682 | MenhirState2683 | MenhirState2669 | MenhirState2649 | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState121 | MenhirState2472 | MenhirState128 | MenhirState141 | MenhirState143 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState158 | MenhirState161 | MenhirState162 | MenhirState2449 | MenhirState163 | MenhirState170 | MenhirState2442 | MenhirState2439 | MenhirState324 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1814 | MenhirState1816 | MenhirState1805 | MenhirState1807 | MenhirState1809 | MenhirState1799 | MenhirState1800 | MenhirState1766 | MenhirState1779 | MenhirState1781 | MenhirState1770 | MenhirState1772 | MenhirState1774 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1361 | MenhirState1359 | MenhirState1345 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1288 | MenhirState1289 | MenhirState1284 | MenhirState1269 | MenhirState1277 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState306 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState272 | MenhirState270 | MenhirState245 | MenhirState254 | MenhirState255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5653) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce599 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5654)
    | MenhirState2521 | MenhirState106 | MenhirState2478 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1642 | MenhirState1643 | MenhirState1601 | MenhirState1592 | MenhirState1593 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1490 | MenhirState1500 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState328 | MenhirState329 | MenhirState244 | MenhirState317 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5655) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce599 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5656)
    | _ ->
        _menhir_fail ()) : 'freshtv5658)

and _menhir_reduce557 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_non_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__non_arrowed_non_simple_core_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1173 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 38112 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5651) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2698 | MenhirState2669 | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState143 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState324 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1814 | MenhirState1816 | MenhirState1805 | MenhirState1807 | MenhirState1809 | MenhirState1766 | MenhirState1779 | MenhirState1781 | MenhirState1770 | MenhirState1772 | MenhirState1774 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState272 | MenhirState270 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5647) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce598 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5648)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5649) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__non_arrowed_non_simple_core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce598 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5650)
    | _ ->
        _menhir_fail ()) : 'freshtv5652)

and _menhir_error616 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5645 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce278 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5646)

and _menhir_reduce277 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_lbl_pattern_list)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3352 "reason_parser.mly"
      ( let (fields, closed) = _2 in mkpat(Ppat_record(fields, closed)) )
# 38159 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error521 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5643 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce278 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5644)

and _menhir_reduce284 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3370 "reason_parser.mly"
      ( _2 )
# 38182 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce161 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__pattern_optional_constraint = 
# 3417 "reason_parser.mly"
                                ( _1 )
# 38194 "reason_parser.ml"
     in
    _menhir_goto__pattern_optional_constraint _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run668 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState668

and _menhir_reduce160 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__or_pattern = 
# 3083 "reason_parser.mly"
    ( mkpat(Ppat_or(_1, _3)) )
# 38264 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5641) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__or_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5637) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce547 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5638)
    | MenhirState2578 | MenhirState191 | MenhirState2314 | MenhirState738 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState676 | MenhirState587 | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5639) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__or_pattern) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce547 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5640)
    | _ ->
        _menhir_fail ()) : 'freshtv5642)

and _menhir_run594 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState594 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState594

and _menhir_run592 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState592 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState592

and _menhir_run591 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce637 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce555 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__core_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__core_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1173 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 38421 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5635) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__core_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState666 | MenhirState270 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5631) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce399 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5632)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5633) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce399 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5634)
    | _ ->
        _menhir_fail ()) : 'freshtv5636)

and _menhir_reduce483 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_pattern -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)) = _menhir_stack in
    let _v : 'tv_lbl_pattern_list = 
# 3435 "reason_parser.mly"
                ( [_1], Closed )
# 38455 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error524 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_pattern -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce483 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce551 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern_not_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_pattern_not_ident) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__simple_pattern_not_ident_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1203 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 38472 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5629) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__simple_pattern_not_ident_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5625) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_not_ident_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce729 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5626)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState486 | MenhirState738 | MenhirState1886 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1657 | MenhirState1639 | MenhirState1637 | MenhirState1634 | MenhirState1183 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5627) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__simple_pattern_not_ident_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce729 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5628)
    | _ ->
        _menhir_fail ()) : 'freshtv5630)

and _menhir_goto_simple_non_labeled_expr_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_simple_non_labeled_expr_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5621 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState842
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState842 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce268 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5622)
    | MenhirState953 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5623 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState955
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState955 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce268 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5624)
    | _ ->
        _menhir_fail ()

and _menhir_goto_labeled_simple_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_labeled_simple_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1648 | MenhirState827 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5613 * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce720 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5614)
    | MenhirState1647 | MenhirState820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5615) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce719 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5616)
    | MenhirState2688 | MenhirState940 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5617 * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce720 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5618)
    | MenhirState2687 | MenhirState933 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5619) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_labeled_simple_expr) = _v in
        (_menhir_reduce719 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5620)
    | _ ->
        _menhir_fail ()

and _menhir_goto_label_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_label_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1647 | MenhirState1648 | MenhirState820 | MenhirState827 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5609) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_label_expr) = _v in
        (_menhir_reduce469 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5610)
    | MenhirState2687 | MenhirState2688 | MenhirState933 | MenhirState940 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5611) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_label_expr) = _v in
        (_menhir_reduce469 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5612)
    | _ ->
        _menhir_fail ()

and _menhir_reduce725 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_non_labeled_expr_list_as_tuple = 
# 2828 "reason_parser.mly"
                                                                                                 (_1)
# 38671 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5607) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_simple_non_labeled_expr_list_as_tuple) = _v in
    ((match _menhir_s with
    | MenhirState840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5603 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_non_labeled_expr_list_as_tuple) = _v in
        (_menhir_reduce72 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5604)
    | MenhirState953 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5605 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_simple_non_labeled_expr_list_as_tuple) = _v in
        (_menhir_reduce72 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5606)
    | _ ->
        _menhir_fail ()) : 'freshtv5608)

and _menhir_error2189 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1141 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce416 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_optional_constraint), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_expr_comma_seq_extension = 
# 3147 "reason_parser.mly"
    ( ([], Some _2) )
# 38716 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_seq_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1100 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_expr_comma_seq : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_expr_comma_seq -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1081 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5595 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1083
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1084 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1083
          | BARRBRACKET ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1083
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1083)) : 'freshtv5596)
    | MenhirState2182 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5597 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2183
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1084 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2183
          | BARRBRACKET ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2183
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2183)) : 'freshtv5598)
    | MenhirState415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5599 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2354
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1084 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2354
          | BARRBRACKET ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2354
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2354)) : 'freshtv5600)
    | MenhirState178 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5601 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2428
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run1084 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2428
          | BARRBRACKET ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2428
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2428)) : 'freshtv5602)
    | _ ->
        _menhir_fail ()

and _menhir_reduce412 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_expr_comma_list)), _endpos__2_, _), _, (_3 : 'tv_expr_optional_constraint), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_expr_comma_list = 
# 3130 "reason_parser.mly"
      ( _3 :: _1 )
# 38803 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce413 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_), _endpos__2_), _, (_3 : 'tv_expr_optional_constraint), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_expr_comma_list = 
# 3132 "reason_parser.mly"
      ( [_3; _1] )
# 38814 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error1062 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_type_constraint_right_of_colon : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_constraint_right_of_colon -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1895 | MenhirState1206 | MenhirState1068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5591 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_constraint_right_of_colon) = _v in
        (_menhir_reduce807 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5592)
    | MenhirState1089 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5593 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_constraint_right_of_colon) = _v in
        (_menhir_reduce807 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5594)
    | _ ->
        _menhir_fail ()

and _menhir_error1130 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce476 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)) = _menhir_stack in
    let _v : 'tv_lbl_expr_list = 
# 3173 "reason_parser.mly"
              ( [_1] )
# 38854 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error1119 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce476 _menhir_env (Obj.magic _menhir_stack)

and _menhir_goto_record_expr : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_record_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1113 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5577 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1125 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv5575 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce226 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5576)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1125 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5578)
    | MenhirState2202 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5581 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2204 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv5579 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce226 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5580)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2204 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5582)
    | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5585 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2244 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5583 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce222 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5584)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2244 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5586)
    | MenhirState432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5589 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2331 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv5587 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_record_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce222 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5588)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2331 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5590)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_type_parameters : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_class_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1538 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5553 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 38950 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5551 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 38960 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1539 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1544 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1542 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1542 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1543 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1542 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1542 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1542) : 'freshtv5552)
        | MINUS ->
            _menhir_run1541 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1539 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce828 _menhir_env (Obj.magic _menhir_stack) MenhirState1539
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1539) : 'freshtv5554)
    | MenhirState1626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5559 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 38996 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5555) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1627 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1804 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run1788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1787 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1787) : 'freshtv5556)
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5557 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 39052 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1627 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1630 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1630 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1630) : 'freshtv5558)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run1629 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1628 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce828 _menhir_env (Obj.magic _menhir_stack) MenhirState1627
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1627) : 'freshtv5560)
    | MenhirState2054 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5563 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 39128 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5561 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 39138 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2055 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1790 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2056 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1789 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2056 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2056) : 'freshtv5562)
        | MINUS ->
            _menhir_run1541 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2055 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce828 _menhir_env (Obj.magic _menhir_stack) MenhirState2055
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2055) : 'freshtv5564)
    | MenhirState2061 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5567 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 39174 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5565 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 39184 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2062 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1804 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run1788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2063) : 'freshtv5566)
        | MINUS ->
            _menhir_run1541 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2062 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce828 _menhir_env (Obj.magic _menhir_stack) MenhirState2062
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2062) : 'freshtv5568)
    | MenhirState2662 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5573 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 39240 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5569) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2663 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run1804 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run1788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2698 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2698) : 'freshtv5570)
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5571 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 39296 "reason_parser.ml"
            ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2663 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2681 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2664 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run2677 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2664 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run2673 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2664 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2664 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2664 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2665 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2664 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2664 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2664) : 'freshtv5572)
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run1629 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run1628 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2663 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_reduce828 _menhir_env (Obj.magic _menhir_stack) MenhirState2663
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2663) : 'freshtv5574)
    | _ ->
        _menhir_fail ()

and _menhir_goto_typevar_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_typevar_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState1206 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5531 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5527 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1219 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1219) : 'freshtv5528)
        | QUOTE ->
            _menhir_run1217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5529 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5530)) : 'freshtv5532)
    | MenhirState153 | MenhirState1720 | MenhirState1328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5537 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5533 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1330 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1330) : 'freshtv5534)
        | QUOTE ->
            _menhir_run1217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5535 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5536)) : 'freshtv5538)
    | MenhirState1700 | MenhirState1695 | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5543 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5539 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1579) : 'freshtv5540)
        | QUOTE ->
            _menhir_run1217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5541 * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5542)) : 'freshtv5544)
    | MenhirState1895 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5549 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5545 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1904 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1904) : 'freshtv5546)
        | QUOTE ->
            _menhir_run1217 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5547 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5548)) : 'freshtv5550)
    | _ ->
        _menhir_fail ()

and _menhir_reduce56 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type2), _startpos__1_), _endpos__2_, _, _startpos__2_), _startpos__3_), _endpos__4_, _, (_4 : 'tv_ident), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__core_type = 
# 3862 "reason_parser.mly"
      ( mktyp(Ptyp_alias(_1, _4)) )
# 39591 "reason_parser.ml"
     in
    _menhir_goto__core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce587 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_ident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mty_longident = 
# 4160 "reason_parser.mly"
                                                ( Lident _1 )
# 39603 "reason_parser.ml"
     in
    _menhir_goto_mty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce588 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_mty_longident = 
# 4161 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 39616 "reason_parser.ml"
     in
    _menhir_goto_mty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce591 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_name_tag = 
# 4187 "reason_parser.mly"
                                                ( _2 )
# 39629 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5525) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_name_tag) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState158 | MenhirState161 | MenhirState162 | MenhirState2449 | MenhirState163 | MenhirState306 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState170 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5499 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState173
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | AMPERSAND ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5495) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState173 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5493) = Obj.magic _menhir_stack in
              let (_endpos__1_ : Lexing.position) = _endpos in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_startpos__1_ : Lexing.position) = _startpos in
              ((let _1 = () in
              let _v : 'tv_opt_ampersand = 
# 4001 "reason_parser.mly"
                                                ( true )
# 39663 "reason_parser.ml"
               in
              _menhir_goto_opt_ampersand _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5494)) : 'freshtv5496)
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState173
          | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5497) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState173 in
              ((let _v : 'tv_opt_ampersand = 
# 4002 "reason_parser.mly"
                                                ( false )
# 39675 "reason_parser.ml"
               in
              _menhir_goto_opt_ampersand _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5498)
          | AMPERAMPER | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState173) : 'freshtv5500)
    | MenhirState2457 | MenhirState296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5503 * _menhir_state * 'tv_name_tag_list) * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5501 * _menhir_state * 'tv_name_tag_list) * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_name_tag_list)), _endpos__2_, _, (_2 : 'tv_name_tag), _startpos__2_) = _menhir_stack in
        let _v : 'tv_name_tag_list = 
# 4010 "reason_parser.mly"
                                                ( _2 :: _1 )
# 39689 "reason_parser.ml"
         in
        _menhir_goto_name_tag_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5502)) : 'freshtv5504)
    | MenhirState2456 | MenhirState295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5507 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5505 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
        let _v : 'tv_name_tag_list = 
# 4009 "reason_parser.mly"
                                                ( [_1] )
# 39701 "reason_parser.ml"
         in
        _menhir_goto_name_tag_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5506)) : 'freshtv5508)
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5509 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState550
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState550 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce275 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5510)
    | MenhirState532 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5511 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce275 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5512)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState486 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1639 | MenhirState1637 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5513 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce275 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5514)
    | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState738 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5515 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState646
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState646 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce275 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5516)
    | MenhirState2555 | MenhirState180 | MenhirState187 | MenhirState457 | MenhirState464 | MenhirState470 | MenhirState2268 | MenhirState1647 | MenhirState1648 | MenhirState840 | MenhirState842 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 | MenhirState817 | MenhirState811 | MenhirState805 | MenhirState747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5517 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce204 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5518)
    | MenhirState2687 | MenhirState2688 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState417 | MenhirState2350 | MenhirState419 | MenhirState474 | MenhirState729 | MenhirState730 | MenhirState734 | MenhirState740 | MenhirState746 | MenhirState764 | MenhirState894 | MenhirState895 | MenhirState953 | MenhirState955 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 | MenhirState930 | MenhirState920 | MenhirState910 | MenhirState912 | MenhirState913 | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5519 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce204 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5520)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5521 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState817
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState817 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce204 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5522)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 | MenhirState2574 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5523 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState930
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState930 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce204 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5524)
    | _ ->
        _menhir_fail ()) : 'freshtv5526)

and _menhir_reduce831 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _v : 'tv_typevar_list = 
# 3737 "reason_parser.mly"
                                                ( [_2] )
# 39939 "reason_parser.ml"
     in
    _menhir_goto_typevar_list _menhir_env _menhir_stack _menhir_s _v _startpos

and _menhir_reduce138 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3939 "reason_parser.mly"
      ( mktyp(Ptyp_var _2) )
# 39952 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce158 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3537 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _3), Contravariant )
# 39966 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce156 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3535 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _3), Covariant)
# 39980 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce154 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__optional_type_variable_with_variance = 
# 3533 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _2), Invariant )
# 39993 "reason_parser.ml"
     in
    _menhir_goto__optional_type_variable_with_variance _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce678 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 898 "reason_parser.mly"
       (string * string option)
# 40000 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 898 "reason_parser.mly"
       (string * string option)
# 40006 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _, (_2 : 'tv_primitive_declaration)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_primitive_declaration = 
# 3454 "reason_parser.mly"
                                                ( fst _1 :: _2 )
# 40012 "reason_parser.ml"
     in
    _menhir_goto_primitive_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce530 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_instance_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_instance_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cty__class_instance_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1158 "reason_parser.mly"
 ( {x with pcty_loc = {x.pcty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40025 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5491) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__class_instance_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState1767 | MenhirState1542 | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5487) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_instance_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce349 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5488)
    | MenhirState2707 | MenhirState2056 | MenhirState1826 | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5489) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_instance_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce349 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5490)
    | _ ->
        _menhir_fail ()) : 'freshtv5492)

and _menhir_reduce522 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1193 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40062 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5485) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1738 | MenhirState1663 | MenhirState1632 | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5481) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce342 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5482)
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5483) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce342 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5484)
    | _ ->
        _menhir_fail ()) : 'freshtv5486)

and _menhir_reduce525 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_def -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_fun_def) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_fun_def_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1193 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40098 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5479) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_fun_def_) = _v in
    ((match _menhir_s with
    | MenhirState1639 | MenhirState1637 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5475) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_def_) = _v in
        (_menhir_reduce346 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5476)
    | MenhirState2679 | MenhirState2677 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5477) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_fun_def_) = _v in
        (_menhir_reduce346 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5478)
    | _ ->
        _menhir_fail ()) : 'freshtv5480)

and _menhir_reduce407 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_curried_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_curried_binding) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_curried_binding_return_typed_ = 
# 3026 "reason_parser.mly"
    (_1)
# 40129 "reason_parser.ml"
     in
    _menhir_goto_curried_binding_return_typed_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce62 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_curried_binding_return_typed_) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__curried_binding_return_typed = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3016 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         let (l, o, p) = _1 in mkexp ~ghost:true ~loc (Pexp_fun(l, o, p, _2))
      )
# 40146 "reason_parser.ml"
     in
    _menhir_goto__curried_binding_return_typed _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce63 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 40153 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_curried_binding_return_typed_) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 857 "reason_parser.mly"
       (string)
# 40159 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__curried_binding_return_typed = 
# 3021 "reason_parser.mly"
      ( mkexp(Pexp_newtype(_3, _5)) )
# 40169 "reason_parser.ml"
     in
    _menhir_goto__curried_binding_return_typed _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce405 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 40176 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_curried_binding_return_typed_) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 857 "reason_parser.mly"
       (string)
# 40182 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_curried_binding = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3064 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp ~loc (Pexp_newtype(_3, _5))
      )
# 40196 "reason_parser.ml"
     in
    _menhir_goto_curried_binding _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce404 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_curried_binding_return_typed_) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_curried_binding = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3058 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let (l, o, p) = _1 in
        mkexp ~ghost:true ~loc (Pexp_fun(l, o, p, _2))
      )
# 40213 "reason_parser.ml"
     in
    _menhir_goto_curried_binding _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2334 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5473 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce225 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5474)

and _menhir_reduce224 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_self_pattern_and_structure), _startpos__2_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2722 "reason_parser.mly"
      ( mkexp (Pexp_object _2) )
# 40236 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2265 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5471 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce225 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5472)

and _menhir_error2674 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5469 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce49 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5470)

and _menhir_reduce48 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr_lets_and_rest)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_simple_expr = 
# 1964 "reason_parser.mly"
                                           ( _2 )
# 40268 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1760 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5467 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce49 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5468)

and _menhir_reduce529 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_constructor_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_constructor_type) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cty__class_constructor_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1158 "reason_parser.mly"
 ( {x with pcty_loc = {x.pcty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40289 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5465) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cty__class_constructor_type_) = _v in
    ((match _menhir_s with
    | MenhirState2669 | MenhirState1766 | MenhirState1781 | MenhirState1772 | MenhirState1774 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5461) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_constructor_type_) = _v in
        (_menhir_reduce337 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5462)
    | MenhirState2698 | MenhirState2063 | MenhirState1827 | MenhirState1787 | MenhirState1816 | MenhirState1807 | MenhirState1809 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5463) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cty__class_constructor_type_) = _v in
        (_menhir_reduce337 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5464)
    | _ ->
        _menhir_fail ()) : 'freshtv5466)

and _menhir_run1552 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1552 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5457) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1552 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1558 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState1558
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1558) : 'freshtv5458)
    | VIRTUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5459) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1552 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MUTABLE ->
            _menhir_run1325 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1553 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState1553
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1553) : 'freshtv5460)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1552

and _menhir_run1568 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5441) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5435 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5433 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos__2_ : Lexing.position) = _endpos in
            let (_startpos__2_ : Lexing.position) = _startpos in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _2 = () in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4218 "reason_parser.mly"
                    ( Private, Virtual )
# 40394 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5434)) : 'freshtv5436)
        | LIDENT _ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5437 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4216 "reason_parser.mly"
            ( Private, Concrete )
# 40405 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5438)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5439 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            (raise _eRR : 'freshtv5440)) : 'freshtv5442)
    | VIRTUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5451) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PRIVATE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5445 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5443 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos__2_ : Lexing.position) = _endpos in
            let (_startpos__2_ : Lexing.position) = _startpos in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _2 = () in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4219 "reason_parser.mly"
                    ( Private, Virtual )
# 40439 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5444)) : 'freshtv5446)
        | LIDENT _ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5447 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _endpos__1_, _startpos__1_) = _menhir_stack in
            let _1 = () in
            let _v : 'tv_private_virtual_flags = 
# 4217 "reason_parser.mly"
            ( Public, Virtual )
# 40450 "reason_parser.ml"
             in
            _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5448)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5449 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            (raise _eRR : 'freshtv5450)) : 'freshtv5452)
    | LIDENT _ ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5453) = Obj.magic _menhir_stack in
        ((let _v : 'tv_private_virtual_flags = 
# 4215 "reason_parser.mly"
                 ( Public, Concrete )
# 40465 "reason_parser.ml"
         in
        _menhir_goto_private_virtual_flags _menhir_env _menhir_stack _v) : 'freshtv5454)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5455 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5456)

and _menhir_run1586 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1544 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run1543 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1586 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1586

and _menhir_run1599 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1599 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1599

and _menhir_reduce381 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mark_position_cl__constrained_class_declaration_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_mark_position_cl__constrained_class_declaration_) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_constrained_class_declaration = 
# 1801 "reason_parser.mly"
                                                                                (_1)
# 40538 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5431) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_constrained_class_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5427 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 40551 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrained_class_declaration) = _v in
        (_menhir_reduce339 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5428)
    | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5429 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 40562 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constrained_class_declaration) = _v in
        (_menhir_reduce339 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5430)
    | _ ->
        _menhir_fail ()) : 'freshtv5432)

and _menhir_reduce29 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_class_fun_return -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_class_fun_return) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_fun_binding = 
# 1863 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _2)) )
# 40579 "reason_parser.ml"
     in
    _menhir_goto__class_fun_binding _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1423 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5421 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5419 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_startpos__2_ : Lexing.position) = _startpos in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _v : 'tv_with_type_binder = 
# 3731 "reason_parser.mly"
                   ( Private )
# 40605 "reason_parser.ml"
         in
        _menhir_goto_with_type_binder _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5420)) : 'freshtv5422)
    | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5423 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
        let _1 = () in
        let _v : 'tv_with_type_binder = 
# 3730 "reason_parser.mly"
                   ( Public )
# 40616 "reason_parser.ml"
         in
        _menhir_goto_with_type_binder _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5424)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5425 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5426)

and _menhir_goto_potentially_long_ident_and_optional_type_parameters : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_potentially_long_ident_and_optional_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5405 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PLUSEQ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5401 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1372 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BAR | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
                _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState1372
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1372) : 'freshtv5402)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5403 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5404)) : 'freshtv5406)
    | MenhirState76 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PLUSEQ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5407 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2536 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BAR | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
                _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState2536
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2536) : 'freshtv5408)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5410)) : 'freshtv5412)
    | MenhirState2584 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5417 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PLUSEQ ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2586 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BAR | COLONCOLON | FALSE | LPAREN | TRUE | UIDENT _ ->
                _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState2586
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2586) : 'freshtv5414)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5415 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5416)) : 'freshtv5418)
    | _ ->
        _menhir_fail ()

and _menhir_run1268 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run1298 _menhir_env (Obj.magic _menhir_stack) MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        _menhir_run1295 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOT ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5397 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1268 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce817 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv5398)
    | FALSE ->
        _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1288 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5399 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1268 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run1298 _menhir_env (Obj.magic _menhir_stack) MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON ->
            _menhir_run1295 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1292 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1288 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1287 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_reduce680 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1287) : 'freshtv5400)
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run1284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1269 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1268 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState1268
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1268

and _menhir_reduce675 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 40831 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 40837 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _, (_2 : 'tv_optional_type_parameters)) = _menhir_stack in
    let _v : 'tv_potentially_long_ident_and_optional_type_parameters = 
# 3632 "reason_parser.mly"
                                  (
      let loc = mklocation _startpos__1_ _endpos__1_ in
      let lident_lident_loc = mkloc (Lident _1) loc in
      (lident_lident_loc, _2)
  )
# 40846 "reason_parser.ml"
     in
    _menhir_goto_potentially_long_ident_and_optional_type_parameters _menhir_env _menhir_stack _menhir_s _v

and _menhir_run106 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run2489 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        _menhir_run2486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOT ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5393 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState106 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce817 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv5394)
    | FALSE ->
        _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run2479 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5395 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState106 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BAR ->
            _menhir_run2489 _menhir_env (Obj.magic _menhir_stack) MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONCOLON ->
            _menhir_run2486 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run2483 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2479 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run2475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run108 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_reduce680 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2478) : 'freshtv5396)
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run2475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run108 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState106 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState106
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState106

and _menhir_reduce554 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__structure_item_without_item_extension_sugar) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1188 "reason_parser.mly"
 ( {x with pstr_loc = {x.pstr_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 40963 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5391) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5387) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce794 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5388)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5389) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_without_item_extension_sugar_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce794 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5390)
    | _ ->
        _menhir_fail ()) : 'freshtv5392)

and _menhir_reduce553 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__structure_item) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_str__structure_item_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1188 "reason_parser.mly"
 ( {x with pstr_loc = {x.pstr_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41000 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5385) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_str__structure_item_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5381) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce793 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5382)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5383) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_str__structure_item_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce793 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5384)
    | _ ->
        _menhir_fail ()) : 'freshtv5386)

and _menhir_reduce532 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__curried_binding_return_typed -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__curried_binding_return_typed) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__curried_binding_return_typed_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1168 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41036 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5379) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__curried_binding_return_typed_) = _v in
    ((match _menhir_s with
    | MenhirState1184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5375) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__curried_binding_return_typed_) = _v in
        (_menhir_reduce406 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5376)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5377) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__curried_binding_return_typed_) = _v in
        (_menhir_reduce406 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5378)
    | _ ->
        _menhir_fail ()) : 'freshtv5380)

and _menhir_run1859 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constraints -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1859 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1859

and _menhir_run1359 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constraints -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1359 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1359

and _menhir_run1988 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5373 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1988 in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 41149 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv5371 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 41161 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1989 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1990 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1990) : 'freshtv5372)
        | AND | DOT | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
            _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1989) : 'freshtv5374)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1988

and _menhir_reduce572 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 41190 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 906 "reason_parser.mly"
       (string)
# 41196 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_mod_ext_longident), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_mod_ext2 = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 4156 "reason_parser.mly"
                                            ( lapply (Lident(_1)) _3 _symbolstartpos _endpos )
# 41207 "reason_parser.ml"
     in
    _menhir_goto_mod_ext2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce366 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 41214 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 857 "reason_parser.mly"
       (string)
# 41219 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_clty_longident = 
# 4165 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 41228 "reason_parser.ml"
     in
    _menhir_goto_clty_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1255 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce155 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1256 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1256 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1256 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1256

and _menhir_run1258 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5365 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1260 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1260 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1260) : 'freshtv5366)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5367 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce157 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5368)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5369 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5370)

and _menhir_run1262 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5359 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1264 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1264) : 'freshtv5360)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5361 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce159 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5362)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5363 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5364)

and _menhir_run83 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 41329 "reason_parser.ml"
) * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState83 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState83

and _menhir_run82 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 41346 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run83 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState82 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | DOT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
        _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState82

and _menhir_reduce824 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 41365 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 857 "reason_parser.mly"
       (string)
# 41370 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_longident = 
# 4129 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 41379 "reason_parser.ml"
     in
    _menhir_goto_type_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run196 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5357 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState196 in
        let (_v : (
# 857 "reason_parser.mly"
       (string)
# 41397 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce824 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv5358)
    | UIDENT _v ->
        _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState196 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState196

and _menhir_reduce571 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext2), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_mod_ext_longident), _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_mod_ext2 = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 4155 "reason_parser.mly"
                                              ( lapply _1 _3 _symbolstartpos _endpos )
# 41421 "reason_parser.ml"
     in
    _menhir_goto_mod_ext2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce570 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 41428 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _ _startpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _), _endpos__3_, _, (_3 : (
# 906 "reason_parser.mly"
       (string)
# 41434 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_mod_ext_longident), _startpos__5_) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv_mod_ext2 = let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 4152 "reason_parser.mly"
                                                                  (
     lapply ( Ldot(_1, _3)) _5 _symbolstartpos _endpos
   )
# 41448 "reason_parser.ml"
     in
    _menhir_goto_mod_ext2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run80 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 41455 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce572 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run81 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState81 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState81

and _menhir_reduce853 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_with_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_with_constraint) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_with_constraints = 
# 3692 "reason_parser.mly"
                                                ( [_1] )
# 41480 "reason_parser.ml"
     in
    _menhir_goto_with_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce854 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_with_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_with_constraint) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_with_constraints)), _endpos__2_, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_with_constraints = 
# 3693 "reason_parser.mly"
                                                ( _3 :: _1 )
# 41492 "reason_parser.ml"
     in
    _menhir_goto_with_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce543 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__module_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__module_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1178 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41505 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5355) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__module_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5351) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce586 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5352)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState1965 | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5353) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce586 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5354)
    | _ ->
        _menhir_fail ()) : 'freshtv5356)

and _menhir_reduce545 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_module_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__simple_module_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1178 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41542 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5349) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__simple_module_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5345) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__simple_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce722 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5346)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState1965 | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5347) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__simple_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce722 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5348)
    | _ ->
        _menhir_fail ()) : 'freshtv5350)

and _menhir_reduce541 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_module_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mod__simple_module_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1153 "reason_parser.mly"
 ( {x with pmod_loc = {x.pmod_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41579 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5343) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mod__simple_module_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2644 | MenhirState2645 | MenhirState411 | MenhirState2366 | MenhirState451 | MenhirState2288 | MenhirState756 | MenhirState2172 | MenhirState759 | MenhirState761 | MenhirState2142 | MenhirState2143 | MenhirState2138 | MenhirState2139 | MenhirState2128 | MenhirState2129 | MenhirState1245 | MenhirState1246 | MenhirState799 | MenhirState800 | MenhirState772 | MenhirState773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5339) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__simple_module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce721 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5340)
    | MenhirState403 | MenhirState2395 | MenhirState404 | MenhirState2393 | MenhirState412 | MenhirState2364 | MenhirState2306 | MenhirState2307 | MenhirState2302 | MenhirState2303 | MenhirState2295 | MenhirState2296 | MenhirState2291 | MenhirState2292 | MenhirState762 | MenhirState2165 | MenhirState1476 | MenhirState2025 | MenhirState1945 | MenhirState1947 | MenhirState1948 | MenhirState1962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5341) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__simple_module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce721 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5342)
    | _ ->
        _menhir_fail ()) : 'freshtv5344)

and _menhir_reduce66 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_fun_def) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_labeled_simple_pattern), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = 
# 2514 "reason_parser.mly"
                                       (
      let (l,o,p) = _2 in
      mkexp (Pexp_fun(l, o, p, _3))
    )
# 41619 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce67 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 41626 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ _ (_6 : 'tv_fun_def) ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, (_4 : (
# 857 "reason_parser.mly"
       (string)
# 41632 "reason_parser.ml"
    )), _startpos__4_), _endpos__5_, _startpos__5_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__expr = 
# 2519 "reason_parser.mly"
      ( mkexp (Pexp_newtype(_4, _6)) )
# 41643 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce441 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 41650 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_fun_def) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, (_3 : (
# 857 "reason_parser.mly"
       (string)
# 41656 "reason_parser.ml"
    )), _startpos__3_), _endpos__4_, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_fun_def = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3118 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp ~loc (Pexp_newtype(_3, _5))
      )
# 41670 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce440 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_fun_def) ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_fun_def = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3112 "reason_parser.mly"
      (
       let (l,o,p) = _1 in
       let loc = mklocation _symbolstartpos _endpos in
       mkexp ~ghost:true ~loc (Pexp_fun(l, o, p, _2))
      )
# 41687 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce491 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_case -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_leading_bar_match_case) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_leading_bar_match_cases = 
# 3071 "reason_parser.mly"
                           ( [_1] )
# 41697 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_cases _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce492 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_case -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_leading_bar_match_case) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_leading_bar_match_cases)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_leading_bar_match_cases = 
# 3072 "reason_parser.mly"
                                                   ( _2 :: _1 )
# 41708 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_cases _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1791 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5337 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce35 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5338)

and _menhir_reduce34 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_class_sig_body)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_instance_type = 
# 2219 "reason_parser.mly"
      ( mkcty(Pcty_signature _2) )
# 41731 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1548 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv5335 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce35 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5336)

and _menhir_reduce180 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_), _endpos__2_, _, (_2 : 'tv_opt_semi)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__semi_terminated_seq_expr = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2381 "reason_parser.mly"
                          (
      let loc = mklocation _symbolstartpos _endpos in
      expr_of_let_bindings _1 @@ ghunit ~loc ()
    )
# 41757 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce181 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__semi_terminated_seq_expr_row = 
# 2392 "reason_parser.mly"
                                        (
      let expr = _1 in
      let item_attrs = _2 in
      (* Final item in the sequence - just append item attributes to the
       * expression attributes *)
      {expr with pexp_attributes = item_attrs @ expr.pexp_attributes}
    )
# 41775 "reason_parser.ml"
     in
    _menhir_goto__semi_terminated_seq_expr_row _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error706 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5333 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce283 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5334)

and _menhir_error701 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5331 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce280 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5332)

and _menhir_reduce279 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list_extension), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3357 "reason_parser.mly"
      ( let seq, ext_opt = _2 in
        let loc_rbracket = mklocation _startpos__2_ _endpos__2_ in
        make_real_pat (mktailpat_extension loc_rbracket (List.rev seq) ext_opt) )
# 41809 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error606 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5329 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce280 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5330)

and _menhir_reduce281 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern_comma_list), _startpos__2_), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3364 "reason_parser.mly"
      ( mkpat(Ppat_array(List.rev _2)) )
# 41832 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error599 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5327 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce283 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5328)

and _menhir_reduce578 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_module_binding_body_expr) ->
    let _endpos = _endpos__1_ in
    let _v : 'tv_module_binding_body = 
# 1486 "reason_parser.mly"
                             ( _1 )
# 41851 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce539 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_binding_body_functor -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__module_binding_body_functor) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mod__module_binding_body_functor_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1153 "reason_parser.mly"
 ( {x with pmod_loc = {x.pmod_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41863 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5325) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mod__module_binding_body_functor_) = _v in
    ((match _menhir_s with
    | MenhirState2637 | MenhirState2323 | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5321) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_binding_body_functor_) = _v in
        (_menhir_reduce582 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5322)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5323) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_binding_body_functor_) = _v in
        (_menhir_reduce582 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5324)
    | _ ->
        _menhir_fail ()) : 'freshtv5326)

and _menhir_goto_field_expr_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_field_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState423 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5313 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState425
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run426 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState425
          | GREATERRBRACE ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState425
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState425)) : 'freshtv5314)
    | MenhirState1108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5315 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1109
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run426 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1109
          | GREATERRBRACE ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1109
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1109)) : 'freshtv5316)
    | MenhirState2197 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5317 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2198
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run426 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2198
          | GREATERRBRACE ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2198
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2198)) : 'freshtv5318)
    | MenhirState420 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5319 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2339
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run426 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2339
          | GREATERRBRACE ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2339
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2339)) : 'freshtv5320)
    | _ ->
        _menhir_fail ()

and _menhir_reduce540 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__module_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mod__module_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1153 "reason_parser.mly"
 ( {x with pmod_loc = {x.pmod_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 41968 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5311) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mod__module_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2644 | MenhirState411 | MenhirState451 | MenhirState756 | MenhirState759 | MenhirState2142 | MenhirState2138 | MenhirState2128 | MenhirState1245 | MenhirState799 | MenhirState772 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5307) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce584 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5308)
    | MenhirState403 | MenhirState404 | MenhirState412 | MenhirState2306 | MenhirState2302 | MenhirState2295 | MenhirState2291 | MenhirState762 | MenhirState1476 | MenhirState1945 | MenhirState1948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5309) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mod__module_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce584 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5310)
    | _ ->
        _menhir_fail ()) : 'freshtv5312)

and _menhir_reduce544 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__non_arrowed_module_type) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_mty__non_arrowed_module_type_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1178 "reason_parser.mly"
 ( {x with pmty_loc = {x.pmty_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 42005 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5305) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_mty__non_arrowed_module_type_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5301) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__non_arrowed_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce597 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5302)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState1965 | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5303) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_mty__non_arrowed_module_type_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce597 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5304)
    | _ ->
        _menhir_fail ()) : 'freshtv5306)

and _menhir_reduce556 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type2 -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__core_type2) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_typ__core_type2_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1173 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 42042 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5299) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_typ__core_type2_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState666 | MenhirState272 | MenhirState270 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5295) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type2_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce400 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5296)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5297) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_typ__core_type2_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce400 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5298)
    | _ ->
        _menhir_fail ()) : 'freshtv5300)

and _menhir_goto_package_type_cstrs : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_package_type_cstrs -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5287 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce647 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5288)
    | MenhirState2411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5289 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce650 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5290)
    | MenhirState134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5291 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce647 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5292)
    | MenhirState2470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5293 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_package_type_cstrs) = _v in
        (_menhir_reduce650 _menhir_env (Obj.magic _menhir_stack) _menhir_s _v : 'freshtv5294)
    | _ ->
        _menhir_fail ()

and _menhir_reduce549 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__pattern_without_or -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__pattern_without_or) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__pattern_without_or_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1203 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 42109 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5285) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__pattern_without_or_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5281) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__pattern_without_or_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce667 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5282)
    | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState738 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5283) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_pat__pattern_without_or_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce667 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5284)
    | _ ->
        _menhir_fail ()) : 'freshtv5286)

and _menhir_error2529 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2517 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2429 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5279 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce229 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5280)

and _menhir_reduce228 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_comma_seq)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = 
# 2733 "reason_parser.mly"
      ( mkexp (Pexp_array(List.rev _2)) )
# 42166 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2355 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5277 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce229 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5278)

and _menhir_error2340 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5275 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce239 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5276)

and _menhir_error2199 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5273 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce242 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5274)

and _menhir_error2194 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2184 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5271 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce232 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5272)

and _menhir_reduce822 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_private_flag), _startpos__4_), _endpos__5_, _startpos__5_), _, (_6 : 'tv_label_declarations)), _, (_7 : 'tv_opt_comma)) = _menhir_stack in
    let _8 = () in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__8_ in
    let _v : 'tv_type_kind = 
# 3512 "reason_parser.mly"
      ( (Ptype_record(List.rev _6), _4, Some _2) )
# 42222 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1353 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce818 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos__3_, _startpos__3_), _, (_4 : 'tv_label_declarations)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let _6 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_type_kind = 
# 3504 "reason_parser.mly"
      ( (Ptype_record(List.rev _4), _2, None) )
# 42241 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error1340 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce241 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_field_expr_list)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let _6 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42267 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2780 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let exp = Exp.mk ~loc ~attrs:[] (Pexp_override(List.rev _4)) in
        mkexp(Pexp_open(Fresh, _1, exp))
      )
# 42280 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1110 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5269 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce242 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5270)

and _menhir_reduce417 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_), _, (_2 : 'tv_opt_comma)) = _menhir_stack in
    let _3 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_expr_comma_seq_extension = 
# 3149 "reason_parser.mly"
    ( ([_1], None) )
# 42302 "reason_parser.ml"
     in
    _menhir_goto_expr_comma_seq_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1105 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce231 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__6_ ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_expr_comma_seq)), _, (_5 : 'tv_opt_comma)) = _menhir_stack in
    let _6 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42328 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2739 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let rec_exp = Exp.mk ~loc ~attrs:[] (Pexp_array(List.rev _4)) in
        mkexp(Pexp_open(Fresh, _1, rec_exp))
      )
# 42341 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1095 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (((('freshtv5267 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce232 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5268)

and _menhir_reduce238 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_field_expr_list)), _, (_3 : 'tv_opt_comma)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = 
# 2774 "reason_parser.mly"
      ( mkexp (Pexp_override(List.rev _2)) )
# 42364 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error428 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : (('freshtv5265 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce239 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5266)

and _menhir_goto_tag_field : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_tag_field -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState158 | MenhirState161 | MenhirState2449 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState170 | MenhirState163 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5251 * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
        (_menhir_reduce691 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5252)
    | MenhirState306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5257 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5253 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce145 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5254)
        | BAR ->
            _menhir_reduce691 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5255 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5256)) : 'freshtv5258)
    | MenhirState162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5263 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5259 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce145 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv5260)
        | BAR ->
            _menhir_reduce691 _menhir_env (Obj.magic _menhir_stack)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5261 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_tag_field) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5262)) : 'freshtv5264)
    | _ ->
        _menhir_fail ()

and _menhir_reduce429 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42445 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4105 "reason_parser.mly"
                                                ( "::" )
# 42457 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42467 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3683 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42479 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce430 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42498 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4107 "reason_parser.mly"
                                                ( "false" )
# 42510 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42520 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3683 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42532 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce428 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let ((((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__200_, _startpos__200_), _endpos__2_, _, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _200 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42552 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4104 "reason_parser.mly"
                                                ( "()" )
# 42566 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42576 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3683 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42588 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce431 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _100 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42607 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4108 "reason_parser.mly"
                                                ( "true" )
# 42619 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42629 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3683 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42641 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce427 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 42648 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, (_100 : (
# 906 "reason_parser.mly"
       (string)
# 42654 "reason_parser.ml"
    )), _startpos__100_), _endpos__2_, _, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_extension_constructor_rebind = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42667 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( _1 )
# 42679 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42689 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3683 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         Te.rebind _1 _3 ~loc ~attrs:_4
      )
# 42701 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_rebind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce424 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4105 "reason_parser.mly"
                                                ( "::" )
# 42719 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42729 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42740 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce425 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4107 "reason_parser.mly"
                                                ( "false" )
# 42758 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42768 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42779 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce423 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__200_, _startpos__200_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _200 = () in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4104 "reason_parser.mly"
                                                ( "()" )
# 42800 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42810 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42821 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce426 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4108 "reason_parser.mly"
                                                ( "true" )
# 42839 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42849 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42860 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce422 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 42867 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, (_100 : (
# 906 "reason_parser.mly"
       (string)
# 42873 "reason_parser.ml"
    )), _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _endpos = _endpos__3_ in
    let _v : 'tv_extension_constructor_declaration = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( _1 )
# 42885 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42895 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3677 "reason_parser.mly"
      ( let args, res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Te.decl _1 ~args ?res ~loc ~attrs:_3 )
# 42906 "reason_parser.ml"
     in
    _menhir_goto_extension_constructor_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_label_declarations : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_label_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5243 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1337
        | RBRACE ->
            _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1337
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1337) : 'freshtv5244)
    | MenhirState1351 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5245 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1352
        | RBRACE ->
            _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1352
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1352) : 'freshtv5246)
    | MenhirState2515 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5247 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2516
        | RBRACE ->
            _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2516
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2516) : 'freshtv5248)
    | MenhirState2527 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5249 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            _menhir_run1338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2528
        | RBRACE ->
            _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2528
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2528) : 'freshtv5250)
    | _ ->
        _menhir_fail ()

and _menhir_reduce386 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4105 "reason_parser.mly"
                                                ( "::" )
# 42988 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 42998 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3577 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 43010 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce387 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4107 "reason_parser.mly"
                                                ( "false" )
# 43029 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43039 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3577 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 43051 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce385 : _menhir_env -> ((('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let ((((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__200_, _startpos__200_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _200 = () in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4104 "reason_parser.mly"
                                                ( "()" )
# 43073 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43083 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3577 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 43095 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce388 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4108 "reason_parser.mly"
                                                ( "true" )
# 43114 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43124 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3577 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 43136 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce384 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 43143 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_attributes) ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__100_, (_100 : (
# 906 "reason_parser.mly"
       (string)
# 43149 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_constructor_declaration_leading_bar = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( _1 )
# 43162 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43172 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3577 "reason_parser.mly"
      (
       let args,res = _3 in
       let loc = mklocation _symbolstartpos _endpos in
       Type.constructor _2 ~args ?res ~loc ~attrs:_4
      )
# 43184 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce391 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4105 "reason_parser.mly"
                                                ( "::" )
# 43202 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43212 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3568 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 43225 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce392 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4107 "reason_parser.mly"
                                                ( "false" )
# 43243 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43253 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3568 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 43266 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce390 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let (((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__200_, _, _startpos__200_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _200 = () in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__20_ = _endpos__200_ in
      let _startpos__10_ = _startpos__100_ in
      let _20 = _200 in
      let _10 = _100 in
      let x =
        let _2 = _20 in
        let _1 = _10 in
        
# 4104 "reason_parser.mly"
                                                ( "()" )
# 43287 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__20_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43297 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3568 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 43310 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce393 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _100 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4108 "reason_parser.mly"
                                                ( "true" )
# 43328 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43338 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3568 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 43351 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce336 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_attribute -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_attributes) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_attribute)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_attributes = 
# 4318 "reason_parser.mly"
                         ( _1 :: _2 )
# 43362 "reason_parser.ml"
     in
    _menhir_goto_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce389 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 43369 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_attributes) ->
    let ((_menhir_stack, _endpos__100_, _menhir_s, (_100 : (
# 906 "reason_parser.mly"
       (string)
# 43375 "reason_parser.ml"
    )), _startpos__100_), _endpos__2_, _, (_2 : 'tv_generalized_constructor_arguments)) = _menhir_stack in
    let _endpos = _endpos__3_ in
    let _v : 'tv_constructor_declaration_no_leading_bar = let _1 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4102 "reason_parser.mly"
                                                ( _1 )
# 43387 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43397 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos__100_ in
    let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3568 "reason_parser.mly"
     (
        let args,res = _2 in
        let loc = mklocation _symbolstartpos _endpos in
        Type.constructor _1 ~args ?res ~loc ~attrs:_3
     )
# 43410 "reason_parser.ml"
     in
    _menhir_goto_constructor_declaration_no_leading_bar _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce815 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ (_2 : 'tv_constructor_declarations) ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_kind = 
# 3498 "reason_parser.mly"
      ( (Ptype_variant(_2),  Public, None) )
# 43422 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce819 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ (_4 : 'tv_constructor_declarations) ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_, _startpos__3_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_type_kind = 
# 3506 "reason_parser.mly"
      ( (Ptype_variant(_4), Public,  Some _2) )
# 43435 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce820 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__5_ _ (_5 : 'tv_constructor_declarations) ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _, _startpos__4_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_type_kind = 
# 3508 "reason_parser.mly"
      ( (Ptype_variant(_5), Private, Some _2) )
# 43449 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce816 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : 'tv_constructor_declarations) ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_type_kind = 
# 3500 "reason_parser.mly"
      ( (Ptype_variant(_3), Private, None) )
# 43462 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce786 : _menhir_env -> ((((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_nonrec_flag), _startpos__2_), _, (_3 : 'tv_potentially_long_ident_and_optional_type_parameters)), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_private_flag), _startpos__5_), _, (_6 : 'tv_opt_bar)), _endpos__7_, _, (_7 : 'tv_str_extension_constructors)), _endpos__8_, _, (_8 : 'tv_post_item_attributes)) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__8_ in
    let _v : 'tv_str_type_extension = 
# 3644 "reason_parser.mly"
  (
    if _2 <> Recursive then not_expecting _startpos__2_ _endpos__2_ "nonrec flag";
    let (potentially_long_ident, optional_type_parameters) = _3 in
    Te.mk potentially_long_ident (List.rev _7)
          ~params:optional_type_parameters ~priv:_5 ~attrs:_8
  )
# 43481 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5241) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_str_type_extension) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5237) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce300 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5238)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5239) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_str_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce300 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5240)
    | _ ->
        _menhir_fail ()) : 'freshtv5242)

and _menhir_reduce304 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43525 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1429 "reason_parser.mly"
                                                   (
      let item_attrs = _4 in
      let ident = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_modtype (Mtd.mk ident ~attrs:item_attrs ~loc))
    )
# 43538 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce305 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_type), _startpos__5_), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 43559 "reason_parser.ml"
      
    in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1435 "reason_parser.mly"
                                                                     (
      let ident = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_modtype (Mtd.mk ident ~typ:_5 ~attrs:_6 ~loc))
    )
# 43571 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2325 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 43578 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2319 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce302 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_nonlocal_module_binding_details)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1421 "reason_parser.mly"
                                                                    (
      let (ident, body) = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_module (Mb.mk ident body ~attrs:_4 ~loc))
    )
# 43605 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce517 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_nonlocal_module_binding_details)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv_many_nonlocal_module_bindings = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1490 "reason_parser.mly"
                                                                        (
    let (ident, body) = _4 in
    let loc = mklocation _symbolstartpos _endpos in
    [Mb.mk ident body ~attrs:_5 ~loc]
  )
# 43626 "reason_parser.ml"
     in
    _menhir_goto_many_nonlocal_module_bindings _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2133 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 43633 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_many_class_descriptions : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_class_descriptions -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5235 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5229) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2092 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState2092
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2092) : 'freshtv5230)
    | SEMI ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5231 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_descriptions), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1723 "reason_parser.mly"
                            (
      mksig(Psig_class (List.rev _1))
    )
# 43675 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5232)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5233 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5234)) : 'freshtv5236)

and _menhir_reduce309 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1451 "reason_parser.mly"
                                             (
      let loc = mklocation _symbolstartpos _endpos in
      mkstr(Pstr_include (Incl.mk _2 ~attrs:_3 ~loc))
    )
# 43700 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce297 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1404 "reason_parser.mly"
                              (
      mkstrexp _1 _2
    )
# 43714 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce310 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1455 "reason_parser.mly"
                                        (
    (* No sense in having item_extension_sugar for something that's already an
     * item_extension *)
    mkstr(Pstr_extension (_1, _2))
  )
# 43730 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce325 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_class_declaration = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1793 "reason_parser.mly"
                                                     (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc
  )
# 43748 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5227) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_class_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5223 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_declaration) = _v in
        (_menhir_reduce510 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5224)
    | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5225 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_declaration) = _v in
        (_menhir_reduce510 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5226)
    | _ ->
        _menhir_fail ()) : 'freshtv5228)

and _menhir_reduce327 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_type_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_class_type_declaration = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2326 "reason_parser.mly"
                                                          (
    let (ident, instance_type, virt, class_type_params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Ci.mk ident instance_type ~virt:virt ~params:class_type_params ~attrs:_3 ~loc
  )
# 43787 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5221) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_class_type_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1875 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5217 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_type_declaration) = _v in
        (_menhir_reduce514 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5218)
    | MenhirState2736 | MenhirState2086 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5219 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_class_type_declaration) = _v in
        (_menhir_reduce514 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5220)
    | _ ->
        _menhir_fail ()) : 'freshtv5222)

and _menhir_reduce330 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_nonlocal_module_binding_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_nonlocal_module_bindings = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1501 "reason_parser.mly"
                                                             (
    let (ident, body) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Mb.mk ident body ~attrs:_3 ~loc
  )
# 43826 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5215) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_nonlocal_module_bindings) = _v in
    ((match _menhir_s with
    | MenhirState1870 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5211 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_nonlocal_module_bindings) = _v in
        (_menhir_reduce518 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5212)
    | MenhirState2731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5213 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_nonlocal_module_bindings) = _v in
        (_menhir_reduce518 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5214)
    | _ ->
        _menhir_fail ()) : 'freshtv5216)

and _menhir_reduce331 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_type_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_type_declaration = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3471 "reason_parser.mly"
                                                      (
    let (ident, params, constraints, kind, priv, manifest) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Type.mk ident
      ~params:params ~cstrs:constraints
      ~kind ~priv ?manifest ~attrs:_3 ~loc
  )
# 43867 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5209) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_type_declaration) = _v in
    ((match _menhir_s with
    | MenhirState1853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5205 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_type_declaration) = _v in
        (_menhir_reduce520 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5206)
    | MenhirState2730 | MenhirState2074 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5207 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_type_declaration) = _v in
        (_menhir_reduce520 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5208)
    | _ ->
        _menhir_fail ()) : 'freshtv5210)

and _menhir_reduce509 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_many_class_declarations = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1782 "reason_parser.mly"
                                                         (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    [Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc]
  )
# 43907 "reason_parser.ml"
     in
    _menhir_goto_many_class_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__class_field : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_field -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5203) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_field) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5201) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_field) : 'tv__class_field) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cf__class_field_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1198 "reason_parser.mly"
 ( {x with pcf_loc = {x.pcf_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 43931 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5199) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cf__class_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5197) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cf__class_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5195) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_cf__class_field_) : 'tv_mark_position_cf__class_field_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_class_field = 
# 1977 "reason_parser.mly"
                                            (_1)
# 43952 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5193) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_field) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5191 * Lexing.position * _menhir_state * 'tv_class_field) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_error1752 _menhir_env (Obj.magic _menhir_stack)
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | SEMI ->
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv5189 * Lexing.position * _menhir_state * 'tv_class_field) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          ((let _menhir_stack = (_menhir_stack, _endpos) in
          let _menhir_env = _menhir_discard _menhir_env in
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1744 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INHERIT ->
              _menhir_run1737 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INITIALIZER ->
              _menhir_run1733 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | METHOD ->
              _menhir_run1690 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | VAL ->
              _menhir_run1667 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACE ->
              _menhir_reduce695 _menhir_env (Obj.magic _menhir_stack) MenhirState1753
          | _ ->
              assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1753) : 'freshtv5190)
      | RBRACE ->
          _menhir_reduce696 _menhir_env (Obj.magic _menhir_stack)
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_error1752 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5192)) : 'freshtv5194)) : 'freshtv5196)) : 'freshtv5198)) : 'freshtv5200)) : 'freshtv5202)) : 'freshtv5204)

and _menhir_reduce504 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_rec_flag)), _endpos__3_, _, (_3 : 'tv_let_binding_body)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_let_binding_impl = 
# 2882 "reason_parser.mly"
                                                       (
      (_2, _3, _4)
    )
# 44013 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5187) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_let_binding_impl) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2673 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1632 | MenhirState1663 | MenhirState1161 | MenhirState1176 | MenhirState1167 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5179) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce496 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5180)
    | MenhirState2743 | MenhirState1664 | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5181 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce497 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5182)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState475 | MenhirState2259 | MenhirState2253 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5183) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce496 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5184)
    | MenhirState2255 | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5185 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_let_binding_impl) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce497 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5186)
    | _ ->
        _menhir_fail ()) : 'freshtv5188)

and _menhir_reduce513 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_type_declaration_details)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_many_class_type_declarations = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2315 "reason_parser.mly"
                                                                   (
    let (ident, instance_type, virt, class_type_params) = _3 in
    let loc = mklocation _symbolstartpos _endpos in
    [Ci.mk ident instance_type ~virt:virt ~params:class_type_params ~attrs:_4 ~loc]
  )
# 44073 "reason_parser.ml"
     in
    _menhir_goto_many_class_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce780 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_extension_constructor_declaration)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_exception_declaration = 
# 3586 "reason_parser.mly"
      (
        let ext = _2 in
        {ext with pext_attributes = ext.pext_attributes @ _3}
      )
# 44089 "reason_parser.ml"
     in
    _menhir_goto_str_exception_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce781 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_extension_constructor_rebind)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_str_exception_declaration = 
# 3591 "reason_parser.mly"
      (
        let ext = _2 in
        {ext with pext_attributes = ext.pext_attributes @ _3}
      )
# 44105 "reason_parser.ml"
     in
    _menhir_goto_str_exception_declaration _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce298 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_primitive_declaration)), _endpos__7_, _, (_7 : 'tv_post_item_attributes)) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 44126 "reason_parser.ml"
      
    in
    let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1407 "reason_parser.mly"
                                                                                                (
      let loc = mklocation _symbolstartpos _endpos in
      mkstr
        (Pstr_primitive (Val.mk _2 _4 ~prim:_6 ~attrs:_7 ~loc))
    )
# 44138 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_many_module_rec_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_many_module_rec_declarations -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5177 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
    ((assert (not _menhir_env._menhir_error);
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5171) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run1465 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2082 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2082) : 'freshtv5172)
    | SEMI ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5173 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_module_rec_declarations), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1705 "reason_parser.mly"
                                 (
      mksig(Psig_recmodule (List.rev _1))
    )
# 44176 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5174)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5175 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5176)) : 'freshtv5178)

and _menhir_reduce606 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_override_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_mod_longident), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_open_statement = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 44202 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1739 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        Opn.mk _3 ~override:_2 ~attrs:_4 ~loc
      )
# 44213 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5169) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_open_statement) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5161) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_open_statement) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce306 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5162)
    | MenhirState2563 | MenhirState2565 | MenhirState1251 | MenhirState1408 | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5165) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_open_statement) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5163) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_open_statement) : 'tv_open_statement) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1716 "reason_parser.mly"
                   (
      mksig(Psig_open _1)
    )
# 44250 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv5164)) : 'freshtv5166)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5167) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_open_statement) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce306 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5168)
    | _ ->
        _menhir_fail ()) : 'freshtv5170)

and _menhir_goto__signature_item : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__signature_item -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5159) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__signature_item) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5157) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__signature_item) : 'tv__signature_item) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _v : 'tv_mark_position_sig__signature_item_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1183 "reason_parser.mly"
 ( {x with psig_loc = {x.psig_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 44283 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5155) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_sig__signature_item_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5153) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_sig__signature_item_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5151) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_sig__signature_item_) : 'tv_mark_position_sig__signature_item_) = _v in
    ((let _v : 'tv_signature_item = 
# 1669 "reason_parser.mly"
                                                   ( _1 )
# 44300 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5149) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_signature_item) = _v in
    ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1251 | MenhirState2068 | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5141 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5137 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2051 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run2048 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2039 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2036 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run1409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run1404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run1252 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | RBRACE ->
                _menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState2068
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2068) : 'freshtv5138)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5139 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5140)) : 'freshtv5142)
    | MenhirState2565 | MenhirState2563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5147 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5143 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | CLASS ->
                _menhir_run2051 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXCEPTION ->
                _menhir_run2048 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2039 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2036 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run1409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run1404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run1252 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2565 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EOF ->
                _menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState2565
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2565) : 'freshtv5144)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5145 * _menhir_state * 'tv_signature_item) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5146)) : 'freshtv5148)
    | _ ->
        _menhir_fail ()) : 'freshtv5150)) : 'freshtv5152)) : 'freshtv5154)) : 'freshtv5156)) : 'freshtv5158)) : 'freshtv5160)

and _menhir_reduce519 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_nonrec_flag), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_type_declaration_details)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_many_type_declarations = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = if Pervasives.(!=) _startpos_x_ _endpos_x_ then
        _startpos_x_
      else
        _endpos in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 44423 "reason_parser.ml"
      
    in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3460 "reason_parser.mly"
                                                                           (
      let (ident, params, constraints, kind, priv, manifest) = _3 in
      let loc = mklocation _symbolstartpos _endpos in
      [Type.mk ident
       ~params:params ~cstrs:constraints
       ~kind ~priv ?manifest ~attrs:(add_nonrec _2 _4) ~loc]
  )
# 44437 "reason_parser.ml"
     in
    _menhir_goto_many_type_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce328 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_let_binding_body)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_and_let_binding = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2867 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in
         mklb _2 _3 loc
      )
# 44454 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5135) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_and_let_binding) = _v in
    ((match _menhir_s with
    | MenhirState2742 | MenhirState1662 | MenhirState1166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5131 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_let_binding) = _v in
        (_menhir_reduce506 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5132)
    | MenhirState2252 | MenhirState1885 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5133 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_and_let_binding) = _v in
        (_menhir_reduce506 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv5134)
    | _ ->
        _menhir_fail ()) : 'freshtv5136)

and _menhir_reduce674 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_attribute)), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_post_item_attributes = 
# 4323 "reason_parser.mly"
                                        ( _1 :: _2 )
# 44486 "reason_parser.ml"
     in
    _menhir_goto_post_item_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run1156 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5127 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1156 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5123 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 906 "reason_parser.mly"
       (string)
# 44513 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EQUAL ->
                _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1244) : 'freshtv5124)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5125 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5126)) : 'freshtv5128)
    | OPEN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5129 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1156 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BANG ->
            _menhir_run393 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1157 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _ ->
            _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack) MenhirState1157
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1157) : 'freshtv5130)
    | REC ->
        _menhir_run435 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce686 _menhir_env (Obj.magic _menhir_stack) MenhirState1156
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1156

and _menhir_error1160 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce533 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1168 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 44578 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5121) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5117) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce411 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5118)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 | MenhirState2574 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5119) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce411 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5120)
    | _ ->
        _menhir_fail ()) : 'freshtv5122)

and _menhir_reduce526 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__class_simple_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_simple_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1193 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 44615 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5115) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_simple_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1738 | MenhirState1663 | MenhirState1632 | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5111) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce361 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5112)
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5113) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__class_simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce361 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5114)
    | _ ->
        _menhir_fail ()) : 'freshtv5116)

and _menhir_goto_toplevel_directive : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_toplevel_directive -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5099 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5095 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5093 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            let (_endpos__2_ : Lexing.position) = _endpos in
            ((let (_menhir_stack, _menhir_s, (_1 : 'tv_toplevel_directive)) = _menhir_stack in
            let _2 = () in
            let _v : 'tv__toplevel_phrase = 
# 1126 "reason_parser.mly"
                                        ( _1 )
# 44665 "reason_parser.ml"
             in
            _menhir_goto__toplevel_phrase _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5094)) : 'freshtv5096)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5097 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5098)) : 'freshtv5100)
    | MenhirState2760 | MenhirState2757 | MenhirState2753 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5109 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EOF ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5103 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5101 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, (_1 : 'tv_toplevel_directive)) = _menhir_stack in
            let _2 = () in
            let _v : 'tv__use_file = 
# 1135 "reason_parser.mly"
                                                   ( [_1] )
# 44691 "reason_parser.ml"
             in
            _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5102)) : 'freshtv5104)
        | SEMI ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5105 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_stack = (_menhir_stack, _endpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | ASSERT ->
                _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BACKQUOTE ->
                _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | BANG ->
                _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CHAR _v ->
                _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | CLASS ->
                _menhir_run2660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EOF ->
                _menhir_run2754 _menhir_env (Obj.magic _menhir_stack) MenhirState2757
            | EXCEPTION ->
                _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EXTERNAL ->
                _menhir_run2647 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FLOAT _v ->
                _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FOR ->
                _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | FUN ->
                _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | IF ->
                _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INCLUDE ->
                _menhir_run2644 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT _v ->
                _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT32 _v ->
                _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | INT64 _v ->
                _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LAZY ->
                _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACE ->
                _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACELESS ->
                _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETATATAT ->
                _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETBAR ->
                _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENTPERCENT ->
                _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LET ->
                _menhir_run2635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUS ->
                _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MINUSDOT ->
                _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run2628 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NATIVEINT _v ->
                _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | NEW ->
                _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | OPEN ->
                _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PERCENT ->
                _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUSDOT ->
                _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PREFIXOP _v ->
                _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run2619 _menhir_env (Obj.magic _menhir_stack) MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | STRING _v ->
                _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SWITCH ->
                _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRY ->
                _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TYPE ->
                _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | WHILE ->
                _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2757) : 'freshtv5106)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv5107 * _menhir_state * 'tv_toplevel_directive) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5108)) : 'freshtv5110)
    | _ ->
        _menhir_fail ()

and _menhir_run1983 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 44811 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5091 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 44823 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1983 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1984 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1984) : 'freshtv5092)
    | AND | DOT | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
        _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1983

and _menhir_reduce216 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_), _), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 44862 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2684 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let pat = mkpat (Ppat_var (mkloc "this" loc)) in
        mkexp(Pexp_open (Fresh, _1,
                         mkexp(Pexp_object(Cstr.mk pat [])))) )
# 44875 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run136 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 44882 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce466 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run555 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 44892 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce577 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run755 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run805 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5089 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState755 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState756 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState756 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState756 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState756 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState756 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState756) : 'freshtv5090)
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState755 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState755

and _menhir_run2182 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2182 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2182

and _menhir_run2187 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run2188 _menhir_env (Obj.magic _menhir_stack) MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2187

and _menhir_run2197 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run421 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2197 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2197

and _menhir_run2202 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | DOTDOTDOT ->
        _menhir_run1115 _menhir_env (Obj.magic _menhir_stack) MenhirState2202 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5087 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2202 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce216 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv5088)
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2202 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2202

and _menhir_run635 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState635 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState635

and _menhir_run554 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run555 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState554 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState554

and _menhir_reduce467 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 45255 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos__3_, _, (_3 : (
# 857 "reason_parser.mly"
       (string)
# 45261 "reason_parser.ml"
    )), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_label_longident = 
# 4125 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 45269 "reason_parser.ml"
     in
    _menhir_goto_label_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run11 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 45276 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce577 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_reduce351 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 45285 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _ (_3 : (
# 857 "reason_parser.mly"
       (string)
# 45290 "reason_parser.ml"
  )) _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_class_longident = 
# 4169 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 45299 "reason_parser.ml"
     in
    _menhir_goto_class_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1254 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 45306 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MINUS ->
        _menhir_run1262 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run1258 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run1256 _menhir_env (Obj.magic _menhir_stack) MenhirState1254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run1255 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState1254

and _menhir_run91 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce155 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run92 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState92 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState92 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState92

and _menhir_run96 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5081 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState98 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState98 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState98) : 'freshtv5082)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5083 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce157 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5084)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5085 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5086)

and _menhir_run100 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | QUOTE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5075 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState102 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState102 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState102) : 'freshtv5076)
    | UNDERSCORE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5077 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce159 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv5078)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5079 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5080)

and _menhir_run1631 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1631

and _menhir_run1632 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1657 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1633 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1632 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1632
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1632

and _menhir_run1637 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1637

and _menhir_run1641 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1641 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1641

and _menhir_reduce679 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv_private_flag = 
# 4202 "reason_parser.mly"
                                                ( Public )
# 45551 "reason_parser.ml"
     in
    _menhir_goto_private_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1373 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce680 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce680 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_private_flag = 
# 4203 "reason_parser.mly"
                                                ( Private )
# 45570 "reason_parser.ml"
     in
    _menhir_goto_private_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce589 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv_mutable_flag = 
# 4206 "reason_parser.mly"
                                                ( Immutable )
# 45581 "reason_parser.ml"
     in
    _menhir_goto_mutable_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce590 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mutable_flag = 
# 4207 "reason_parser.mly"
                                                ( Mutable )
# 45594 "reason_parser.ml"
     in
    _menhir_goto_mutable_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_class_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState7 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5055 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce237 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5056)
    | MenhirState109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5057 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce139 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5058)
    | MenhirState120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5059 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState121
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState121 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce139 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5060)
    | MenhirState122 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5061 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce139 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5062)
    | MenhirState181 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5063 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce237 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5064)
    | MenhirState243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5065 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState244
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState244 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce139 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv5066)
    | MenhirState1641 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5067 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1642
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | AS | COLON | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI ->
              _menhir_reduce15 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1642)) : 'freshtv5068)
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1632 | MenhirState1663 | MenhirState1738 | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5069 * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce47 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5070)
    | MenhirState2681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5071 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2682
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2682 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
              _menhir_reduce15 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2682)) : 'freshtv5072)
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5073 * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce47 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5074)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lident_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lident_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1208 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5035 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 45790 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv5033 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 45796 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 45801 "reason_parser.ml"
        )), _startpos__1_), _, (_2 : 'tv_lident_list)) = _menhir_stack in
        let _v : 'tv_lident_list = 
# 2878 "reason_parser.mly"
                                      ( _1 :: _2 )
# 45806 "reason_parser.ml"
         in
        _menhir_goto_lident_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv5034)) : 'freshtv5036)
    | MenhirState1207 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5041 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5037 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1211) : 'freshtv5038)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5039 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5040)) : 'freshtv5042)
    | MenhirState1721 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv5047 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((((('freshtv5043 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1723 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1723) : 'freshtv5044)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((((('freshtv5045 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5046)) : 'freshtv5048)
    | MenhirState1896 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv5053 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | DOT ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5049 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACKET ->
                _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETGREATER ->
                _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETLESS ->
                _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESS ->
                _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSDOTDOTGREATER ->
                _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LESSGREATER ->
                _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | SHARP ->
                _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1898) : 'freshtv5050)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ((('freshtv5051 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5052)) : 'freshtv5054)
    | _ ->
        _menhir_fail ()

and _menhir_run246 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState246 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState246 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState246

and _menhir_error2648 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2040 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1710 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1710 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1710

and _menhir_error1479 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1470 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1189 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5031 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1189 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5027 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 46142 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv5023 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 46154 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run1195 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run1193 _menhir_env (Obj.magic _menhir_stack) MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run1189 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1192) : 'freshtv5024)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv5025 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 46219 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5026)) : 'freshtv5028)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv5029 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5030)) : 'freshtv5032)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1189

and _menhir_run1193 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1193 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1193

and _menhir_run1195 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1195 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1195

and _menhir_reduce837 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos__3_, _, (_3 : 'tv_val_ident), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_val_longident = 
# 4114 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 46362 "reason_parser.ml"
     in
    _menhir_goto_val_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce164 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__pattern_without_or = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 46381 "reason_parser.ml"
      
    in
    
# 3265 "reason_parser.mly"
      ( mkpat(Ppat_alias(_1, _3)) )
# 46387 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce269 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 46405 "reason_parser.ml"
      
    in
    
# 3333 "reason_parser.mly"
      ( mkpat(Ppat_var _1) )
# 46411 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce836 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_val_ident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_val_longident = 
# 4113 "reason_parser.mly"
                                                ( Lident _1 )
# 46423 "reason_parser.ml"
     in
    _menhir_goto_val_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_single_attr_id : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_single_attr_id -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2032 | MenhirState1474 | MenhirState455 | MenhirState453 | MenhirState439 | MenhirState396 | MenhirState333 | MenhirState189 | MenhirState174 | MenhirState71 | MenhirState19 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5017 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error70 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5015 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AND ->
                  _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AS ->
                  _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ASSERT ->
                  _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BEGIN ->
                  _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CONSTRAINT ->
                  _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DO ->
                  _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DONE ->
                  _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOWNTO ->
                  _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ELSE ->
                  _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | END ->
                  _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTION ->
                  _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTOR ->
                  _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IN ->
                  _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INHERIT ->
                  _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INITIALIZER ->
                  _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | METHOD ->
                  _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MUTABLE ->
                  _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NONREC ->
                  _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OBJECT ->
                  _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | REC ->
                  _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SIG ->
                  _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRUCT ->
                  _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | THEN ->
                  _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TO ->
                  _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VAL ->
                  _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VIRTUAL ->
                  _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHEN ->
                  _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WITH ->
                  _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState71 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState71) : 'freshtv5016)
          | ASSERT | BACKQUOTE | BANG | CHAR _ | CLASS | COLON | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | IF | INCLUDE | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LET | LIDENT _ | LPAREN | MINUS | MINUSDOT | MODULE | NATIVEINT _ | NEW | OPEN | PERCENT | PLUS | PLUSDOT | PREFIXOP _ | QUESTION | RBRACKET | STRING _ | SWITCH | TRUE | TRY | TYPE | UIDENT _ | WHILE ->
              _menhir_reduce332 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error70 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5018)
    | MenhirState389 | MenhirState337 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5021 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error388 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv5019 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position) = Obj.magic _menhir_stack in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AND ->
                  _menhir_run387 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AS ->
                  _menhir_run386 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ASSERT ->
                  _menhir_run385 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BEGIN ->
                  _menhir_run384 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CLASS ->
                  _menhir_run383 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CONSTRAINT ->
                  _menhir_run382 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DO ->
                  _menhir_run381 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DONE ->
                  _menhir_run380 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOWNTO ->
                  _menhir_run379 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | ELSE ->
                  _menhir_run378 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | END ->
                  _menhir_run377 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXTERNAL ->
                  _menhir_run375 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run374 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run372 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTION ->
                  _menhir_run371 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUNCTOR ->
                  _menhir_run370 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run369 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IN ->
                  _menhir_run368 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INCLUDE ->
                  _menhir_run367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INHERIT ->
                  _menhir_run366 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INITIALIZER ->
                  _menhir_run365 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run364 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run363 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run362 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | METHOD ->
                  _menhir_run361 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run360 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MUTABLE ->
                  _menhir_run359 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run358 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NONREC ->
                  _menhir_run357 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OBJECT ->
                  _menhir_run356 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OPEN ->
                  _menhir_run355 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run354 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PRIVATE ->
                  _menhir_run353 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | REC ->
                  _menhir_run352 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SIG ->
                  _menhir_run351 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRUCT ->
                  _menhir_run350 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run349 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | THEN ->
                  _menhir_run348 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TO ->
                  _menhir_run347 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run346 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run345 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  _menhir_run344 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run343 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VAL ->
                  _menhir_run342 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | VIRTUAL ->
                  _menhir_run341 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHEN ->
                  _menhir_run340 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run339 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WITH ->
                  _menhir_run338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState389) : 'freshtv5020)
          | ASSERT | BACKQUOTE | BANG | CHAR _ | CLASS | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | IF | INCLUDE | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LET | LIDENT _ | LPAREN | MINUS | MINUSDOT | MODULE | NATIVEINT _ | NEW | OPEN | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | TYPE | UIDENT _ | WHILE ->
              _menhir_reduce332 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error388 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv5022)
    | _ ->
        _menhir_fail ()

and _menhir_reduce536 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_expr) _startpos_x_ ->
    let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__simple_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1168 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 46691 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv5013) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__simple_expr_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState2555 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState180 | MenhirState184 | MenhirState187 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState457 | MenhirState464 | MenhirState470 | MenhirState2274 | MenhirState473 | MenhirState2268 | MenhirState2270 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState747 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1647 | MenhirState1648 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState840 | MenhirState842 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 | MenhirState816 | MenhirState817 | MenhirState811 | MenhirState805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5009) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce718 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5010)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 | MenhirState2687 | MenhirState2688 | MenhirState2574 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState417 | MenhirState2350 | MenhirState418 | MenhirState419 | MenhirState2277 | MenhirState474 | MenhirState2229 | MenhirState729 | MenhirState730 | MenhirState734 | MenhirState2213 | MenhirState740 | MenhirState746 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState764 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState793 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState894 | MenhirState895 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState953 | MenhirState955 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 | MenhirState929 | MenhirState930 | MenhirState920 | MenhirState910 | MenhirState912 | MenhirState913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5011) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_expr_) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce718 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv5012)
    | _ ->
        _menhir_fail ()) : 'freshtv5014)

and _menhir_run1667 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run393 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1667 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5007) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1667 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MUTABLE ->
            _menhir_run1325 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1668 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState1668
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1668) : 'freshtv5008)
    | LIDENT _ | MUTABLE ->
        _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack) MenhirState1667
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1667

and _menhir_run1690 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run393 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1690 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ | PRIVATE | VIRTUAL ->
        _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack) MenhirState1690
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1690

and _menhir_run1733 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | EQUALGREATER ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5003 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1734 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1734) : 'freshtv5004)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5005 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv5006)

and _menhir_run1737 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run393 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1737 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS | FUN | LBRACE | LBRACKETPERCENT | LIDENT _ | LPAREN | UIDENT _ ->
        _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack) MenhirState1737
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1737

and _menhir_run1744 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1744 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1744

and _menhir_goto_label_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_label_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState135 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4967 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error140 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4965 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState141 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState141) : 'freshtv4966)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error140 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4968)
    | MenhirState240 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4971 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error241 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4969 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState242 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState242) : 'freshtv4970)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error241 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4972)
    | MenhirState516 | MenhirState615 | MenhirState525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4975 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error530 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4973 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState531 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState531) : 'freshtv4974)
          | COMMA | RBRACE ->
              _menhir_reduce482 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error530 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4976)
    | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4981 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : (('freshtv4977 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4978)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4979 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1018 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1018) : 'freshtv4980)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce212 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4982)
    | MenhirState815 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4985 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1058 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4983 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1059 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1059) : 'freshtv4984)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce212 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4986)
    | MenhirState1131 | MenhirState1117 | MenhirState1120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4989 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1122 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4987 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1123 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1123) : 'freshtv4988)
          | COMMA | RBRACE ->
              _menhir_reduce475 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1122 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4990)
    | MenhirState432 | MenhirState475 | MenhirState2202 | MenhirState1113 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4993 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1133 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4991 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1134) : 'freshtv4992)
          | COMMA ->
              _menhir_reduce475 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1133 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4994)
    | MenhirState1420 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1421
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run1262 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run1258 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run1256 _menhir_env (Obj.magic _menhir_stack) MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run1255 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState1421) : 'freshtv4996)
    | MenhirState1972 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1973
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | MINUS ->
              _menhir_run1262 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run1258 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run1256 _menhir_env (Obj.magic _menhir_stack) MenhirState1973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run1255 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1973 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState1973) : 'freshtv4998)
    | MenhirState736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4999 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce212 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5000)
    | MenhirState472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv5001 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce212 _menhir_env (Obj.magic _menhir_stack) : 'freshtv5002)
    | _ ->
        _menhir_fail ()

and _menhir_reduce260 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_module_expr = 
# 1266 "reason_parser.mly"
      ( mkmod (Pmod_structure []) )
# 47514 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce444 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 47521 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 906 "reason_parser.mly"
       (string)
# 47527 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_functor_arg_name = 
# 1224 "reason_parser.mly"
               ( _1 )
# 47534 "reason_parser.ml"
     in
    _menhir_goto_functor_arg_name _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_functor_arg : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_functor_arg -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2301 | MenhirState758 | MenhirState2137 | MenhirState1944 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4959 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_functor_arg) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4957 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_functor_arg) : 'tv_functor_arg) = _v in
        let (_startpos__2_ : Lexing.position) = _startpos in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_functor_args), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _v : 'tv_functor_args = 
# 1230 "reason_parser.mly"
      ( _2 :: _1 )
# 47557 "reason_parser.ml"
         in
        _menhir_goto_functor_args _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv4958)) : 'freshtv4960)
    | MenhirState2637 | MenhirState2323 | MenhirState443 | MenhirState757 | MenhirState1244 | MenhirState1943 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4963) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_functor_arg) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4961) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_functor_arg) : 'tv_functor_arg) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _v : 'tv_functor_args = 
# 1232 "reason_parser.mly"
      ( [ _1 ] )
# 47575 "reason_parser.ml"
         in
        _menhir_goto_functor_args _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv4962)) : 'freshtv4964)
    | _ ->
        _menhir_fail ()

and _menhir_reduce835 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ _startpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_operator)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_val_ident = 
# 4070 "reason_parser.mly"
                                                ( _2 )
# 47591 "reason_parser.ml"
     in
    _menhir_goto_val_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run174 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState174 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState174

and _menhir_run333 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState333 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState333

and _menhir_reduce449 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _menhir_s _startpos_x0_ ->
    let x0 = () in
    let _endpos = _endpos_x0_ in
    let _v : 'tv_generalized_constructor_arguments = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 47830 "reason_parser.ml"
      
    in
    
# 3606 "reason_parser.mly"
      ( Location.raise_errorf ~loc:_1.loc "The of keyword is not used in Reason" )
# 47836 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce363 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_class_type_parameters = 
# 1877 "reason_parser.mly"
              ( [] )
# 47845 "reason_parser.ml"
     in
    _menhir_goto_class_type_parameters _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce202 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_constant) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_expr = 
# 2639 "reason_parser.mly"
             ( mkexp (Pexp_constant _1) )
# 47856 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce707 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : 'tv_constant) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_signed_constant = 
# 4048 "reason_parser.mly"
                                           ( _1 )
# 47867 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_payload : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_payload -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4901 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4897 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce458 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4898)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4899 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4900)) : 'freshtv4902)
    | MenhirState2033 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4907 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4903 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce437 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4904)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4905 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4906)) : 'freshtv4908)
    | MenhirState456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4913 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4909 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce437 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4910)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4911 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4912)) : 'freshtv4914)
    | MenhirState454 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4919 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4915 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce458 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4916)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4917 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4918)) : 'freshtv4920)
    | MenhirState440 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4925 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4921 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce457 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4922)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4923 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4924)) : 'freshtv4926)
    | MenhirState397 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4931 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4927 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce457 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4928)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4929 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4930)) : 'freshtv4932)
    | MenhirState334 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4937 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4933 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4934)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4935 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4936)) : 'freshtv4938)
    | MenhirState190 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4943 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4939 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce421 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4940)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4941 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4942)) : 'freshtv4944)
    | MenhirState175 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4949 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4945 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce334 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4946)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4947 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4948)) : 'freshtv4950)
    | MenhirState73 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4955 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACKET ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4951 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce421 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4952)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv4953 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) * _menhir_state * 'tv_payload) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4954)) : 'freshtv4956)
    | _ ->
        _menhir_fail ()

and _menhir_reduce252 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_structure)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_expr = 
# 1240 "reason_parser.mly"
      ( mkmod(Pmod_structure(_2)) )
# 48078 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1961 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce260 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run109 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState109 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState109 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState109

and _menhir_run117 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce449 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1492 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1492 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1492

and _menhir_run1493 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RBRACKET ->
        _menhir_run515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1493
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1493

and _menhir_run1498 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1498 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1498

and _menhir_error2567 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error2106 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4895 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce266 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4896)

and _menhir_reduce265 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__3_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_signature)) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_module_type = 
# 1548 "reason_parser.mly"
      ( mkmty(Pmty_signature _2) )
# 48194 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2100 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4893 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
    ((let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce266 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4894)

and _menhir_reduce705 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_signature_item) * Lexing.position) * _menhir_state * 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_signature_item)), _endpos__2_), _, (_3 : 'tv_signature)) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_signature = 
# 1666 "reason_parser.mly"
                                  ( _1 :: _3 )
# 48214 "reason_parser.ml"
     in
    _menhir_goto_signature _menhir_env _menhir_stack _menhir_s _v

and _menhir_run445 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4891) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_functor_arg_name = 
# 1225 "reason_parser.mly"
               ( "_" )
# 48232 "reason_parser.ml"
     in
    _menhir_goto_functor_arg_name _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4892)

and _menhir_run1249 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 48239 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1249 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
    | EQUALGREATER | LBRACKETAT | RPAREN | WITH ->
        _menhir_reduce453 _menhir_env (Obj.magic _menhir_stack)
    | COLON ->
        _menhir_reduce444 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1249

and _menhir_run402 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4887 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState403 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState403) : 'freshtv4888)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4889 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4890)

and _menhir_run1250 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1250 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1250 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1250 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1250 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1250 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1249 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1250 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run445 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1250 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1250

and _menhir_run1251 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run2051 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run2048 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2039 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2036 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run1404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1252 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState1251
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1251

and _menhir_goto_meth_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_meth_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4875 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4871 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce142 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4872)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4873 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4874)) : 'freshtv4876)
    | MenhirState284 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4881 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | GREATER ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4877 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce142 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv4878)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4879 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4880)) : 'freshtv4882)
    | MenhirState288 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4885 * _menhir_state * 'tv_field) * Lexing.position * _menhir_state) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4883 * _menhir_state * 'tv_field) * Lexing.position * _menhir_state) * _menhir_state * 'tv_meth_list) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_field)), _endpos__2_, _), _, (_3 : 'tv_meth_list)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_meth_list = 
# 4024 "reason_parser.mly"
                                              ( let (f, c) = _3 in (_1 :: f, c) )
# 48408 "reason_parser.ml"
         in
        _menhir_goto_meth_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4884)) : 'freshtv4886)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_bar : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_opt_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4857 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState158 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState158) : 'freshtv4858)
    | MenhirState159 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4859 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState161) : 'freshtv4860)
    | MenhirState291 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4861 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState292 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState292) : 'freshtv4862)
    | MenhirState301 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4863 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState303 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState303) : 'freshtv4864)
    | MenhirState1374 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4865 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run1389 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1386 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1382 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1379 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1375 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1375 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1375) : 'freshtv4866)
    | MenhirState2537 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4867 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run1524 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2538 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run1518 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2538 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2538 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run1505 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2538 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run1490 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2538 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2538) : 'freshtv4868)
    | MenhirState2587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4869 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLONCOLON ->
            _menhir_run2608 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run2604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2599 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run2595 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run2589 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2588) : 'freshtv4870)
    | _ ->
        _menhir_fail ()

and _menhir_reduce717 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 871 "reason_parser.mly"
       (nativeint)
# 48636 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 871 "reason_parser.mly"
       (nativeint)
# 48641 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4058 "reason_parser.mly"
                                           ( Const_nativeint _2 )
# 48650 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce716 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 841 "reason_parser.mly"
       (int64)
# 48657 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 841 "reason_parser.mly"
       (int64)
# 48662 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4057 "reason_parser.mly"
                                           ( Const_int64 _2 )
# 48671 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce715 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 840 "reason_parser.mly"
       (int32)
# 48678 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 840 "reason_parser.mly"
       (int32)
# 48683 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4056 "reason_parser.mly"
                                           ( Const_int32 _2 )
# 48692 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce713 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 839 "reason_parser.mly"
       (int)
# 48699 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 839 "reason_parser.mly"
       (int)
# 48704 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4054 "reason_parser.mly"
                                           ( Const_int _2 )
# 48713 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce714 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 821 "reason_parser.mly"
       (string)
# 48720 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 821 "reason_parser.mly"
       (string)
# 48725 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4055 "reason_parser.mly"
                                           ( Const_float _2 )
# 48734 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce712 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 871 "reason_parser.mly"
       (nativeint)
# 48741 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 871 "reason_parser.mly"
       (nativeint)
# 48746 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4053 "reason_parser.mly"
                                           ( Const_nativeint(Nativeint.neg _2) )
# 48755 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce711 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 841 "reason_parser.mly"
       (int64)
# 48762 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 841 "reason_parser.mly"
       (int64)
# 48767 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4052 "reason_parser.mly"
                                           ( Const_int64(Int64.neg _2) )
# 48776 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce710 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 840 "reason_parser.mly"
       (int32)
# 48783 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 840 "reason_parser.mly"
       (int32)
# 48788 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4051 "reason_parser.mly"
                                           ( Const_int32(Int32.neg _2) )
# 48797 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce708 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 839 "reason_parser.mly"
       (int)
# 48804 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 839 "reason_parser.mly"
       (int)
# 48809 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4049 "reason_parser.mly"
                                           ( Const_int(- _2) )
# 48818 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce709 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 821 "reason_parser.mly"
       (string)
# 48825 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ (_2 : (
# 821 "reason_parser.mly"
       (string)
# 48830 "reason_parser.ml"
  )) _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_signed_constant = 
# 4050 "reason_parser.mly"
                                           ( Const_float("-" ^ _2) )
# 48839 "reason_parser.ml"
     in
    _menhir_goto_signed_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce332 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_single_attr_id), _startpos_x0_) = _menhir_stack in
    let _endpos = _endpos_x0_ in
    let _v : 'tv_attr_id = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 48856 "reason_parser.ml"
      
    in
    
# 4302 "reason_parser.mly"
                           ( _1 )
# 48862 "reason_parser.ml"
     in
    _menhir_goto_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce646 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mty_longident), _startpos_x0_) = _menhir_stack in
    let _v : 'tv_package_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 48878 "reason_parser.ml"
      
    in
    
# 3976 "reason_parser.mly"
                          ( (_1, []) )
# 48884 "reason_parser.ml"
     in
    _menhir_goto_package_type _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_type_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2683 | MenhirState2682 | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState2472 | MenhirState121 | MenhirState158 | MenhirState161 | MenhirState2449 | MenhirState162 | MenhirState1800 | MenhirState1799 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1345 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1289 | MenhirState1284 | MenhirState1277 | MenhirState1269 | MenhirState313 | MenhirState306 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState255 | MenhirState254 | MenhirState170 | MenhirState163 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4845 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce141 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4846)
    | MenhirState193 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4847 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce276 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4848)
    | MenhirState212 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4849 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce276 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4850)
    | MenhirState2698 | MenhirState2669 | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState2465 | MenhirState2463 | MenhirState143 | MenhirState153 | MenhirState2442 | MenhirState2439 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1816 | MenhirState1814 | MenhirState1809 | MenhirState1807 | MenhirState1805 | MenhirState1766 | MenhirState1781 | MenhirState1779 | MenhirState1774 | MenhirState1772 | MenhirState1770 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1361 | MenhirState1359 | MenhirState1328 | MenhirState1330 | MenhirState1268 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState324 | MenhirState272 | MenhirState270 | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4851 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState254
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState254 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce141 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4852)
    | MenhirState2521 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState1643 | MenhirState1642 | MenhirState1593 | MenhirState1592 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1500 | MenhirState1490 | MenhirState329 | MenhirState328 | MenhirState317 | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4853 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce141 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4854)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4855 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState328
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKET ->
              _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETGREATER ->
              _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETLESS ->
              _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESS ->
              _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSDOTDOTGREATER ->
              _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LESSGREATER ->
              _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | QUOTE ->
              _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState328 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETPERCENTPERCENT | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce141 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4856)
    | _ ->
        _menhir_fail ()

and _menhir_goto__non_arrowed_simple_core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2698 | MenhirState2682 | MenhirState2683 | MenhirState2669 | MenhirState2649 | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState121 | MenhirState2472 | MenhirState128 | MenhirState141 | MenhirState143 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState158 | MenhirState161 | MenhirState162 | MenhirState2449 | MenhirState163 | MenhirState170 | MenhirState2439 | MenhirState2442 | MenhirState324 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1814 | MenhirState1816 | MenhirState1805 | MenhirState1807 | MenhirState1809 | MenhirState1799 | MenhirState1800 | MenhirState1766 | MenhirState1779 | MenhirState1781 | MenhirState1770 | MenhirState1772 | MenhirState1774 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1359 | MenhirState1361 | MenhirState1268 | MenhirState1345 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1288 | MenhirState1289 | MenhirState1284 | MenhirState1269 | MenhirState1277 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1068 | MenhirState1071 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState306 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState245 | MenhirState270 | MenhirState272 | MenhirState254 | MenhirState255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4841) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce558 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4842)
    | MenhirState106 | MenhirState2521 | MenhirState2478 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1642 | MenhirState1643 | MenhirState1601 | MenhirState1592 | MenhirState1593 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1490 | MenhirState1500 | MenhirState1501 | MenhirState1498 | MenhirState1089 | MenhirState1092 | MenhirState1087 | MenhirState722 | MenhirState681 | MenhirState687 | MenhirState328 | MenhirState329 | MenhirState244 | MenhirState317 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4843) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce558 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4844)
    | _ ->
        _menhir_fail ()

and _menhir_goto__non_arrowed_non_simple_core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_non_simple_core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2698 | MenhirState2669 | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState143 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState2439 | MenhirState2442 | MenhirState324 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1814 | MenhirState1816 | MenhirState1805 | MenhirState1807 | MenhirState1809 | MenhirState1766 | MenhirState1779 | MenhirState1781 | MenhirState1770 | MenhirState1772 | MenhirState1774 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1359 | MenhirState1361 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1068 | MenhirState1071 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState245 | MenhirState270 | MenhirState272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4837) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce557 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4838)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1089 | MenhirState1092 | MenhirState1087 | MenhirState722 | MenhirState681 | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4839) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_non_simple_core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce557 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4840)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lbl_pattern_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4827 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error521 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4825 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce277 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4826)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error521 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4828)
    | MenhirState525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4831 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4829 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)), _endpos__2_), _, (_3 : 'tv_lbl_pattern_list)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_lbl_pattern_list = 
# 3439 "reason_parser.mly"
      ( let (fields, closed) = _3 in _1 :: fields, closed )
# 49059 "reason_parser.ml"
         in
        _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4830)) : 'freshtv4832)
    | MenhirState615 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4835 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error616 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4833 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lbl_pattern_list) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce277 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4834)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error616 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4836)
    | _ ->
        _menhir_fail ()

and _menhir_reduce482 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_) = _menhir_stack in
    let _v : 'tv_lbl_pattern = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 49095 "reason_parser.ml"
      
    in
    
# 3445 "reason_parser.mly"
      (
        (_1, pat_of_label _1)
      )
# 49103 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce272 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_signed_constant), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3342 "reason_parser.mly"
      ( mkpat(Ppat_constant _1) )
# 49115 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_pattern : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState592 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4723 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState593
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState593 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARRBRACKET | COMMA | SEMI ->
              _menhir_reduce659 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState593)) : 'freshtv4724)
    | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4725 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce160 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4726)
    | MenhirState511 | MenhirState588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4727 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState602
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState602 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BARRBRACKET | COMMA | SEMI ->
              _menhir_reduce658 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState602)) : 'freshtv4728)
    | MenhirState611 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4729 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState612
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState612 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET | SEMI ->
              _menhir_reduce661 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState612)) : 'freshtv4730)
    | MenhirState610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4731 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState613
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState613 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET | SEMI ->
              _menhir_reduce662 _menhir_env (Obj.magic _menhir_stack)
          | COMMA ->
              _menhir_reduce659 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState613)) : 'freshtv4732)
    | MenhirState514 | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4733 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState614
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState614 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | RBRACKET | SEMI ->
              _menhir_reduce660 _menhir_env (Obj.magic _menhir_stack)
          | COMMA ->
              _menhir_reduce658 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState614)) : 'freshtv4734)
    | MenhirState676 | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4737 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState665
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState665 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4735 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState665 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState666) : 'freshtv4736)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce161 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4738)
    | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4739 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce160 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4740)
    | MenhirState495 | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4745 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState678
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState678 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState678 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState681 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState681) : 'freshtv4742)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4743 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState678 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce284 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4744)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce161 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4746)
    | MenhirState531 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4747 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState699
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState699 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COMMA | RBRACE ->
              _menhir_reduce481 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState699)) : 'freshtv4748)
    | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4753 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState719
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run594 _menhir_env (Obj.magic _menhir_stack) MenhirState719 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4749 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState719 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState722 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState722) : 'freshtv4750)
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4751 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState719 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce284 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4752)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce161 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4754)
    | MenhirState738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4757 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState739
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState739 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4755 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState739 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState740 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState740) : 'freshtv4756)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState739)) : 'freshtv4758)
    | MenhirState767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4787 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState768
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState768 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUALGREATER | WHEN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4785 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_pattern_with_bar = 
# 3085 "reason_parser.mly"
                               (
  _2
)
# 49493 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4783) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : 'tv_pattern_with_bar) = _v in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
              match _menhir_s with
              | MenhirState2552 | MenhirState2551 | MenhirState2434 | MenhirState2433 | MenhirState2347 | MenhirState2346 | MenhirState2149 | MenhirState766 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4765 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | EQUALGREATER ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4759 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LET ->
                          _menhir_run1156 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PERCENT ->
                          _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2147 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2147) : 'freshtv4760)
                  | WHEN ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4761 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState770 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState770) : 'freshtv4762)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4763 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4764)) : 'freshtv4766)
              | MenhirState1038 | MenhirState898 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4773 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | EQUALGREATER ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4767 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1036) : 'freshtv4768)
                  | WHEN ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4769 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1032) : 'freshtv4770)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4771 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4772)) : 'freshtv4774)
              | MenhirState2235 | MenhirState476 ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4781 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | EQUALGREATER ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4775 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2233 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2233) : 'freshtv4776)
                  | WHEN ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4777 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | ASSERT ->
                          _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BACKQUOTE ->
                          _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | BANG ->
                          _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | CHAR _v ->
                          _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FALSE ->
                          _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FLOAT _v ->
                          _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FOR ->
                          _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | FUN ->
                          _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | IF ->
                          _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT _v ->
                          _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT32 _v ->
                          _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | INT64 _v ->
                          _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LAZY ->
                          _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACELESS ->
                          _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKET ->
                          _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETBAR ->
                          _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LIDENT _v ->
                          _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUS ->
                          _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | MINUSDOT ->
                          _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NATIVEINT _v ->
                          _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | NEW ->
                          _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUS ->
                          _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PLUSDOT ->
                          _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | PREFIXOP _v ->
                          _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | STRING _v ->
                          _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | SWITCH ->
                          _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRUE ->
                          _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | TRY ->
                          _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | WHILE ->
                          _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2229 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2229) : 'freshtv4778)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : 'freshtv4779 * _menhir_state * 'tv_pattern_with_bar) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4780)) : 'freshtv4782)
              | _ ->
                  _menhir_fail ()) : 'freshtv4784)) : 'freshtv4786)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState768)) : 'freshtv4788)
    | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4791 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState909
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState909 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | IN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState909 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState910 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState910) : 'freshtv4790)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState909)) : 'freshtv4792)
    | MenhirState1634 | MenhirState1183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4795 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1236
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState1236 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4793 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1236 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1237 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1237) : 'freshtv4794)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1236)) : 'freshtv4796)
    | MenhirState1657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4801 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1658
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState1658 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4799 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1658 in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4797 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _startpos = _startpos__1_ in
              let _endpos = _endpos__3_ in
              let _v : 'tv__class_self_pattern = 
# 1910 "reason_parser.mly"
      ( _2 )
# 50213 "reason_parser.ml"
               in
              _menhir_goto__class_self_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4798)) : 'freshtv4800)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1658)) : 'freshtv4802)
    | MenhirState2314 | MenhirState1886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4805 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1921
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState1921 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4803 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1921 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1922) : 'freshtv4804)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1921)) : 'freshtv4806)
    | MenhirState191 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4811 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2421
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState2421 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | WHEN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4807 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2421 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2422) : 'freshtv4808)
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4809 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_payload = 
# 4355 "reason_parser.mly"
                     ( PPat (_2, None) )
# 50412 "reason_parser.ml"
               in
              _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4810)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2421)) : 'freshtv4812)
    | MenhirState2578 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4823 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2579
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BAR ->
              _menhir_run668 _menhir_env (Obj.magic _menhir_stack) MenhirState2579 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4821 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2579 in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4819 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
              let (_ : _menhir_state) = _menhir_s in
              ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1071 "reason_parser.mly"
      (Parsetree.pattern)
# 50441 "reason_parser.ml"
              ) = 
# 1145 "reason_parser.mly"
                ( apply_mapper_chain_to_pattern _1 default_mapper_chain)
# 50445 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4817) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1071 "reason_parser.mly"
      (Parsetree.pattern)
# 50453 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4815) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1071 "reason_parser.mly"
      (Parsetree.pattern)
# 50461 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4813) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1071 "reason_parser.mly"
      (Parsetree.pattern)
# 50469 "reason_parser.ml"
              )) : (
# 1071 "reason_parser.mly"
      (Parsetree.pattern)
# 50473 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv4814)) : 'freshtv4816)) : 'freshtv4818)) : 'freshtv4820)) : 'freshtv4822)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2579)) : 'freshtv4824)
    | _ ->
        _menhir_fail ()

and _menhir_goto_pattern_comma_list : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_pattern_comma_list -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4713 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState590
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState590
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState590
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState590) : 'freshtv4714)
    | MenhirState514 | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4719 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState609
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4717 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState609 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOTDOT ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4715 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
                  let (_menhir_s : _menhir_state) = MenhirState610 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | EXCEPTION ->
                      _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SHARP ->
                      _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState611 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState611) : 'freshtv4716)
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState610) : 'freshtv4718)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState609)) : 'freshtv4720)
    | MenhirState511 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4721 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState705
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              _menhir_run592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState705
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState705
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState705) : 'freshtv4722)
    | _ ->
        _menhir_fail ()

and _menhir_goto_pattern_comma_list_extension : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_pattern_comma_list_extension -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState605
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState605
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState605) : 'freshtv4710)
    | MenhirState514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4711 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState700
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              _menhir_run591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState700
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState700) : 'freshtv4712)
    | _ ->
        _menhir_fail ()

and _menhir_goto__core_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState153 | MenhirState2439 | MenhirState2442 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1359 | MenhirState1361 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1068 | MenhirState1071 | MenhirState1066 | MenhirState666 | MenhirState245 | MenhirState270 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4705) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce555 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4706)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1089 | MenhirState1092 | MenhirState1087 | MenhirState722 | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4707) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce555 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4708)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lbl_pattern : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_pattern -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4703 * _menhir_state * 'tv_lbl_pattern) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_error524 _menhir_env (Obj.magic _menhir_stack)
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | COMMA ->
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv4701 * _menhir_state * 'tv_lbl_pattern) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          ((let _menhir_stack = (_menhir_stack, _endpos) in
          let _menhir_env = _menhir_discard _menhir_env in
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LIDENT _v ->
              _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState525 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState525 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4699 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState525 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | COMMA ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4697) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState526 in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4698)
              | RBRACE ->
                  _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState526
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState526) : 'freshtv4700)
          | RBRACE ->
              _menhir_reduce484 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState525) : 'freshtv4702)
      | RBRACE ->
          _menhir_reduce483 _menhir_env (Obj.magic _menhir_stack)
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_error524 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4704)

and _menhir_goto__simple_pattern_not_ident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_pattern_not_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4693) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern_not_ident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce551 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4694)
    | MenhirState2663 | MenhirState2704 | MenhirState2677 | MenhirState2679 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState476 | MenhirState2238 | MenhirState480 | MenhirState484 | MenhirState2224 | MenhirState486 | MenhirState738 | MenhirState1886 | MenhirState1887 | MenhirState1912 | MenhirState1891 | MenhirState1627 | MenhirState1823 | MenhirState1703 | MenhirState1707 | MenhirState1714 | MenhirState1657 | MenhirState1637 | MenhirState1639 | MenhirState1634 | MenhirState1183 | MenhirState1184 | MenhirState1227 | MenhirState1188 | MenhirState1192 | MenhirState1199 | MenhirState898 | MenhirState1041 | MenhirState902 | MenhirState906 | MenhirState1027 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4695) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_pattern_not_ident) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce551 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4696)
    | _ ->
        _menhir_fail ()

and _menhir_run934 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 50785 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4689 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 50797 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState938 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState938) : 'freshtv4690)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4691 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 50856 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState935 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState935) : 'freshtv4692)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce723 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_simple_non_labeled_expr_list = 
# 2823 "reason_parser.mly"
      ( [_1] )
# 50918 "reason_parser.ml"
     in
    _menhir_goto_simple_non_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce724 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_non_labeled_expr_list), _startpos__1_), _endpos__2_, _, (_2 : 'tv_less_aggressive_simple_expression), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_simple_non_labeled_expr_list = 
# 2825 "reason_parser.mly"
      ( _2 :: _1 )
# 50930 "reason_parser.ml"
     in
    _menhir_goto_simple_non_labeled_expr_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce468 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_labeled_simple_expr = 
# 2844 "reason_parser.mly"
      ( ("", _1) )
# 50941 "reason_parser.ml"
     in
    _menhir_goto_labeled_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce464 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 50948 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 50954 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_less_aggressive_simple_expression), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_label_expr = 
# 2852 "reason_parser.mly"
      ( (_1, _3) )
# 50961 "reason_parser.ml"
     in
    _menhir_goto_label_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce465 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 50968 "reason_parser.ml"
) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 50974 "reason_parser.ml"
    )), _startpos__1_), _endpos__3_, _, (_3 : 'tv_less_aggressive_simple_expression), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_label_expr = 
# 2857 "reason_parser.mly"
      ( ("?" ^ _1, _3) )
# 50981 "reason_parser.ml"
     in
    _menhir_goto_label_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_run821 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 50988 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4685 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 51000 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState825 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState825) : 'freshtv4686)
    | EXPLICITLY_PASSED_OPTIONAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4687 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 51059 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState822 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState822) : 'freshtv4688)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce537 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_non_labeled_expr_list_as_tuple -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__simple_non_labeled_expr_list_as_tuple) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1168 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 51121 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4683) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) = _v in
    ((match _menhir_s with
    | MenhirState840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4679) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) = _v in
        (_menhir_reduce725 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4680)
    | MenhirState953 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4681) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp__simple_non_labeled_expr_list_as_tuple_) = _v in
        (_menhir_reduce725 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4682)
    | _ ->
        _menhir_fail ()) : 'freshtv4684)

and _menhir_reduce212 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 51161 "reason_parser.ml"
      
    in
    
# 2670 "reason_parser.mly"
      ( mkexp(Pexp_field(_1, _3)) )
# 51167 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_expr_optional_constraint : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_expr_optional_constraint -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState755 | MenhirState798 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4637 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1062 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4635 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1063 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1063) : 'freshtv4636)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1062 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4638)
    | MenhirState1063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4639 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce413 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4640)
    | MenhirState1076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4641 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce412 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4642)
    | MenhirState178 | MenhirState415 | MenhirState2182 | MenhirState1081 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4645 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4643 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_expr_optional_constraint), _startpos__1_) = _menhir_stack in
        let _v : 'tv_expr_comma_seq = 
# 3141 "reason_parser.mly"
      ( [_1] )
# 51282 "reason_parser.ml"
         in
        _menhir_goto_expr_comma_seq _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4644)) : 'freshtv4646)
    | MenhirState1084 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4649 * _menhir_state * 'tv_expr_comma_seq) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4647 * _menhir_state * 'tv_expr_comma_seq) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_expr_comma_seq)), _endpos__2_, _), _, (_3 : 'tv_expr_optional_constraint), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_expr_comma_seq = 
# 3139 "reason_parser.mly"
      ( _3::_1 )
# 51295 "reason_parser.ml"
         in
        _menhir_goto_expr_comma_seq _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4648)) : 'freshtv4650)
    | MenhirState1099 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4653 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1100 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4651 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce416 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4652)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1100 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4654)
    | MenhirState418 | MenhirState1103 | MenhirState1098 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4657 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1102
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4655 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1102 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOTDOT ->
                  _menhir_run1099 _menhir_env (Obj.magic _menhir_stack) MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1103 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RBRACKET ->
                  _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1103) : 'freshtv4656)
          | RBRACKET ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1102
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1102)) : 'freshtv4658)
    | MenhirState1115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4661 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1116 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4659 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1117 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1117 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1117) : 'freshtv4660)
          | RBRACE ->
              _menhir_reduce688 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1116 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4662)
    | MenhirState15 | MenhirState184 | MenhirState408 | MenhirState771 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4665 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1141 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4663 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1142 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1142) : 'freshtv4664)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1141 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4666)
    | MenhirState1142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4667 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce413 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4668)
    | MenhirState1147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4669 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce412 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4670)
    | MenhirState2188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4673 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2189 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4671 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce416 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4672)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2189 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4674)
    | MenhirState416 | MenhirState2192 | MenhirState2187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4677 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2191
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4675 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2191 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | DOTDOTDOT ->
                  _menhir_run2188 _menhir_env (Obj.magic _menhir_stack) MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2192 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RBRACKET ->
                  _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2192) : 'freshtv4676)
          | RBRACKET ->
              _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2191
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2191)) : 'freshtv4678)
    | _ ->
        _menhir_fail ()

and _menhir_reduce809 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_type_constraint_right_of_colon = 
# 3245 "reason_parser.mly"
                                          ( (Some _1, None) )
# 51666 "reason_parser.ml"
     in
    _menhir_goto_type_constraint_right_of_colon _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce688 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_optional_constraint), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_record_expr = 
# 3168 "reason_parser.mly"
                                                           ( (Some _2, []))
# 51677 "reason_parser.ml"
     in
    _menhir_goto_record_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_lbl_expr_list : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState1117 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4625 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_lbl_expr_list) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4623 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_lbl_expr_list) : 'tv_lbl_expr_list) = _v in
        ((let (((_menhir_stack, _menhir_s, _startpos__1_), _, (_2 : 'tv_expr_optional_constraint), _startpos__2_), _endpos__3_) = _menhir_stack in
        let _3 = () in
        let _1 = () in
        let _v : 'tv_record_expr = 
# 3169 "reason_parser.mly"
                                                           ( (Some _2, _4) )
# 51699 "reason_parser.ml"
         in
        _menhir_goto_record_expr _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4624)) : 'freshtv4626)
    | MenhirState1120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4629 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_lbl_expr_list) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4627 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_lbl_expr_list) : 'tv_lbl_expr_list) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_lbl_expr_list = 
# 3174 "reason_parser.mly"
                                  ( _1 :: _3 )
# 51716 "reason_parser.ml"
         in
        _menhir_goto_lbl_expr_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4628)) : 'freshtv4630)
    | MenhirState1131 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4633 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_lbl_expr_list) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4631 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_lbl_expr_list) : 'tv_lbl_expr_list) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_lbl_expr_list_that_is_not_a_single_punned_field = 
# 3210 "reason_parser.mly"
      ( _1::_3 )
# 51733 "reason_parser.ml"
         in
        _menhir_goto_lbl_expr_list_that_is_not_a_single_punned_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4632)) : 'freshtv4634)
    | _ ->
        _menhir_fail ()

and _menhir_reduce475 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_) = _menhir_stack in
    let _v : 'tv_lbl_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 51751 "reason_parser.ml"
      
    in
    
# 3181 "reason_parser.mly"
      (
        (_1, exp_of_label _1)
      )
# 51759 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_lbl_expr : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1131 | MenhirState1120 | MenhirState1117 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4617 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1119 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4615 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1120 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1120 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RBRACE ->
                  _menhir_reduce478 _menhir_env (Obj.magic _menhir_stack)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1120) : 'freshtv4616)
          | RBRACE ->
              _menhir_reduce476 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1119 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4618)
    | MenhirState432 | MenhirState475 | MenhirState2202 | MenhirState1113 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4621 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1130 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COMMA ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4619 * _menhir_state * 'tv_lbl_expr) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1131 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1131 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1131) : 'freshtv4620)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1130 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4622)
    | _ ->
        _menhir_fail ()

and _menhir_goto_lbl_expr_list_that_is_not_a_single_punned_field : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_lbl_expr_list_that_is_not_a_single_punned_field -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4613) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_lbl_expr_list_that_is_not_a_single_punned_field) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4611) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_lbl_expr_list_that_is_not_a_single_punned_field) : 'tv_lbl_expr_list_that_is_not_a_single_punned_field) = _v in
    ((let _v : 'tv_record_expr = 
# 3170 "reason_parser.mly"
                                                           ( (None, _1) )
# 51843 "reason_parser.ml"
     in
    _menhir_goto_record_expr _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4612)) : 'freshtv4614)

and _menhir_goto_ident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState92 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4507 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce154 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4508)
    | MenhirState98 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4509 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce156 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4510)
    | MenhirState102 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4511 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce158 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4512)
    | MenhirState115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4513 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce138 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4514)
    | MenhirState124 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4515 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce138 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4516)
    | MenhirState154 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4519 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ('freshtv4517 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
          ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv4518)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT | QUOTE ->
              _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce138 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4520)
    | MenhirState164 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4521 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce591 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4522)
    | MenhirState250 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4523 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce588 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4524)
    | MenhirState2631 | MenhirState129 | MenhirState2391 | MenhirState2389 | MenhirState2368 | MenhirState2362 | MenhirState2360 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2173 | MenhirState2158 | MenhirState2161 | MenhirState2155 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1955 | MenhirState1958 | MenhirState1952 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 | MenhirState801 | MenhirState775 | MenhirState246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4525 * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4526)
    | MenhirState277 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4527 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce56 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4528)
    | MenhirState399 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4531 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState400
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4529 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState400 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState401 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState401 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run234 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState401 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState401) : 'freshtv4530)
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState400
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState400) : 'freshtv4532)
    | MenhirState465 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4533 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce591 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4534)
    | MenhirState504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4535 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce588 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4536)
    | MenhirState233 | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState1965 | MenhirState2005 | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4537 * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce587 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4538)
    | MenhirState692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4539 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce56 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4540)
    | MenhirState1217 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4543 * _menhir_state * 'tv_typevar_list * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4541 * _menhir_state * 'tv_typevar_list * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, (_1 : 'tv_typevar_list), _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_ident), _startpos__3_) = _menhir_stack in
        let _2 = () in
        let _startpos = _startpos__1_ in
        let _v : 'tv_typevar_list = 
# 3738 "reason_parser.mly"
                                                ( _3 :: _1 )
# 51967 "reason_parser.ml"
         in
        _menhir_goto_typevar_list _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv4542)) : 'freshtv4544)
    | MenhirState1256 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4545 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce154 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4546)
    | MenhirState1260 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4547 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce156 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4548)
    | MenhirState1264 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4549 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce158 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4550)
    | MenhirState1405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4553 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1406
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4551 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1406 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1407 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1407 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1407) : 'freshtv4552)
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1406
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1406) : 'freshtv4554)
    | MenhirState1576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4557 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : ('freshtv4555 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
          (_menhir_reduce138 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4556)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT | QUOTE ->
              _menhir_reduce831 _menhir_env (Obj.magic _menhir_stack)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce138 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4558)
    | MenhirState1617 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4585 * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4583 * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__type_variable = 
# 3552 "reason_parser.mly"
                                                ( mktyp (Ptyp_var _2) )
# 52046 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4581) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv__type_variable) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4579) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv__type_variable) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4577) = Obj.magic _menhir_stack in
        let (_endpos_x_ : Lexing.position) = _endpos in
        let ((x : 'tv__type_variable) : 'tv__type_variable) = _v in
        let (_startpos_x_ : Lexing.position) = _startpos in
        ((let _v : 'tv_mark_position_typ__type_variable_ = let _endpos = _endpos_x_ in
        let _symbolstartpos = _startpos_x_ in
        
# 1173 "reason_parser.mly"
 ( {x with ptyp_loc = {x.ptyp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 52068 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4575) = _menhir_stack in
        let (_v : 'tv_mark_position_typ__type_variable_) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4573) = Obj.magic _menhir_stack in
        let (_v : 'tv_mark_position_typ__type_variable_) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4571) = Obj.magic _menhir_stack in
        let ((_1 : 'tv_mark_position_typ__type_variable_) : 'tv_mark_position_typ__type_variable_) = _v in
        ((let _v : 'tv_type_variable = 
# 3550 "reason_parser.mly"
                                                 (_1)
# 52082 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4569) = _menhir_stack in
        let (_v : 'tv_type_variable) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4567 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
        let (_v : 'tv_type_variable) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4565 * _menhir_state * 'tv_type_variance) = Obj.magic _menhir_stack in
        let ((_2 : 'tv_type_variable) : 'tv_type_variable) = _v in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_type_variance)) = _menhir_stack in
        let _v : 'tv_type_parameter = 
# 3542 "reason_parser.mly"
                                                  ( _2, _1 )
# 52097 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4563) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_parameter) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4561 * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_type_parameter) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4559 * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_type_parameter) : 'tv_type_parameter) = _v in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_class_type_parameters)) = _menhir_stack in
        let _v : 'tv_class_type_parameters = 
# 1878 "reason_parser.mly"
                                         (
      _2::_1
    )
# 52117 "reason_parser.ml"
         in
        _menhir_goto_class_type_parameters _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4560)) : 'freshtv4562)) : 'freshtv4564)) : 'freshtv4566)) : 'freshtv4568)) : 'freshtv4570)) : 'freshtv4572)) : 'freshtv4574)) : 'freshtv4576)) : 'freshtv4578)) : 'freshtv4580)) : 'freshtv4582)) : 'freshtv4584)) : 'freshtv4586)
    | MenhirState2619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4605 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2620
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | FALSE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4589 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2620 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4587 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4180 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_bool false) )
# 52146 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4588)) : 'freshtv4590)
          | INT _v ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4593 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2620 in
              let (_v : (
# 839 "reason_parser.mly"
       (int)
# 52157 "reason_parser.ml"
              )) = _v in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4591 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let ((_3 : (
# 839 "reason_parser.mly"
       (int)
# 52168 "reason_parser.ml"
              )) : (
# 839 "reason_parser.mly"
       (int)
# 52172 "reason_parser.ml"
              )) = _v in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4177 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_int _3) )
# 52180 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4592)) : 'freshtv4594)
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2620 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2620 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4597 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2620 in
              let (_v : (
# 898 "reason_parser.mly"
       (string * string option)
# 52195 "reason_parser.ml"
              )) = _v in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4595 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let ((_3 : (
# 898 "reason_parser.mly"
       (string * string option)
# 52206 "reason_parser.ml"
              )) : (
# 898 "reason_parser.mly"
       (string * string option)
# 52210 "reason_parser.ml"
              )) = _v in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4176 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_string (fst _3)) )
# 52218 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4596)) : 'freshtv4598)
          | TRUE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4601 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2620 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4599 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos__3_ : Lexing.position) = _endpos in
              let (_ : _menhir_state) = _menhir_s in
              let (_startpos__3_ : Lexing.position) = _startpos in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _3 = () in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4181 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_bool true) )
# 52239 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4600)) : 'freshtv4602)
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2620 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | EOF | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4603 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              ((let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4175 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_none) )
# 52252 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4604)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2620)) : 'freshtv4606)
    | MenhirState2629 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4609 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2630
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4607 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2630 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2631 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2631) : 'freshtv4608)
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2630
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2630) : 'freshtv4610)
    | _ ->
        _menhir_fail ()

and _menhir_goto_primitive_declaration : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_primitive_declaration -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1483 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4497 * Lexing.position * _menhir_state * (
# 898 "reason_parser.mly"
       (string * string option)
# 52309 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce678 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4498)
    | MenhirState1482 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4499 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1485
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1485
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1485) : 'freshtv4500)
    | MenhirState2044 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4501 * Lexing.position * _menhir_state * (
# 898 "reason_parser.mly"
       (string * string option)
# 52329 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce678 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4502)
    | MenhirState2043 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4503 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2046
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2046
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2046) : 'freshtv4504)
    | MenhirState2651 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4505 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2652
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2652
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2652) : 'freshtv4506)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_sig_fields : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_sig_fields -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4495 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1609
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | SEMI ->
          let (_menhir_env : _menhir_env) = _menhir_env in
          let (_menhir_stack : 'freshtv4493 * Lexing.position * _menhir_state * 'tv_class_sig_fields) = Obj.magic _menhir_stack in
          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
          let (_menhir_s : _menhir_state) = MenhirState1609 in
          ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
          let _menhir_env = _menhir_discard _menhir_env in
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1599 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INHERIT ->
              _menhir_run1586 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETATATAT ->
              _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENTPERCENT ->
              _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | METHOD ->
              _menhir_run1568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | VAL ->
              _menhir_run1552 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1610 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce637 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv4494)
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState1609) : 'freshtv4496)

and _menhir_goto__class_instance_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_instance_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1767 | MenhirState1542 | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4489) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_instance_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce530 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4490)
    | MenhirState2707 | MenhirState2056 | MenhirState1826 | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4491) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_instance_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce530 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4492)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_sig_field : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_sig_field -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4487) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_sig_field) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4485) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_sig_field) : 'tv__class_sig_field) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_ctf__class_sig_field_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1163 "reason_parser.mly"
 ( {x with pctf_loc = {x.pctf_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 52437 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4483) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_ctf__class_sig_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4481) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_ctf__class_sig_field_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4479) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_ctf__class_sig_field_) : 'tv_mark_position_ctf__class_sig_field_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_class_sig_field = 
# 2251 "reason_parser.mly"
                                                      (_1)
# 52458 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4477) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_sig_field) = _v in
    ((match _menhir_s with
    | MenhirState1610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4471 * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_sig_field) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4469 * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos__3_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_class_sig_field) : 'tv_class_sig_field) = _v in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_sig_fields)), _endpos__2_, _) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_class_sig_fields = 
# 2248 "reason_parser.mly"
                                        ( _3 :: _1 )
# 52483 "reason_parser.ml"
         in
        _menhir_goto_class_sig_fields _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4470)) : 'freshtv4472)
    | MenhirState1551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4475) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_sig_field) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4473) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_class_sig_field) : 'tv_class_sig_field) = _v in
        ((let _endpos = _endpos__1_ in
        let _v : 'tv_class_sig_fields = 
# 2247 "reason_parser.mly"
                  ( [_1] )
# 52501 "reason_parser.ml"
         in
        _menhir_goto_class_sig_fields _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4474)) : 'freshtv4476)
    | _ ->
        _menhir_fail ()) : 'freshtv4478)) : 'freshtv4480)) : 'freshtv4482)) : 'freshtv4484)) : 'freshtv4486)) : 'freshtv4488)

and _menhir_goto__class_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1632 | MenhirState1663 | MenhirState1738 | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4465) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce522 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4466)
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4467) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce522 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4468)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_fun_def : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_fun_def -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1637 | MenhirState1639 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4461) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_def) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce525 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4462)
    | MenhirState2677 | MenhirState2679 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4463) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_fun_def) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce525 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4464)
    | _ ->
        _menhir_fail ()

and _menhir_goto_value : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_value -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4459 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1674
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1674
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1674) : 'freshtv4460)

and _menhir_goto_curried_binding : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_curried_binding -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1227 | MenhirState1188 | MenhirState1192 | MenhirState1199 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4451) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding) = _v in
        (_menhir_reduce407 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4452)
    | MenhirState1912 | MenhirState1891 | MenhirState1707 | MenhirState1714 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4453) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding) = _v in
        (_menhir_reduce407 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4454)
    | MenhirState1703 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4457 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4455 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__4_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_curried_binding) : 'tv_curried_binding) = _v in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_) = _menhir_stack in
        let _endpos = _endpos__4_ in
        let _v : 'tv_method_ = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 52605 "reason_parser.ml"
          
        in
        let _startpos__3_ = _startpos_x0_ in
        let _endpos = _endpos__4_ in
        let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
          _startpos__1_
        else
          if Pervasives.(!=) _startpos__2_ _endpos__2_ then
            _startpos__2_
          else
            _startpos__3_ in
        
# 2041 "reason_parser.mly"
      ( _3, _2,
        Cfk_concrete (_1, let loc = mklocation _symbolstartpos _endpos in mkexp ~ghost:true ~loc (Pexp_poly (_4, None))) )
# 52621 "reason_parser.ml"
         in
        _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4456)) : 'freshtv4458)
    | _ ->
        _menhir_fail ()

and _menhir_goto_curried_binding_return_typed_ : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_curried_binding_return_typed_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1199 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4435 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce404 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4436)
    | MenhirState1192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4437 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 52642 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce405 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4438)
    | MenhirState1188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 52653 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce63 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4440)
    | MenhirState1227 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4441 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce62 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4442)
    | MenhirState1714 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4443 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce404 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4444)
    | MenhirState1707 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4445 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 52678 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce405 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4446)
    | MenhirState1891 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4447 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 52689 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce63 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4448)
    | MenhirState1912 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4449 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_curried_binding_return_typed_) = _v in
        (_menhir_reduce62 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4450)
    | _ ->
        _menhir_fail ()

and _menhir_goto_method_ : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_method_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ('freshtv4433 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1731
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1731
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1731) : 'freshtv4434)

and _menhir_goto_parent_binder : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_parent_binder -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : ((('freshtv4431 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1742
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | LBRACKETATAT ->
          _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1742
      | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
          _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1742) : 'freshtv4432)

and _menhir_goto_semi_terminated_class_fields : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_semi_terminated_class_fields -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4425 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_class_fields) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4423 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_semi_terminated_class_fields) : 'tv_semi_terminated_class_fields) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_self_pattern), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv_class_self_pattern_and_structure = 
# 1915 "reason_parser.mly"
    ( Cstr.mk _1 _2 )
# 52755 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4421) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_self_pattern_and_structure) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        match _menhir_s with
        | MenhirState2673 | MenhirState1632 | MenhirState1663 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv4411 * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv4409 * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_self_pattern_and_structure), _startpos__1_) = _menhir_stack in
            let _startpos = _startpos__1_ in
            let _endpos = _endpos__1_ in
            let _v : 'tv__class_expr_lets_and_rest = 
# 1892 "reason_parser.mly"
                                     ( mkclass(Pcl_structure _1))
# 52776 "reason_parser.ml"
             in
            _menhir_goto__class_expr_lets_and_rest _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4410)) : 'freshtv4412)
        | MenhirState475 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4415 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error2265 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce224 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4414)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error2265 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4416)
        | MenhirState432 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error2334 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4417 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_self_pattern_and_structure * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce224 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4418)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error2334 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4420)
        | _ ->
            _menhir_fail ()) : 'freshtv4422)) : 'freshtv4424)) : 'freshtv4426)
    | MenhirState1753 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4429 * Lexing.position * _menhir_state * 'tv_class_field) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_semi_terminated_class_fields) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4427 * Lexing.position * _menhir_state * 'tv_class_field) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__3_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_semi_terminated_class_fields) : 'tv_semi_terminated_class_fields) = _v in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_field)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_semi_terminated_class_fields = 
# 2000 "reason_parser.mly"
                                                    ( _1::_3 )
# 52834 "reason_parser.ml"
         in
        _menhir_goto_semi_terminated_class_fields _menhir_env _menhir_stack _endpos _menhir_s _v) : 'freshtv4428)) : 'freshtv4430)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_expr_lets_and_rest : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_expr_lets_and_rest -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4407) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_expr_lets_and_rest) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4405) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_expr_lets_and_rest) : 'tv__class_expr_lets_and_rest) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__class_expr_lets_and_rest_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos_x_ _endpos_x_ then
      _startpos_x_
    else
      _endpos in
    
# 1193 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 52863 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4403) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_expr_lets_and_rest_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4401) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__class_expr_lets_and_rest_) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4399) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_cl__class_expr_lets_and_rest_) : 'tv_mark_position_cl__class_expr_lets_and_rest_) = _v in
    ((let _endpos = _endpos__1_ in
    let _v : 'tv_class_expr_lets_and_rest = 
# 1890 "reason_parser.mly"
                                                                      (_1)
# 52884 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4397) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_expr_lets_and_rest) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4387 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4385 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_class_expr_lets_and_rest)) = _menhir_stack in
        let _2 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_expr_lets_and_rest = 
# 1894 "reason_parser.mly"
                                               (
    class_of_let_bindings _1 _3
  )
# 52907 "reason_parser.ml"
         in
        _menhir_goto__class_expr_lets_and_rest _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv4386)) : 'freshtv4388)
    | MenhirState1632 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4391 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1760 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4389 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce48 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4390)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1760 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4392)
    | MenhirState2673 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4395 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2674 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4393 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr_lets_and_rest) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce48 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4394)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2674 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4396)
    | _ ->
        _menhir_fail ()) : 'freshtv4398)) : 'freshtv4400)) : 'freshtv4402)) : 'freshtv4404)) : 'freshtv4406)) : 'freshtv4408)

and _menhir_goto__class_constructor_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_constructor_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2669 | MenhirState1766 | MenhirState1781 | MenhirState1772 | MenhirState1774 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4381) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_constructor_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce529 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4382)
    | MenhirState2698 | MenhirState2063 | MenhirState1827 | MenhirState1787 | MenhirState1816 | MenhirState1807 | MenhirState1809 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4383) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_constructor_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce529 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4384)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_declaration_details : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_declaration_details -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1534 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4373 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1842
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1842
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1842) : 'freshtv4374)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4375 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1882
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1882
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1882) : 'freshtv4376)
    | MenhirState2660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4377 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2717
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2717
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2717) : 'freshtv4378)
    | MenhirState2738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4379 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2739
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2739
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2739) : 'freshtv4380)
    | _ ->
        _menhir_fail ()

and _menhir_goto_class_self_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_self_type -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4371 * Lexing.position * _menhir_state * 'tv_class_self_type) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1551
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | CONSTRAINT ->
          _menhir_run1599 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | INHERIT ->
          _menhir_run1586 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETATATAT ->
          _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETPERCENTPERCENT ->
          _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | METHOD ->
          _menhir_run1568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | VAL ->
          _menhir_run1552 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1551 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | RBRACE | SEMI ->
          _menhir_reduce358 _menhir_env (Obj.magic _menhir_stack) MenhirState1551
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1551)) : 'freshtv4372)

and _menhir_reduce527 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__constrained_class_declaration -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x_ _menhir_s (x : 'tv__constrained_class_declaration) _startpos_x_ ->
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_cl__constrained_class_declaration_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1193 "reason_parser.mly"
 ( {x with pcl_loc = {x.pcl_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 53063 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv4369) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_cl__constrained_class_declaration_) = _v in
    ((match _menhir_s with
    | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4365) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__constrained_class_declaration_) = _v in
        (_menhir_reduce381 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4366)
    | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4367) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_cl__constrained_class_declaration_) = _v in
        (_menhir_reduce381 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4368)
    | _ ->
        _menhir_fail ()) : 'freshtv4370)

and _menhir_goto_class_fun_return : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_class_fun_return -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1823 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4361 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_return) = _v in
        (_menhir_reduce29 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4362)
    | MenhirState2704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4363 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_fun_return) = _v in
        (_menhir_reduce29 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4364)
    | _ ->
        _menhir_fail ()

and _menhir_goto_type_kind : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_type_kind -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2077 | MenhirState1267 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4357 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53117 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) MenhirState1357 : 'freshtv4358)
    | MenhirState105 | MenhirState1856 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4359 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53125 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) MenhirState1857 : 'freshtv4360)
    | _ ->
        _menhir_fail ()

and _menhir_run396 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState396 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState396

and _menhir_run439 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState439 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState439

and _menhir_goto_optional_type_parameters : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_optional_type_parameters -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState90 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4337 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53362 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState105
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run106 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState105 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) MenhirState105
          | PLUSEQ ->
              _menhir_reduce675 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState105)) : 'freshtv4338)
    | MenhirState1254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4339 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53384 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1267
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run1268 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1267 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) MenhirState1267
          | PLUSEQ ->
              _menhir_reduce675 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1267)) : 'freshtv4340)
    | MenhirState1367 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4343 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4341 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_type_strictly_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_optional_type_parameters)) = _menhir_stack in
        let _v : 'tv_potentially_long_ident_and_optional_type_parameters = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 53416 "reason_parser.ml"
          
        in
        
# 3637 "reason_parser.mly"
                                                             ((_1, _2))
# 53422 "reason_parser.ml"
         in
        _menhir_goto_potentially_long_ident_and_optional_type_parameters _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4342)) : 'freshtv4344)
    | MenhirState1421 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4347 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1422
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONEQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4345 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1422 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1425 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1425) : 'freshtv4346)
          | EQUAL ->
              _menhir_run1423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1422)) : 'freshtv4348)
    | MenhirState1855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4349 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53484 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1856
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run106 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1856 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) MenhirState1856
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1856)) : 'freshtv4350)
    | MenhirState1973 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4353 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1974
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLONEQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4351 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1974 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1975 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1975) : 'freshtv4352)
          | EQUAL ->
              _menhir_run1423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1974 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1974)) : 'freshtv4354)
    | MenhirState2076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4355 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53558 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2077
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              _menhir_run1268 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2077 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | CONSTRAINT | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) MenhirState2077
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2077)) : 'freshtv4356)
    | _ ->
        _menhir_fail ()

and _menhir_goto__structure_item_without_item_extension_sugar : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item_without_item_extension_sugar -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4333) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce554 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4334)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4335) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item_without_item_extension_sugar) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce554 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4336)
    | _ ->
        _menhir_fail ()

and _menhir_goto__structure_item : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__structure_item -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4329) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce553 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4330)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4331) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__structure_item) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce553 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4332)
    | _ ->
        _menhir_fail ()

and _menhir_goto__curried_binding_return_typed : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__curried_binding_return_typed -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState1184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4325) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__curried_binding_return_typed) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce532 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4326)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4327) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__curried_binding_return_typed) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce532 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4328)
    | _ ->
        _menhir_fail ()

and _menhir_reduce500 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp_expr_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__7_ _ (_7 : 'tv_mark_position_exp_expr_) ->
    let (((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__2_, _, _startpos__2_), _, (_3 : 'tv_typevar_list), _startpos__3_), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_), _endpos__6_, _startpos__6_) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _2 = () in
    let _endpos = _endpos__7_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1213 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 53659 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2918 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        (mkpat ~ghost:true ~loc (Ppat_constraint(_1, mktyp ~ghost:true (Ptyp_poly(List.rev _3,_5)))), _7)
      )
# 53671 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce501 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position -> Lexing.position -> _menhir_state -> 'tv_mark_position_exp_expr_ -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__8_ _ (_8 : 'tv_mark_position_exp_expr_) ->
    let ((((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_lident_list)), _endpos__6_, _, (_6 : 'tv_core_type), _startpos__6_), _endpos__7_, _startpos__7_) = _menhir_stack in
    let _7 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _endpos = _endpos__8_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1213 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 53693 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__8_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2970 "reason_parser.mly"
     (
       let exp, poly = wrap_type_annotation _4 _6 _8 in
       let loc = mklocation _symbolstartpos _endpos in
       (mkpat ~ghost:true ~loc (Ppat_constraint(_1, poly)), exp)
     )
# 53706 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_let_binding_body : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_let_binding_body -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4317 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1239
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1239
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1239) : 'freshtv4318)
    | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4319 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1635
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1635
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1635) : 'freshtv4320)
    | MenhirState1886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4321 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1924
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1924
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1924) : 'freshtv4322)
    | MenhirState2314 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4323 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2315
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2315
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2315) : 'freshtv4324)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constraints : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1357 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4309 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53774 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1358
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1359 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1358 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | SEMI ->
              _menhir_reduce811 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1358)) : 'freshtv4310)
    | MenhirState1428 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4311 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1429
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1359 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1429 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce849 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1429)) : 'freshtv4312)
    | MenhirState1857 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4313 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 53810 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1858
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1858 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
              _menhir_reduce811 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1858)) : 'freshtv4314)
    | MenhirState1978 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv4315 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1979
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | CONSTRAINT ->
              _menhir_run1859 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1979 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AND | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce849 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1979)) : 'freshtv4316)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mod_ext_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mod_ext_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState78 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4233 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 53853 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState79
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState79
          | RPAREN ->
              _menhir_run80 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState79 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState79)) : 'freshtv4234)
    | MenhirState83 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4237 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 53873 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState84
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState84
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4235 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 53887 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState84 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce570 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4236)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState84)) : 'freshtv4238)
    | MenhirState87 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4241 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState88
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState88
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4239 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState88 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4240)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState88)) : 'freshtv4242)
    | MenhirState2698 | MenhirState2682 | MenhirState2683 | MenhirState2669 | MenhirState2649 | MenhirState2608 | MenhirState2604 | MenhirState2600 | MenhirState2595 | MenhirState2589 | MenhirState2570 | MenhirState2479 | MenhirState121 | MenhirState2472 | MenhirState128 | MenhirState141 | MenhirState143 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState158 | MenhirState161 | MenhirState162 | MenhirState2449 | MenhirState163 | MenhirState170 | MenhirState2442 | MenhirState2439 | MenhirState324 | MenhirState2220 | MenhirState2063 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1827 | MenhirState1787 | MenhirState1814 | MenhirState1816 | MenhirState1805 | MenhirState1807 | MenhirState1809 | MenhirState1799 | MenhirState1800 | MenhirState1766 | MenhirState1779 | MenhirState1781 | MenhirState1770 | MenhirState1772 | MenhirState1774 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1710 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1389 | MenhirState1386 | MenhirState1383 | MenhirState1379 | MenhirState1376 | MenhirState1361 | MenhirState1359 | MenhirState1345 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1312 | MenhirState1309 | MenhirState1306 | MenhirState1302 | MenhirState1299 | MenhirState1295 | MenhirState1292 | MenhirState1288 | MenhirState1289 | MenhirState1284 | MenhirState1269 | MenhirState1277 | MenhirState1278 | MenhirState1275 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1195 | MenhirState1071 | MenhirState1068 | MenhirState1066 | MenhirState1022 | MenhirState666 | MenhirState306 | MenhirState313 | MenhirState307 | MenhirState303 | MenhirState292 | MenhirState272 | MenhirState270 | MenhirState245 | MenhirState254 | MenhirState255 | MenhirState193 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4243 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState195
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run196 _menhir_env (Obj.magic _menhir_stack) MenhirState195
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState195)) : 'freshtv4244)
    | MenhirState2521 | MenhirState106 | MenhirState2478 | MenhirState2503 | MenhirState2500 | MenhirState2497 | MenhirState2493 | MenhirState2490 | MenhirState2486 | MenhirState2483 | MenhirState2480 | MenhirState2475 | MenhirState108 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1642 | MenhirState1643 | MenhirState1601 | MenhirState1592 | MenhirState1593 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1524 | MenhirState1518 | MenhirState1512 | MenhirState1505 | MenhirState1490 | MenhirState1500 | MenhirState1501 | MenhirState1498 | MenhirState1092 | MenhirState1089 | MenhirState1087 | MenhirState722 | MenhirState687 | MenhirState681 | MenhirState328 | MenhirState329 | MenhirState244 | MenhirState317 | MenhirState212 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4249 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState215
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4247 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState215 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4245 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState216 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 53958 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce824 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv4246)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState216 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState216) : 'freshtv4248)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState215)) : 'freshtv4250)
    | MenhirState2631 | MenhirState129 | MenhirState2391 | MenhirState2389 | MenhirState2368 | MenhirState2362 | MenhirState2360 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2173 | MenhirState2158 | MenhirState2161 | MenhirState2155 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1955 | MenhirState1958 | MenhirState1952 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 | MenhirState801 | MenhirState775 | MenhirState246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4255 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState249
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4253 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState249 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState250 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4251 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState250 in
                  let (_v : (
# 906 "reason_parser.mly"
       (string)
# 53999 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LPAREN ->
                      _menhir_run83 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState251 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | DOT ->
                      _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)
                  | AND | COLONGREATER | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
                      _menhir_reduce453 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState251) : 'freshtv4252)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState250) : 'freshtv4254)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState249)) : 'freshtv4256)
    | MenhirState233 | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState2116 | MenhirState1250 | MenhirState1965 | MenhirState2005 | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4261 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState503
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4259 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState503 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState504 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4257 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState504 in
                  let (_v : (
# 906 "reason_parser.mly"
       (string)
# 54050 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LPAREN ->
                      _menhir_run83 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | DOT ->
                      _menhir_reduce574 _menhir_env (Obj.magic _menhir_stack)
                  | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
                      _menhir_reduce453 _menhir_env (Obj.magic _menhir_stack)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState505) : 'freshtv4258)
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState504) : 'freshtv4260)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState503)) : 'freshtv4262)
    | MenhirState118 | MenhirState1271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4263 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 54080 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1272
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run196 _menhir_env (Obj.magic _menhir_stack) MenhirState1272
          | RPAREN ->
              _menhir_run80 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1272)) : 'freshtv4264)
    | MenhirState2584 | MenhirState76 | MenhirState1253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4275 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1397
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4273 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1397 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4271 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1398 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 54119 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4269 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos__3_ : Lexing.position) = _endpos in
                  let (_ : _menhir_state) = _menhir_s in
                  let ((_3 : (
# 857 "reason_parser.mly"
       (string)
# 54130 "reason_parser.ml"
                  )) : (
# 857 "reason_parser.mly"
       (string)
# 54134 "reason_parser.ml"
                  )) = _v in
                  let (_startpos__3_ : Lexing.position) = _startpos in
                  ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _) = _menhir_stack in
                  let _2 = () in
                  let _startpos = _startpos__1_ in
                  let _endpos = _endpos__3_ in
                  let _v : 'tv_type_strictly_longident = 
# 4139 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 54144 "reason_parser.ml"
                   in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4267) = _menhir_stack in
                  let (_endpos : Lexing.position) = _endpos in
                  let (_menhir_s : _menhir_state) = _menhir_s in
                  let (_v : 'tv_type_strictly_longident) = _v in
                  let (_startpos : Lexing.position) = _startpos in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv4265 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) = Obj.magic _menhir_stack in
                  ((assert (not _menhir_env._menhir_error);
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | MINUS ->
                      _menhir_run1262 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1367 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run1258 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1367 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | QUOTE ->
                      _menhir_run1256 _menhir_env (Obj.magic _menhir_stack) MenhirState1367 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UNDERSCORE ->
                      _menhir_run1255 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1367 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState1367) : 'freshtv4266)) : 'freshtv4268)) : 'freshtv4270)) : 'freshtv4272)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1398 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1398) : 'freshtv4274)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1397)) : 'freshtv4276)
    | MenhirState1432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4277 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 54183 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1433
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1433
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce852 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1433)) : 'freshtv4278)
    | MenhirState1435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4279 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1436
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1436
          | AND | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
              _menhir_reduce851 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1436)) : 'freshtv4280)
    | MenhirState1767 | MenhirState1542 | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4285 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1587
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4283 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1587 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4281 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1588 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 54238 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce366 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv4282)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1588) : 'freshtv4284)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1587)) : 'freshtv4286)
    | MenhirState2707 | MenhirState2056 | MenhirState1826 | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4291 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1794
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv4289 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1794 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4287 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1795 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 54277 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce366 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv4288)
              | UIDENT _v ->
                  _menhir_run82 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1795 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1795) : 'freshtv4290)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1794)) : 'freshtv4292)
    | MenhirState1984 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4295 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 54297 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1985
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1985
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4293 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 54311 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1985 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce572 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4294)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1985)) : 'freshtv4296)
    | MenhirState1982 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4297 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 54327 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1987
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run1988 _menhir_env (Obj.magic _menhir_stack) MenhirState1987
          | AND | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce852 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1987)) : 'freshtv4298)
    | MenhirState1990 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4301 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 54347 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1991
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1991
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4299 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 54361 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1991 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce570 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4300)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1991)) : 'freshtv4302)
    | MenhirState1994 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4305 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1995
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run81 _menhir_env (Obj.magic _menhir_stack) MenhirState1995
          | RPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4303 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1995 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce571 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv4304)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1995)) : 'freshtv4306)
    | MenhirState1998 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4307 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1999
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run1988 _menhir_env (Obj.magic _menhir_stack) MenhirState1999
          | AND | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
              _menhir_reduce851 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1999)) : 'freshtv4308)
    | _ ->
        _menhir_fail ()

and _menhir_reduce575 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext2), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mod_ext_longident = 
# 4148 "reason_parser.mly"
             ( _1 )
# 54421 "reason_parser.ml"
     in
    _menhir_goto_mod_ext_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_with_constraint : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_with_constraint -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4225 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce854 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4226)
    | MenhirState1419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4227) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce853 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4228)
    | MenhirState2001 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4229 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce854 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4230)
    | MenhirState1971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4231) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_with_constraint) = _v in
        (_menhir_reduce853 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4232)
    | _ ->
        _menhir_fail ()

and _menhir_reduce125 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_type), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_with_constraints)) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__module_type = 
# 1570 "reason_parser.mly"
                                      (
    (* See note above about why WITH constraints aren't considered
     * non-arrowed.
     * We might just consider unifying the syntax for record extension with
     * module extension/WITH constraints.
     *
     *    mod MyModule = {
     *       ModuleToInclude...
     *    };
     *
     *    let module CreateFactory
     *               (Spec: ContainerSpec)
     *               :{DescriptorFactoryIntf.S with
     *                  type props = Spec.props and type dependencies = Spec.props} =>
     *
     *)
    mkmty(Pmty_with(_1, List.rev _3))
  )
# 54485 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto__module_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4221) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce543 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4222)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState1250 | MenhirState2116 | MenhirState1965 | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4223) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce543 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4224)
    | _ ->
        _menhir_fail ()

and _menhir_goto__simple_module_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4217) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce545 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4218)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState1250 | MenhirState2116 | MenhirState1965 | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4219) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce545 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4220)
    | _ ->
        _menhir_fail ()

and _menhir_goto__simple_module_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2644 | MenhirState2645 | MenhirState411 | MenhirState2366 | MenhirState451 | MenhirState2288 | MenhirState756 | MenhirState2172 | MenhirState759 | MenhirState761 | MenhirState2142 | MenhirState2143 | MenhirState2138 | MenhirState2139 | MenhirState2128 | MenhirState2129 | MenhirState1245 | MenhirState1246 | MenhirState799 | MenhirState800 | MenhirState772 | MenhirState773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4213) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce541 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4214)
    | MenhirState403 | MenhirState2395 | MenhirState404 | MenhirState2393 | MenhirState412 | MenhirState2364 | MenhirState2306 | MenhirState2307 | MenhirState2302 | MenhirState2303 | MenhirState2295 | MenhirState2296 | MenhirState2291 | MenhirState2292 | MenhirState762 | MenhirState2165 | MenhirState1476 | MenhirState2025 | MenhirState1945 | MenhirState1947 | MenhirState1948 | MenhirState1962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4215) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce541 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4216)
    | _ ->
        _menhir_fail ()

and _menhir_goto_fun_def : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_fun_def -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4197 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce440 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4198)
    | MenhirState906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4199 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 54570 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce441 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4200)
    | MenhirState902 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4201 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 54581 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce67 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4202)
    | MenhirState1041 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4203 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce66 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4204)
    | MenhirState2224 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4205 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce440 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4206)
    | MenhirState484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4207 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 54606 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce441 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4208)
    | MenhirState480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4209 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 54617 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce67 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4210)
    | MenhirState2238 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4211 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_fun_def) = _v in
        (_menhir_reduce66 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4212)
    | _ ->
        _menhir_fail ()

and _menhir_goto_leading_bar_match_case : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_leading_bar_match_case -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1038 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4189 * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce492 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4190)
    | MenhirState898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4191) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce491 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4192)
    | MenhirState2235 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4193 * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce492 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4194)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4195) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_leading_bar_match_case) = _v in
        (_menhir_reduce491 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4196)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_semi : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_opt_semi -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4153 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error599 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4151 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce281 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4152)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error599 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4154)
    | MenhirState605 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4157 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error606 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4155 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce279 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4156)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error606 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4158)
    | MenhirState700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4161 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error701 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4159 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce279 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4160)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error701 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4162)
    | MenhirState705 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4165 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error706 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4163 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce281 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4164)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error706 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4166)
    | MenhirState1175 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4167 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce181 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4168)
    | MenhirState1166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4169 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce180 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4170)
    | MenhirState1609 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4183 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4181 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_self_type)), _endpos__2_, _, (_2 : 'tv_class_sig_fields)), _endpos__3_, _, (_3 : 'tv_opt_semi)) = _menhir_stack in
        let _v : 'tv_class_sig_body = 
# 2233 "reason_parser.mly"
    ( Csig.mk _1 (List.rev _2 ))
# 54760 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4179) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_class_sig_body) = _v in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
        match _menhir_s with
        | MenhirState1544 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4173 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error1548 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4171 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce34 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4172)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error1548 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4174)
        | MenhirState1790 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv4177 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
            ((if _menhir_env._menhir_error then
              _menhir_error1791 _menhir_env (Obj.magic _menhir_stack)
            else
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | RBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv4175 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_class_sig_body) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce34 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4176)
              | _ ->
                  (assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_error1791 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4178)
        | _ ->
            _menhir_fail ()) : 'freshtv4180)) : 'freshtv4182)) : 'freshtv4184)
    | MenhirState2258 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4185 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce181 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4186)
    | MenhirState2252 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4187 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state * 'tv_opt_semi) = Obj.magic _menhir_stack in
        (_menhir_reduce180 _menhir_env (Obj.magic _menhir_stack) : 'freshtv4188)
    | _ ->
        _menhir_fail ()

and _menhir_reduce218 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54834 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54846 "reason_parser.ml"
      
    in
    
# 2695 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "(") (with_txt _5 ")") )
# 54852 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce220 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54873 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 54885 "reason_parser.ml"
      
    in
    
# 2702 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "[") (with_txt _5 "]") )
# 54891 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_module_binding_body_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_module_binding_body_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState2637 | MenhirState2323 | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4147) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_expr) = _v in
        (_menhir_reduce578 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4148)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4149) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_module_binding_body_expr) = _v in
        (_menhir_reduce578 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4150)
    | _ ->
        _menhir_fail ()

and _menhir_goto__module_binding_body_functor : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_binding_body_functor -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2637 | MenhirState2323 | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4143) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_binding_body_functor) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce539 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4144)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4145) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_binding_body_functor) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce539 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4146)
    | _ ->
        _menhir_fail ()

and _menhir_goto_field_expr : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_field_expr -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    match _menhir_s with
    | MenhirState426 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4137 * _menhir_state * 'tv_field_expr_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_field_expr) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4135 * _menhir_state * 'tv_field_expr_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : 'tv_field_expr) : 'tv_field_expr) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_field_expr_list)), _endpos__2_, _) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_field_expr_list = 
# 3240 "reason_parser.mly"
      ( _3::_1 )
# 54954 "reason_parser.ml"
         in
        _menhir_goto_field_expr_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4136)) : 'freshtv4138)
    | MenhirState420 | MenhirState2197 | MenhirState1108 | MenhirState423 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4141) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_field_expr) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4139) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_field_expr) : 'tv_field_expr) = _v in
        ((let _v : 'tv_field_expr_list = 
# 3238 "reason_parser.mly"
     ( [_1] )
# 54969 "reason_parser.ml"
         in
        _menhir_goto_field_expr_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4140)) : 'freshtv4142)
    | _ ->
        _menhir_fail ()

and _menhir_goto__module_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__module_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2644 | MenhirState411 | MenhirState451 | MenhirState756 | MenhirState759 | MenhirState2142 | MenhirState2138 | MenhirState2128 | MenhirState1245 | MenhirState799 | MenhirState772 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4131) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce540 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4132)
    | MenhirState403 | MenhirState404 | MenhirState412 | MenhirState2306 | MenhirState2302 | MenhirState2295 | MenhirState2291 | MenhirState762 | MenhirState1476 | MenhirState1945 | MenhirState1948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4133) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__module_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce540 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4134)
    | _ ->
        _menhir_fail ()

and _menhir_goto__non_arrowed_module_type : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__non_arrowed_module_type -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2631 | MenhirState2304 | MenhirState2293 | MenhirState449 | MenhirState2140 | MenhirState1247 | MenhirState2122 | MenhirState2125 | MenhirState2113 | MenhirState1407 | MenhirState2036 | MenhirState1466 | MenhirState1414 | MenhirState1442 | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4127) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce544 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4128)
    | MenhirState401 | MenhirState2168 | MenhirState1248 | MenhirState1250 | MenhirState2116 | MenhirState1965 | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4129) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__non_arrowed_module_type) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce544 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4130)
    | _ ->
        _menhir_fail ()

and _menhir_goto__core_type2 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__core_type2 -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2649 | MenhirState2570 | MenhirState2479 | MenhirState118 | MenhirState128 | MenhirState141 | MenhirState2465 | MenhirState2463 | MenhirState153 | MenhirState2439 | MenhirState2442 | MenhirState2041 | MenhirState2027 | MenhirState1917 | MenhirState1895 | MenhirState1904 | MenhirState1898 | MenhirState1859 | MenhirState1744 | MenhirState1720 | MenhirState1723 | MenhirState1599 | MenhirState1545 | MenhirState1480 | MenhirState1471 | MenhirState1427 | MenhirState1425 | MenhirState1359 | MenhirState1361 | MenhirState1268 | MenhirState1328 | MenhirState1330 | MenhirState1287 | MenhirState1288 | MenhirState1278 | MenhirState1275 | MenhirState1271 | MenhirState1232 | MenhirState1206 | MenhirState1219 | MenhirState1211 | MenhirState1068 | MenhirState1071 | MenhirState1066 | MenhirState666 | MenhirState245 | MenhirState270 | MenhirState272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4123) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type2) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce556 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4124)
    | MenhirState106 | MenhirState2478 | MenhirState242 | MenhirState2406 | MenhirState327 | MenhirState1977 | MenhirState1975 | MenhirState1861 | MenhirState1700 | MenhirState1695 | MenhirState1680 | MenhirState1671 | MenhirState1601 | MenhirState1575 | MenhirState1579 | MenhirState1566 | MenhirState1561 | MenhirState1556 | MenhirState1501 | MenhirState1498 | MenhirState1089 | MenhirState1092 | MenhirState1087 | MenhirState722 | MenhirState681 | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4125) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__core_type2) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce556 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4126)
    | _ ->
        _menhir_fail ()

and _menhir_reduce649 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_package_type_cstr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _menhir_s, (_1 : 'tv_package_type_cstr)) = _menhir_stack in
    let _v : 'tv_package_type_cstrs = 
# 3983 "reason_parser.mly"
                      ( [_1] )
# 55047 "reason_parser.ml"
     in
    _menhir_goto_package_type_cstrs _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto__pattern_without_or : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__pattern_without_or -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4119) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__pattern_without_or) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce549 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4120)
    | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState738 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4121) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__pattern_without_or) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce549 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv4122)
    | _ ->
        _menhir_fail ()

and _menhir_goto_opt_comma : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_opt_comma -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4057 * _menhir_state * 'tv_field) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4055 * _menhir_state * 'tv_field) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_field)), _, (_2 : 'tv_opt_comma)) = _menhir_stack in
        let _v : 'tv_meth_list = 
# 4025 "reason_parser.mly"
                                                 ( [_1], Closed )
# 55086 "reason_parser.ml"
         in
        _menhir_goto_meth_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4056)) : 'freshtv4058)
    | MenhirState425 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4061 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error428 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4059 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce238 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4060)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error428 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4062)
    | MenhirState526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4065 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4063 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)), _endpos__2_), _endpos__3_, _, _startpos__3_), _, (_4 : 'tv_opt_comma)) = _menhir_stack in
        let _3 = () in
        let _2 = () in
        let _v : 'tv_lbl_pattern_list = 
# 3437 "reason_parser.mly"
                                           ( [_1], Open )
# 55118 "reason_parser.ml"
         in
        _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4064)) : 'freshtv4066)
    | MenhirState1083 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4069 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1095 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4067 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce231 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4068)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1095 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4070)
    | MenhirState1102 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4073 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1105 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4071 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce417 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4072)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1105 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4074)
    | MenhirState1109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4077 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1110 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4075 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce241 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4076)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1110 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4078)
    | MenhirState1337 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4081 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1340 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4079 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce818 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4080)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1340 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4082)
    | MenhirState1352 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv4085 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1353 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv4083 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce822 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4084)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1353 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4086)
    | MenhirState2183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4089 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2184 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4087 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce231 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4088)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2184 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4090)
    | MenhirState2191 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4093 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2194 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv4091 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce417 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4092)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2194 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4094)
    | MenhirState2198 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4097 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2199 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4095 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce241 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4096)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2199 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4098)
    | MenhirState2339 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4101 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2340 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | GREATERRBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4099 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce238 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4100)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2340 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4102)
    | MenhirState2354 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4105 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2355 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4103 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce228 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4104)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2355 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4106)
    | MenhirState2428 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4109 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2429 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BARRBRACKET ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv4107 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce228 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4108)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2429 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4110)
    | MenhirState2516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv4113 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2517 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv4111 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce818 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4112)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2517 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4114)
    | MenhirState2528 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv4117 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2529 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((((('freshtv4115 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) * _menhir_state * 'tv_opt_comma) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce822 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv4116)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2529 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv4118)
    | _ ->
        _menhir_fail ()

and _menhir_goto_attributes : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1280 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3939 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 55366 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce389 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3940)
    | MenhirState1282 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3941 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce336 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3942)
    | MenhirState1285 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3943 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce393 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3944)
    | MenhirState1290 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3945 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce390 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3946)
    | MenhirState1293 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3947 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce392 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3948)
    | MenhirState1296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3949 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce391 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3950)
    | MenhirState1300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3951 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 55412 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce384 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3952)
    | MenhirState1303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3953 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce388 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3954)
    | MenhirState1307 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3955 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce385 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3956)
    | MenhirState1310 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3957 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce387 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3958)
    | MenhirState1313 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3959 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce386 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3960)
    | MenhirState1332 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3973 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 55451 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3971 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 55460 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos__5_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_5 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mutable_flag), _startpos__1_), _endpos_x0_, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 55468 "reason_parser.ml"
        )), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_poly_type)) = _menhir_stack in
        let _3 = () in
        let _v : 'tv_label_declaration = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 55480 "reason_parser.ml"
          
        in
        let _startpos__2_ = _startpos_x0_ in
        let _endpos = _endpos__5_ in
        let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
          _startpos__1_
        else
          _startpos__2_ in
        
# 3623 "reason_parser.mly"
      (
       let loc = mklocation _symbolstartpos _endpos in
       Type.field _2 _4 ~mut:_1 ~attrs:_5 ~loc
      )
# 55495 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3969) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_label_declaration) = _v in
        ((match _menhir_s with
        | MenhirState1338 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3963 * _menhir_state * 'tv_label_declarations) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : 'tv_label_declaration) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3961 * _menhir_state * 'tv_label_declarations) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
            let (_ : _menhir_state) = _menhir_s in
            let ((_3 : 'tv_label_declaration) : 'tv_label_declaration) = _v in
            ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_label_declarations)), _endpos__2_, _) = _menhir_stack in
            let _2 = () in
            let _v : 'tv_label_declarations = 
# 3618 "reason_parser.mly"
                                                 ( _3 :: _1 )
# 55516 "reason_parser.ml"
             in
            _menhir_goto_label_declarations _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3962)) : 'freshtv3964)
        | MenhirState2527 | MenhirState2515 | MenhirState1351 | MenhirState1324 ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3967) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : 'tv_label_declaration) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3965) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let ((_1 : 'tv_label_declaration) : 'tv_label_declaration) = _v in
            ((let _v : 'tv_label_declarations = 
# 3617 "reason_parser.mly"
                                                ( [_1] )
# 55531 "reason_parser.ml"
             in
            _menhir_goto_label_declarations _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3966)) : 'freshtv3968)
        | _ ->
            _menhir_fail ()) : 'freshtv3970)) : 'freshtv3972)) : 'freshtv3974)
    | MenhirState1377 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3975 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 55541 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce422 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3976)
    | MenhirState1380 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3977 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce426 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3978)
    | MenhirState1384 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3979 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce423 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3980)
    | MenhirState1387 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce425 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3982)
    | MenhirState1390 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce424 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3984)
    | MenhirState1494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3985 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 55580 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce427 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3986)
    | MenhirState1496 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3987 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce336 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3988)
    | MenhirState1503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3989 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 55598 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce422 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3990)
    | MenhirState1507 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3991 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce431 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3992)
    | MenhirState1509 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3993 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce426 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3994)
    | MenhirState1514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce428 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3996)
    | MenhirState1516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce423 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3998)
    | MenhirState1520 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce430 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4000)
    | MenhirState1522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4001 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce425 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4002)
    | MenhirState1526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4003 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce429 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4004)
    | MenhirState1528 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4005 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce424 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4006)
    | MenhirState2441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4009 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4007 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        let (_endpos__4_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_), _, (_2 : 'tv_opt_ampersand)), _endpos__3_, _, (_3 : 'tv_amper_type_list)) = _menhir_stack in
        let _v : 'tv_tag_field = 
# 3996 "reason_parser.mly"
      ( Rtag (_1, _4, _2, List.rev _3) )
# 55675 "reason_parser.ml"
         in
        _menhir_goto_tag_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4008)) : 'freshtv4010)
    | MenhirState173 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4013 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4011 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_2 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
        let _v : 'tv_tag_field = 
# 3998 "reason_parser.mly"
      ( Rtag (_1, _2, true, []) )
# 55693 "reason_parser.ml"
         in
        _menhir_goto_tag_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4012)) : 'freshtv4014)
    | MenhirState2459 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4023 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4021 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        let (_endpos__4_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let ((_4 : 'tv_attributes) : 'tv_attributes) = _v in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_label), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_poly_type)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_field = 
# 4030 "reason_parser.mly"
                                               ( (_1, _4, _3) )
# 55712 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4019) = _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_field) = _v in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv4017 * _menhir_state * 'tv_field) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COMMA ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv4015 * _menhir_state * 'tv_field) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState287 in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | DOTDOT ->
                _menhir_run149 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState288
            | LIDENT _v ->
                _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState288 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | GREATER ->
                _menhir_reduce635 _menhir_env (Obj.magic _menhir_stack)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState288) : 'freshtv4016)
        | GREATER ->
            _menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState287
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState287) : 'freshtv4018)) : 'freshtv4020)) : 'freshtv4022)) : 'freshtv4024)
    | MenhirState2473 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4025 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 55754 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce389 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4026)
    | MenhirState2476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4027 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce393 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4028)
    | MenhirState2481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4029 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce390 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4030)
    | MenhirState2484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4031 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce392 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4032)
    | MenhirState2487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv4033 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce391 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4034)
    | MenhirState2491 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4035 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 55793 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce384 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4036)
    | MenhirState2494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4037 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce388 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4038)
    | MenhirState2498 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4039 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce385 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4040)
    | MenhirState2501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4041 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce387 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4042)
    | MenhirState2504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4043 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce386 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4044)
    | MenhirState2593 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4045 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 55832 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce427 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4046)
    | MenhirState2597 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4047 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce431 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4048)
    | MenhirState2602 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv4049 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce428 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4050)
    | MenhirState2606 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4051 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce430 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4052)
    | MenhirState2610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv4053 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_attributes) = _v in
        (_menhir_reduce429 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv4054)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constructor_declarations : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constructor_declarations -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    match _menhir_s with
    | MenhirState1287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3923 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce816 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3924)
    | MenhirState1346 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3925 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce820 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3926)
    | MenhirState1344 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3927 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce819 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3928)
    | MenhirState1268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3929 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce815 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3930)
    | MenhirState2478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3931 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce816 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3932)
    | MenhirState2522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3933 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce820 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3934)
    | MenhirState2520 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3935 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce819 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3936)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3937 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constructor_declarations) = _v in
        (_menhir_reduce815 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3938)
    | _ ->
        _menhir_fail ()

and _menhir_reduce813 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_core_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_type_kind = 
# 3494 "reason_parser.mly"
      ( (Ptype_abstract, Public, Some _2) )
# 55939 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_goto_generalized_constructor_arguments : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_generalized_constructor_arguments -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1345 | MenhirState1269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3863 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 55952 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1280
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1280
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1280) : 'freshtv3864)
    | MenhirState1284 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3865 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1285
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1285
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1285) : 'freshtv3866)
    | MenhirState1289 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3867 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1290
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1290
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1290) : 'freshtv3868)
    | MenhirState1292 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3869 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1293
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1293
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1293) : 'freshtv3870)
    | MenhirState1295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3871 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1296
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1296
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1296) : 'freshtv3872)
    | MenhirState1299 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3873 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 56016 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1300
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1300
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1300) : 'freshtv3874)
    | MenhirState1302 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3875 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1303
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1303
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1303) : 'freshtv3876)
    | MenhirState1306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3877 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1307
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1307
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1307) : 'freshtv3878)
    | MenhirState1309 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3879 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1310
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1310
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1310) : 'freshtv3880)
    | MenhirState1312 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3881 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1313
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1313
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1313) : 'freshtv3882)
    | MenhirState2589 | MenhirState1376 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3883 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 56080 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1377
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1377
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1377) : 'freshtv3884)
    | MenhirState2595 | MenhirState1379 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3885 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1380
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1380
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1380) : 'freshtv3886)
    | MenhirState2600 | MenhirState1383 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3887 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1384
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1384
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1384) : 'freshtv3888)
    | MenhirState2604 | MenhirState1386 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3889 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1387
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1387
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1387) : 'freshtv3890)
    | MenhirState2608 | MenhirState1389 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3891 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1390
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState1390
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1390) : 'freshtv3892)
    | MenhirState1490 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3893 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 56144 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1503
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1503
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1503) : 'freshtv3894)
    | MenhirState1505 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3895 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1509
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1509
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1509) : 'freshtv3896)
    | MenhirState1512 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3897 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1516
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1516
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1516) : 'freshtv3898)
    | MenhirState1518 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3899 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1522
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1522
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1522) : 'freshtv3900)
    | MenhirState1524 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3901 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1528
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1528
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1528) : 'freshtv3902)
    | MenhirState2521 | MenhirState108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3903 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 56208 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2473
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2473
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2473) : 'freshtv3904)
    | MenhirState2475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3905 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2476
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2476
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2476) : 'freshtv3906)
    | MenhirState2480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3907 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2481
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2481
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2481) : 'freshtv3908)
    | MenhirState2483 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3909 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2484
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2484
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2484) : 'freshtv3910)
    | MenhirState2486 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3911 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2487
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2487
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2487) : 'freshtv3912)
    | MenhirState2490 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3913 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 56272 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2491
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2491
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2491) : 'freshtv3914)
    | MenhirState2493 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3915 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2494
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2494
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2494) : 'freshtv3916)
    | MenhirState2497 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3917 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2498
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2498
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2498) : 'freshtv3918)
    | MenhirState2500 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3919 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2501
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2501
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2501) : 'freshtv3920)
    | MenhirState2503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3921 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2504
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState2504
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2504) : 'freshtv3922)
    | _ ->
        _menhir_fail ()

and _menhir_goto_post_item_attributes : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_post_item_attributes -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1159 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3629 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1160 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3627 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1156 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1161 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1161) : 'freshtv3628)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1160 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3630)
    | MenhirState1174 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3633 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1175
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3631 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1175 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1156 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1176 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce637 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3632)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState1175) : 'freshtv3634)
    | MenhirState1181 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3635 * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce674 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3636)
    | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3637 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce328 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3638)
    | MenhirState1369 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3639 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce519 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3640)
    | MenhirState1392 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3649 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3647 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_nonrec_flag), _startpos__2_), _, (_3 : 'tv_potentially_long_ident_and_optional_type_parameters)), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_private_flag), _startpos__5_), _, (_6 : 'tv_opt_bar)), _endpos__7_, _, (_7 : 'tv_sig_extension_constructors)), _endpos__8_, _, (_8 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__8_ in
        let _v : 'tv_sig_type_extension = 
# 3655 "reason_parser.mly"
  (
    if _2 <> Recursive then not_expecting _startpos__2_ _endpos__2_ "nonrec flag";
    let (potentially_long_ident, optional_type_parameters) = _3 in
    Te.mk potentially_long_ident (List.rev _7)
          ~params:optional_type_parameters ~priv:_5 ~attrs:_8
  )
# 56553 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3645) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3643) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_type_extension) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3641) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_sig_type_extension) : 'tv_sig_type_extension) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1682 "reason_parser.mly"
                       (
      mksig(Psig_typext _1)
    )
# 56580 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3642)) : 'freshtv3644)) : 'freshtv3646)) : 'freshtv3648)) : 'freshtv3650)
    | MenhirState1402 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3651 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce606 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3652)
    | MenhirState1460 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3655 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 56592 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3653 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 56598 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 56603 "reason_parser.ml"
        )), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos_x1_, _, (x1 : 'tv_mod_longident), _startpos_x1_), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__signature_item = let _5 =
          let _endpos_x_ = _endpos_x1_ in
          let _startpos_x_ = _startpos_x1_ in
          let x = x1 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56619 "reason_parser.ml"
          
        in
        let _endpos__5_ = _endpos_x1_ in
        let _startpos__5_ = _startpos_x1_ in
        let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56633 "reason_parser.ml"
          
        in
        let _endpos = _endpos__6_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1692 "reason_parser.mly"
                                                                               (
      let loc = mklocation _symbolstartpos _endpos in
      let loc_mod = mklocation _startpos__5_ _endpos__5_ in
      mksig(
        Psig_module (
          Md.mk
            _3
            (Mty.alias ~loc:loc_mod _5)
            ~attrs:_6
            ~loc
        )
      )
    )
# 56653 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3654)) : 'freshtv3656)
    | MenhirState1462 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3659 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 56661 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 56667 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 56672 "reason_parser.ml"
        )), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_module_declaration)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv__signature_item = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56687 "reason_parser.ml"
          
        in
        let _endpos = _endpos__5_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1688 "reason_parser.mly"
                                                                      (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_module (Md.mk _3 _4 ~attrs:_5 ~loc))
    )
# 56698 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3658)) : 'freshtv3660)
    | MenhirState1468 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3663 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3661 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_rec_declaration_details)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _3 = () in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv_many_module_rec_declarations = let _endpos = _endpos__5_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1760 "reason_parser.mly"
                                                                       (
      let (ident, body) = _4 in
      let loc = mklocation _symbolstartpos _endpos in
      [Md.mk ident body ~attrs:_5 ~loc]
    )
# 56721 "reason_parser.ml"
         in
        _menhir_goto_many_module_rec_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3662)) : 'freshtv3664)
    | MenhirState1472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3667 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3665 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _3 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv__signature_item = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 56743 "reason_parser.ml"
          
        in
        let _endpos = _endpos__5_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1671 "reason_parser.mly"
                                                               (
        let loc = mklocation _symbolstartpos _endpos in
        mksig(Psig_value (Val.mk _2 _4 ~attrs:_5 ~loc))
    )
# 56754 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3666)) : 'freshtv3668)
    | MenhirState1485 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3669 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce298 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3670)
    | MenhirState1487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3671 * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce674 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3672)
    | MenhirState1530 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3673 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce781 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3674)
    | MenhirState1532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3675 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce780 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3676)
    | MenhirState1563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3677 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_value_type)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_sig_field = 
# 2259 "reason_parser.mly"
                                        (
      mkctf_attrs (Pctf_val _2) _3
    )
# 56787 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3678)) : 'freshtv3680)
    | MenhirState1581 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3683 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3681 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), (_2 : 'tv_private_virtual_flags)), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_poly_type)), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__class_sig_field = 
# 2263 "reason_parser.mly"
       (
        let (p, v) = _2 in
        mkctf_attrs (Pctf_method (_3, p, v, _5)) _6
       )
# 56806 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3682)) : 'freshtv3684)
    | MenhirState1595 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3687 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_instance_type), _startpos__2_), _endpos__3_, _, (_3 : 'tv_item_attribute)), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : 'tv__class_sig_field = 
# 2258 "reason_parser.mly"
      ( mkctf_attrs (Pctf_inherit _2) (_3::_4) )
# 56821 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3686)) : 'freshtv3688)
    | MenhirState1603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_constrain_field)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_sig_field = 
# 2268 "reason_parser.mly"
       ( mkctf_attrs (Pctf_constraint _2) _3 )
# 56836 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3690)) : 'freshtv3692)
    | MenhirState1606 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3695 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3693 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__class_sig_field = 
# 2270 "reason_parser.mly"
       ( mkctf_attrs (Pctf_extension _1) _2 )
# 56850 "reason_parser.ml"
         in
        _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3694)) : 'freshtv3696)
    | MenhirState1623 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3697 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce513 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3698)
    | MenhirState1635 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3699 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce504 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3700)
    | MenhirState1674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_value)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_field = 
# 1982 "reason_parser.mly"
      ( mkcf_attrs (Pcf_val _2) _3 )
# 56873 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3702)) : 'freshtv3704)
    | MenhirState1731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3707 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3705 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_method_)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_field = 
# 1984 "reason_parser.mly"
      ( mkcf_attrs (Pcf_method _2) _3 )
# 56888 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3706)) : 'freshtv3708)
    | MenhirState1735 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3711 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : 'tv__class_field = 
# 1988 "reason_parser.mly"
      ( mkcf_attrs (Pcf_initializer _3) _4 )
# 56904 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3710)) : 'freshtv3712)
    | MenhirState1742 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3715 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3713 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_override_flag), _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_expr), _startpos__3_), _endpos__4_, _, (_4 : 'tv_parent_binder)), _endpos__5_, _, (_5 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__5_ in
        let _v : 'tv__class_field = 
# 1980 "reason_parser.mly"
      ( mkcf_attrs (Pcf_inherit (_2, _3, _4)) _5 )
# 56919 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3714)) : 'freshtv3716)
    | MenhirState1745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3719 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3717 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_constrain_field)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__class_field = 
# 1986 "reason_parser.mly"
      ( mkcf_attrs (Pcf_constraint _2) _3 )
# 56934 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3718)) : 'freshtv3720)
    | MenhirState1749 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3723 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3721 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__class_field = 
# 1990 "reason_parser.mly"
      ( mkcf_attrs (Pcf_extension _1) _2 )
# 56948 "reason_parser.ml"
         in
        _menhir_goto__class_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3722)) : 'freshtv3724)
    | MenhirState1842 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3725 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce509 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3726)
    | MenhirState1867 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3727 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce331 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3728)
    | MenhirState1872 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3729 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce330 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3730)
    | MenhirState1877 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3731 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce327 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3732)
    | MenhirState1882 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3733 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce325 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3734)
    | MenhirState1924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3735 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce328 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3736)
    | MenhirState1932 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3737 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce310 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3738)
    | MenhirState1935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3739 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce297 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3740)
    | MenhirState2025 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce309 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3742)
    | MenhirState2037 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3745 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3743 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv__signature_item = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1719 "reason_parser.mly"
                                             (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_include (Incl.mk _2 ~attrs:_3 ~loc))
    )
# 57004 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3744)) : 'freshtv3746)
    | MenhirState2046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3749 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3747 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_core_type), _startpos__4_), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_primitive_declaration)), _endpos__7_, _, (_7 : 'tv_post_item_attributes)) = _menhir_stack in
        let _5 = () in
        let _3 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__7_ in
        let _v : 'tv__signature_item = let _2 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 57027 "reason_parser.ml"
          
        in
        let _endpos = _endpos__7_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1675 "reason_parser.mly"
                                                                                                (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_value (Val.mk _2 _4 ~prim:_6 ~attrs:_7 ~loc))
    )
# 57038 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3748)) : 'freshtv3750)
    | MenhirState2049 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3759 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3757 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_extension_constructor_declaration)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv_sig_exception_declaration = 
# 3598 "reason_parser.mly"
      (
        let ext = _2 in
        {ext with pext_attributes = ext.pext_attributes @ _3}
      )
# 57056 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3755) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3753) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_sig_exception_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3751) = Obj.magic _menhir_stack in
        let (_endpos__1_ : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : 'tv_sig_exception_declaration) : 'tv_sig_exception_declaration) = _v in
        let (_startpos__1_ : Lexing.position) = _startpos in
        ((let _startpos = _startpos__1_ in
        let _endpos = _endpos__1_ in
        let _v : 'tv__signature_item = 
# 1685 "reason_parser.mly"
                              (
      mksig(Psig_exception _1)
    )
# 57083 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3752)) : 'freshtv3754)) : 'freshtv3756)) : 'freshtv3758)) : 'freshtv3760)
    | MenhirState2058 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3761 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce513 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3762)
    | MenhirState2065 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3765 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3763 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_description_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__3_ in
        let _v : 'tv_many_class_descriptions = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 2292 "reason_parser.mly"
                                                         (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    [Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc]
  )
# 57108 "reason_parser.ml"
         in
        _menhir_goto_many_class_descriptions _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3764)) : 'freshtv3766)
    | MenhirState2078 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3767 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce331 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3768)
    | MenhirState2083 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3777 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3775 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_rec_declaration_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_and_module_rec_declaration = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1771 "reason_parser.mly"
                                                            (
      let (ident, body) = _2 in
      let loc = mklocation _symbolstartpos _endpos in
      Md.mk ident body ~attrs:_3 ~loc
    )
# 57132 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3773) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_module_rec_declaration) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3771 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_module_rec_declaration) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3769 * Lexing.position * _menhir_state * 'tv_many_module_rec_declarations * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let ((_2 : 'tv_and_module_rec_declaration) : 'tv_and_module_rec_declaration) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_module_rec_declarations), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv_many_module_rec_declarations = 
# 1765 "reason_parser.mly"
                                                             (
      _2::_1
    )
# 57154 "reason_parser.ml"
         in
        _menhir_goto_many_module_rec_declarations _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3770)) : 'freshtv3772)) : 'freshtv3774)) : 'freshtv3776)) : 'freshtv3778)
    | MenhirState2088 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3779 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce327 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3780)
    | MenhirState2093 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3789 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3787 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_description_details)), _endpos__3_, _, (_3 : 'tv_post_item_attributes)) = _menhir_stack in
        let _1 = () in
        let _endpos = _endpos__3_ in
        let _v : 'tv_and_class_description = let _endpos = _endpos__3_ in
        let _symbolstartpos = _startpos__1_ in
        
# 2302 "reason_parser.mly"
                                                       (
    let (ident, binding, virt, params) = _2 in
    let loc = mklocation _symbolstartpos _endpos in
    Ci.mk ident binding ~virt ~params ~attrs:_3 ~loc
  )
# 57178 "reason_parser.ml"
         in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3785) = _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_class_description) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3783 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_v : 'tv_and_class_description) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3781 * Lexing.position * _menhir_state * 'tv_many_class_descriptions * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let ((_2 : 'tv_and_class_description) : 'tv_and_class_description) = _v in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_descriptions), _startpos__1_) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv_many_class_descriptions = 
# 2297 "reason_parser.mly"
                                                  (
    _2 :: _1
  )
# 57200 "reason_parser.ml"
         in
        _menhir_goto_many_class_descriptions _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3782)) : 'freshtv3784)) : 'freshtv3786)) : 'freshtv3788)) : 'freshtv3790)
    | MenhirState2096 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3793 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3791 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_item_extension), _startpos__1_), _endpos__2_, _, (_2 : 'tv_post_item_attributes)) = _menhir_stack in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__2_ in
        let _v : 'tv__signature_item = 
# 1729 "reason_parser.mly"
                                        (
      mksig(Psig_extension (_1, _2))
    )
# 57216 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3792)) : 'freshtv3794)
    | MenhirState2103 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3797 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3795 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_type), _startpos__5_), _endpos__6_, _, (_6 : 'tv_post_item_attributes)) = _menhir_stack in
        let _4 = () in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__6_ in
        let _v : 'tv__signature_item = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 57239 "reason_parser.ml"
          
        in
        let _endpos = _endpos__6_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1712 "reason_parser.mly"
                                                                     (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_modtype (Mtd.mk _3 ~typ:_5 ~loc ~attrs:_6))
    )
# 57250 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3796)) : 'freshtv3798)
    | MenhirState1406 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3801 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3799 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos_x0_, _, (x0 : 'tv_ident), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_post_item_attributes)) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _endpos = _endpos__4_ in
        let _v : 'tv__signature_item = let _3 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 57272 "reason_parser.ml"
          
        in
        let _endpos = _endpos__4_ in
        let _symbolstartpos = _startpos__1_ in
        
# 1708 "reason_parser.mly"
                                                   (
      let loc = mklocation _symbolstartpos _endpos in
      mksig(Psig_modtype (Mtd.mk _3 ~attrs:_4 ~loc))
    )
# 57283 "reason_parser.ml"
         in
        _menhir_goto__signature_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3800)) : 'freshtv3802)
    | MenhirState2132 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3805 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 57291 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2133 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3803 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 57303 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run1156 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2134 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2134) : 'freshtv3804)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2133 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3806)
    | MenhirState2257 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3809 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2258
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3807 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState2258 in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2259 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERAMPER | AMPERSAND | AND | AS | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | MATCH | METHOD | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | THEN | TILDE | TO | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WITH ->
                  _menhir_reduce637 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3808)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState2258) : 'freshtv3810)
    | MenhirState2310 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3811 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce517 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3812)
    | MenhirState2312 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3813 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce302 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3814)
    | MenhirState2315 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3815 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce504 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3816)
    | MenhirState438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3819 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2319 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3817 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2320 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2320) : 'freshtv3818)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2319 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3820)
    | MenhirState2324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3823 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 57590 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2325 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (((('freshtv3821 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 57602 "reason_parser.ml"
              ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_stack = (_menhir_stack, _endpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LET ->
                  _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2326 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2326) : 'freshtv3822)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2325 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3824)
    | MenhirState2396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce305 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3826)
    | MenhirState400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce304 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3828)
    | MenhirState395 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3829 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce606 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3830)
    | MenhirState2533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3831 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce519 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3832)
    | MenhirState2539 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3833 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce786 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3834)
    | MenhirState2612 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((('freshtv3835 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce786 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3836)
    | MenhirState2632 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3837 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce305 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3838)
    | MenhirState2630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3839 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce304 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3840)
    | MenhirState2640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3841 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce517 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3842)
    | MenhirState2642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3843 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce302 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3844)
    | MenhirState2645 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3845 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce309 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3846)
    | MenhirState2652 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3847 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce298 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3848)
    | MenhirState2655 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3849 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce781 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3850)
    | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3851 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce780 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3852)
    | MenhirState2717 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3853 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce509 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3854)
    | MenhirState2733 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3855 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce330 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3856)
    | MenhirState2739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3857 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce325 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3858)
    | MenhirState2745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3859 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce310 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3860)
    | MenhirState2748 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3861 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce297 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3862)
    | _ ->
        _menhir_fail ()

and _menhir_goto__expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState475 | MenhirState2253 | MenhirState2255 | MenhirState2259 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState771 | MenhirState1147 | MenhirState1142 | MenhirState1134 | MenhirState1115 | MenhirState1123 | MenhirState1081 | MenhirState1084 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1047 | MenhirState1051 | MenhirState816 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState833 | MenhirState886 | MenhirState860 | MenhirState880 | MenhirState882 | MenhirState884 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState866 | MenhirState870 | MenhirState868 | MenhirState864 | MenhirState862 | MenhirState858 | MenhirState856 | MenhirState850 | MenhirState854 | MenhirState852 | MenhirState835 | MenhirState837 | MenhirState847 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3623) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce533 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3624)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 | MenhirState2574 | MenhirState2422 | MenhirState2374 | MenhirState2376 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2187 | MenhirState2192 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState770 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1167 | MenhirState1171 | MenhirState1176 | MenhirState791 | MenhirState1138 | MenhirState1098 | MenhirState1103 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1032 | MenhirState1034 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1008 | MenhirState1013 | MenhirState1015 | MenhirState1011 | MenhirState1006 | MenhirState929 | MenhirState946 | MenhirState999 | MenhirState973 | MenhirState993 | MenhirState995 | MenhirState997 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState979 | MenhirState983 | MenhirState981 | MenhirState977 | MenhirState975 | MenhirState971 | MenhirState969 | MenhirState963 | MenhirState967 | MenhirState965 | MenhirState948 | MenhirState950 | MenhirState960 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3625) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce533 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3626)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_simple_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1632 | MenhirState1663 | MenhirState1738 | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3619) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce526 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3620)
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3621) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__class_simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce526 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3622)
    | _ ->
        _menhir_fail ()

and _menhir_goto_mod_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_mod_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState2709 | MenhirState2705 | MenhirState2700 | MenhirState2664 | MenhirState2680 | MenhirState2681 | MenhirState122 | MenhirState120 | MenhirState7 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3539 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState9
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3537 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState9 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3535 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState10 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 57838 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce351 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv3536)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState10 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState10) : 'freshtv3538)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState9)) : 'freshtv3540)
    | MenhirState2673 | MenhirState2665 | MenhirState1831 | MenhirState1824 | MenhirState1819 | MenhirState1630 | MenhirState1631 | MenhirState1738 | MenhirState1663 | MenhirState1632 | MenhirState1640 | MenhirState1641 | MenhirState243 | MenhirState181 | MenhirState109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3545 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState111
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3543 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState111 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3541 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState112 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 57877 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce351 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos) : 'freshtv3542)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState112 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState112) : 'freshtv3544)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState111)) : 'freshtv3546)
    | MenhirState736 | MenhirState2202 | MenhirState1972 | MenhirState1420 | MenhirState1113 | MenhirState924 | MenhirState240 | MenhirState135 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3551 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState137
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3549 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState137 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3547 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState138 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 57916 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce467 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3548)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState138 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState138) : 'freshtv3550)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState137)) : 'freshtv3552)
    | MenhirState394 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3553 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState395
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run554 _menhir_env (Obj.magic _menhir_stack) MenhirState395
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState395
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState395) : 'freshtv3554)
    | MenhirState437 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3555 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState438
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState438
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState438
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState438) : 'freshtv3556)
    | MenhirState472 | MenhirState1131 | MenhirState1120 | MenhirState1117 | MenhirState815 | MenhirState615 | MenhirState525 | MenhirState516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3561 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState518
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3559 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState518 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3557 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState519 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 57984 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  _menhir_reduce467 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3558)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState519 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState519) : 'freshtv3560)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState518)) : 'freshtv3562)
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1525 | MenhirState1519 | MenhirState1513 | MenhirState1506 | MenhirState1491 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState563 | MenhirState564 | MenhirState540 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3563 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState553
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run554 _menhir_env (Obj.magic _menhir_stack) MenhirState553
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce374 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3564)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2609 | MenhirState2605 | MenhirState2601 | MenhirState2596 | MenhirState2590 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState486 | MenhirState738 | MenhirState1886 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1657 | MenhirState1639 | MenhirState1637 | MenhirState1634 | MenhirState1183 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState649 | MenhirState650 | MenhirState642 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3565 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState634
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState634
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce374 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3566)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState2555 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState180 | MenhirState184 | MenhirState187 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState457 | MenhirState464 | MenhirState470 | MenhirState2274 | MenhirState473 | MenhirState2268 | MenhirState2270 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1647 | MenhirState1648 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState840 | MenhirState842 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 | MenhirState816 | MenhirState817 | MenhirState811 | MenhirState805 | MenhirState747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3569 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState753
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3567 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState753 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run2202 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run2197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run2187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run2182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run755 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run555 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState754 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState754) : 'freshtv3568)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce374 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3570)
    | MenhirState2644 | MenhirState2645 | MenhirState411 | MenhirState2366 | MenhirState451 | MenhirState2288 | MenhirState756 | MenhirState2172 | MenhirState759 | MenhirState761 | MenhirState2142 | MenhirState2143 | MenhirState2138 | MenhirState2139 | MenhirState2128 | MenhirState2129 | MenhirState1245 | MenhirState1246 | MenhirState799 | MenhirState800 | MenhirState772 | MenhirState773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3571 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState780
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState780
          | AND | COLON | EOF | EQUAL | EQUALGREATER | LBRACE | LBRACKETAT | LBRACKETATAT | LBRACKETPERCENT | LPAREN | RPAREN | SEMI | UIDENT _ | WITH ->
              _menhir_reduce251 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState780)) : 'freshtv3572)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 | MenhirState2687 | MenhirState2688 | MenhirState2574 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState417 | MenhirState2350 | MenhirState418 | MenhirState419 | MenhirState2277 | MenhirState474 | MenhirState2229 | MenhirState729 | MenhirState730 | MenhirState734 | MenhirState2213 | MenhirState740 | MenhirState746 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState764 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState894 | MenhirState895 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState953 | MenhirState955 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 | MenhirState929 | MenhirState930 | MenhirState920 | MenhirState910 | MenhirState912 | MenhirState913 | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3589 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState796
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3587 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState796 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3575 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState797 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | DOTDOTDOT ->
                      _menhir_run1115 _menhir_env (Obj.magic _menhir_stack) MenhirState1113 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run136 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1113 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | RBRACE ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : (('freshtv3573 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_menhir_s : _menhir_state) = MenhirState1113 in
                      ((let _menhir_env = _menhir_discard _menhir_env in
                      _menhir_reduce216 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv3574)
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1113 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1113) : 'freshtv3576)
              | LBRACELESS ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3577 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState797 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run421 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1108 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1108) : 'freshtv3578)
              | LBRACKET ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3579 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState797 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | DOTDOTDOT ->
                      _menhir_run1099 _menhir_env (Obj.magic _menhir_stack) MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1098 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1098) : 'freshtv3580)
              | LBRACKETBAR ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3581 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState797 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1081 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1081) : 'freshtv3582)
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState797 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3585 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState797 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | AMPERAMPER ->
                      _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | AMPERSAND ->
                      _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | ASSERT ->
                      _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BACKQUOTE ->
                      _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BANG ->
                      _menhir_run805 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | BARBAR ->
                      _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | CHAR _v ->
                      _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | COLONEQUAL ->
                      _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FALSE ->
                      _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FLOAT _v ->
                      _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FOR ->
                      _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | FUN ->
                      _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | GREATER ->
                      _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | IF ->
                      _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP0 _v ->
                      _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP1 _v ->
                      _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP2 _v ->
                      _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP3 _v ->
                      _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INFIXOP4 _v ->
                      _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT _v ->
                      _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT32 _v ->
                      _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | INT64 _v ->
                      _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LAZY ->
                      _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACE ->
                      _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACELESS ->
                      _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKET ->
                      _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETBAR ->
                      _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LBRACKETPERCENT ->
                      _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESS ->
                      _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSDOTDOTGREATER ->
                      _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LESSGREATER ->
                      _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LIDENT _v ->
                      _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | LPAREN ->
                      _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUS ->
                      _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MINUSDOT ->
                      _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | MODULE ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : (('freshtv3583 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_menhir_s : _menhir_state) = MenhirState798 in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | FUN ->
                          _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACE ->
                          _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LBRACKETPERCENT ->
                          _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | LPAREN ->
                          _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState799 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | UIDENT _v ->
                          _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState799 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState799) : 'freshtv3584)
                  | NATIVEINT _v ->
                      _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | NEW ->
                      _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | OR ->
                      _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PERCENT ->
                      _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUS ->
                      _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSDOT ->
                      _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PLUSEQ ->
                      _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | PREFIXOP _v ->
                      _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STAR ->
                      _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | STRING _v ->
                      _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | SWITCH ->
                      _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRUE ->
                      _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | TRY ->
                      _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | UIDENT _v ->
                      _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | WHILE ->
                      _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState798 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState798) : 'freshtv3586)
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState797 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState797) : 'freshtv3588)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce374 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3590)
    | MenhirState1158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3591 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1159
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1159
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1159
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1159) : 'freshtv3592)
    | MenhirState1401 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3593 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1402
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1402
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1402
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1402) : 'freshtv3594)
    | MenhirState1430 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3597 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1434
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1434
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3595 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1434 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1435 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1435) : 'freshtv3596)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1434)) : 'freshtv3598)
    | MenhirState1459 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv3599 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 58508 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1460
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1460
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1460
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1460) : 'freshtv3600)
    | MenhirState1980 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3603 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1997
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run635 _menhir_env (Obj.magic _menhir_stack) MenhirState1997
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3601 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1997 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | UIDENT _v ->
                  _menhir_run1983 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1998 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1998) : 'freshtv3602)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1997)) : 'freshtv3604)
    | MenhirState403 | MenhirState2395 | MenhirState404 | MenhirState2393 | MenhirState412 | MenhirState2364 | MenhirState2306 | MenhirState2307 | MenhirState2302 | MenhirState2303 | MenhirState2295 | MenhirState2296 | MenhirState2291 | MenhirState2292 | MenhirState762 | MenhirState2165 | MenhirState1476 | MenhirState2025 | MenhirState1945 | MenhirState1947 | MenhirState1948 | MenhirState1962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3605 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2016
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              _menhir_run554 _menhir_env (Obj.magic _menhir_stack) MenhirState2016
          | AND | COLON | EOF | EQUALGREATER | LBRACE | LBRACKETAT | LBRACKETATAT | LBRACKETPERCENT | LPAREN | RBRACE | RBRACKET | RPAREN | SEMI | UIDENT _ | WITH ->
              _menhir_reduce251 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2016)) : 'freshtv3606)
    | MenhirState432 | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3611 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2247
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3609 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2247 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACE ->
                  _menhir_run2202 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run2197 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run2187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run2182 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3607 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState2248 in
                  let (_v : (
# 857 "reason_parser.mly"
       (string)
# 58599 "reason_parser.ml"
                  )) = _v in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                      _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)
                  | COLON | COMMA ->
                      _menhir_reduce467 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3608)
              | LPAREN ->
                  _menhir_run755 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run555 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2248 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2248) : 'freshtv3610)
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce374 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3612)
    | MenhirState2620 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3617 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2626
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | DOT ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3613 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState2626 in
              ((let _menhir_stack = (_menhir_stack, _menhir_s) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LIDENT _v ->
                  _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2627 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run11 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2627 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2627) : 'freshtv3614)
          | EOF | SEMI ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3615 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
              ((let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_ident), _startpos__2_), _endpos__3_, _, (_3 : 'tv_mod_longident), _startpos__3_) = _menhir_stack in
              let _1 = () in
              let _v : 'tv_toplevel_directive = 
# 4179 "reason_parser.mly"
                                ( Ptop_dir(_2, Pdir_ident _3) )
# 58654 "reason_parser.ml"
               in
              _menhir_goto_toplevel_directive _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3616)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2626)) : 'freshtv3618)
    | _ ->
        _menhir_fail ()

and _menhir_goto_nonrec_flag : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_nonrec_flag -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState74 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3529 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3527) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState76 in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 58682 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | MINUS ->
                _menhir_run100 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | PLUS ->
                _menhir_run96 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | QUOTE ->
                _menhir_run92 _menhir_env (Obj.magic _menhir_stack) MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UNDERSCORE ->
                _menhir_run91 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState90 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | VAL | VIRTUAL | WHEN | WHILE | WITH ->
                _menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState90) : 'freshtv3528)
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState76 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState76) : 'freshtv3530)
    | MenhirState1252 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3531 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run1254 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1253 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1253) : 'freshtv3532)
    | MenhirState2583 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3533 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run1254 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2584 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2584 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2584) : 'freshtv3534)
    | _ ->
        _menhir_fail ()

and _menhir_goto_additive : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_additive -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState475 | MenhirState2253 | MenhirState2255 | MenhirState2259 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState771 | MenhirState1147 | MenhirState1142 | MenhirState1134 | MenhirState1115 | MenhirState1123 | MenhirState1081 | MenhirState1084 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1047 | MenhirState1051 | MenhirState816 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState833 | MenhirState886 | MenhirState860 | MenhirState880 | MenhirState882 | MenhirState884 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState866 | MenhirState870 | MenhirState868 | MenhirState864 | MenhirState862 | MenhirState858 | MenhirState856 | MenhirState850 | MenhirState854 | MenhirState852 | MenhirState835 | MenhirState847 | MenhirState837 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3523 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState847 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState847) : 'freshtv3524)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 | MenhirState2574 | MenhirState2422 | MenhirState2374 | MenhirState2376 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2187 | MenhirState2192 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState770 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1167 | MenhirState1171 | MenhirState1176 | MenhirState791 | MenhirState1138 | MenhirState1098 | MenhirState1103 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1032 | MenhirState1034 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1008 | MenhirState1013 | MenhirState1015 | MenhirState1011 | MenhirState1006 | MenhirState929 | MenhirState946 | MenhirState999 | MenhirState973 | MenhirState993 | MenhirState995 | MenhirState997 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState979 | MenhirState983 | MenhirState981 | MenhirState977 | MenhirState975 | MenhirState971 | MenhirState969 | MenhirState963 | MenhirState967 | MenhirState965 | MenhirState948 | MenhirState960 | MenhirState950 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3525 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState960 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState960) : 'freshtv3526)
    | _ ->
        _menhir_fail ()

and _menhir_goto_override_flag : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_override_flag -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState392 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3495 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState394 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState394) : 'freshtv3496)
    | MenhirState436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3497 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState437 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState437) : 'freshtv3498)
    | MenhirState1157 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3499 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1158 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1158) : 'freshtv3500)
    | MenhirState1400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3501 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1401 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1401) : 'freshtv3502)
    | MenhirState1667 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3509 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | MUTABLE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3507 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1676 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | VIRTUAL ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3503 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | LIDENT _v ->
                    _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1678 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1678) : 'freshtv3504)
            | LIDENT _ ->
                _menhir_reduce590 _menhir_env (Obj.magic _menhir_stack)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3505 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3506)) : 'freshtv3508)
        | LIDENT _ ->
            _menhir_reduce589 _menhir_env (Obj.magic _menhir_stack) MenhirState1676
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1676) : 'freshtv3510)
    | MenhirState1690 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3519 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | PRIVATE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3515 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1691 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | VIRTUAL ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3511 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | LIDENT _v ->
                    _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1698 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1698) : 'freshtv3512)
            | LIDENT _ ->
                _menhir_reduce680 _menhir_env (Obj.magic _menhir_stack)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : ('freshtv3513 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3514)) : 'freshtv3516)
        | VIRTUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3517 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1691 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | PRIVATE ->
                _menhir_run1373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1692 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _ ->
                _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState1692
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1692) : 'freshtv3518)
        | LIDENT _ ->
            _menhir_reduce679 _menhir_env (Obj.magic _menhir_stack) MenhirState1691
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1691) : 'freshtv3520)
    | MenhirState1737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3521 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | CLASS ->
            _menhir_run1641 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run1637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run1632 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1631 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1738) : 'freshtv3522)
    | _ ->
        _menhir_fail ()

and _menhir_reduce350 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 59087 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 857 "reason_parser.mly"
       (string)
# 59092 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_longident = 
# 4168 "reason_parser.mly"
                                                ( Lident _1 )
# 59099 "reason_parser.ml"
     in
    _menhir_goto_class_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_subtractive : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_subtractive -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1477 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState816 | MenhirState813 | MenhirState798 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3491 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState813 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState813) : 'freshtv3492)
    | MenhirState2753 | MenhirState2760 | MenhirState2757 | MenhirState2743 | MenhirState2582 | MenhirState2574 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState929 | MenhirState922 | MenhirState907 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3493 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState922 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState922) : 'freshtv3494)
    | _ ->
        _menhir_fail ()

and _menhir_run1704 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3489 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1704 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3485 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 59373 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3481 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 59385 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run1710 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run1708 _menhir_env (Obj.magic _menhir_stack) MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1707 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1707) : 'freshtv3482)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3483 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 59450 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3484)) : 'freshtv3486)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3487 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3488)) : 'freshtv3490)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1704 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1704

and _menhir_run1708 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1708 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1708

and _menhir_run1208 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 59550 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run1208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1208 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3479 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 59564 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 59569 "reason_parser.ml"
        )), _startpos__1_) = _menhir_stack in
        let _v : 'tv_lident_list = 
# 2877 "reason_parser.mly"
                                      ( [_1] )
# 59574 "reason_parser.ml"
         in
        _menhir_goto_lident_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3480)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1208

and _menhir_error1699 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1694 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1066 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _menhir_stack = (_menhir_stack, _menhir_s) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1066 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1066

and _menhir_run1068 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1068 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1068

and _menhir_error1679 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1670 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run1576 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1576 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1576

and _menhir_error1574 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1565 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error1560 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_run107 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce140 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run243 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState243 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState243

and _menhir_run115 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState115

and _menhir_run245 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run246 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState245

and _menhir_run323 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 59783 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3477 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 59795 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState324 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState324) : 'freshtv3478)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce823 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run282 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce144 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run283 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce143 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run284 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | DOTDOT ->
        _menhir_run149 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState284
    | LIDENT _v ->
        _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState284 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState284

and _menhir_run291 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run157 _menhir_env (Obj.magic _menhir_stack) MenhirState291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState291
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState291

and _menhir_run301 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run157 _menhir_env (Obj.magic _menhir_stack) MenhirState301 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3475 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState301 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce149 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv3476)
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState301
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState301

and _menhir_run306 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3473 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState306 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState307 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState307) : 'freshtv3474)
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState306 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState306

and _menhir_error1555 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error945 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_error832 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_reduce243 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_label), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = 
# 2788 "reason_parser.mly"
      ( mkexp(Pexp_send(_1, _3)) )
# 60002 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run154 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState154 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState154 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState154

and _menhir_error152 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s

and _menhir_goto_val_ident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_val_ident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState2555 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState180 | MenhirState184 | MenhirState187 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState457 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState2268 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState475 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1939 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1846 | MenhirState1477 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1648 | MenhirState1647 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState842 | MenhirState840 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState827 | MenhirState820 | MenhirState825 | MenhirState822 | MenhirState817 | MenhirState816 | MenhirState813 | MenhirState798 | MenhirState811 | MenhirState805 | MenhirState747 | MenhirState470 | MenhirState464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3401 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce836 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3402)
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState610 | MenhirState611 | MenhirState603 | MenhirState594 | MenhirState592 | MenhirState588 | MenhirState574 | MenhirState533 | MenhirState564 | MenhirState563 | MenhirState550 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3403 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce269 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3404)
    | MenhirState577 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3405 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce164 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3406)
    | MenhirState2704 | MenhirState2663 | MenhirState2679 | MenhirState2677 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2238 | MenhirState476 | MenhirState480 | MenhirState2224 | MenhirState484 | MenhirState1912 | MenhirState1887 | MenhirState1891 | MenhirState1823 | MenhirState1627 | MenhirState1703 | MenhirState1714 | MenhirState1707 | MenhirState1657 | MenhirState1639 | MenhirState1637 | MenhirState1227 | MenhirState1184 | MenhirState1188 | MenhirState1199 | MenhirState1192 | MenhirState1041 | MenhirState898 | MenhirState902 | MenhirState1027 | MenhirState906 | MenhirState908 | MenhirState767 | MenhirState738 | MenhirState486 | MenhirState711 | MenhirState676 | MenhirState668 | MenhirState655 | MenhirState587 | MenhirState650 | MenhirState649 | MenhirState646 | MenhirState642 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3407 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce269 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3408)
    | MenhirState658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3409 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce164 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3410)
    | MenhirState2760 | MenhirState2757 | MenhirState2753 | MenhirState2743 | MenhirState2582 | MenhirState2688 | MenhirState2687 | MenhirState2620 | MenhirState2574 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState2350 | MenhirState417 | MenhirState418 | MenhirState419 | MenhirState2277 | MenhirState474 | MenhirState2229 | MenhirState729 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState1161 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState955 | MenhirState953 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState940 | MenhirState933 | MenhirState938 | MenhirState935 | MenhirState930 | MenhirState929 | MenhirState922 | MenhirState907 | MenhirState920 | MenhirState913 | MenhirState912 | MenhirState910 | MenhirState895 | MenhirState894 | MenhirState793 | MenhirState764 | MenhirState746 | MenhirState740 | MenhirState734 | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3411 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce836 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3412)
    | MenhirState2627 | MenhirState797 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3413 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce837 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3414)
    | MenhirState1634 | MenhirState1183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3433 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1184
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3417 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1184 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run154 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3415 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1206 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run1208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1207 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1207) : 'freshtv3416)
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1206 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1206) : 'freshtv3418)
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState1184
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3419) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1184 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1204 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1204) : 'freshtv3420)
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3431) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1184 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AMPERAMPER ->
                  _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERSAND ->
                  _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BARBAR ->
                  _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONEQUAL ->
                  _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | GREATER ->
                  _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP0 _v ->
                  _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP1 _v ->
                  _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP2 _v ->
                  _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP3 _v ->
                  _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP4 _v ->
                  _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSEQ ->
                  _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RPAREN ->
                  _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STAR ->
                  _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv3429 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1185 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_v : (
# 857 "reason_parser.mly"
       (string)
# 60342 "reason_parser.ml"
                      )) = _v in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | RPAREN ->
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3421 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 60354 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                          ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                          let _menhir_env = _menhir_discard _menhir_env in
                          let _tok = _menhir_env._menhir_token in
                          match _tok with
                          | BACKQUOTE ->
                              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | CHAR _v ->
                              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | COLON ->
                              _menhir_run1195 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | EQUALGREATER ->
                              _menhir_run1193 _menhir_env (Obj.magic _menhir_stack) MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FALSE ->
                              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FLOAT _v ->
                              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT _v ->
                              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT32 _v ->
                              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT64 _v ->
                              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACE ->
                              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKET ->
                              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETBAR ->
                              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETPERCENT ->
                              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LIDENT _v ->
                              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LPAREN ->
                              _menhir_run1189 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | MINUS ->
                              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | NATIVEINT _v ->
                              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | PLUS ->
                              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | SHARP ->
                              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | STRING _v ->
                              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | TRUE ->
                              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UIDENT _v ->
                              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UNDERSCORE ->
                              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | _ ->
                              assert (not _menhir_env._menhir_error);
                              _menhir_env._menhir_error <- true;
                              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1188) : 'freshtv3422)
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3423 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 60419 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3424)) : 'freshtv3426)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3428)) : 'freshtv3430)
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1185 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1185) : 'freshtv3432)
          | MINUS ->
              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce269 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1184)) : 'freshtv3434)
    | MenhirState1409 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3437 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1470 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3435 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1471 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1471) : 'freshtv3436)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1470 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3438)
    | MenhirState1478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3441 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1479 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1480) : 'freshtv3440)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1479 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3442)
    | MenhirState2314 | MenhirState1886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3461 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1887
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3445 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1887 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run154 _menhir_env (Obj.magic _menhir_stack) MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ('freshtv3443 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1895 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run1208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1896 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1896) : 'freshtv3444)
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1895 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1895) : 'freshtv3446)
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState1887
          | EQUALGREATER ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3447) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = MenhirState1887 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1893 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1893) : 'freshtv3448)
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3459) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1887 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | AMPERAMPER ->
                  _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | AMPERSAND ->
                  _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BARBAR ->
                  _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | COLONEQUAL ->
                  _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | EXCEPTION ->
                  _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | GREATER ->
                  _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP0 _v ->
                  _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP1 _v ->
                  _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP2 _v ->
                  _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP3 _v ->
                  _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INFIXOP4 _v ->
                  _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MODULE ->
                  _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | OR ->
                  _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PERCENT ->
                  _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSEQ ->
                  _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | RPAREN ->
                  _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STAR ->
                  _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : 'freshtv3457 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1888 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3453 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                      let (_v : (
# 857 "reason_parser.mly"
       (string)
# 60846 "reason_parser.ml"
                      )) = _v in
                      let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                      ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                      let _menhir_env = _menhir_discard _menhir_env in
                      let _tok = _menhir_env._menhir_token in
                      match _tok with
                      | RPAREN ->
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3449 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 60858 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                          let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                          ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                          let _menhir_env = _menhir_discard _menhir_env in
                          let _tok = _menhir_env._menhir_token in
                          match _tok with
                          | BACKQUOTE ->
                              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | CHAR _v ->
                              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | COLON ->
                              _menhir_run1710 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | EQUALGREATER ->
                              _menhir_run1708 _menhir_env (Obj.magic _menhir_stack) MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FALSE ->
                              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | FLOAT _v ->
                              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT _v ->
                              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT32 _v ->
                              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | INT64 _v ->
                              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACE ->
                              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKET ->
                              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETBAR ->
                              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LBRACKETPERCENT ->
                              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LIDENT _v ->
                              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | LPAREN ->
                              _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | MINUS ->
                              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | NATIVEINT _v ->
                              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | PLUS ->
                              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | SHARP ->
                              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | STRING _v ->
                              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | TRUE ->
                              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UIDENT _v ->
                              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | UNDERSCORE ->
                              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1891 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                          | _ ->
                              assert (not _menhir_env._menhir_error);
                              _menhir_env._menhir_error <- true;
                              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1891) : 'freshtv3450)
                      | _ ->
                          assert (not _menhir_env._menhir_error);
                          _menhir_env._menhir_error <- true;
                          let (_menhir_env : _menhir_env) = _menhir_env in
                          let (_menhir_stack : (('freshtv3451 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 60923 "reason_parser.ml"
                          ) * Lexing.position) = Obj.magic _menhir_stack in
                          ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3452)) : 'freshtv3454)
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      let (_menhir_env : _menhir_env) = _menhir_env in
                      let (_menhir_stack : ('freshtv3455 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                      ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3456)) : 'freshtv3458)
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1888) : 'freshtv3460)
          | MINUS ->
              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1887 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AS | BAR | COLONCOLON | EQUAL | LBRACKETAT ->
              _menhir_reduce269 _menhir_env (Obj.magic _menhir_stack)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1887)) : 'freshtv3462)
    | MenhirState2039 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3465 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2040 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3463 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2041 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2041) : 'freshtv3464)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2040 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3466)
    | MenhirState2248 | MenhirState754 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3467 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce837 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3468)
    | MenhirState2647 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3471 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2648 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3469 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2649) : 'freshtv3470)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2648 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3472)
    | _ ->
        _menhir_fail ()

and _menhir_goto_rec_flag : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_rec_flag -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2635 | MenhirState1156 | MenhirState1633 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3397 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1634 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1634) : 'freshtv3398)
    | MenhirState434 | MenhirState441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3399 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2314 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2314) : 'freshtv3400)
    | _ ->
        _menhir_fail ()

and _menhir_reduce779 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4297 "reason_parser.mly"
         ( "with" )
# 61200 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce778 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4296 "reason_parser.mly"
          ( "while" )
# 61212 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce777 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4295 "reason_parser.mly"
         ( "when" )
# 61224 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce776 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4294 "reason_parser.mly"
            ( "virtual" )
# 61236 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce775 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4293 "reason_parser.mly"
        ( "val" )
# 61248 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce731 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 61255 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 906 "reason_parser.mly"
       (string)
# 61260 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4249 "reason_parser.mly"
           ( _1 )
# 61267 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce774 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4292 "reason_parser.mly"
         ( "type" )
# 61279 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce773 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4291 "reason_parser.mly"
        ( "try" )
# 61291 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce772 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4290 "reason_parser.mly"
         ( "true" )
# 61303 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce771 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4289 "reason_parser.mly"
       ( "to" )
# 61315 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce770 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4288 "reason_parser.mly"
         ( "then" )
# 61327 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce757 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4275 "reason_parser.mly"
           ( "match" )
# 61339 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce769 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4287 "reason_parser.mly"
           ( "struct" )
# 61351 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce768 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4286 "reason_parser.mly"
        ( "sig" )
# 61363 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce767 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4285 "reason_parser.mly"
        ( "rec" )
# 61375 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce766 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4284 "reason_parser.mly"
            ( "private" )
# 61387 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce765 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4283 "reason_parser.mly"
       ( "or" )
# 61399 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce764 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4282 "reason_parser.mly"
         ( "open" )
# 61411 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce763 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4281 "reason_parser.mly"
           ( "object" )
# 61423 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce762 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4280 "reason_parser.mly"
           ( "nonrec" )
# 61435 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce761 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4279 "reason_parser.mly"
        ( "new" )
# 61447 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce760 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4278 "reason_parser.mly"
            ( "mutable" )
# 61459 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce759 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4277 "reason_parser.mly"
           ( "module" )
# 61471 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce758 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4276 "reason_parser.mly"
           ( "method" )
# 61483 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce730 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 61490 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 857 "reason_parser.mly"
       (string)
# 61495 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4248 "reason_parser.mly"
           ( _1 )
# 61502 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce756 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4274 "reason_parser.mly"
        ( "let" )
# 61514 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce755 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4273 "reason_parser.mly"
         ( "lazy" )
# 61526 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce754 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4272 "reason_parser.mly"
                ( "initializer" )
# 61538 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce753 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4271 "reason_parser.mly"
            ( "inherit" )
# 61550 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce752 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4270 "reason_parser.mly"
            ( "include" )
# 61562 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce751 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4269 "reason_parser.mly"
       ( "in" )
# 61574 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce750 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4268 "reason_parser.mly"
       ( "if" )
# 61586 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce749 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4267 "reason_parser.mly"
            ( "functor" )
# 61598 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce748 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4266 "reason_parser.mly"
             ( "function" )
# 61610 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce747 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4265 "reason_parser.mly"
        ( "fun" )
# 61622 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce746 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4264 "reason_parser.mly"
        ( "for" )
# 61634 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce745 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4263 "reason_parser.mly"
          ( "false" )
# 61646 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce744 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4262 "reason_parser.mly"
             ( "external" )
# 61658 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce743 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4261 "reason_parser.mly"
              ( "exception" )
# 61670 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce742 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4260 "reason_parser.mly"
        ( "end" )
# 61682 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce741 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4259 "reason_parser.mly"
         ( "else" )
# 61694 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce740 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4258 "reason_parser.mly"
           ( "downto" )
# 61706 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce739 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4257 "reason_parser.mly"
         ( "done" )
# 61718 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce738 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4256 "reason_parser.mly"
       ( "do" )
# 61730 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce737 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4255 "reason_parser.mly"
               ( "constraint" )
# 61742 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce736 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4254 "reason_parser.mly"
          ( "class" )
# 61754 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce735 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4253 "reason_parser.mly"
          ( "begin" )
# 61766 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce734 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4252 "reason_parser.mly"
           ( "assert" )
# 61778 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce733 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4251 "reason_parser.mly"
       ( "as" )
# 61790 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce732 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_single_attr_id = 
# 4250 "reason_parser.mly"
        ( "and" )
# 61802 "reason_parser.ml"
     in
    _menhir_goto_single_attr_id _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce375 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constr_longident = 
# 4118 "reason_parser.mly"
                                                ( Lident "[]" )
# 61816 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce434 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 61823 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 61829 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _v : 'tv_field_expr = let _endpos = _endpos__1_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3228 "reason_parser.mly"
    (
      let loc = mklocation _symbolstartpos _endpos in
      let lident_loc = mkloc _1 loc in
      let lident_lident_loc = mkloc (Lident _1) loc in
      (lident_loc, mkexp (Pexp_ident lident_lident_loc))
    )
# 61841 "reason_parser.ml"
     in
    _menhir_goto_field_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto__simple_expr : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__simple_expr -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState2555 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState180 | MenhirState184 | MenhirState187 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState457 | MenhirState464 | MenhirState470 | MenhirState2274 | MenhirState473 | MenhirState2268 | MenhirState2270 | MenhirState475 | MenhirState2253 | MenhirState2255 | MenhirState2259 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState747 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1647 | MenhirState1648 | MenhirState771 | MenhirState1147 | MenhirState1142 | MenhirState1134 | MenhirState1115 | MenhirState1123 | MenhirState1081 | MenhirState1084 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1047 | MenhirState1051 | MenhirState816 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState833 | MenhirState886 | MenhirState860 | MenhirState880 | MenhirState882 | MenhirState884 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState866 | MenhirState870 | MenhirState868 | MenhirState864 | MenhirState862 | MenhirState858 | MenhirState856 | MenhirState850 | MenhirState854 | MenhirState852 | MenhirState835 | MenhirState837 | MenhirState847 | MenhirState840 | MenhirState842 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 | MenhirState817 | MenhirState811 | MenhirState805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3393) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce536 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3394)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 | MenhirState2687 | MenhirState2688 | MenhirState2574 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState2422 | MenhirState2374 | MenhirState2376 | MenhirState416 | MenhirState417 | MenhirState2350 | MenhirState418 | MenhirState419 | MenhirState2277 | MenhirState474 | MenhirState2229 | MenhirState729 | MenhirState730 | MenhirState734 | MenhirState2213 | MenhirState740 | MenhirState746 | MenhirState2187 | MenhirState2192 | MenhirState2188 | MenhirState763 | MenhirState764 | MenhirState2147 | MenhirState770 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1167 | MenhirState1171 | MenhirState1176 | MenhirState791 | MenhirState1138 | MenhirState793 | MenhirState1098 | MenhirState1103 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState894 | MenhirState895 | MenhirState897 | MenhirState1036 | MenhirState1032 | MenhirState1034 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1008 | MenhirState1013 | MenhirState1015 | MenhirState1011 | MenhirState1006 | MenhirState929 | MenhirState946 | MenhirState999 | MenhirState973 | MenhirState993 | MenhirState995 | MenhirState997 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState979 | MenhirState983 | MenhirState981 | MenhirState977 | MenhirState975 | MenhirState971 | MenhirState969 | MenhirState963 | MenhirState967 | MenhirState965 | MenhirState948 | MenhirState950 | MenhirState960 | MenhirState953 | MenhirState955 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 | MenhirState930 | MenhirState920 | MenhirState910 | MenhirState912 | MenhirState913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3395) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_expr) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce536 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3396)
    | _ ->
        _menhir_fail ()

and _menhir_goto__class_self_pattern : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv__class_self_pattern -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3391) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__class_self_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3389) = Obj.magic _menhir_stack in
    let (_endpos_x_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((x : 'tv__class_self_pattern) : 'tv__class_self_pattern) = _v in
    let (_startpos_x_ : Lexing.position) = _startpos in
    ((let _startpos = _startpos_x_ in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_pat__class_self_pattern_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos_x_ _endpos_x_ then
      _startpos_x_
    else
      _endpos in
    
# 1203 "reason_parser.mly"
 ( {x with ppat_loc = {x.ppat_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 61891 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3387) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__class_self_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3385) = Obj.magic _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_pat__class_self_pattern_) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3383) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_mark_position_pat__class_self_pattern_) : 'tv_mark_position_pat__class_self_pattern_) = _v in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_class_self_pattern = 
# 1899 "reason_parser.mly"
                                                           (_1)
# 61916 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3381) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_self_pattern) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3379 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
    ((if _menhir_env._menhir_error then
      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1666
    else
      let _tok = _menhir_env._menhir_token in
      match _tok with
      | CONSTRAINT ->
          _menhir_run1744 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | INHERIT ->
          _menhir_run1737 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | INITIALIZER ->
          _menhir_run1733 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETATATAT ->
          _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | LBRACKETPERCENTPERCENT ->
          _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | METHOD ->
          _menhir_run1690 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | VAL ->
          _menhir_run1667 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1666 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
      | RBRACE ->
          _menhir_reduce695 _menhir_env (Obj.magic _menhir_stack) MenhirState1666
      | _ ->
          (assert (not _menhir_env._menhir_error);
          _menhir_env._menhir_error <- true;
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1666)) : 'freshtv3380)) : 'freshtv3382)) : 'freshtv3384)) : 'freshtv3386)) : 'freshtv3388)) : 'freshtv3390)) : 'freshtv3392)

and _menhir_reduce466 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 61956 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 61962 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_label_longident = 
# 4124 "reason_parser.mly"
                                                ( Lident _1 )
# 61969 "reason_parser.ml"
     in
    _menhir_goto_label_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1245 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1245 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1245

and _menhir_run1247 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1247 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1247 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1247

and _menhir_run2164 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce260 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run444 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3375 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState444 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3373 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos__2_ : Lexing.position) = _endpos in
        let (_ : _menhir_state) = _menhir_s in
        let (_startpos__2_ : Lexing.position) = _startpos in
        ((let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
        let _2 = () in
        let _1 = () in
        let _startpos = _startpos__1_ in
        let _v : 'tv_functor_arg = 
# 1218 "reason_parser.mly"
 ( mkloc "*" (mklocation _startpos__2_ _endpos__2_), None )
# 62047 "reason_parser.ml"
         in
        _menhir_goto_functor_arg _menhir_env _menhir_stack _menhir_s _v _startpos) : 'freshtv3374)) : 'freshtv3376)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3377) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState444 in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 62058 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce444 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3378)
    | UNDERSCORE ->
        _menhir_run445 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState444 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState444

and _menhir_run213 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 62074 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce823 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce376 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ _startpos__2_ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constr_longident = 
# 4119 "reason_parser.mly"
                                                ( Lident "()" )
# 62091 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce607 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 62098 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 882 "reason_parser.mly"
       (string)
# 62104 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _v : 'tv_operator = 
# 4097 "reason_parser.mly"
                                                ( _1 )
# 62109 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce615 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4080 "reason_parser.mly"
                                                ( "+." )
# 62122 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 62128 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce614 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4079 "reason_parser.mly"
                                                ( "+" )
# 62141 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 62147 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_run223 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 871 "reason_parser.mly"
       (nativeint)
# 62154 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce717 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run224 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 841 "reason_parser.mly"
       (int64)
# 62163 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce716 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run225 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 840 "reason_parser.mly"
       (int32)
# 62172 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce715 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run226 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 839 "reason_parser.mly"
       (int)
# 62181 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce713 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run227 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 821 "reason_parser.mly"
       (string)
# 62190 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce714 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_reduce290 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 62199 "reason_parser.ml"
) * Lexing.position -> Lexing.position -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__4_ _startpos__4_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos_x0_, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 62205 "reason_parser.ml"
    )), _startpos_x0_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 62221 "reason_parser.ml"
      
    in
    
# 3383 "reason_parser.mly"
      ( mkpat(Ppat_unpack _3) )
# 62227 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run234 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 62234 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState234 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
    | EOF | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | RPAREN | SEMI | WITH ->
        _menhir_reduce453 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState234

and _menhir_reduce617 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4082 "reason_parser.mly"
                                                ( "-." )
# 62261 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 62267 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce616 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__10_, _menhir_s, _startpos__10_) = _menhir_stack in
    let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4081 "reason_parser.mly"
                                                ( "-" )
# 62280 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 62286 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_run490 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 871 "reason_parser.mly"
       (nativeint)
# 62293 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce712 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run491 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 841 "reason_parser.mly"
       (int64)
# 62302 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce711 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run492 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 840 "reason_parser.mly"
       (int32)
# 62311 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce710 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run493 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 839 "reason_parser.mly"
       (int)
# 62320 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce708 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run494 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 821 "reason_parser.mly"
       (string)
# 62329 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce709 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run496 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3369 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 62348 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv3363 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 62360 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState499 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run234 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState499 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState499) : 'freshtv3364)
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv3365 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 62381 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce290 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv3366)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv3367 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 62394 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3368)) : 'freshtv3370)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3371 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3372)

and _menhir_run533 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run496 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState533 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState533

and _menhir_reduce608 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_operator = 
# 4098 "reason_parser.mly"
                                                ( "!" )
# 62516 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v

and _menhir_goto_operator : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_operator -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState184 | MenhirState408 | MenhirState755 | MenhirState495 | MenhirState533 | MenhirState579 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3355 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3351 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce835 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv3352)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3353 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3354)) : 'freshtv3356)
    | MenhirState15 | MenhirState210 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState771 | MenhirState798 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3361 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3357 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce835 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv3358)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3359 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_operator) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3360)) : 'freshtv3362)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constr_longident : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constr_longident -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState532 | MenhirState563 | MenhirState564 | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3315 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce274 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3316)
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3317 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState563
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce274 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3318)
    | MenhirState2663 | MenhirState2704 | MenhirState2677 | MenhirState2679 | MenhirState476 | MenhirState2238 | MenhirState480 | MenhirState484 | MenhirState2224 | MenhirState486 | MenhirState1887 | MenhirState1912 | MenhirState1891 | MenhirState1627 | MenhirState1823 | MenhirState1703 | MenhirState1707 | MenhirState1714 | MenhirState1637 | MenhirState1639 | MenhirState1184 | MenhirState1227 | MenhirState1188 | MenhirState1192 | MenhirState1199 | MenhirState898 | MenhirState1041 | MenhirState902 | MenhirState906 | MenhirState1027 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3319 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce274 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3320)
    | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState738 | MenhirState1886 | MenhirState1657 | MenhirState1634 | MenhirState1183 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3321 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState649
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState649 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACELESS | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce274 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3322)
    | MenhirState2555 | MenhirState180 | MenhirState187 | MenhirState457 | MenhirState464 | MenhirState470 | MenhirState2268 | MenhirState747 | MenhirState1647 | MenhirState1648 | MenhirState840 | MenhirState842 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 | MenhirState817 | MenhirState811 | MenhirState805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3323 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce203 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3324)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState475 | MenhirState2253 | MenhirState2255 | MenhirState2259 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState771 | MenhirState1147 | MenhirState1142 | MenhirState1134 | MenhirState1115 | MenhirState1123 | MenhirState1081 | MenhirState1084 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1047 | MenhirState1051 | MenhirState816 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState833 | MenhirState886 | MenhirState860 | MenhirState880 | MenhirState882 | MenhirState884 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState866 | MenhirState870 | MenhirState868 | MenhirState864 | MenhirState862 | MenhirState858 | MenhirState856 | MenhirState850 | MenhirState854 | MenhirState852 | MenhirState835 | MenhirState847 | MenhirState837 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3325 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState840
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState840 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce203 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3326)
    | MenhirState2687 | MenhirState2688 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState417 | MenhirState2350 | MenhirState419 | MenhirState474 | MenhirState729 | MenhirState730 | MenhirState734 | MenhirState740 | MenhirState746 | MenhirState764 | MenhirState793 | MenhirState894 | MenhirState895 | MenhirState953 | MenhirState955 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 | MenhirState930 | MenhirState920 | MenhirState910 | MenhirState912 | MenhirState913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3327 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce203 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3328)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 | MenhirState2574 | MenhirState2422 | MenhirState2374 | MenhirState2376 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2187 | MenhirState2192 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState770 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1167 | MenhirState1171 | MenhirState1176 | MenhirState791 | MenhirState1138 | MenhirState1098 | MenhirState1103 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1032 | MenhirState1034 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1008 | MenhirState1013 | MenhirState1015 | MenhirState1011 | MenhirState1006 | MenhirState929 | MenhirState946 | MenhirState999 | MenhirState973 | MenhirState993 | MenhirState995 | MenhirState997 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState979 | MenhirState983 | MenhirState981 | MenhirState977 | MenhirState975 | MenhirState971 | MenhirState969 | MenhirState963 | MenhirState967 | MenhirState965 | MenhirState948 | MenhirState960 | MenhirState950 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3329 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState953
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState953 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce203 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3330)
    | MenhirState1491 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3331 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 62800 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1494
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1494
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1494) : 'freshtv3332)
    | MenhirState1506 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3333 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1507
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1507
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1507) : 'freshtv3334)
    | MenhirState1513 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3335 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1514
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1514
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1514) : 'freshtv3336)
    | MenhirState1519 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1520
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1520
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1520) : 'freshtv3338)
    | MenhirState1525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3339 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1526
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run333 _menhir_env (Obj.magic _menhir_stack) MenhirState1526
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1526) : 'freshtv3340)
    | MenhirState2590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3341 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 62864 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2593
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2593
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2593) : 'freshtv3342)
    | MenhirState2596 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3343 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2597
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2597
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2597) : 'freshtv3344)
    | MenhirState2601 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3345 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2602
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2602
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2602) : 'freshtv3346)
    | MenhirState2605 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3347 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2606
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2606
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2606) : 'freshtv3348)
    | MenhirState2609 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3349 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2610
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETAT ->
              _menhir_run174 _menhir_env (Obj.magic _menhir_stack) MenhirState2610
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2610) : 'freshtv3350)
    | _ ->
        _menhir_fail ()

and _menhir_run580 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce614 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run581 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce616 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run1270 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce449 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run2591 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2591 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2591

and _menhir_run2592 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RBRACKET ->
        _menhir_run604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2592
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2592

and _menhir_run1275 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1275

and _menhir_fail : unit -> 'a =
  fun () ->
    Printf.fprintf Pervasives.stderr "Internal failure -- please contact the parser generator's developers.\n%!";
    assert false

and _menhir_goto_virtual_flag : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_virtual_flag -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1876 | MenhirState1536 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3287 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3283 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 63028 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce363 _menhir_env (Obj.magic _menhir_stack) MenhirState1538) : 'freshtv3284)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3285 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3286)) : 'freshtv3288)
    | MenhirState1558 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3289 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1559 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1559) : 'freshtv3290)
    | MenhirState1881 | MenhirState1534 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3295 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3291 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 63066 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce363 _menhir_env (Obj.magic _menhir_stack) MenhirState1626) : 'freshtv3292)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3293 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3294)) : 'freshtv3296)
    | MenhirState2087 | MenhirState2052 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3301 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3297 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 63092 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce363 _menhir_env (Obj.magic _menhir_stack) MenhirState2054) : 'freshtv3298)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3299 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3300)) : 'freshtv3302)
    | MenhirState2092 | MenhirState2051 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3307 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3303 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 63118 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce363 _menhir_env (Obj.magic _menhir_stack) MenhirState2061) : 'freshtv3304)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3305 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3306)) : 'freshtv3308)
    | MenhirState2738 | MenhirState2660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3313 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3309 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 63144 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce363 _menhir_env (Obj.magic _menhir_stack) MenhirState2662) : 'freshtv3310)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3311 * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3312)) : 'freshtv3314)
    | _ ->
        _menhir_fail ()

and _menhir_goto_constant : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'tv_constant -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    match _menhir_s with
    | MenhirState210 | MenhirState2417 | MenhirState477 | MenhirState481 | MenhirState1888 | MenhirState1704 | MenhirState1185 | MenhirState1189 | MenhirState899 | MenhirState903 | MenhirState487 | MenhirState495 | MenhirState713 | MenhirState511 | MenhirState514 | MenhirState531 | MenhirState532 | MenhirState533 | MenhirState603 | MenhirState610 | MenhirState611 | MenhirState588 | MenhirState592 | MenhirState594 | MenhirState574 | MenhirState540 | MenhirState563 | MenhirState564 | MenhirState550 | MenhirState545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3275) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce707 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3276)
    | MenhirState2663 | MenhirState2704 | MenhirState2677 | MenhirState2679 | MenhirState2578 | MenhirState191 | MenhirState2415 | MenhirState2314 | MenhirState476 | MenhirState2238 | MenhirState480 | MenhirState484 | MenhirState2224 | MenhirState486 | MenhirState738 | MenhirState1886 | MenhirState1887 | MenhirState1912 | MenhirState1891 | MenhirState1627 | MenhirState1823 | MenhirState1703 | MenhirState1707 | MenhirState1714 | MenhirState1657 | MenhirState1637 | MenhirState1639 | MenhirState1634 | MenhirState1183 | MenhirState1184 | MenhirState1227 | MenhirState1188 | MenhirState1192 | MenhirState1199 | MenhirState898 | MenhirState1041 | MenhirState902 | MenhirState906 | MenhirState1027 | MenhirState908 | MenhirState767 | MenhirState711 | MenhirState676 | MenhirState587 | MenhirState668 | MenhirState655 | MenhirState642 | MenhirState649 | MenhirState650 | MenhirState646 | MenhirState619 | MenhirState630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3277) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce707 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3278)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState2555 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState180 | MenhirState184 | MenhirState187 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState457 | MenhirState464 | MenhirState470 | MenhirState2274 | MenhirState473 | MenhirState2268 | MenhirState2270 | MenhirState475 | MenhirState2253 | MenhirState2255 | MenhirState2259 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState747 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1846 | MenhirState1929 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1647 | MenhirState1648 | MenhirState771 | MenhirState1147 | MenhirState1142 | MenhirState1134 | MenhirState1115 | MenhirState1123 | MenhirState1081 | MenhirState1084 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1047 | MenhirState1051 | MenhirState816 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState833 | MenhirState886 | MenhirState860 | MenhirState880 | MenhirState882 | MenhirState884 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState866 | MenhirState870 | MenhirState868 | MenhirState864 | MenhirState862 | MenhirState858 | MenhirState856 | MenhirState850 | MenhirState854 | MenhirState852 | MenhirState835 | MenhirState847 | MenhirState837 | MenhirState840 | MenhirState842 | MenhirState820 | MenhirState827 | MenhirState825 | MenhirState822 | MenhirState817 | MenhirState811 | MenhirState805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3279) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce202 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3280)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2582 | MenhirState2743 | MenhirState2687 | MenhirState2688 | MenhirState2574 | MenhirState2 | MenhirState6 | MenhirState17 | MenhirState177 | MenhirState2422 | MenhirState2374 | MenhirState2376 | MenhirState416 | MenhirState417 | MenhirState2350 | MenhirState418 | MenhirState419 | MenhirState2277 | MenhirState474 | MenhirState2229 | MenhirState729 | MenhirState730 | MenhirState734 | MenhirState2213 | MenhirState740 | MenhirState746 | MenhirState2187 | MenhirState2192 | MenhirState2188 | MenhirState763 | MenhirState764 | MenhirState2147 | MenhirState770 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1167 | MenhirState1171 | MenhirState1176 | MenhirState791 | MenhirState1138 | MenhirState793 | MenhirState1098 | MenhirState1103 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState894 | MenhirState895 | MenhirState897 | MenhirState1036 | MenhirState1032 | MenhirState1034 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1008 | MenhirState1013 | MenhirState1015 | MenhirState1011 | MenhirState1006 | MenhirState929 | MenhirState946 | MenhirState999 | MenhirState973 | MenhirState993 | MenhirState995 | MenhirState997 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState979 | MenhirState983 | MenhirState981 | MenhirState977 | MenhirState975 | MenhirState971 | MenhirState969 | MenhirState963 | MenhirState967 | MenhirState965 | MenhirState948 | MenhirState960 | MenhirState950 | MenhirState953 | MenhirState955 | MenhirState933 | MenhirState940 | MenhirState938 | MenhirState935 | MenhirState930 | MenhirState920 | MenhirState910 | MenhirState912 | MenhirState913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3281) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_constant) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce202 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3282)
    | _ ->
        _menhir_fail ()

and _menhir_reduce454 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 63201 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 857 "reason_parser.mly"
       (string)
# 63206 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_ident = 
# 4065 "reason_parser.mly"
                                                ( _1 )
# 63213 "reason_parser.ml"
     in
    _menhir_goto_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_goto_structure : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_structure -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3231 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3229 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_structure_item), _startpos_x0_), _endpos__2_, _startpos__2_), _, (_3 : 'tv_structure)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_structure = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 63237 "reason_parser.ml"
          
        in
        
# 1377 "reason_parser.mly"
                                           (
    syntax_error_str _1.loc "Statement has to end with a semicolon" :: _3
  )
# 63245 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3230)) : 'freshtv3232)
    | MenhirState1939 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3235 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3233 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_), _endpos__2_), _, (_3 : 'tv_structure)) = _menhir_stack in
        let _2 = () in
        let _v : 'tv_structure = 
# 1380 "reason_parser.mly"
                                  (
      let effective_loc = mklocation _startpos__1_ _endpos__2_ in
      set_structure_item_location _1 effective_loc :: _3
  )
# 63261 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3234)) : 'freshtv3236)
    | MenhirState1477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3241 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3237 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce252 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3238)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3239 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3240)) : 'freshtv3242)
    | MenhirState73 | MenhirState175 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState440 | MenhirState454 | MenhirState456 | MenhirState2033 | MenhirState1475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3245 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3243 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, (_1 : 'tv_structure)) = _menhir_stack in
        let _v : 'tv_payload = 
# 4353 "reason_parser.mly"
              ( PStr _1 )
# 63292 "reason_parser.ml"
         in
        _menhir_goto_payload _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3244)) : 'freshtv3246)
    | MenhirState452 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3251 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RBRACE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3247 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce252 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3248)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3249 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3250)) : 'freshtv3252)
    | MenhirState2557 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3255 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3253 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _), _, (_3 : 'tv_structure)) = _menhir_stack in
        let _2 = () in
        let x0 = () in
        let _v : 'tv_structure = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 63331 "reason_parser.ml"
          
        in
        
# 1375 "reason_parser.mly"
                                 ( syntax_error_str _1.loc "Invalid statement" :: _3 )
# 63337 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3254)) : 'freshtv3256)
    | MenhirState1 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3259 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3257 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_structure)) = _menhir_stack in
        let x0 = () in
        let _v : 'tv_structure = let _1 =
          let _endpos_x_ = _endpos_x0_ in
          let _startpos_x_ = _startpos_x0_ in
          let x = x0 in
          let _endpos = _endpos_x_ in
          let _symbolstartpos = _startpos_x_ in
          
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 63356 "reason_parser.ml"
          
        in
        
# 1374 "reason_parser.mly"
                            ( syntax_error_str _1.loc "Invalid statement" :: _2 )
# 63362 "reason_parser.ml"
         in
        _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3258)) : 'freshtv3260)
    | MenhirState0 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3273 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
        ((assert (not _menhir_env._menhir_error);
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | EOF ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3269 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3267 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, (_1 : 'tv_structure)) = _menhir_stack in
            let _2 = () in
            let _v : (
# 1059 "reason_parser.mly"
      (Parsetree.structure)
# 63381 "reason_parser.ml"
            ) = 
# 1116 "reason_parser.mly"
                                         ( apply_mapper_chain_to_structure _1 default_mapper_chain )
# 63385 "reason_parser.ml"
             in
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3265) = _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : (
# 1059 "reason_parser.mly"
      (Parsetree.structure)
# 63393 "reason_parser.ml"
            )) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3263) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let (_v : (
# 1059 "reason_parser.mly"
      (Parsetree.structure)
# 63401 "reason_parser.ml"
            )) = _v in
            ((let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3261) = Obj.magic _menhir_stack in
            let (_menhir_s : _menhir_state) = _menhir_s in
            let ((_1 : (
# 1059 "reason_parser.mly"
      (Parsetree.structure)
# 63409 "reason_parser.ml"
            )) : (
# 1059 "reason_parser.mly"
      (Parsetree.structure)
# 63413 "reason_parser.ml"
            )) = _v in
            (Obj.magic _1 : 'freshtv3262)) : 'freshtv3264)) : 'freshtv3266)) : 'freshtv3268)) : 'freshtv3270)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : 'freshtv3271 * _menhir_state * 'tv_structure) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3272)) : 'freshtv3274)
    | _ ->
        _menhir_fail ()

and _menhir_run110 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 63429 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce350 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run186 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce376 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run411 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState411 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState411

and _menhir_run443 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 63464 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3225) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState443 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2293 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2293 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2293 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2293) : 'freshtv3226)
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3227) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState443 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2291 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2291 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2291) : 'freshtv3228)
    | LPAREN ->
        _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState443 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState443

and _menhir_run515 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce375 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s

and _menhir_run2188 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2188 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2188

and _menhir_run404 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run1961 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState404 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3223 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState404 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState405 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState405) : 'freshtv3224)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState404

and _menhir_run1477 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState1477
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1477

and _menhir_run1943 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1943 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1943

and _menhir_run481 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3221 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState481 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3217 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 63938 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3213 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 63950 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run2220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run2218 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState484 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState484) : 'freshtv3214)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv3215 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 64015 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3216)) : 'freshtv3218)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3219 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3220)) : 'freshtv3222)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState481 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState481

and _menhir_run2218 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2218 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2218

and _menhir_run2220 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2220 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2220

and _menhir_run1490 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 64152 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3211 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 64166 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1490 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1491 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1491 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1491) : 'freshtv3212)
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1490 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1490
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1490

and _menhir_run1505 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3209 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1505 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1506 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1506) : 'freshtv3210)
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1505 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1505
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1505

and _menhir_run1511 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3205 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3203 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState1512 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FALSE ->
                _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run1493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1513 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1513) : 'freshtv3204)
        | LBRACKET ->
            _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1512 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1512
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1512) : 'freshtv3206)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3207 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3208)

and _menhir_run1518 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3201 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1518 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1519 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1519) : 'freshtv3202)
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1518 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1518
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1518

and _menhir_run1524 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1498 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3199 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1524 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1525 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run1493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1525 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1525 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1525 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1525 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1525) : 'freshtv3200)
    | LBRACKET ->
        _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run117 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run109 _menhir_env (Obj.magic _menhir_stack) MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1524 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | RBRACE | RBRACKET | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1524
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1524

and _menhir_run93 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 64511 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce453 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run94 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 64521 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce454 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run184 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run805 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState184 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState184

and _menhir_goto_signature : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv_signature -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3175 * _menhir_state * 'tv_signature_item) * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        (_menhir_reduce705 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3176)
    | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3179 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2100 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3177 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce265 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3178)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2100 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3180)
    | MenhirState1251 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3183 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2106 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | RBRACE ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ('freshtv3181 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              ((let _menhir_env = _menhir_discard _menhir_env in
              _menhir_reduce265 _menhir_env (Obj.magic _menhir_stack) _endpos) : 'freshtv3182)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2106 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3184)
    | MenhirState2565 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3185 * _menhir_state * 'tv_signature_item) * Lexing.position) * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        (_menhir_reduce705 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3186)
    | MenhirState2563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3197 * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error2567 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EOF ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3195 * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3193 * _menhir_state * 'tv_signature) = Obj.magic _menhir_stack in
              ((let (_menhir_stack, _menhir_s, (_1 : 'tv_signature)) = _menhir_stack in
              let _2 = () in
              let _v : (
# 1061 "reason_parser.mly"
      (Parsetree.signature)
# 64708 "reason_parser.ml"
              ) = 
# 1119 "reason_parser.mly"
                                         ( apply_mapper_chain_to_signature _1 default_mapper_chain )
# 64712 "reason_parser.ml"
               in
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3191) = _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1061 "reason_parser.mly"
      (Parsetree.signature)
# 64720 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3189) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let (_v : (
# 1061 "reason_parser.mly"
      (Parsetree.signature)
# 64728 "reason_parser.ml"
              )) = _v in
              ((let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3187) = Obj.magic _menhir_stack in
              let (_menhir_s : _menhir_state) = _menhir_s in
              let ((_1 : (
# 1061 "reason_parser.mly"
      (Parsetree.signature)
# 64736 "reason_parser.ml"
              )) : (
# 1061 "reason_parser.mly"
      (Parsetree.signature)
# 64740 "reason_parser.ml"
              )) = _v in
              (Obj.magic _1 : 'freshtv3188)) : 'freshtv3190)) : 'freshtv3192)) : 'freshtv3194)) : 'freshtv3196)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error2567 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3198)
    | _ ->
        _menhir_fail ()

and _menhir_run1412 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3165 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1442 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1457) : 'freshtv3166)
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3171 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 64780 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3167 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 64792 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LBRACE ->
                _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKETPERCENT ->
                _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LIDENT _v ->
                _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | MODULE ->
                _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1414 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1414 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1414) : 'freshtv3168)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv3169 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 64823 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3170)) : 'freshtv3172)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3173 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3174)

and _menhir_run1442 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1442 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1442 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1442 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1442

and _menhir_run1465 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 64861 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3161 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 64873 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACE ->
            _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1466 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1466 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1466 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1466) : 'freshtv3162)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3163 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 64904 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3164)

and _menhir_run130 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 64912 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState130 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOT ->
        _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
    | AND | COLONGREATER | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
        _menhir_reduce453 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState130

and _menhir_run450 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3157 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FUN ->
            _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState451 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState451) : 'freshtv3158)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3159 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3160)

and _menhir_run1248 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1251 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1250 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run402 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1249 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run445 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1248 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1248

and _menhir_run1408 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run2051 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run2048 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2039 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2036 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run1404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1252 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState1408
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1408

and _menhir_run1376 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 65028 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1376 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1376
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1376

and _menhir_run1379 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1379 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1379
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1379

and _menhir_run1382 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3153 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1383 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1383
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1383) : 'freshtv3154)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3155 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3156)

and _menhir_run1386 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1386 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1386
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1386

and _menhir_run1389 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1389 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1389
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1389

and _menhir_reduce140 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3943 "reason_parser.mly"
      ( mktyp(Ptyp_any) )
# 65268 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run78 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 65275 "reason_parser.ml"
) * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState78 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState78

and _menhir_run129 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState129 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState129

and _menhir_reduce144 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3951 "reason_parser.mly"
      ( mktyp(Ptyp_object ([], Closed)) )
# 65312 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce143 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3949 "reason_parser.mly"
      ( mktyp(Ptyp_object ([], Open)) )
# 65324 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run148 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 65331 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce460 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run149 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3151) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    ((let _1 = () in
    let _v : 'tv_meth_list = 
# 4026 "reason_parser.mly"
                                                ( [], Open )
# 65349 "reason_parser.ml"
     in
    _menhir_goto_meth_list _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3152)

and _menhir_reduce632 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_opt_bar = 
# 4226 "reason_parser.mly"
                                                ( () )
# 65358 "reason_parser.ml"
     in
    _menhir_goto_opt_bar _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce149 : _menhir_env -> 'ttv_tail * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_simple_core_type = 
# 3965 "reason_parser.mly"
      ( mktyp(Ptyp_variant([], Open, None)) )
# 65372 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run157 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3149) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_opt_bar = 
# 4227 "reason_parser.mly"
                                                ( () )
# 65387 "reason_parser.ml"
     in
    _menhir_goto_opt_bar _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3150)

and _menhir_run122 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState122 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState122 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState122

and _menhir_reduce271 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3340 "reason_parser.mly"
      ( mkpat(Ppat_any) )
# 65414 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run144 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 65421 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce823 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run199 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 871 "reason_parser.mly"
       (nativeint)
# 65431 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce717 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run200 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 841 "reason_parser.mly"
       (int64)
# 65440 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce716 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run201 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 840 "reason_parser.mly"
       (int32)
# 65449 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce715 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run202 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 839 "reason_parser.mly"
       (int)
# 65458 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce713 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run203 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 821 "reason_parser.mly"
       (string)
# 65467 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce714 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run205 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 871 "reason_parser.mly"
       (nativeint)
# 65476 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce712 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run206 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 841 "reason_parser.mly"
       (int64)
# 65485 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce711 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run207 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 840 "reason_parser.mly"
       (int32)
# 65494 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce710 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run208 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 839 "reason_parser.mly"
       (int)
# 65503 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce708 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_run209 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> (
# 821 "reason_parser.mly"
       (string)
# 65512 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce709 _menhir_env (Obj.magic _menhir_stack) _endpos _v _startpos

and _menhir_reduce282 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3366 "reason_parser.mly"
      ( mkpat(Ppat_array []) )
# 65528 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run512 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run226 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run225 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run224 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run223 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3147 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3148)

and _menhir_run513 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run494 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run491 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run490 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3145 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3146)

and _menhir_run517 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 65583 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce466 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run487 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState487 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState487

and _menhir_error70 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce332 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error238 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce646 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce139 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__non_arrowed_simple_core_type = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65716 "reason_parser.ml"
      
    in
    
# 3941 "reason_parser.mly"
      ( mktyp(Ptyp_class(_2, [])) )
# 65722 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce135 : _menhir_env -> (('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_non_simple_core_type = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65741 "reason_parser.ml"
      
    in
    
# 3924 "reason_parser.mly"
      ( mktyp(Ptyp_class(_2, List.rev _3)) )
# 65747 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_non_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce823 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 65754 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 65760 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_type_longident = 
# 4128 "reason_parser.mly"
                                                ( Lident _1 )
# 65767 "reason_parser.ml"
     in
    _menhir_goto_type_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce141 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_type_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__non_arrowed_simple_core_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65785 "reason_parser.ml"
      
    in
    
# 3945 "reason_parser.mly"
      ( mktyp(Ptyp_constr(_1, [])) )
# 65791 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce134 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_type_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__non_arrowed_non_simple_core_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65809 "reason_parser.ml"
      
    in
    
# 3922 "reason_parser.mly"
      ( mktyp(Ptyp_constr(_1, List.rev _2)) )
# 65815 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_non_simple_core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce57 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_core_type), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__core_type2 = 
# 3879 "reason_parser.mly"
      ( _1 )
# 65827 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error421 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 65834 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce434 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce236 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65854 "reason_parser.ml"
      
    in
    
# 2767 "reason_parser.mly"
      (
        let bang = {_1 with txt="!"} in
        mkexp(Pexp_apply(mkoperator bang, ["",_2]))
      )
# 65863 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce244 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 895 "reason_parser.mly"
       (string)
# 65870 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _endpos_x0_, _, (x0 : (
# 895 "reason_parser.mly"
       (string)
# 65876 "reason_parser.ml"
    )), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_simple_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__simple_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65889 "reason_parser.ml"
      
    in
    
# 2790 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 65895 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce484 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_lbl_pattern) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_pattern)), _endpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_pattern_list = 
# 3436 "reason_parser.mly"
                      ( [_1], Closed )
# 65906 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error530 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce482 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error544 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce272 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce173 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern_without_or), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = 
# 3306 "reason_parser.mly"
      ( mkpat(Ppat_exception _2) )
# 65927 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce275 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_pattern_not_ident = 
# 3348 "reason_parser.mly"
      ( mkpat(Ppat_variant(_1, None)) )
# 65939 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce274 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern_not_ident = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65957 "reason_parser.ml"
      
    in
    
# 3346 "reason_parser.mly"
      ( mkpat(Ppat_construct(_1, None)) )
# 65963 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce166 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_simple_pattern_list)) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__pattern_without_or = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 65981 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3277 "reason_parser.mly"
    (
      match is_pattern_list_single_any _2 with
        | Some singleAnyPat ->
            mkpat
              (Ppat_construct(_1, Some singleAnyPat))
        | None ->
          let loc = mklocation _symbolstartpos _endpos in
          let argPattern = simple_pattern_list_to_tuple ~loc _2 in
          mkExplicitArityTuplePat (Ppat_construct(_1, Some argPattern))
    )
# 65999 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce656 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern = 
# 3257 "reason_parser.mly"
                       ( _1 )
# 66011 "reason_parser.ml"
     in
    _menhir_goto_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce168 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos_x0_, _, _startpos_x0_), _endpos__3_, _, (_3 : 'tv_pattern_without_or), _startpos__3_) = _menhir_stack in
    let x0 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__pattern_without_or = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66030 "reason_parser.ml"
      
    in
    
# 3292 "reason_parser.mly"
      ( Location.raise_errorf ~loc:_2.loc ":: is not supported in Reason, please use [hd, ...tl] instead" )
# 66036 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce659 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_comma_list), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_pattern_comma_list = 
# 3424 "reason_parser.mly"
                                                  ( _3 :: _1 )
# 66049 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce658 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern_comma_list = 
# 3423 "reason_parser.mly"
                                                ( [_1] )
# 66061 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce661 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_comma_list), _startpos__1_), _endpos__2_, _), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_pattern), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv_pattern_comma_list_extension = 
# 3430 "reason_parser.mly"
                                                ( (_1, Some _4) )
# 66075 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce662 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_comma_list), _startpos__1_), _endpos__2_, _), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_pattern_comma_list_extension = 
# 3431 "reason_parser.mly"
                                                ( (_3 :: _1, None) )
# 66088 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce660 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_pattern_comma_list_extension = 
# 3429 "reason_parser.mly"
                                                ( [_1], None )
# 66100 "reason_parser.ml"
     in
    _menhir_goto_pattern_comma_list_extension _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce169 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _ _startpos_x0_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__pattern_without_or = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66120 "reason_parser.ml"
      
    in
    
# 3294 "reason_parser.mly"
      ( expecting_pat (with_txt _3 "pattern") )
# 66126 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce165 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _ _startpos_x0_ ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern_without_or), _startpos__1_), _endpos__2_, _, _startpos__2_) = _menhir_stack in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__pattern_without_or = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66146 "reason_parser.ml"
      
    in
    
# 3267 "reason_parser.mly"
      ( expecting_pat (with_txt _3 "identifier") )
# 66152 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce55 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type2), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__core_type = 
# 3860 "reason_parser.mly"
      ( _1 )
# 66164 "reason_parser.ml"
     in
    _menhir_goto__core_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce60 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_core_type2), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type2), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__core_type2 = 
# 3885 "reason_parser.mly"
      ( mktyp(Ptyp_arrow("", _1, _3)) )
# 66177 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce481 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_pattern), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_pattern = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66194 "reason_parser.ml"
      
    in
    
# 3443 "reason_parser.mly"
      ( (_1,_3) )
# 66200 "reason_parser.ml"
     in
    _menhir_goto_lbl_pattern _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce286 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_pattern_not_ident = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66220 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66232 "reason_parser.ml"
      
    in
    
# 3374 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _3 ")") )
# 66238 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce289 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x0_ _ _startpos_x0_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_) = _menhir_stack in
    let x0 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_pattern_not_ident = let _4 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66259 "reason_parser.ml"
      
    in
    
# 3381 "reason_parser.mly"
      ( expecting_pat (with_txt _4 "type") )
# 66265 "reason_parser.ml"
     in
    _menhir_goto__simple_pattern_not_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce77 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_pattern), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_simple_expr), _startpos__4_), _, (_5 : 'tv_direction_flag)), _endpos__6_, _, (_6 : 'tv_simple_expr), _startpos__6_), _endpos__7_, _, (_7 : 'tv_simple_expr), _startpos__7_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = 
# 2553 "reason_parser.mly"
      ( mkexp(Pexp_for(_2, _4, _6, _5, _7)) )
# 66279 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce107 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2591 "reason_parser.mly"
      ( mkexp (Pexp_assert _2) )
# 66292 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce204 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_expr = 
# 2648 "reason_parser.mly"
      ( mkexp (Pexp_variant (_1, None)) )
# 66304 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce73 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_name_tag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2543 "reason_parser.mly"
    (
      mkexp(Pexp_variant(_1, Some _2))
    )
# 66318 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce64 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__expr = 
# 2511 "reason_parser.mly"
      ( _1 )
# 66330 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce495 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_less_aggressive_simple_expression = 
# 2818 "reason_parser.mly"
              (_1)
# 66342 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3143) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_less_aggressive_simple_expression) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState475 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3119 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState820
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run821 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState820 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce64 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3120)
    | MenhirState822 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3121 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 66409 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce465 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3122)
    | MenhirState825 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3123 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 66417 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce464 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3124)
    | MenhirState1647 | MenhirState1648 | MenhirState820 | MenhirState827 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3125 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce468 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3126)
    | MenhirState842 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3127 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce724 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3128)
    | MenhirState840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3129 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce723 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3130)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 | MenhirState2574 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3131 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState933
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | BANG ->
              _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACELESS ->
              _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run934 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NEW ->
              _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PREFIXOP _v ->
              _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState933 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BAR | BARBAR | BARRBRACKET | BEGIN | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | LAZY | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRUCT | SWITCH | THEN | TILDE | TO | TRY | TYPE | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce64 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv3132)
    | MenhirState935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3133 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 66489 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce465 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3134)
    | MenhirState938 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3135 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 66497 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce464 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3136)
    | MenhirState2687 | MenhirState2688 | MenhirState933 | MenhirState940 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3137 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce468 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3138)
    | MenhirState955 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3139 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce724 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3140)
    | MenhirState953 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3141 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce723 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3142)
    | _ ->
        _menhir_fail ()) : 'freshtv3144)

and _menhir_error821 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 66518 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce65 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_less_aggressive_simple_expression), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_labeled_expr_list)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2513 "reason_parser.mly"
      ( mkexp(Pexp_apply(_1, List.rev _2)) )
# 66531 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce106 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66550 "reason_parser.ml"
      
    in
    
# 2589 "reason_parser.mly"
      ( mkexp(Pexp_setinstvar(_1, _3)) )
# 66556 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce88 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4083 "reason_parser.mly"
                                                ( "*" )
# 66575 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66585 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66591 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce83 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 836 "reason_parser.mly"
       (string)
# 66598 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 836 "reason_parser.mly"
       (string)
# 66604 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4078 "reason_parser.mly"
                                                ( _1 )
# 66617 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66627 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66633 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce203 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_constr_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66651 "reason_parser.ml"
      
    in
    
# 2643 "reason_parser.mly"
    (
      mkexp (Pexp_construct (_1, None))
    )
# 66659 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce268 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_non_labeled_expr_list), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__simple_non_labeled_expr_list_as_tuple = 
# 2830 "reason_parser.mly"
                               (
    mkexp (Pexp_tuple(List.rev _1))
  )
# 66673 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3117) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__simple_non_labeled_expr_list_as_tuple) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3113) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_non_labeled_expr_list_as_tuple) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce537 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3114)
    | MenhirState953 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3115) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__simple_non_labeled_expr_list_as_tuple) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce537 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3116)
    | _ ->
        _menhir_fail ()) : 'freshtv3118)

and _menhir_reduce101 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_additive), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66715 "reason_parser.ml"
      
    in
    
# 2568 "reason_parser.mly"
      (
        mkuplus _1 _2
      )
# 66723 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce96 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4091 "reason_parser.mly"
                                                ( "+=" )
# 66742 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66752 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66758 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce97 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4092 "reason_parser.mly"
                                                ( "%" )
# 66777 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66787 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66793 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce82 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 835 "reason_parser.mly"
       (string)
# 66800 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 835 "reason_parser.mly"
       (string)
# 66806 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4077 "reason_parser.mly"
                                                ( _1 )
# 66819 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66829 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66835 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce85 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4080 "reason_parser.mly"
                                                ( "+." )
# 66854 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66864 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66870 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce84 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4079 "reason_parser.mly"
                                                ( "+" )
# 66889 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66899 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66905 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce91 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4086 "reason_parser.mly"
                                                ( "or" )
# 66924 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66934 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66940 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce87 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4082 "reason_parser.mly"
                                                ( "-." )
# 66959 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 66969 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 66975 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce86 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4081 "reason_parser.mly"
                                                ( "-" )
# 66994 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67004 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67010 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce98 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4093 "reason_parser.mly"
                                                ( "<>" )
# 67029 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67039 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67045 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce81 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 834 "reason_parser.mly"
       (string)
# 67052 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 834 "reason_parser.mly"
       (string)
# 67058 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4076 "reason_parser.mly"
                                                ( _1 )
# 67071 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67081 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67087 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce80 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 833 "reason_parser.mly"
       (string)
# 67094 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 833 "reason_parser.mly"
       (string)
# 67100 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4075 "reason_parser.mly"
                                                ( _1 )
# 67113 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67123 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67129 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce99 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4094 "reason_parser.mly"
                                                ( "<..>" )
# 67148 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67158 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67164 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce89 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4084 "reason_parser.mly"
                                                ( "<" )
# 67183 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67193 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67199 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce79 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 832 "reason_parser.mly"
       (string)
# 67206 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, (_100 : (
# 832 "reason_parser.mly"
       (string)
# 67212 "reason_parser.ml"
    )), _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4074 "reason_parser.mly"
                                                ( _1 )
# 67225 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67235 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67241 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce90 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4085 "reason_parser.mly"
                                                ( ">" )
# 67260 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67270 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67276 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce92 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4087 "reason_parser.mly"
                                                ( "||" )
# 67295 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67305 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67311 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce93 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4088 "reason_parser.mly"
                                                ( "&" )
# 67330 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67340 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67346 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce94 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4089 "reason_parser.mly"
                                                ( "&&" )
# 67365 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67375 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67381 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce95 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__100_, _, _startpos__100_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _100 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = let _2 =
      let _endpos__10_ = _endpos__100_ in
      let _startpos__10_ = _startpos__100_ in
      let _10 = _100 in
      let x =
        let _1 = _10 in
        
# 4090 "reason_parser.mly"
                                                ( ":=" )
# 67400 "reason_parser.ml"
        
      in
      let _endpos_x_ = _endpos__10_ in
      let _startpos_x_ = _startpos__10_ in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67410 "reason_parser.ml"
      
    in
    
# 2562 "reason_parser.mly"
      ( mkinfix _1 _2 _3 )
# 67416 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce217 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _, _startpos__5_) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2690 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "Array" "get")),
                         ["",_1; "",_4])) )
# 67436 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce103 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _, _startpos__5_), _endpos__6_, _startpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2574 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "Array" "set")),
                         ["",_1; "",_4; "",_7])) )
# 67457 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce109 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2618 "reason_parser.mly"
                                                       (
      (* Should use ghost expressions, but not sure how that would work with source maps *)
      (* So ? will become true and : becomes false for now*)
      let loc_question = mklocation _startpos__2_ _endpos__2_ in
      let loc_colon = mklocation _startpos__4_ _endpos__4_ in
      let fauxTruePat =
        Pat.mk ~loc:loc_question (Ppat_construct({txt = Lident "true"; loc = loc_question}, None)) in
      let fauxFalsePat =
        Pat.mk ~loc:loc_colon (Ppat_construct({txt = Lident "false"; loc = loc_colon}, None)) in
      let fauxMatchCaseTrue = Exp.case fauxTruePat _3 in
      let fauxMatchCaseFalse = Exp.case fauxFalsePat _5 in
      mkexp (Pexp_match (_1, [fauxMatchCaseTrue; fauxMatchCaseFalse]))
    )
# 67483 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce219 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2697 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "String" "get")),
                         ["",_1; "",_4])) )
# 67503 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce104 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _), _endpos__6_, _startpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2579 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp(Pexp_apply(mkexp ~ghost:true ~loc (Pexp_ident(array_function ~loc "String" "set")),
                         ["",_1; "",_4; "",_7])) )
# 67524 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce221 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _) = _menhir_stack in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__simple_expr = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2704 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        bigarray_get ~loc _1 _4
      )
# 67544 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce105 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_), _endpos__5_, _), _endpos__6_, _startpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__expr = let _endpos = _endpos__7_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2584 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        bigarray_set ~loc _1 _4 _7
      )
# 67565 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1058 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce212 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce102 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_expr), _startpos__1_), _), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67589 "reason_parser.ml"
      
    in
    
# 2572 "reason_parser.mly"
      ( mkexp(Pexp_setfield(_1, _3, _5)) )
# 67595 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce100 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_subtractive), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67613 "reason_parser.ml"
      
    in
    
# 2564 "reason_parser.mly"
      (
        mkuminus _1 _2
      )
# 67621 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce635 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
    let _1 = () in
    let _v : 'tv_opt_comma = 
# 4231 "reason_parser.mly"
                                                ( () )
# 67632 "reason_parser.ml"
     in
    _menhir_goto_opt_comma _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce419 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _v : 'tv_expr_optional_constraint = 
# 3160 "reason_parser.mly"
                         ( _1 )
# 67643 "reason_parser.ml"
     in
    _menhir_goto_expr_optional_constraint _menhir_env _menhir_stack _menhir_s _v _startpos

and _menhir_error1091 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce809 _menhir_env (Obj.magic _menhir_stack)

and _menhir_error1116 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce688 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce478 : _menhir_env -> ('ttv_tail * _menhir_state * 'tv_lbl_expr) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, (_1 : 'tv_lbl_expr)), _endpos__2_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_expr_list = 
# 3175 "reason_parser.mly"
                    ( [_1] )
# 67662 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr_list _menhir_env _menhir_stack _menhir_s _v

and _menhir_error1122 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce475 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce474 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_lbl_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67683 "reason_parser.ml"
      
    in
    
# 3179 "reason_parser.mly"
      ( (_1, _3) )
# 67689 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce602 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_non_punned_lbl_expression = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67706 "reason_parser.ml"
      
    in
    
# 3188 "reason_parser.mly"
      ( (_1, _3) )
# 67712 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3111) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_punned_lbl_expression) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3109) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_non_punned_lbl_expression) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3107) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv_non_punned_lbl_expression) : 'tv_non_punned_lbl_expression) = _v in
    ((let _v : 'tv_lbl_expr_list_that_is_not_a_single_punned_field = 
# 3208 "reason_parser.mly"
     ( [_1] )
# 67729 "reason_parser.ml"
     in
    _menhir_goto_lbl_expr_list_that_is_not_a_single_punned_field _menhir_env _menhir_stack _menhir_s _v) : 'freshtv3108)) : 'freshtv3110)) : 'freshtv3112)

and _menhir_reduce453 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 67736 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 906 "reason_parser.mly"
       (string)
# 67742 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_ident = 
# 4064 "reason_parser.mly"
                                                ( _1 )
# 67749 "reason_parser.ml"
     in
    _menhir_goto_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce677 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 898 "reason_parser.mly"
       (string * string option)
# 67756 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 898 "reason_parser.mly"
       (string * string option)
# 67762 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_primitive_declaration = 
# 3453 "reason_parser.mly"
                                                ( [fst _1] )
# 67768 "reason_parser.ml"
     in
    _menhir_goto_primitive_declaration _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce450 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_generalized_constructor_arguments = 
# 3607 "reason_parser.mly"
                                                         ( (List.rev _1, None) )
# 67779 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce358 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_class_sig_fields = 
# 2246 "reason_parser.mly"
                                        ( [] )
# 67789 "reason_parser.ml"
     in
    _menhir_goto_class_sig_fields _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce32 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_clty_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__class_instance_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67807 "reason_parser.ml"
      
    in
    
# 2215 "reason_parser.mly"
      ( mkcty(Pcty_constr (_1, [])) )
# 67813 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce33 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_clty_longident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_instance_type = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67831 "reason_parser.ml"
      
    in
    
# 2217 "reason_parser.mly"
      ( mkcty(Pcty_constr (_1, List.rev _2)) )
# 67837 "reason_parser.ml"
     in
    _menhir_goto__class_instance_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce40 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_instance_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_sig_field = 
# 2256 "reason_parser.mly"
      ( mkctf_attrs (Pctf_inherit _2) [] )
# 67850 "reason_parser.ml"
     in
    _menhir_goto__class_sig_field _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce362 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 67857 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 67863 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_instance_type), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_class_type_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67876 "reason_parser.ml"
      
    in
    
# 2334 "reason_parser.mly"
                                                                                (
      (_2, _5, _1, List.rev _3)
    )
# 67884 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3105) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_class_type_declaration_details) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState1536 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3097 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1623
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1623
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1623) : 'freshtv3098)
    | MenhirState1876 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3099 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1877
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1877
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1877) : 'freshtv3100)
    | MenhirState2052 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3101 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2058
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2058
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2058) : 'freshtv3102)
    | MenhirState2087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3103 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2088
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2088
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2088) : 'freshtv3104)
    | _ ->
        _menhir_fail ()) : 'freshtv3106)

and _menhir_reduce15 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__class_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67959 "reason_parser.ml"
      
    in
    
# 1948 "reason_parser.mly"
                                  (
      mkclass(Pcl_constr(_2, []))
    )
# 67967 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce16 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_class_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_non_arrowed_simple_core_type_list)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_expr = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 67986 "reason_parser.ml"
      
    in
    
# 1952 "reason_parser.mly"
                                                                    (
      mkclass(Pcl_constr(_2, List.rev _3))
    )
# 67994 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce11 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_simple_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_expr = 
# 1921 "reason_parser.mly"
      ( _1 )
# 68006 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce13 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_simple_expr), _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_labeled_expr_list)) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_expr = 
# 1932 "reason_parser.mly"
      ( mkclass(Pcl_apply(_1, List.rev _2)) )
# 68018 "reason_parser.ml"
     in
    _menhir_goto__class_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce30 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_labeled_simple_pattern), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_class_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__class_fun_def = 
# 1885 "reason_parser.mly"
      ( let (l,o,p) = _1 in mkclass(Pcl_fun(l, o, p, _3)) )
# 68031 "reason_parser.ml"
     in
    _menhir_goto__class_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce842 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_value = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68049 "reason_parser.ml"
      
    in
    
# 2024 "reason_parser.mly"
      ( _3, _2, Cfk_concrete (_1, _5) )
# 68055 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce843 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_mutable_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, (_4 : 'tv_type_constraint)), _endpos__5_, _startpos__5_), _endpos__6_, _, (_6 : 'tv_expr), _startpos__6_) = _menhir_stack in
    let _5 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_value = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68073 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2026 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        let e = ghexp_constraint loc _6 _4 in
        _3, _2, Cfk_concrete (_1, e)
      )
# 68092 "reason_parser.ml"
     in
    _menhir_goto_value _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce403 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_curried_binding = 
# 3056 "reason_parser.mly"
      ( _2 )
# 68104 "reason_parser.ml"
     in
    _menhir_goto_curried_binding _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce408 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_non_arrowed_core_type), _startpos__2_), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_curried_binding_return_typed_ = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3045 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        ghexp_constraint loc _4 (Some _2, None)
      )
# 68122 "reason_parser.ml"
     in
    _menhir_goto_curried_binding_return_typed_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce568 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_method_ = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68140 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__5_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2048 "reason_parser.mly"
      ( _3, _2,
        Cfk_concrete (_1, let loc = mklocation _symbolstartpos _endpos in mkexp ~ghost:true ~loc (Pexp_poly(_5, None))) )
# 68156 "reason_parser.ml"
     in
    _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce569 : _menhir_env -> ((((((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, _startpos__5_), _, (_6 : 'tv_lident_list)), _endpos__8_, _, (_8 : 'tv_core_type), _startpos__8_), _endpos__9_, _startpos__9_), _endpos__10_, _, (_10 : 'tv_expr), _startpos__10_) = _menhir_stack in
    let _9 = () in
    let _7 = () in
    let _5 = () in
    let _4 = () in
    let _endpos = _endpos__10_ in
    let _v : 'tv_method_ = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68177 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__10_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2054 "reason_parser.mly"
      (

        (* For non, methods we'd create a pattern binding:
           ((Ppat_constraint(mkpatvar ..., Ptyp_poly (typeVars, poly_type_varified))),
            exp_with_newtypes_constrained_by_non_varified)

           For methods, we create:
           Pexp_poly (Pexp_constraint (methodFunWithNewtypes, non_varified), Some (Ptyp_poly newTypes varified))
         *)
        let (exp_with_newtypes_constrained_by_non_varified, poly_type_varified) =
          wrap_type_annotation _6 _8 _10 in
        (
          _3, _2,
          Cfk_concrete (
            _1,
            let loc = mklocation _symbolstartpos _endpos in
            mkexp ~ghost:true ~loc (Pexp_poly(exp_with_newtypes_constrained_by_non_varified, Some poly_type_varified))
          )
        )
      )
# 68211 "reason_parser.ml"
     in
    _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce567 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_override_flag), _startpos__1_), _endpos__2_, _, (_2 : 'tv_private_flag), _startpos__2_), _endpos_x0_, _, (x0 : 'tv_label), _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_poly_type)), _endpos__6_, _startpos__6_), _endpos__7_, _, (_7 : 'tv_expr), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _4 = () in
    let _endpos = _endpos__7_ in
    let _v : 'tv_method_ = let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68230 "reason_parser.ml"
      
    in
    let _startpos__3_ = _startpos_x0_ in
    let _endpos = _endpos__7_ in
    let _symbolstartpos = if Pervasives.(!=) _startpos__1_ _endpos__1_ then
      _startpos__1_
    else
      if Pervasives.(!=) _startpos__2_ _endpos__2_ then
        _startpos__2_
      else
        _startpos__3_ in
    
# 2045 "reason_parser.mly"
      ( _3, _2,
        Cfk_concrete (_1, let loc = mklocation _symbolstartpos _endpos in mkexp ~ghost:true ~loc (Pexp_poly(_7, Some _5))) )
# 68246 "reason_parser.ml"
     in
    _menhir_goto_method_ _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce652 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_parent_binder = 
# 2007 "reason_parser.mly"
          ( None )
# 68256 "reason_parser.ml"
     in
    _menhir_goto_parent_binder _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce695 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_semi_terminated_class_fields = 
# 1998 "reason_parser.mly"
                          ([])
# 68266 "reason_parser.ml"
     in
    _menhir_goto_semi_terminated_class_fields _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce19 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_class_expr), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__class_expr_lets_and_rest = 
# 1893 "reason_parser.mly"
               (_1)
# 68278 "reason_parser.ml"
     in
    _menhir_goto__class_expr_lets_and_rest _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce7 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_instance_type), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__class_constructor_type = 
# 2196 "reason_parser.mly"
      ( _2 )
# 68291 "reason_parser.ml"
     in
    _menhir_goto__class_constructor_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce340 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 68298 "reason_parser.ml"
) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_virtual_flag)), _endpos_x0_, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 68304 "reason_parser.ml"
    )), _startpos_x0_), _, (_3 : 'tv_class_type_parameters)), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_class_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_class_declaration_details = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68317 "reason_parser.ml"
      
    in
    
# 1828 "reason_parser.mly"
    (
       (_2, _5, _1, List.rev _3)
    )
# 68325 "reason_parser.ml"
     in
    _menhir_goto_class_declaration_details _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce355 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos__0_) = Obj.magic _menhir_stack in
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_class_self_type = let _endpos = _endpos__0_ in
    let _symbolstartpos = _endpos in
    
# 2240 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        Typ.mk ~loc Ptyp_any
       )
# 68341 "reason_parser.ml"
     in
    _menhir_goto_class_self_type _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce54 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_constructor_type)), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_class_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__constrained_class_declaration = 
# 1804 "reason_parser.mly"
      ( mkclass(Pcl_constraint(_4, _2)) )
# 68355 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3095) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__constrained_class_declaration) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((match _menhir_s with
    | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3091) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__constrained_class_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce527 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3092)
    | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3093) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv__constrained_class_declaration) = _v in
        let (_startpos : Lexing.position) = _startpos in
        (_menhir_reduce527 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos : 'freshtv3094)
    | _ ->
        _menhir_fail ()) : 'freshtv3096)

and _menhir_reduce347 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_class_fun_return = 
# 1868 "reason_parser.mly"
      ( _2 )
# 68391 "reason_parser.ml"
     in
    _menhir_goto_class_fun_return _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce348 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_non_arrowed_class_constructor_type)), _startpos__3_), _endpos__4_, _, (_4 : 'tv_class_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_class_fun_return = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1870 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkclass ~loc (Pcl_constraint(_4, _2))
      )
# 68409 "reason_parser.ml"
     in
    _menhir_goto_class_fun_return _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce299 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1412 "reason_parser.mly"
                           (
      mkstr(Pstr_type (List.rev _1))
    )
# 68423 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce640 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_optional_type_parameters = 
# 3515 "reason_parser.mly"
                                   ( [] )
# 68433 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameters _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce812 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_type_kind = 
# 3492 "reason_parser.mly"
      ( (Ptype_abstract, Public, None) )
# 68443 "reason_parser.ml"
     in
    _menhir_goto_type_kind _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce811 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 68450 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 68456 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_optional_type_parameters)), _endpos__3_, _, (_3 : 'tv_type_kind)), _endpos__4_, _, (_4 : 'tv_constraints)) = _menhir_stack in
    let _endpos = _endpos__4_ in
    let _v : 'tv_type_declaration_details = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68468 "reason_parser.ml"
      
    in
    
# 3482 "reason_parser.mly"
      ( let (kind, priv, manifest) = _3 in
        (_1, _2, List.rev _4, kind, priv, manifest)
       )
# 68476 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3089) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_type_declaration_details) = _v in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v) in
    match _menhir_s with
    | MenhirState2584 | MenhirState1253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3081 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1369
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState1369
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1369) : 'freshtv3082)
    | MenhirState1854 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3083 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1867
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState1867
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1867) : 'freshtv3084)
    | MenhirState2075 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv3085 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2078
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run439 _menhir_env (Obj.magic _menhir_stack) MenhirState2078
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2078) : 'freshtv3086)
    | MenhirState76 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3087 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2533
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | LBRACKETATAT ->
              _menhir_run396 _menhir_env (Obj.magic _menhir_stack) MenhirState2533
          | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
              _menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2533) : 'freshtv3088)
    | _ ->
        _menhir_fail ()) : 'freshtv3090)

and _menhir_reduce641 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_optional_type_parameter_list)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_optional_type_parameters = 
# 3516 "reason_parser.mly"
                                   ( List.rev _1 )
# 68543 "reason_parser.ml"
     in
    _menhir_goto_optional_type_parameters _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce303 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_nonlocal_module_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1426 "reason_parser.mly"
                                  (
      mkstr(Pstr_recmodule(List.rev _1))
    )
# 68557 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce308 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_type_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1447 "reason_parser.mly"
                                 (
      (* Each declaration has their own preceeding post_item_attributes *)
      mkstr(Pstr_class_type (List.rev _1))
    )
# 68572 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce307 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_many_class_declarations), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item_without_item_extension_sugar = 
# 1443 "reason_parser.mly"
                            (
      (* Each declaration has their own preceeding post_item_attributes *)
      mkstr(Pstr_class (List.rev _1))
    )
# 68587 "reason_parser.ml"
     in
    _menhir_goto__structure_item_without_item_extension_sugar _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce296 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_let_bindings), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv__structure_item = 
# 1396 "reason_parser.mly"
                 ( val_of_let_bindings _1 )
# 68599 "reason_parser.ml"
     in
    _menhir_goto__structure_item _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce61 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__curried_binding_return_typed = let _endpos = _endpos__2_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3010 "reason_parser.mly"
      (
          let loc = mklocation _symbolstartpos _endpos in
          let nil = { txt = Lident "()"; loc = make_ghost_loc loc } in
          mkexp ~ghost:true ~loc (Pexp_fun("", None, mkpat ~ghost:true ~loc (Ppat_construct (nil, None)), _2))
      )
# 68618 "reason_parser.ml"
     in
    _menhir_goto__curried_binding_return_typed _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce538 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x_, _menhir_s, (x : 'tv_expr), _startpos_x_) = _menhir_stack in
    let _endpos = _endpos_x_ in
    let _v : 'tv_mark_position_exp_expr_ = let _endpos = _endpos_x_ in
    let _symbolstartpos = _startpos_x_ in
    
# 1168 "reason_parser.mly"
 ( {x with pexp_loc = {x.pexp_loc with loc_start = _symbolstartpos; loc_end = _endpos}} )
# 68631 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3079) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_mark_position_exp_expr_) = _v in
    ((match _menhir_s with
    | MenhirState1213 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3071 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce501 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3072)
    | MenhirState1221 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3073 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce500 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3074)
    | MenhirState1900 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv3075 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce501 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3076)
    | MenhirState1906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv3077 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _endpos in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : 'tv_mark_position_exp_expr_) = _v in
        (_menhir_reduce500 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v : 'freshtv3078)
    | _ ->
        _menhir_fail ()) : 'freshtv3080)

and _menhir_reduce498 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_val_ident), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_type_constraint)), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_let_binding_body = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1213 "reason_parser.mly"
    ( let loc = mklocation _symbolstartpos _endpos in
      mkpat ~loc (Ppat_var (mkloc x loc)) )
# 68685 "reason_parser.ml"
      
    in
    let _startpos__1_ = _startpos_x0_ in
    let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2909 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        (_1, ghexp_constraint loc _4 _2)
      )
# 68697 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce503 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_simple_pattern_not_ident), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_core_type), _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_let_binding_body = let _endpos = _endpos__5_ in
    let _symbolstartpos = _startpos__1_ in
    
# 2989 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        (mkpat ~loc (Ppat_constraint(_1, _3)), _5)
      )
# 68715 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce502 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_pattern), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_let_binding_body = 
# 2987 "reason_parser.mly"
      ( (_1, _3) )
# 68727 "reason_parser.ml"
     in
    _menhir_goto_let_binding_body _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce117 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _, (_2 : 'tv_functor_args), _startpos__2_), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__4_ in
    let _v : 'tv__module_expr = 
# 1282 "reason_parser.mly"
      ( mkFunctorThatReturns _2 _4 )
# 68741 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce383 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_constraints = 
# 3488 "reason_parser.mly"
                                                ( [] )
# 68751 "reason_parser.ml"
     in
    _menhir_goto_constraints _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce849 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_label_longident), _startpos_x0_), _endpos__3_, _, (_3 : 'tv_optional_type_parameters)), _, (_4 : 'tv_with_type_binder)), _endpos__5_, _, (_5 : 'tv_core_type), _startpos__5_), _endpos__6_, _, (_6 : 'tv_constraints)) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__6_ in
    let _v : 'tv_with_constraint = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68769 "reason_parser.ml"
      
    in
    let _endpos = _endpos__6_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3697 "reason_parser.mly"
      (
         let loc = mklocation _symbolstartpos _endpos in

         Pwith_type
          (_2,
           (Type.mk {_2 with txt=Longident.last _2.txt}
              ~params:_3
              ~cstrs:(List.rev _6)
              ~manifest:_5
              ~priv:_4
              ~loc)) )
# 68787 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce573 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 68794 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 906 "reason_parser.mly"
       (string)
# 68800 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mod_ext_longident = 
# 4146 "reason_parser.mly"
                                                ( Lident _1 )
# 68807 "reason_parser.ml"
     in
    _menhir_goto_mod_ext_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce852 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 68814 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : (
# 906 "reason_parser.mly"
       (string)
# 68820 "reason_parser.ml"
    )), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos_x1_, _, (x1 : 'tv_mod_ext_longident), _startpos_x1_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos_x1_ in
    let _v : 'tv_with_constraint = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68834 "reason_parser.ml"
      
    in
    let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68846 "reason_parser.ml"
      
    in
    
# 3727 "reason_parser.mly"
      ( Pwith_modsubst (_2, _4) )
# 68852 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce574 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 68859 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_ext_longident), _startpos__1_), _), _endpos__3_, _, (_3 : (
# 906 "reason_parser.mly"
       (string)
# 68865 "reason_parser.ml"
    )), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv_mod_ext_longident = 
# 4147 "reason_parser.mly"
                                                ( Ldot(_1, _3) )
# 68873 "reason_parser.ml"
     in
    _menhir_goto_mod_ext_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error1993 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce575 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce851 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_mod_longident), _startpos_x0_), _endpos__3_, _, _startpos__3_), _endpos_x1_, _, (x1 : 'tv_mod_ext_longident), _startpos_x1_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos_x1_ in
    let _v : 'tv_with_constraint = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68896 "reason_parser.ml"
      
    in
    let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68908 "reason_parser.ml"
      
    in
    
# 3725 "reason_parser.mly"
      ( Pwith_module (_2, _4) )
# 68914 "reason_parser.ml"
     in
    _menhir_goto_with_constraint _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2000 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_with_constraints -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce125 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce128 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_type), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_type), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__module_type = let _endpos = _endpos__3_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1615 "reason_parser.mly"
                                                          (
      (**
       * In OCaml, this is invalid:
       * module MyFunctor: functor MT -> (sig end) = functor MT -> (struct end);;
       *
       * Not only must curried functor args have annotations, but functor
       * annotations must include *names* for each argument in a functor type.
       *
       * module MyFunctor: functor (MT:MT) -> (sig end) = functor (MT:MT) -> (struct end)
       *
       * In Reason, we will parse the functor type:
       *
       *    (AB:MT) -> ReturnSig
       *
       * as in:
       *                   /----------------\
       * module MyFunctor: (A:B) => ReturnSig = functor (C:D) => {}
       *
       * But only for the sake of compatibility with existing OCaml code (the
       * ability to "view" OCaml code in Reason's syntax without loss of
       * information.) Do not write identifiers in functor argument type
       * positions - you wouldn't do it with functions, and they are
       * meaningless in functors.
       *
       *  But for sake of consistency (and for sake of a syntax that truly
       *  unifies functor syntax with function syntax, the following "sugars"
       *  will be parsed (and printed):
       *
       *   A => B => C
       *
       * Is parsed into:
       *
       * functor (_:A) -> functor (_:B) -> C
       *
       * And a dummy "_" is inserted into the parse tree where no name has been
       * provided.
       *
       *   {SomeSig} => {} => {}
       *
       * Is parsed into:
       *
       * (_:SomeSig) => (_:{}) => {}
       *
       *
       *)
      let loc = mklocation _symbolstartpos _endpos in
      mkmty(Pmty_functor(ghloc ~loc "_", Some _1, _3))
    )
# 68980 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce251 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos_x0_, _menhir_s, (x0 : 'tv_mod_longident), _startpos_x0_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x0_ in
    let _v : 'tv__simple_module_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 68998 "reason_parser.ml"
      
    in
    
# 1238 "reason_parser.mly"
      ( mkmod(Pmod_ident _1) )
# 69004 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce127 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos_x0_, _, (x0 : 'tv_functor_arg_name), _startpos_x0_), _endpos__3_, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_), _endpos__5_, _, _startpos__5_), _startpos__6_), _endpos__7_, _, (_7 : 'tv_module_type), _startpos__7_) = _menhir_stack in
    let _6 = () in
    let _5 = () in
    let _3 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__7_ in
    let _v : 'tv__module_type = let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69026 "reason_parser.ml"
      
    in
    
# 1601 "reason_parser.mly"
                                                                                                       (
    (* Why does this rule cause a conflict with core_type2? It has nothing to do
     * with it.
     *
     * Update: I'm guessing it has something to do with the fact that this isn't
     * parsed correctly because the => is considered part of "type dependenences" :
     *
     *    let module CreateFactory
     *               (Spec: ContainerSpec)
     *               :DescriptorFactoryIntf.S with
     *                  type props = Spec.props and type dependencies = Spec.props =>
     *)
      mkmty(Pmty_functor(_2, Some _4, _7))
     )
# 69045 "reason_parser.ml"
     in
    _menhir_goto__module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce263 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_module_type = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69065 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69077 "reason_parser.ml"
      
    in
    
# 1544 "reason_parser.mly"
      ( unclosed_mty (with_txt _1 "(") (with_txt _3 ")"))
# 69083 "reason_parser.ml"
     in
    _menhir_goto__simple_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce119 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_module_expr), _startpos__1_), _endpos_x0_, _, _startpos_x0_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69103 "reason_parser.ml"
      
    in
    let _2 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69115 "reason_parser.ml"
      
    in
    
# 1286 "reason_parser.mly"
      ( unclosed_mod (with_txt _2 "(") (with_txt _4 ")") )
# 69121 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce254 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_type), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _3 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_module_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69142 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69154 "reason_parser.ml"
      
    in
    
# 1244 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _5 ")"))
# 69160 "reason_parser.ml"
     in
    _menhir_goto__simple_module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce249 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_expr), _startpos__5_), _endpos__6_, _, _startpos__6_) = _menhir_stack in
    let x1 = () in
    let _6 = () in
    let _4 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _7 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69183 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69195 "reason_parser.ml"
      
    in
    
# 2807 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "(") (with_txt _7 ")"))
# 69201 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce215 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_), _), _endpos_x0_, _, _startpos_x0_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69222 "reason_parser.ml"
      
    in
    let _3 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69234 "reason_parser.ml"
      
    in
    
# 2682 "reason_parser.mly"
      ( unclosed_exp (with_txt _3 "(") (with_txt _5 ")") )
# 69240 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run927 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce218 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1004 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce220 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce438 : _menhir_env -> ('ttv_tail * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_fun_def = 
# 3105 "reason_parser.mly"
                      ( _2 )
# 69262 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce439 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_non_arrowed_core_type), _startpos__2_), _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_fun_def = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 3107 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkexp ~loc (Pexp_constraint(_4, _2))
      )
# 69280 "reason_parser.ml"
     in
    _menhir_goto_fun_def _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce488 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _endpos = _endpos__5_ in
    let _v : 'tv_leading_bar_match_case = 
# 3094 "reason_parser.mly"
      ( Exp.case _1 ~guard:_3 _5 )
# 69293 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_case _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce487 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_pattern_with_bar)), _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _endpos = _endpos__3_ in
    let _v : 'tv_leading_bar_match_case = 
# 3090 "reason_parser.mly"
                                       (
      Exp.case _1 _3
    )
# 69307 "reason_parser.ml"
     in
    _menhir_goto_leading_bar_match_case _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce68 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_leading_bar_match_cases)) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2524 "reason_parser.mly"
      ( mkexp (Pexp_function(List.rev _2)) )
# 69320 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce374 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_mod_longident), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constr_longident = 
# 4117 "reason_parser.mly"
                                                ( _1 )
# 69332 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce636 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_opt_semi = 
# 4234 "reason_parser.mly"
                                                ( () )
# 69342 "reason_parser.ml"
     in
    _menhir_goto_opt_semi _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce637 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__1_ in
    let _v : 'tv_opt_semi = 
# 4235 "reason_parser.mly"
                                                ( () )
# 69354 "reason_parser.ml"
     in
    _menhir_goto_opt_semi _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce75 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_expr), _startpos__3_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = 
# 2549 "reason_parser.mly"
      ( mkexp (Pexp_ifthenelse(_2, _3, None)) )
# 69367 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce74 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_expr), _startpos__3_), _endpos__4_, _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__expr = 
# 2547 "reason_parser.mly"
      ( mkexp(Pexp_ifthenelse(_2, _3, Some _5)) )
# 69381 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run889 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce218 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1049 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce220 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce108 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__expr = 
# 2593 "reason_parser.mly"
      ( mkexp (Pexp_lazy _2) )
# 69404 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce580 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_) = _menhir_stack in
    let _1 = () in
    let _endpos = _endpos__2_ in
    let _v : 'tv_module_binding_body_expr = 
# 1466 "reason_parser.mly"
      ( _2 )
# 69416 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce581 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_module_type), _startpos__2_), _endpos__3_, _, _startpos__3_), _endpos__4_, _, (_4 : 'tv_module_expr), _startpos__4_) = _menhir_stack in
    let _3 = () in
    let _1 = () in
    let _endpos = _endpos__4_ in
    let _v : 'tv_module_binding_body_expr = let _endpos = _endpos__4_ in
    let _symbolstartpos = _startpos__1_ in
    
# 1468 "reason_parser.mly"
      (
        let loc = mklocation _symbolstartpos _endpos in
        mkmod ~loc (Pmod_constraint(_4, _2))
      )
# 69434 "reason_parser.ml"
     in
    _menhir_goto_module_binding_body_expr _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce111 : _menhir_env -> (('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _menhir_s, (_1 : 'tv_functor_args), _startpos__1_), _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__module_binding_body_functor = 
# 1476 "reason_parser.mly"
                                          (
      mkFunctorThatReturns _1 _3
    )
# 69449 "reason_parser.ml"
     in
    _menhir_goto__module_binding_body_functor _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce112 : _menhir_env -> (((('ttv_tail * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _menhir_s, (_1 : 'tv_functor_args), _startpos__1_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_module_type), _startpos__3_), _startpos__4_), _endpos__5_, _, (_5 : 'tv_module_expr), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__module_binding_body_functor = 
# 1479 "reason_parser.mly"
                                                                        (
      let loc = mklocation _startpos__5_ _endpos__5_ in
      mkFunctorThatReturns _1 (mkmod ~loc (Pmod_constraint(_5, _3)))
    )
# 69466 "reason_parser.ml"
     in
    _menhir_goto__module_binding_body_functor _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce433 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 69473 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 857 "reason_parser.mly"
       (string)
# 69479 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _v : 'tv_field_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69491 "reason_parser.ml"
      
    in
    
# 3226 "reason_parser.mly"
    ( (_1, _3) )
# 69497 "reason_parser.ml"
     in
    _menhir_goto_field_expr _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce247 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_), _endpos__4_, _, _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69519 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69531 "reason_parser.ml"
      
    in
    
# 2799 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "(") (with_txt _5 ")") )
# 69537 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce211 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _, (_2 : 'tv_expr_comma_list)) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69557 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69569 "reason_parser.ml"
      
    in
    
# 2668 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "(") (with_txt _3 ")") )
# 69575 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce206 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_expr), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69595 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69607 "reason_parser.ml"
      
    in
    
# 2652 "reason_parser.mly"
      ( unclosed_exp (with_txt _1 "(") (with_txt _3 ")") )
# 69613 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce123 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let (((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_) = _menhir_stack in
    let x1 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _4 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69634 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69646 "reason_parser.ml"
      
    in
    
# 1294 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _4 ")") )
# 69652 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce122 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69674 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69686 "reason_parser.ml"
      
    in
    
# 1292 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _5 ")") )
# 69692 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce121 : _menhir_env -> ((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _, (_3 : 'tv_expr), _startpos__3_), _endpos__4_, _, _startpos__4_) = _menhir_stack in
    let x1 = () in
    let _4 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _5 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69714 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69726 "reason_parser.ml"
      
    in
    
# 1290 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _5 ")") )
# 69732 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce120 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_module_expr), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__module_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69752 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69764 "reason_parser.ml"
      
    in
    
# 1288 "reason_parser.mly"
      ( unclosed_mod (with_txt _1 "(") (with_txt _3 ")") )
# 69770 "reason_parser.ml"
     in
    _menhir_goto__module_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce132 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_module_expr), _startpos__3_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__non_arrowed_module_type = 
# 1535 "reason_parser.mly"
      ( mkmty(Pmty_typeof _3) )
# 69784 "reason_parser.ml"
     in
    _menhir_goto__non_arrowed_module_type _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce58 : _menhir_env -> ((((('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 69791 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 69797 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _endpos__4_, _, _startpos__4_), _startpos__5_), _endpos__6_, _, (_6 : 'tv_core_type2), _startpos__6_) = _menhir_stack in
    let _5 = () in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__6_ in
    let _v : 'tv__core_type2 = 
# 3881 "reason_parser.mly"
      ( mktyp(Ptyp_arrow("?" ^ _1 , mkoption _3, _6)) )
# 69807 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce59 : _menhir_env -> (((('ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 69814 "reason_parser.ml"
) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((((_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 69820 "reason_parser.ml"
    )), _startpos__1_), _endpos__2_, _startpos__2_), _endpos__3_, _, (_3 : 'tv_non_arrowed_core_type), _startpos__3_), _, _startpos__4_), _endpos__5_, _, (_5 : 'tv_core_type2), _startpos__5_) = _menhir_stack in
    let _4 = () in
    let _2 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__5_ in
    let _v : 'tv__core_type2 = 
# 3883 "reason_parser.mly"
      ( mktyp(Ptyp_arrow(_1, _3, _5)) )
# 69829 "reason_parser.ml"
     in
    _menhir_goto__core_type2 _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_error2410 : _menhir_env -> 'ttv_tail * _menhir_state * 'tv_package_type_cstr -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce649 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce171 : _menhir_env -> (((((('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let (((((((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, _startpos__2_), _endpos__3_, _startpos__3_), _endpos__4_, _startpos__4_), _endpos__5_, _, (_5 : 'tv_pattern_without_or), _startpos__5_), _endpos__6_, _), _endpos__7_, _, (_7 : 'tv_pattern_without_or), _startpos__7_) = _menhir_stack in
    let x1 = () in
    let _6 = () in
    let _4 = () in
    let _3 = () in
    let _2 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__pattern_without_or = let _8 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69857 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69869 "reason_parser.ml"
      
    in
    
# 3302 "reason_parser.mly"
      ( unclosed_pat (with_txt _1 "(") (with_txt _8 ")") )
# 69875 "reason_parser.ml"
     in
    _menhir_goto__pattern_without_or _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce235 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 69882 "reason_parser.ml"
) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, (x0 : (
# 882 "reason_parser.mly"
       (string)
# 69888 "reason_parser.ml"
    )), _startpos_x0_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_) = _menhir_stack in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 69901 "reason_parser.ml"
      
    in
    
# 2757 "reason_parser.mly"
      (
        mkexp(Pexp_apply(mkoperator _1, ["",_2]))
      )
# 69909 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce634 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_opt_comma = 
# 4230 "reason_parser.mly"
                                                ( () )
# 69918 "reason_parser.ml"
     in
    _menhir_goto_opt_comma _menhir_env _menhir_stack _menhir_s _v

and _menhir_reduce335 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_attributes = 
# 4317 "reason_parser.mly"
               ( [] )
# 69928 "reason_parser.ml"
     in
    _menhir_goto_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce394 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declarations_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constructor_declarations = 
# 3561 "reason_parser.mly"
                                                                                ( List.rev _1 )
# 69939 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce395 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declaration_no_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constructor_declarations = 
# 3562 "reason_parser.mly"
                                                                                ( [_1] )
# 69950 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce396 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_constructor_declaration_no_leading_bar)), _endpos__2_, _, (_2 : 'tv_constructor_declarations_leading_bar)) = _menhir_stack in
    let _endpos = _endpos__2_ in
    let _v : 'tv_constructor_declarations = 
# 3563 "reason_parser.mly"
                                                                                 ( _1 :: List.rev _2 )
# 69961 "reason_parser.ml"
     in
    _menhir_goto_constructor_declarations _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_error2519 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    _menhir_reduce813 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce448 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_generalized_constructor_arguments = 
# 3604 "reason_parser.mly"
                                                ( ([],None) )
# 69975 "reason_parser.ml"
     in
    _menhir_goto_generalized_constructor_arguments _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce673 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos) = Obj.magic _menhir_stack in
    let _v : 'tv_post_item_attributes = 
# 4322 "reason_parser.mly"
               ( [] )
# 69985 "reason_parser.ml"
     in
    _menhir_goto_post_item_attributes _menhir_env _menhir_stack _endpos _menhir_s _v

and _menhir_reduce76 : _menhir_env -> (('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (((_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_), _endpos__2_, _, (_2 : 'tv_simple_expr), _startpos__2_), _endpos__3_, _, (_3 : 'tv_simple_expr), _startpos__3_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__3_ in
    let _v : 'tv__expr = 
# 2551 "reason_parser.mly"
      ( mkexp (Pexp_while(_2, _3)) )
# 69998 "reason_parser.ml"
     in
    _menhir_goto__expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run1 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SEMI ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3069 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1 in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CLASS ->
            _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXTERNAL ->
            _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INCLUDE ->
            _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETATATAT ->
            _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENTPERCENT ->
            _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LET ->
            _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OPEN ->
            _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TYPE ->
            _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EOF | RBRACE | RBRACKET ->
            _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState2557
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2557) : 'freshtv3070)
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF | RBRACE | RBRACKET ->
        _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState1
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1

and _menhir_reduce53 : _menhir_env -> ('ttv_tail * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos_x1_ _ _startpos_x1_ ->
    let ((_menhir_stack, _endpos_x0_, _menhir_s, _startpos_x0_), _endpos__2_, _, (_2 : 'tv_class_expr), _startpos__2_) = _menhir_stack in
    let x1 = () in
    let x0 = () in
    let _startpos = _startpos_x0_ in
    let _endpos = _endpos_x1_ in
    let _v : 'tv__class_simple_expr = let _3 =
      let _endpos_x_ = _endpos_x1_ in
      let _startpos_x_ = _startpos_x1_ in
      let x = x1 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 70222 "reason_parser.ml"
      
    in
    let _1 =
      let _endpos_x_ = _endpos_x0_ in
      let _startpos_x_ = _startpos_x0_ in
      let x = x0 in
      let _endpos = _endpos_x_ in
      let _symbolstartpos = _startpos_x_ in
      
# 1208 "reason_parser.mly"
 ( mkloc x (mklocation _symbolstartpos _endpos) )
# 70234 "reason_parser.ml"
      
    in
    
# 1974 "reason_parser.mly"
      ( unclosed_cl (with_txt _1 "(") (with_txt _3 ")") )
# 70240 "reason_parser.ml"
     in
    _menhir_goto__class_simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce576 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 70247 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 906 "reason_parser.mly"
       (string)
# 70253 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_mod_longident = 
# 4142 "reason_parser.mly"
                                                ( Lident _1 )
# 70260 "reason_parser.ml"
     in
    _menhir_goto_mod_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce604 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv_nonrec_flag = 
# 4194 "reason_parser.mly"
                                                ( Recursive )
# 70271 "reason_parser.ml"
     in
    _menhir_goto_nonrec_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run75 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3067) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_nonrec_flag = 
# 4195 "reason_parser.mly"
                                                ( Nonrecursive )
# 70289 "reason_parser.ml"
     in
    _menhir_goto_nonrec_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3068)

and _menhir_reduce378 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constr_longident = 
# 4121 "reason_parser.mly"
                                                ( Lident "true" )
# 70301 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce369 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 898 "reason_parser.mly"
       (string * string option)
# 70308 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 898 "reason_parser.mly"
       (string * string option)
# 70313 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4041 "reason_parser.mly"
                                      ( let (s, d) = _1 in Const_string (s, d) )
# 70320 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce322 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_additive = 
# 4243 "reason_parser.mly"
                                                ( "+." )
# 70333 "reason_parser.ml"
     in
    _menhir_goto_additive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce321 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_additive = 
# 4242 "reason_parser.mly"
                                                ( "+" )
# 70346 "reason_parser.ml"
     in
    _menhir_goto_additive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run338 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce779 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run339 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce778 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run340 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce777 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run341 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce776 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run342 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce775 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run343 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 70378 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce731 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run344 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce774 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run345 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce773 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run346 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce772 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run347 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce771 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run348 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce770 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run349 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce757 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run350 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce769 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run351 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce768 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run352 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce767 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run353 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce766 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run354 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce765 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run355 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce764 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run356 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce763 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run357 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce762 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run358 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce761 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run359 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce760 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run360 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce759 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run361 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce758 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run362 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 70477 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce730 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run363 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce756 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run364 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce755 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run365 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce754 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run366 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce753 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run367 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce752 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run368 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce751 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run369 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce750 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run370 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce749 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run371 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce748 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run372 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce747 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run373 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce746 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run374 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce745 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run375 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce744 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run376 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce743 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run377 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce742 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run378 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce741 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run379 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce740 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run380 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce739 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run381 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce738 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run382 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce737 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run383 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce736 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run384 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce735 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run385 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce734 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run386 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce733 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run387 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce732 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_reduce644 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv_override_flag = 
# 4222 "reason_parser.mly"
                                                ( Fresh )
# 70615 "reason_parser.ml"
     in
    _menhir_goto_override_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run393 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3065) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_override_flag = 
# 4223 "reason_parser.mly"
                                                ( Override )
# 70633 "reason_parser.ml"
     in
    _menhir_goto_override_flag _menhir_env _menhir_stack _endpos _menhir_s _v _startpos) : 'freshtv3066)

and _menhir_run8 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 70640 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce350 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_reduce373 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 871 "reason_parser.mly"
       (nativeint)
# 70649 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 871 "reason_parser.mly"
       (nativeint)
# 70654 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4045 "reason_parser.mly"
                                      ( Const_nativeint _1 )
# 70661 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce796 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_subtractive = 
# 4239 "reason_parser.mly"
                                                ( "-." )
# 70674 "reason_parser.ml"
     in
    _menhir_goto_subtractive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce795 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_subtractive = 
# 4238 "reason_parser.mly"
                                                ( "-" )
# 70687 "reason_parser.ml"
     in
    _menhir_goto_subtractive _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run187 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 882 "reason_parser.mly"
       (string)
# 70694 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState187 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce607 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState187

and _menhir_run409 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_reduce615 _menhir_env (Obj.magic _menhir_stack)
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce322 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3063 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3064)

and _menhir_run410 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        _menhir_reduce614 _menhir_env (Obj.magic _menhir_stack)
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce321 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3061 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3062)

and _menhir_run772 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState772 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState772 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState772

and _menhir_run787 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce796 _menhir_env (Obj.magic _menhir_stack)
    | RPAREN ->
        _menhir_reduce617 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3059 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3060)

and _menhir_run788 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT | BACKQUOTE | BANG | CHAR _ | FALSE | FLOAT _ | FOR | FUN | IF | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | MINUSDOT | NATIVEINT _ | NEW | PLUS | PLUSDOT | PREFIXOP _ | STRING _ | SWITCH | TRUE | TRY | UIDENT _ | WHILE ->
        _menhir_reduce795 _menhir_env (Obj.magic _menhir_stack)
    | RPAREN ->
        _menhir_reduce616 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3057 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv3058)

and _menhir_run805 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState805 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce608 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState805

and _menhir_reduce460 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 70901 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 70907 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_label = 
# 4033 "reason_parser.mly"
                                                ( _1 )
# 70914 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv3055) = _menhir_stack in
    let (_endpos : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv_label) = _v in
    let (_startpos : Lexing.position) = _startpos in
    ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    match _menhir_s with
    | MenhirState288 | MenhirState284 | MenhirState147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2997 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error152 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv2995 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run154 _menhir_env (Obj.magic _menhir_stack) MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState153 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState153) : 'freshtv2996)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error152 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv2998)
    | MenhirState743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2999 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce243 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3000)
    | MenhirState749 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3001 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce243 _menhir_env (Obj.magic _menhir_stack) : 'freshtv3002)
    | MenhirState0 | MenhirState1 | MenhirState2557 | MenhirState15 | MenhirState73 | MenhirState175 | MenhirState178 | MenhirState184 | MenhirState190 | MenhirState334 | MenhirState397 | MenhirState405 | MenhirState408 | MenhirState413 | MenhirState415 | MenhirState422 | MenhirState432 | MenhirState2326 | MenhirState2320 | MenhirState440 | MenhirState452 | MenhirState454 | MenhirState456 | MenhirState2274 | MenhirState473 | MenhirState2270 | MenhirState475 | MenhirState2259 | MenhirState2255 | MenhirState2253 | MenhirState2233 | MenhirState2231 | MenhirState2222 | MenhirState2218 | MenhirState2210 | MenhirState737 | MenhirState2182 | MenhirState755 | MenhirState2033 | MenhirState1475 | MenhirState1477 | MenhirState1939 | MenhirState1929 | MenhirState1846 | MenhirState1922 | MenhirState1919 | MenhirState1909 | MenhirState1906 | MenhirState1900 | MenhirState1893 | MenhirState1734 | MenhirState1728 | MenhirState1725 | MenhirState1718 | MenhirState1712 | MenhirState1708 | MenhirState1688 | MenhirState1685 | MenhirState1147 | MenhirState1142 | MenhirState771 | MenhirState1134 | MenhirState1123 | MenhirState1115 | MenhirState1084 | MenhirState1081 | MenhirState798 | MenhirState813 | MenhirState1059 | MenhirState1056 | MenhirState1051 | MenhirState1047 | MenhirState1045 | MenhirState1002 | MenhirState925 | MenhirState891 | MenhirState886 | MenhirState884 | MenhirState882 | MenhirState880 | MenhirState878 | MenhirState876 | MenhirState874 | MenhirState872 | MenhirState870 | MenhirState868 | MenhirState866 | MenhirState864 | MenhirState862 | MenhirState860 | MenhirState858 | MenhirState856 | MenhirState854 | MenhirState852 | MenhirState850 | MenhirState847 | MenhirState837 | MenhirState835 | MenhirState833 | MenhirState816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3005 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error832 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3003 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState833 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState833) : 'freshtv3004)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error832 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3006)
    | MenhirState2753 | MenhirState2757 | MenhirState2760 | MenhirState2743 | MenhirState2582 | MenhirState2574 | MenhirState2422 | MenhirState2376 | MenhirState2374 | MenhirState416 | MenhirState418 | MenhirState2277 | MenhirState2229 | MenhirState2213 | MenhirState2192 | MenhirState2187 | MenhirState2188 | MenhirState763 | MenhirState2147 | MenhirState1155 | MenhirState2134 | MenhirState1949 | MenhirState1161 | MenhirState1237 | MenhirState1234 | MenhirState1224 | MenhirState1221 | MenhirState1213 | MenhirState1204 | MenhirState1197 | MenhirState1193 | MenhirState1176 | MenhirState1171 | MenhirState1167 | MenhirState770 | MenhirState1138 | MenhirState791 | MenhirState1103 | MenhirState1098 | MenhirState1099 | MenhirState1076 | MenhirState1063 | MenhirState1053 | MenhirState893 | MenhirState897 | MenhirState1036 | MenhirState1034 | MenhirState1032 | MenhirState1024 | MenhirState907 | MenhirState922 | MenhirState1018 | MenhirState1015 | MenhirState1013 | MenhirState1011 | MenhirState1008 | MenhirState1006 | MenhirState999 | MenhirState997 | MenhirState995 | MenhirState993 | MenhirState991 | MenhirState989 | MenhirState987 | MenhirState985 | MenhirState983 | MenhirState981 | MenhirState979 | MenhirState977 | MenhirState975 | MenhirState973 | MenhirState971 | MenhirState969 | MenhirState967 | MenhirState965 | MenhirState963 | MenhirState960 | MenhirState950 | MenhirState948 | MenhirState946 | MenhirState929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3009 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error945 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3007 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState946 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState946) : 'freshtv3008)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error945 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3010)
    | MenhirState1554 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3013 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1555 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3011 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1556 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1556) : 'freshtv3012)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1555 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3014)
    | MenhirState1559 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3017 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1560 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3015 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1561 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1561) : 'freshtv3016)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1560 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3018)
    | MenhirState1552 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3021 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1565 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : 'freshtv3019 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1566 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1566) : 'freshtv3020)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1565 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3022)
    | MenhirState1573 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3025 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1574 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3023 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run1576 _menhir_env (Obj.magic _menhir_stack) MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1575 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1575) : 'freshtv3024)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1574 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3026)
    | MenhirState1669 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3029 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1670 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3027 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1671 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1671) : 'freshtv3028)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1670 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3030)
    | MenhirState1678 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3033 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1679 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv3031 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run115 _menhir_env (Obj.magic _menhir_stack) MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1680 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1680) : 'freshtv3032)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1679 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3034)
    | MenhirState1683 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3037 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1684
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              _menhir_run1068 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1684 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLONGREATER ->
              _menhir_run1066 _menhir_env (Obj.magic _menhir_stack) MenhirState1684
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3035 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1684 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1685 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1685) : 'freshtv3036)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1684)) : 'freshtv3038)
    | MenhirState1693 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3041 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1694 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv3039 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run1576 _menhir_env (Obj.magic _menhir_stack) MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1695 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1695) : 'freshtv3040)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1694 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3042)
    | MenhirState1698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv3045 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_error1699 _menhir_env (Obj.magic _menhir_stack)
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : ((('freshtv3043 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run306 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run301 _menhir_env (Obj.magic _menhir_stack) MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run291 _menhir_env (Obj.magic _menhir_stack) MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run284 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run283 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run282 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run323 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run1576 _menhir_env (Obj.magic _menhir_stack) MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run243 _menhir_env (Obj.magic _menhir_stack) MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run107 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1700 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1700) : 'freshtv3044)
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_error1699 _menhir_env (Obj.magic _menhir_stack))) : 'freshtv3046)
    | MenhirState1702 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv3053 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((if _menhir_env._menhir_error then
          _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1703
        else
          let _tok = _menhir_env._menhir_token in
          match _tok with
          | BACKQUOTE ->
              _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | CHAR _v ->
              _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | COLON ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3049 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1703 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | LBRACKET ->
                  _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETGREATER ->
                  _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETLESS ->
                  _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESS ->
                  _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSDOTDOTGREATER ->
                  _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LESSGREATER ->
                  _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | QUOTE ->
                  _menhir_run154 _menhir_env (Obj.magic _menhir_stack) MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SHARP ->
                  _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TYPE ->
                  let (_menhir_env : _menhir_env) = _menhir_env in
                  let (_menhir_stack : ((('freshtv3047 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                  let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                  let (_menhir_s : _menhir_state) = MenhirState1720 in
                  let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                  ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                  let _menhir_env = _menhir_discard _menhir_env in
                  let _tok = _menhir_env._menhir_token in
                  match _tok with
                  | LIDENT _v ->
                      _menhir_run1208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1721 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                  | _ ->
                      assert (not _menhir_env._menhir_error);
                      _menhir_env._menhir_error <- true;
                      _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1721) : 'freshtv3048)
              | UIDENT _v ->
                  _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UNDERSCORE ->
                  _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1720 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1720) : 'freshtv3050)
          | EQUAL ->
              let (_menhir_env : _menhir_env) = _menhir_env in
              let (_menhir_stack : (('freshtv3051 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
              let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
              let (_menhir_s : _menhir_state) = MenhirState1703 in
              let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
              ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
              let _menhir_env = _menhir_discard _menhir_env in
              let _tok = _menhir_env._menhir_token in
              match _tok with
              | ASSERT ->
                  _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BACKQUOTE ->
                  _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | BANG ->
                  _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | CHAR _v ->
                  _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FALSE ->
                  _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FLOAT _v ->
                  _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FOR ->
                  _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | FUN ->
                  _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | IF ->
                  _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT _v ->
                  _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT32 _v ->
                  _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | INT64 _v ->
                  _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LAZY ->
                  _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACE ->
                  _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACELESS ->
                  _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKET ->
                  _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETBAR ->
                  _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LBRACKETPERCENT ->
                  _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LIDENT _v ->
                  _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | LPAREN ->
                  _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUS ->
                  _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | MINUSDOT ->
                  _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NATIVEINT _v ->
                  _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | NEW ->
                  _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUS ->
                  _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PLUSDOT ->
                  _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | PREFIXOP _v ->
                  _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | STRING _v ->
                  _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | SWITCH ->
                  _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRUE ->
                  _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | TRY ->
                  _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | UIDENT _v ->
                  _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | WHILE ->
                  _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1718 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
              | _ ->
                  assert (not _menhir_env._menhir_error);
                  _menhir_env._menhir_error <- true;
                  _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1718) : 'freshtv3052)
          | EQUALGREATER ->
              _menhir_run1708 _menhir_env (Obj.magic _menhir_stack) MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FALSE ->
              _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | FLOAT _v ->
              _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT _v ->
              _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT32 _v ->
              _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | INT64 _v ->
              _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACE ->
              _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKET ->
              _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETBAR ->
              _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LBRACKETPERCENT ->
              _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LIDENT _v ->
              _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | LPAREN ->
              _menhir_run1704 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | MINUS ->
              _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | NATIVEINT _v ->
              _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | PLUS ->
              _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | SHARP ->
              _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | STRING _v ->
              _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | TRUE ->
              _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UIDENT _v ->
              _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | UNDERSCORE ->
              _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1703 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
          | _ ->
              (assert (not _menhir_env._menhir_error);
              _menhir_env._menhir_error <- true;
              _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1703)) : 'freshtv3054)
    | _ ->
        _menhir_fail ()) : 'freshtv3056)

and _menhir_reduce834 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 71867 "reason_parser.ml"
) * Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack ->
    let (_menhir_stack, _endpos__1_, _menhir_s, (_1 : (
# 857 "reason_parser.mly"
       (string)
# 71873 "reason_parser.ml"
    )), _startpos__1_) = _menhir_stack in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_val_ident = 
# 4069 "reason_parser.mly"
                                                ( _1 )
# 71880 "reason_parser.ml"
     in
    _menhir_goto_val_ident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce686 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_rec_flag = 
# 4190 "reason_parser.mly"
                                                ( Nonrecursive )
# 71889 "reason_parser.ml"
     in
    _menhir_goto_rec_flag _menhir_env _menhir_stack _menhir_s _v

and _menhir_run435 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2993) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_rec_flag = 
# 4191 "reason_parser.mly"
                                                ( Recursive )
# 71905 "reason_parser.ml"
     in
    _menhir_goto_rec_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2994)

and _menhir_run2637 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 71912 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2637 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2637

and _menhir_reduce230 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = 
# 2737 "reason_parser.mly"
      ( mkexp (Pexp_array []) )
# 71940 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run20 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce779 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run21 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce778 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run22 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce777 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run23 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce776 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run24 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce775 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run25 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 71972 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce731 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run26 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce774 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run27 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce773 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run28 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce772 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run29 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce771 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run30 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce770 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run31 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce757 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run32 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce769 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run33 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce768 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run34 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce767 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run35 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce766 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run36 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce765 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run37 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce764 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run38 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce763 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run39 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce762 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run40 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce761 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run41 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce760 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run42 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce759 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run43 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce758 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run44 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 72071 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce730 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run45 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce756 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run46 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce755 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run47 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce754 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run48 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce753 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run49 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce752 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run50 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce751 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run51 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce750 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run52 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce749 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run53 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce748 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run54 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce747 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run55 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce746 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run56 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce745 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run57 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce744 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run58 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce743 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run59 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce742 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run60 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce741 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run61 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce740 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run62 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce739 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run63 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce738 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run64 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce737 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run65 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce736 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run66 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce735 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run67 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce734 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run68 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce733 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run69 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce732 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run604 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce375 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s

and _menhir_run1099 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1099 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1099

and _menhir_run421 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 72287 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2991 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 72299 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState422 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState422) : 'freshtv2992)
    | COMMA | GREATERRBRACE ->
        _menhir_reduce434 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_error421 _menhir_env (Obj.magic _menhir_stack)

and _menhir_reduce240 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__2_ _ ->
    let (_menhir_stack, _endpos__1_, _menhir_s, _startpos__1_) = _menhir_stack in
    let _2 = () in
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__2_ in
    let _v : 'tv__simple_expr = 
# 2778 "reason_parser.mly"
      ( mkexp (Pexp_override []))
# 72394 "reason_parser.ml"
     in
    _menhir_goto__simple_expr _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce38 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_, _endpos__0_) = Obj.magic _menhir_stack in
    let (_, _startpos) = Obj.magic _menhir_stack in
    let _endpos = _startpos in
    let _v : 'tv__class_self_pattern = let _endpos = _endpos__0_ in
    let _symbolstartpos = _endpos in
    
# 1902 "reason_parser.mly"
    (
    let loc = mklocation _symbolstartpos _endpos in
    mkpat (Ppat_var (mkloc "this" loc))
  )
# 72411 "reason_parser.ml"
     in
    _menhir_goto__class_self_pattern _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run433 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 72418 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLONCOLON | COLONEQUAL | COLONGREATER | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)
    | COLON | COMMA ->
        _menhir_reduce466 _menhir_env (Obj.magic _menhir_stack)
    | EQUAL ->
        _menhir_reduce460 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run434 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2987 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState434 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | UIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2983 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 906 "reason_parser.mly"
       (string)
# 72455 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1247 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EQUAL ->
                _menhir_run1245 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2323 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2323) : 'freshtv2984)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2985 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2986)) : 'freshtv2988)
    | OPEN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2989 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState434 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BANG ->
            _menhir_run393 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState436 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _ ->
            _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack) MenhirState436
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState436) : 'freshtv2990)
    | REC ->
        _menhir_run435 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState434 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce686 _menhir_env (Obj.magic _menhir_stack) MenhirState434
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState434

and _menhir_run1115 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1115 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1115

and _menhir_run1657 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1657 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1657

and _menhir_reduce372 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 841 "reason_parser.mly"
       (int64)
# 72643 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 841 "reason_parser.mly"
       (int64)
# 72648 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4044 "reason_parser.mly"
                                      ( Const_int64 _1 )
# 72655 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce371 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 840 "reason_parser.mly"
       (int32)
# 72662 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 840 "reason_parser.mly"
       (int32)
# 72667 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4043 "reason_parser.mly"
                                      ( Const_int32 _1 )
# 72674 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce367 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 839 "reason_parser.mly"
       (int)
# 72681 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 839 "reason_parser.mly"
       (int)
# 72686 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4039 "reason_parser.mly"
                                      ( Const_int _1 )
# 72693 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run412 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run2164 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState412 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2981 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState412 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | ASSERT ->
            _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FOR ->
            _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FUN ->
            _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | IF ->
            _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACELESS ->
            _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NEW ->
            _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SWITCH ->
            _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRY ->
            _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | WHILE ->
            _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState413 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState413) : 'freshtv2982)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState412

and _menhir_run452 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACE ->
        _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState452
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState452

and _menhir_run757 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run444 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState757 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState757

and _menhir_run211 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce271 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run903 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2979 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState903 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2975 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 857 "reason_parser.mly"
       (string)
# 73030 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | RPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2971 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 73042 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | COLON ->
                    _menhir_run1022 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EQUALGREATER ->
                    _menhir_run907 _menhir_env (Obj.magic _menhir_stack) MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState906 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState906) : 'freshtv2972)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2973 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 73107 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2974)) : 'freshtv2976)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2977 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2978)) : 'freshtv2980)
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState903 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState903

and _menhir_run907 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState907 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState907

and _menhir_run1022 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1022 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1022

and _menhir_run185 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2969) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4083 "reason_parser.mly"
                                                ( "*" )
# 73255 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 73261 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2970)

and _menhir_run212 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run213 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState212 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState212

and _menhir_run218 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce376 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run219 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 882 "reason_parser.mly"
       (string)
# 73288 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce607 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run220 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2967) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4091 "reason_parser.mly"
                                                ( "+=" )
# 73309 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 73315 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2968)

and _menhir_run221 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce615 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run222 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run227 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run226 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run225 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run224 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run223 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce614 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2965 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2966)

and _menhir_run228 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2963) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4092 "reason_parser.mly"
                                                ( "%" )
# 73365 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 73371 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2964)

and _menhir_run229 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2961) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4086 "reason_parser.mly"
                                                ( "or" )
# 73389 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 73395 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2962)

and _menhir_run230 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | UIDENT _v ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2957 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_v : (
# 906 "reason_parser.mly"
       (string)
# 73412 "reason_parser.ml"
        )) = _v in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv2951 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 73424 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run234 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState233 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState233) : 'freshtv2952)
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv2953 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 73445 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_env = _menhir_discard _menhir_env in
            _menhir_reduce290 _menhir_env (Obj.magic _menhir_stack) _endpos _startpos) : 'freshtv2954)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : (('freshtv2955 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 73458 "reason_parser.ml"
            ) * Lexing.position) = Obj.magic _menhir_stack in
            ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2956)) : 'freshtv2958)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2959 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2960)

and _menhir_run488 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce617 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run489 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run494 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run493 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run492 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run491 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run490 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_reduce616 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2949 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2950)

and _menhir_run495 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2947 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState495 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2943 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2939 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EXCEPTION ->
                    _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState711 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState711) : 'freshtv2940)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2941 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2942)) : 'freshtv2944)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2945 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2946)) : 'freshtv2948)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run496 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState495 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState495

and _menhir_run188 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 73700 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run508 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2937) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4093 "reason_parser.mly"
                                                ( "<>" )
# 73721 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 73727 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2938)

and _menhir_run509 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2935) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4094 "reason_parser.mly"
                                                ( "<..>" )
# 73745 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 73751 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2936)

and _menhir_run510 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2933) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4084 "reason_parser.mly"
                                                ( "<" )
# 73769 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 73775 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2934)

and _menhir_run511 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2931 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState511 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce282 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2932)
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState511 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState511

and _menhir_run514 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState514 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState514

and _menhir_run516 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState516 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState516 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState516

and _menhir_run532 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run533 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState532 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState532

and _menhir_run534 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 836 "reason_parser.mly"
       (string)
# 73973 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2929) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 836 "reason_parser.mly"
       (string)
# 73984 "reason_parser.ml"
    )) : (
# 836 "reason_parser.mly"
       (string)
# 73988 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4078 "reason_parser.mly"
                                                ( _1 )
# 73996 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74002 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2930)

and _menhir_run535 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 835 "reason_parser.mly"
       (string)
# 74009 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2927) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 835 "reason_parser.mly"
       (string)
# 74020 "reason_parser.ml"
    )) : (
# 835 "reason_parser.mly"
       (string)
# 74024 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4077 "reason_parser.mly"
                                                ( _1 )
# 74032 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74038 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2928)

and _menhir_run536 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 834 "reason_parser.mly"
       (string)
# 74045 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2925) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 834 "reason_parser.mly"
       (string)
# 74056 "reason_parser.ml"
    )) : (
# 834 "reason_parser.mly"
       (string)
# 74060 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4076 "reason_parser.mly"
                                                ( _1 )
# 74068 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74074 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2926)

and _menhir_run537 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 833 "reason_parser.mly"
       (string)
# 74081 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2923) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 833 "reason_parser.mly"
       (string)
# 74092 "reason_parser.ml"
    )) : (
# 833 "reason_parser.mly"
       (string)
# 74096 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4075 "reason_parser.mly"
                                                ( _1 )
# 74104 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74110 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2924)

and _menhir_run538 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 832 "reason_parser.mly"
       (string)
# 74117 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2921) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_10 : (
# 832 "reason_parser.mly"
       (string)
# 74128 "reason_parser.ml"
    )) : (
# 832 "reason_parser.mly"
       (string)
# 74132 "reason_parser.ml"
    )) = _v in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4074 "reason_parser.mly"
                                                ( _1 )
# 74140 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74146 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2922)

and _menhir_run539 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2919) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4085 "reason_parser.mly"
                                                ( ">" )
# 74164 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74170 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2920)

and _menhir_run540 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState540 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState540

and _menhir_run568 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2917) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4090 "reason_parser.mly"
                                                ( ":=" )
# 74245 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74251 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2918)

and _menhir_run569 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2915) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4087 "reason_parser.mly"
                                                ( "||" )
# 74269 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74275 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2916)

and _menhir_run570 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce608 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run571 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2913) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4088 "reason_parser.mly"
                                                ( "&" )
# 74299 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74305 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2914)

and _menhir_run572 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2911) = Obj.magic _menhir_stack in
    let (_endpos__10_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__10_ : Lexing.position) = _startpos in
    ((let _10 = () in
    let _v : 'tv_operator = let _1 =
      let _1 = _10 in
      
# 4089 "reason_parser.mly"
                                                ( "&&" )
# 74323 "reason_parser.ml"
      
    in
    
# 4099 "reason_parser.mly"
                                                ( _1 )
# 74329 "reason_parser.ml"
     in
    _menhir_goto_operator _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2912)

and _menhir_run485 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 74336 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2909 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 74348 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState486 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState486) : 'freshtv2910)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run767 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState767 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState767

and _menhir_reduce370 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 821 "reason_parser.mly"
       (string)
# 74465 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 821 "reason_parser.mly"
       (string)
# 74470 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4042 "reason_parser.mly"
                                      ( Const_float _1 )
# 74477 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_reduce377 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s _startpos__1_ ->
    let _1 = () in
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constr_longident = 
# 4120 "reason_parser.mly"
                                                ( Lident "false" )
# 74489 "reason_parser.ml"
     in
    _menhir_goto_constr_longident _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run660 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run581 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run580 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState660

and _menhir_run2589 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 74553 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2907 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 74567 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2589 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2590 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2590) : 'freshtv2908)
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2589 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2589
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2589

and _menhir_run2595 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2905 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2595 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2596 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2596 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2596 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2596 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2596 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2596) : 'freshtv2906)
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2595 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2595
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2595

and _menhir_run2599 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | RPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2901 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | COLON ->
            _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EQUAL ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2899 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2600 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | FALSE ->
                _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LBRACKET ->
                _menhir_run2592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | LPAREN ->
                _menhir_run2591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | TRUE ->
                _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2601 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | UIDENT _v ->
                _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2601 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2601) : 'freshtv2900)
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OF ->
            _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2600 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
            _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2600
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2600) : 'freshtv2902)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2903 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2904)

and _menhir_run2604 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2897 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2604 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2605 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2605 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2605 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2605 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2605 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2605) : 'freshtv2898)
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2604 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2604
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2604

and _menhir_run2608 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLON ->
        _menhir_run1275 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EQUAL ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2895 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2608 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | FALSE ->
            _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2609 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run2592 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2609 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run2591 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2609 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2609 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2609 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2609) : 'freshtv2896)
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OF ->
        _menhir_run1270 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2608 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR | EOF | LBRACKETAT | LBRACKETATAT | SEMI ->
        _menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2608
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2608

and _menhir_goto__use_file : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv__use_file -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2893) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__use_file) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2891) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv__use_file) : 'tv__use_file) = _v in
    ((let _v : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74922 "reason_parser.ml"
    ) = 
# 1129 "reason_parser.mly"
                    (apply_mapper_chain_to_use_file _1 default_mapper_chain)
# 74926 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2889) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74934 "reason_parser.ml"
    )) = _v in
    ((match _menhir_s with
    | MenhirState2753 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2879) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74944 "reason_parser.ml"
        )) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2877) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let ((_1 : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74952 "reason_parser.ml"
        )) : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74956 "reason_parser.ml"
        )) = _v in
        (Obj.magic _1 : 'freshtv2878)) : 'freshtv2880)
    | MenhirState2757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2883 * _menhir_state * 'tv_toplevel_directive) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74966 "reason_parser.ml"
        )) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2881 * _menhir_state * 'tv_toplevel_directive) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74974 "reason_parser.ml"
        )) : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74978 "reason_parser.ml"
        )) = _v in
        ((let ((_menhir_stack, _menhir_s, (_1 : 'tv_toplevel_directive)), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv__use_file = 
# 1133 "reason_parser.mly"
                                                   ( _1 :: _3 )
# 74985 "reason_parser.ml"
         in
        _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2882)) : 'freshtv2884)
    | MenhirState2760 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2887 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        let (_v : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 74995 "reason_parser.ml"
        )) = _v in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2885 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        let (_ : _menhir_state) = _menhir_s in
        let ((_3 : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 75003 "reason_parser.ml"
        )) : (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 75007 "reason_parser.ml"
        )) = _v in
        ((let ((_menhir_stack, _endpos__1_, _menhir_s, (_1 : 'tv_structure_item), _startpos__1_), _endpos__2_) = _menhir_stack in
        let _2 = () in
        let _v : 'tv__use_file = 
# 1132 "reason_parser.mly"
                                                   ( Ptop_def[_1] :: _3 )
# 75014 "reason_parser.ml"
         in
        _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2886)) : 'freshtv2888)
    | _ ->
        _menhir_fail ()) : 'freshtv2890)) : 'freshtv2892)) : 'freshtv2894)

and _menhir_reduce847 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_virtual_flag = 
# 4210 "reason_parser.mly"
                                                ( Concrete )
# 75025 "reason_parser.ml"
     in
    _menhir_goto_virtual_flag _menhir_env _menhir_stack _menhir_s _v

and _menhir_run1535 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2875) = Obj.magic _menhir_stack in
    let (_endpos__1_ : Lexing.position) = _endpos in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_startpos__1_ : Lexing.position) = _startpos in
    ((let _1 = () in
    let _v : 'tv_virtual_flag = 
# 4211 "reason_parser.mly"
                                                ( Virtual )
# 75041 "reason_parser.ml"
     in
    _menhir_goto_virtual_flag _menhir_env _menhir_stack _menhir_s _v) : 'freshtv2876)

and _menhir_run2052 : _menhir_env -> 'ttv_tail * Lexing.position * _menhir_state * Lexing.position -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | VIRTUAL ->
        _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2052 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState2052
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2052

and _menhir_reduce368 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 800 "reason_parser.mly"
       (char)
# 75063 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos__1_ _menhir_s (_1 : (
# 800 "reason_parser.mly"
       (char)
# 75068 "reason_parser.ml"
  )) _startpos__1_ ->
    let _startpos = _startpos__1_ in
    let _endpos = _endpos__1_ in
    let _v : 'tv_constant = 
# 4040 "reason_parser.mly"
                                      ( Const_char _1 )
# 75075 "reason_parser.ml"
     in
    _menhir_goto_constant _menhir_env _menhir_stack _endpos _menhir_s _v _startpos

and _menhir_run125 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 75082 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce453 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run126 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 75092 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce454 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run15 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run805 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run772 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState15 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState15

and _menhir_reduce787 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_structure = 
# 1373 "reason_parser.mly"
                ( [] )
# 75218 "reason_parser.ml"
     in
    _menhir_goto_structure _menhir_env _menhir_stack _menhir_s _v

and _menhir_run2 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2

and _menhir_run16 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 75278 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce576 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run74 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | NONREC ->
        _menhir_run75 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState74 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce604 _menhir_env (Obj.magic _menhir_stack) MenhirState74

and _menhir_run17 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState17 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState17

and _menhir_run176 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce378 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run177 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState177 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState177

and _menhir_run179 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 898 "reason_parser.mly"
       (string * string option)
# 75410 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce369 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run180 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 882 "reason_parser.mly"
       (string)
# 75419 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState180 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState180

and _menhir_run392 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run393 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState392 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _ ->
        _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack) MenhirState392
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState392

and _menhir_run181 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run110 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState181 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState181

and _menhir_run183 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 871 "reason_parser.mly"
       (nativeint)
# 75506 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce373 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run398 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2871 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState399 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState399 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState399) : 'freshtv2872)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2873 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2874)

and _menhir_run408 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run805 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2869 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState408 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2865 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2861 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | ASSERT ->
                    _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BANG ->
                    _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FOR ->
                    _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FUN ->
                    _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | IF ->
                    _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACELESS ->
                    _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUSDOT ->
                    _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NEW ->
                    _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUSDOT ->
                    _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PREFIXOP _v ->
                    _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SWITCH ->
                    _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRY ->
                    _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | WHILE ->
                    _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2374 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2374) : 'freshtv2862)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2863 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2864)) : 'freshtv2866)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2867 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2868)) : 'freshtv2870)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run411 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run186 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState408 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState408

and _menhir_run414 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 75773 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)
    | EQUAL ->
        _menhir_reduce460 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run441 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2859 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState441 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | REC ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2857 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState442 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2309 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2309) : 'freshtv2858)
        | UIDENT _v ->
            _menhir_run443 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState442 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState442) : 'freshtv2860)
    | REC ->
        _menhir_run435 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState441 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce686 _menhir_env (Obj.magic _menhir_stack) MenhirState441
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState441

and _menhir_run453 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState453 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState453

and _menhir_run189 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState189 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState189

and _menhir_run415 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2855 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState415 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce230 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2856)
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState415

and _menhir_run455 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState455 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState455

and _menhir_run416 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run2188 _menhir_env (Obj.magic _menhir_stack) MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run515 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState416 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState416

and _menhir_run423 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | GREATERRBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2853 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState423 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce240 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2854)
    | LIDENT _v ->
        _menhir_run421 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState423 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState423

and _menhir_run432 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1657 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run1115 _menhir_env (Obj.magic _menhir_stack) MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState432 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState432
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState432

and _menhir_run457 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState457 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState457

and _menhir_run458 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 841 "reason_parser.mly"
       (int64)
# 76493 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce372 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run459 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 840 "reason_parser.mly"
       (int32)
# 76502 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce371 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run460 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 839 "reason_parser.mly"
       (int)
# 76511 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce367 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run1476 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run1943 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run1477 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState1476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1476

and _menhir_run474 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState474

and _menhir_run476 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2851 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState476 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | AMPERAMPER ->
            _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERSAND ->
            _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BARBAR ->
            _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONEQUAL ->
            _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP0 _v ->
            _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP1 _v ->
            _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP2 _v ->
            _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP3 _v ->
            _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP4 _v ->
            _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OR ->
            _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSEQ ->
            _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RPAREN ->
            _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STAR ->
            _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TYPE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2849 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState477 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2845 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_v : (
# 857 "reason_parser.mly"
       (string)
# 76738 "reason_parser.ml"
                )) = _v in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | RPAREN ->
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv2841 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 76750 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                    ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                    let _menhir_env = _menhir_discard _menhir_env in
                    let _tok = _menhir_env._menhir_token in
                    match _tok with
                    | BACKQUOTE ->
                        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | CHAR _v ->
                        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | COLON ->
                        _menhir_run2220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | EQUALGREATER ->
                        _menhir_run2218 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FALSE ->
                        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FLOAT _v ->
                        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT _v ->
                        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT32 _v ->
                        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT64 _v ->
                        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACE ->
                        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKET ->
                        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETBAR ->
                        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETPERCENT ->
                        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LIDENT _v ->
                        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LPAREN ->
                        _menhir_run481 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | MINUS ->
                        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | NATIVEINT _v ->
                        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | PLUS ->
                        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | SHARP ->
                        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | STRING _v ->
                        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | TRUE ->
                        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UIDENT _v ->
                        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UNDERSCORE ->
                        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState480 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | _ ->
                        assert (not _menhir_env._menhir_error);
                        _menhir_env._menhir_error <- true;
                        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState480) : 'freshtv2842)
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv2843 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 76815 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2844)) : 'freshtv2846)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2847 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2848)) : 'freshtv2850)
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState477) : 'freshtv2852)
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState476 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState476

and _menhir_run738 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState738 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState738

and _menhir_run461 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 821 "reason_parser.mly"
       (string)
# 76915 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce370 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run462 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce377 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run1478 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1478 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1478 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1478

and _menhir_run1489 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        _menhir_run1524 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1489 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run1518 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1489 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1489 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run1505 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1489 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1490 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1489 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1489

and _menhir_run1534 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2839 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1534 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | VIRTUAL ->
            _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1536 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _ ->
            _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState1536
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1536) : 'freshtv2840)
    | VIRTUAL ->
        _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1534 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState1534
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1534

and _menhir_run463 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 800 "reason_parser.mly"
       (char)
# 76998 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce368 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run464 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState464

and _menhir_run465 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run94 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState465 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run93 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState465 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState465

and _menhir_run811 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run184 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState811 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState811

and _menhir_reduce704 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let _v : 'tv_signature = 
# 1665 "reason_parser.mly"
                         ( [] )
# 77130 "reason_parser.ml"
     in
    _menhir_goto_signature _menhir_env _menhir_stack _menhir_s _v

and _menhir_run1252 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | NONREC ->
        _menhir_run75 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1252 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce604 _menhir_env (Obj.magic _menhir_stack) MenhirState1252

and _menhir_run1404 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2835 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1405 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1405) : 'freshtv2836)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2837 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2838)

and _menhir_run1409 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1409 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1409 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2833 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1409 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | REC ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run1465 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1464 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1464) : 'freshtv2826)
        | UIDENT _v ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2829 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_v : (
# 906 "reason_parser.mly"
       (string)
# 77218 "reason_parser.ml"
            )) = _v in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | COLON ->
                _menhir_run1442 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | EQUAL ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 77232 "reason_parser.ml"
                ) * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_menhir_s : _menhir_state) = MenhirState1411 in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1459 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1459) : 'freshtv2828)
            | LPAREN ->
                _menhir_run1412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1411 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1411) : 'freshtv2830)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2831 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2832)) : 'freshtv2834)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1409

and _menhir_run2036 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACE ->
        _menhir_run1408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1248 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run450 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2036 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run130 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2036 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2036

and _menhir_run2039 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2039 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2039 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2039

and _menhir_run2048 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        _menhir_run1389 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run1386 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run1382 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run1379 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run1376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2048 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2048

and _menhir_run2051 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        _menhir_run2052 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2051 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState2051
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2051

and _menhir_run119 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce140 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run77 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 77349 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LPAREN ->
        _menhir_run78 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState77 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AND | DOT | EOF | EQUAL | EQUALGREATER | LBRACKETAT | LBRACKETATAT | RPAREN | SEMI | WITH ->
        _menhir_reduce573 _menhir_env (Obj.magic _menhir_stack)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState77

and _menhir_run120 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState120 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState120

and _menhir_run124 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState124 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState124 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState124

and _menhir_run128 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run129 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState128 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState128

and _menhir_run142 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 77437 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2823 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 77449 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState143 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState143) : 'freshtv2824)
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce823 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run145 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce144 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run146 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce143 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run147 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | DOTDOT ->
        _menhir_run149 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState147
    | LIDENT _v ->
        _menhir_run148 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState147 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState147

and _menhir_run156 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run157 _menhir_env (Obj.magic _menhir_stack) MenhirState156 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState156
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState156

and _menhir_run159 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BAR ->
        _menhir_run157 _menhir_env (Obj.magic _menhir_stack) MenhirState159 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2821 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState159 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce149 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2822)
    | BACKQUOTE | LBRACKET | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LIDENT _ | LPAREN | QUOTE | SHARP | UIDENT _ | UNDERSCORE ->
        _menhir_reduce632 _menhir_env (Obj.magic _menhir_stack) MenhirState159
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState159

and _menhir_run162 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2819 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState162 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | BACKQUOTE ->
            _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETGREATER ->
            _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETLESS ->
            _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | QUOTE ->
            _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState163 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState163) : 'freshtv2820)
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run122 _menhir_env (Obj.magic _menhir_stack) MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState162 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState162

and _menhir_run192 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce271 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run193 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run144 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState193 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState193

and _menhir_run198 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run203 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run202 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run201 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run200 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run199 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2817 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2818)

and _menhir_run204 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FLOAT _v ->
        _menhir_run209 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run208 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run207 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run206 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run205 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2815 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2816)

and _menhir_run210 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2813 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState210 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2809 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2805 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | EXCEPTION ->
                    _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SHARP ->
                    _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UNDERSCORE ->
                    _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2415 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2415) : 'freshtv2806)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv2807 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2808)) : 'freshtv2810)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv2811 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2812)) : 'freshtv2814)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState210 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState210

and _menhir_run18 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 77898 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run588 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2803 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState588 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce282 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv2804)
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState588 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState588

and _menhir_run603 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run513 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run512 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603
    | SHARP ->
        _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState603 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState603

and _menhir_run615 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run517 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState615 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState615 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState615

and _menhir_run619 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run487 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState619 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState619

and _menhir_run642 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState642 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState642

and _menhir_goto__toplevel_phrase : _menhir_env -> 'ttv_tail -> _menhir_state -> 'tv__toplevel_phrase -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _v ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2801) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : 'tv__toplevel_phrase) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2799) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : 'tv__toplevel_phrase) : 'tv__toplevel_phrase) = _v in
    ((let _v : (
# 1063 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 78166 "reason_parser.ml"
    ) = 
# 1122 "reason_parser.mly"
                                  (apply_mapper_chain_to_toplevel_phrase _1 default_mapper_chain)
# 78170 "reason_parser.ml"
     in
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2797) = _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : (
# 1063 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 78178 "reason_parser.ml"
    )) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2795) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let (_v : (
# 1063 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 78186 "reason_parser.ml"
    )) = _v in
    ((let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv2793) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    let ((_1 : (
# 1063 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 78194 "reason_parser.ml"
    )) : (
# 1063 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 78198 "reason_parser.ml"
    )) = _v in
    (Obj.magic _1 : 'freshtv2794)) : 'freshtv2796)) : 'freshtv2798)) : 'freshtv2800)) : 'freshtv2802)

and _menhir_errorcase : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    match _menhir_s with
    | MenhirState2760 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv55 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv56)
    | MenhirState2757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv57 * _menhir_state * 'tv_toplevel_directive) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv58)
    | MenhirState2753 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv59) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv60)
    | MenhirState2748 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv61 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv62)
    | MenhirState2745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv63 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv64)
    | MenhirState2743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv65 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv66)
    | MenhirState2742 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv67 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv68)
    | MenhirState2739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv69 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv70)
    | MenhirState2738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv71 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv72)
    | MenhirState2737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv73 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv74)
    | MenhirState2736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv75 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv76)
    | MenhirState2733 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv77 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv78)
    | MenhirState2732 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv79 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv80)
    | MenhirState2731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv81 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv82)
    | MenhirState2730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv83 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv84)
    | MenhirState2717 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv85 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv86)
    | MenhirState2710 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv87 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv88)
    | MenhirState2709 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv89 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv90)
    | MenhirState2707 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv91 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv92)
    | MenhirState2706 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv93 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv94)
    | MenhirState2705 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv95 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv96)
    | MenhirState2704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv97 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv98)
    | MenhirState2701 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv99 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv100)
    | MenhirState2700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv101 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv102)
    | MenhirState2698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv103 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv104)
    | MenhirState2697 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv105 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 78334 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv106)
    | MenhirState2690 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv107 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv108)
    | MenhirState2688 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv109 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv110)
    | MenhirState2687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv111 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv112)
    | MenhirState2683 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv113 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv114)
    | MenhirState2682 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv115 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv116)
    | MenhirState2681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv117 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv118)
    | MenhirState2680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv119 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv120)
    | MenhirState2679 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv121 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv122)
    | MenhirState2677 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv123 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv124)
    | MenhirState2673 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv125 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState2673 : 'freshtv126)
    | MenhirState2669 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv127 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv128)
    | MenhirState2666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv131 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv129 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2666 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce53 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv130)) : 'freshtv132)
    | MenhirState2665 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv133 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv134)
    | MenhirState2664 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv135 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 78412 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv136)
    | MenhirState2663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv137 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 78421 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv138)
    | MenhirState2662 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv139 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 78430 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv140)
    | MenhirState2660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv141 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv142)
    | MenhirState2657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv143 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv144)
    | MenhirState2655 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv145 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv146)
    | MenhirState2654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv147 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv148)
    | MenhirState2652 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv149 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv150)
    | MenhirState2651 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv151 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv152)
    | MenhirState2649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv153 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv154)
    | MenhirState2647 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv155 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv156)
    | MenhirState2645 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv157 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv158)
    | MenhirState2644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv159 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv160)
    | MenhirState2642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv161 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv162)
    | MenhirState2640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv163 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv164)
    | MenhirState2639 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv165 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv166)
    | MenhirState2637 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv167 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 78504 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv168)
    | MenhirState2636 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv169 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv170)
    | MenhirState2635 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv171 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv172)
    | MenhirState2632 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv173 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv174)
    | MenhirState2631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv175 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv176)
    | MenhirState2630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv177 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv178)
    | MenhirState2629 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv179 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv180)
    | MenhirState2627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv181 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv182)
    | MenhirState2626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv183 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv184)
    | MenhirState2620 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv185 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv186)
    | MenhirState2619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv187 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv188)
    | MenhirState2613 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv189 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv190)
    | MenhirState2612 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv191 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv192)
    | MenhirState2610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv193 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv194)
    | MenhirState2609 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv195 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv196)
    | MenhirState2608 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv197 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv198)
    | MenhirState2606 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv199 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv200)
    | MenhirState2605 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv201 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv202)
    | MenhirState2604 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv203 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv204)
    | MenhirState2602 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv205 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv206)
    | MenhirState2601 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv207 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv208)
    | MenhirState2600 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv209 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv210)
    | MenhirState2597 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv211 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv212)
    | MenhirState2596 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv213 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv214)
    | MenhirState2595 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv215 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv216)
    | MenhirState2593 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv217 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 78633 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv218)
    | MenhirState2592 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv219 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv220)
    | MenhirState2591 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv221 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv222)
    | MenhirState2590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv223 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 78652 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv224)
    | MenhirState2589 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv225 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 78661 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv226)
    | MenhirState2588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv227 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv228)
    | MenhirState2587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv229 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv230)
    | MenhirState2586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv231 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv232)
    | MenhirState2584 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv233 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv234)
    | MenhirState2583 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv235 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv236)
    | MenhirState2582 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv237) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv238)
    | MenhirState2579 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv239 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv240)
    | MenhirState2578 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv241) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv242)
    | MenhirState2576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv243 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv244)
    | MenhirState2574 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv245) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv246)
    | MenhirState2570 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv247) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv248)
    | MenhirState2565 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv249 * _menhir_state * 'tv_signature_item) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv250)
    | MenhirState2563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv251) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv252)
    | MenhirState2557 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv253 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2557 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv254)
    | MenhirState2556 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv255 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce76 _menhir_env (Obj.magic _menhir_stack) : 'freshtv256)
    | MenhirState2555 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv257 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv258)
    | MenhirState2554 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv259 * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 78743 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv260)
    | MenhirState2552 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv261 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv262)
    | MenhirState2551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv263 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv264)
    | MenhirState2548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv265 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv266)
    | MenhirState2540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv267 * Lexing.position * _menhir_state * 'tv_str_extension_constructors) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv268)
    | MenhirState2539 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv269 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_str_extension_constructors) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2539 : 'freshtv270)
    | MenhirState2538 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv271 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv272)
    | MenhirState2537 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv273 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv274)
    | MenhirState2536 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv275 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv276)
    | MenhirState2533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv277 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2533 : 'freshtv278)
    | MenhirState2528 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv279 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv280)
    | MenhirState2527 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv281 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv282)
    | MenhirState2523 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv283 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv284)
    | MenhirState2522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv285 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv286)
    | MenhirState2521 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv287 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 78815 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2521 : 'freshtv288)
    | MenhirState2520 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv289 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error2519 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv290)
    | MenhirState2516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv291 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv292)
    | MenhirState2515 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv293 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv294)
    | MenhirState2511 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv295 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        (_menhir_reduce396 _menhir_env (Obj.magic _menhir_stack) : 'freshtv296)
    | MenhirState2510 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv297 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        (_menhir_reduce395 _menhir_env (Obj.magic _menhir_stack) : 'freshtv298)
    | MenhirState2507 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv299 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        (_menhir_reduce394 _menhir_env (Obj.magic _menhir_stack) : 'freshtv300)
    | MenhirState2504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv301 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2504 : 'freshtv302)
    | MenhirState2503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv303 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2503 : 'freshtv304)
    | MenhirState2501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv305 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2501 : 'freshtv306)
    | MenhirState2500 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv307 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2500 : 'freshtv308)
    | MenhirState2498 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv309 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2498 : 'freshtv310)
    | MenhirState2497 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv311 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2497 : 'freshtv312)
    | MenhirState2494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv313 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2494 : 'freshtv314)
    | MenhirState2493 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv315 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2493 : 'freshtv316)
    | MenhirState2491 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv317 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 78882 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2491 : 'freshtv318)
    | MenhirState2490 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv319 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 78890 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2490 : 'freshtv320)
    | MenhirState2487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv321 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2487 : 'freshtv322)
    | MenhirState2486 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv323 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2486 : 'freshtv324)
    | MenhirState2484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv325 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2484 : 'freshtv326)
    | MenhirState2483 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv327 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2483 : 'freshtv328)
    | MenhirState2481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv329 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2481 : 'freshtv330)
    | MenhirState2480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv331 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2480 : 'freshtv332)
    | MenhirState2479 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv333 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv334)
    | MenhirState2478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv335 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv336)
    | MenhirState2476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2476 : 'freshtv338)
    | MenhirState2475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv339 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState2475 : 'freshtv340)
    | MenhirState2473 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv341 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 78940 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState2473 : 'freshtv342)
    | MenhirState2472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv343 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv344)
    | MenhirState2470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv345 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv346)
    | MenhirState2466 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv347 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 78958 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv348)
    | MenhirState2465 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv349 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 78967 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv350)
    | MenhirState2464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv351 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 78976 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv352)
    | MenhirState2463 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv353 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 78985 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv354)
    | MenhirState2461 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv355 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 78994 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv356)
    | MenhirState2459 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv357 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv358)
    | MenhirState2457 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv359 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv360)
    | MenhirState2456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv361 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv362)
    | MenhirState2449 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv363 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv364)
    | MenhirState2442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv365 * Lexing.position * _menhir_state * 'tv_amper_type_list) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv366)
    | MenhirState2441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv367 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) * Lexing.position * _menhir_state * 'tv_amper_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv368)
    | MenhirState2439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv369 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * _menhir_state * 'tv_opt_ampersand) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv370)
    | MenhirState2434 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv371 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv372)
    | MenhirState2433 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv373 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv374)
    | MenhirState2432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv375 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv376)
    | MenhirState2428 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv377 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2428 : 'freshtv378)
    | MenhirState2426 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv379 * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 79057 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce235 _menhir_env (Obj.magic _menhir_stack) : 'freshtv380)
    | MenhirState2423 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv381 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv382)
    | MenhirState2422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv383 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv384)
    | MenhirState2421 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv385 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv386)
    | MenhirState2418 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv389 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv387 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2418 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce171 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv388)) : 'freshtv390)
    | MenhirState2417 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv391 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv392)
    | MenhirState2416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv393 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv394)
    | MenhirState2415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv395 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv396)
    | MenhirState2411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv397 * _menhir_state * 'tv_package_type_cstr) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error2410 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv398)
    | MenhirState2407 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv399 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 79110 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce59 _menhir_env (Obj.magic _menhir_stack) : 'freshtv400)
    | MenhirState2406 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv401 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 79118 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv402)
    | MenhirState2405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv403 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 79127 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce58 _menhir_env (Obj.magic _menhir_stack) : 'freshtv404)
    | MenhirState2396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv405 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2396 : 'freshtv406)
    | MenhirState2395 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv407 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce132 _menhir_env (Obj.magic _menhir_stack) : 'freshtv408)
    | MenhirState2393 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2393 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce120 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv410)) : 'freshtv412)
    | MenhirState2391 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv415 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2391 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce121 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv414)) : 'freshtv416)
    | MenhirState2389 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv417 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2389 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce122 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv418)) : 'freshtv420)
    | MenhirState2387 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv423 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv421 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2387 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce123 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv422)) : 'freshtv424)
    | MenhirState2382 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2382 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce206 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv426)) : 'freshtv428)
    | MenhirState2379 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv431 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv429 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2379 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce211 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv430)) : 'freshtv432)
    | MenhirState2377 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv433 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv434)
    | MenhirState2376 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv435 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv436)
    | MenhirState2375 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv437 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv438)
    | MenhirState2374 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv440)
    | MenhirState2368 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv443 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv441 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2368 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce247 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv442)) : 'freshtv444)
    | MenhirState2366 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv445 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv446)
    | MenhirState2364 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv449 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv447 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2364 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce120 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv448)) : 'freshtv450)
    | MenhirState2362 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv453 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv451 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2362 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce121 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv452)) : 'freshtv454)
    | MenhirState2360 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv457 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv455 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2360 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce122 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv456)) : 'freshtv458)
    | MenhirState2358 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv461 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv459 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2358 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce123 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv460)) : 'freshtv462)
    | MenhirState2354 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv463 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2354 : 'freshtv464)
    | MenhirState2351 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv465 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv466)
    | MenhirState2350 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv467 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv468)
    | MenhirState2347 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv469 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv470)
    | MenhirState2346 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv471 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv472)
    | MenhirState2343 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv473 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv474)
    | MenhirState2339 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv475 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2339 : 'freshtv476)
    | MenhirState2337 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv477 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 79311 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce433 _menhir_env (Obj.magic _menhir_stack) : 'freshtv478)
    | MenhirState2326 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv479 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 79319 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv480)
    | MenhirState2324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv481 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 79328 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv482)
    | MenhirState2323 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv483 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 79337 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv484)
    | MenhirState2320 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv485 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv486)
    | MenhirState2315 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv487 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2315 : 'freshtv488)
    | MenhirState2314 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv489 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv490)
    | MenhirState2312 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv491 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2312 : 'freshtv492)
    | MenhirState2310 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv493 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2310 : 'freshtv494)
    | MenhirState2309 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv495 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv496)
    | MenhirState2307 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv497 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce112 _menhir_env (Obj.magic _menhir_stack) : 'freshtv498)
    | MenhirState2306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv499 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv500)
    | MenhirState2304 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv501 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv502)
    | MenhirState2303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv503 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce111 _menhir_env (Obj.magic _menhir_stack) : 'freshtv504)
    | MenhirState2302 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv505 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv506)
    | MenhirState2301 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv507 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv508)
    | MenhirState2296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv509 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce581 _menhir_env (Obj.magic _menhir_stack) : 'freshtv510)
    | MenhirState2295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv511 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv512)
    | MenhirState2294 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv513 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv514)
    | MenhirState2293 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv515 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv516)
    | MenhirState2292 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv517 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce580 _menhir_env (Obj.magic _menhir_stack) : 'freshtv518)
    | MenhirState2291 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv519 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv520)
    | MenhirState2289 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv521 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv522)
    | MenhirState2288 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv523 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv524)
    | MenhirState2281 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv525 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce108 _menhir_env (Obj.magic _menhir_stack) : 'freshtv526)
    | MenhirState2278 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv527 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv528)
    | MenhirState2277 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv529 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv530)
    | MenhirState2275 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv531 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1049 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2275 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv532)
    | MenhirState2274 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv533 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv534)
    | MenhirState2272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv535 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2272 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv536)
    | MenhirState2271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv537 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce74 _menhir_env (Obj.magic _menhir_stack) : 'freshtv538)
    | MenhirState2270 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv539 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv540)
    | MenhirState2269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv541 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce75 _menhir_env (Obj.magic _menhir_stack) : 'freshtv542)
    | MenhirState2268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv543 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv544)
    | MenhirState2259 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv545 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce637 _menhir_env (Obj.magic _menhir_stack) : 'freshtv546)
    | MenhirState2258 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv547 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState2258 : 'freshtv548)
    | MenhirState2257 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv549 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2257 : 'freshtv550)
    | MenhirState2255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv551 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv552)
    | MenhirState2253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv553 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce637 _menhir_env (Obj.magic _menhir_stack) : 'freshtv554)
    | MenhirState2252 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv555 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState2252 : 'freshtv556)
    | MenhirState2248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv557 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv558)
    | MenhirState2247 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv559 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce374 _menhir_env (Obj.magic _menhir_stack) : 'freshtv560)
    | MenhirState2238 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv561 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv562)
    | MenhirState2235 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv563 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        (_menhir_reduce68 _menhir_env (Obj.magic _menhir_stack) : 'freshtv564)
    | MenhirState2234 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv565 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce487 _menhir_env (Obj.magic _menhir_stack) : 'freshtv566)
    | MenhirState2233 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv567 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv568)
    | MenhirState2232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv569 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce488 _menhir_env (Obj.magic _menhir_stack) : 'freshtv570)
    | MenhirState2231 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv571 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv572)
    | MenhirState2230 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv573 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv574)
    | MenhirState2229 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv575 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv576)
    | MenhirState2224 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv577 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv578)
    | MenhirState2223 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv579 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce439 _menhir_env (Obj.magic _menhir_stack) : 'freshtv580)
    | MenhirState2222 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv581 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv582)
    | MenhirState2221 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv583 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv584)
    | MenhirState2220 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv585 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv586)
    | MenhirState2219 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv587 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce438 _menhir_env (Obj.magic _menhir_stack) : 'freshtv588)
    | MenhirState2218 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv589 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv590)
    | MenhirState2217 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv591 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 79588 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv592)
    | MenhirState2214 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv593 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv594)
    | MenhirState2213 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv595 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv596)
    | MenhirState2211 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv597 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1004 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2211 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv598)
    | MenhirState2210 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv599 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv600)
    | MenhirState2208 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv601 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run927 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2208 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv602)
    | MenhirState2202 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv603 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv604)
    | MenhirState2198 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv605 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2198 : 'freshtv606)
    | MenhirState2197 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv607 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv608)
    | MenhirState2192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv609 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv610)
    | MenhirState2191 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv611 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv612)
    | MenhirState2188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv613 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv614)
    | MenhirState2187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv615 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv616)
    | MenhirState2183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv617 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState2183 : 'freshtv618)
    | MenhirState2182 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv619 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv620)
    | MenhirState2179 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv623 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv621 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2179 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce215 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv622)) : 'freshtv624)
    | MenhirState2177 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv625 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv626)
    | MenhirState2173 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv629 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv627 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2173 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce249 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv628)) : 'freshtv630)
    | MenhirState2172 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv631 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv632)
    | MenhirState2169 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv635 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv633 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2169 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce254 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv634)) : 'freshtv636)
    | MenhirState2168 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv637 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv638)
    | MenhirState2165 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv641 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv639 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2165 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce119 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv640)) : 'freshtv642)
    | MenhirState2161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv643 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv644)
    | MenhirState2158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv645 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv646)
    | MenhirState2155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv647 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv648)
    | MenhirState2153 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv649 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv650)
    | MenhirState2149 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv651 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_leading_bar_match_cases_to_sequence_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv652)
    | MenhirState2147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv653 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv654)
    | MenhirState2144 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv655 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv656)
    | MenhirState2143 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv657 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv658)
    | MenhirState2142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv659 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_module_type * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv660)
    | MenhirState2140 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv661 * _menhir_state * 'tv_functor_args * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv662)
    | MenhirState2139 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv663 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv664)
    | MenhirState2138 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv665 * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv666)
    | MenhirState2137 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv667 * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv668)
    | MenhirState2134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv669 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 79783 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv670)
    | MenhirState2132 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv671 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 79792 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv672)
    | MenhirState2129 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv673 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv674)
    | MenhirState2128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv675 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv676)
    | MenhirState2127 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv677 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv678)
    | MenhirState2126 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv680)
    | MenhirState2125 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv681 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv682)
    | MenhirState2123 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv683 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv684)
    | MenhirState2122 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv686)
    | MenhirState2118 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv689 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv687 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2118 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce263 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv688)) : 'freshtv690)
    | MenhirState2117 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv691 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce127 _menhir_env (Obj.magic _menhir_stack) : 'freshtv692)
    | MenhirState2116 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv694)
    | MenhirState2114 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv695 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv696)
    | MenhirState2113 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv697 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv698)
    | MenhirState2109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv701 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv699 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2109 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce263 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv700)) : 'freshtv702)
    | MenhirState2103 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv703 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv704)
    | MenhirState2096 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv705 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv706)
    | MenhirState2093 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv707 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv708)
    | MenhirState2092 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv709 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv710)
    | MenhirState2088 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv711 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv712)
    | MenhirState2087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv713 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv714)
    | MenhirState2086 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv715 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv716)
    | MenhirState2083 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv717 * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv718)
    | MenhirState2082 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv719 * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv720)
    | MenhirState2078 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv721 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv722)
    | MenhirState2077 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv723 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 79923 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv724)
    | MenhirState2076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv725 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 79932 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv726)
    | MenhirState2075 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv727 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv728)
    | MenhirState2074 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv729 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv730)
    | MenhirState2068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv731 * _menhir_state * 'tv_signature_item) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState2068 : 'freshtv732)
    | MenhirState2065 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv733 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_description_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv734)
    | MenhirState2063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv735 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 79960 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv736)
    | MenhirState2062 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv737 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 79969 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv738)
    | MenhirState2061 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv739 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 79978 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv740)
    | MenhirState2058 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv742)
    | MenhirState2057 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv743 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 79992 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv744)
    | MenhirState2056 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv745 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 80001 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv746)
    | MenhirState2055 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv747 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 80010 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv748)
    | MenhirState2054 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv749 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 80019 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv750)
    | MenhirState2052 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv751 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv752)
    | MenhirState2051 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv753 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv754)
    | MenhirState2049 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv755 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv756)
    | MenhirState2048 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv757 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv758)
    | MenhirState2046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv759 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv760)
    | MenhirState2044 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv761 * Lexing.position * _menhir_state * (
# 898 "reason_parser.mly"
       (string * string option)
# 80053 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv762)
    | MenhirState2043 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv763 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv764)
    | MenhirState2041 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv765 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv766)
    | MenhirState2039 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv767 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv768)
    | MenhirState2037 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv769 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv770)
    | MenhirState2036 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv771 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv772)
    | MenhirState2033 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv773 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2033 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv774)
    | MenhirState2032 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv775 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv776)
    | MenhirState2027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv777 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv778)
    | MenhirState2025 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv779 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState2025 : 'freshtv780)
    | MenhirState2016 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv781 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce251 _menhir_env (Obj.magic _menhir_stack) : 'freshtv782)
    | MenhirState2006 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv783 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce128 _menhir_env (Obj.magic _menhir_stack) : 'freshtv784)
    | MenhirState2005 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv785 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv786)
    | MenhirState2001 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv787 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error2000 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv788)
    | MenhirState1999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv789 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce851 _menhir_env (Obj.magic _menhir_stack) : 'freshtv790)
    | MenhirState1998 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv791 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv792)
    | MenhirState1997 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv793 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv794)
    | MenhirState1995 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv795 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv796)
    | MenhirState1994 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv797 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error1993 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv798)
    | MenhirState1991 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv799 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80147 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv800)
    | MenhirState1990 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv801 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80156 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv802)
    | MenhirState1989 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv803 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80165 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce574 _menhir_env (Obj.magic _menhir_stack) : 'freshtv804)
    | MenhirState1988 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv805 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv806)
    | MenhirState1987 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv807 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80178 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce852 _menhir_env (Obj.magic _menhir_stack) : 'freshtv808)
    | MenhirState1985 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv809 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80186 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv810)
    | MenhirState1984 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv811 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80195 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv812)
    | MenhirState1983 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv813 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80204 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce573 _menhir_env (Obj.magic _menhir_stack) : 'freshtv814)
    | MenhirState1982 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv815 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 80212 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv816)
    | MenhirState1980 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv817 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv818)
    | MenhirState1979 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv819 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        (_menhir_reduce849 _menhir_env (Obj.magic _menhir_stack) : 'freshtv820)
    | MenhirState1978 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv821 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) MenhirState1978 : 'freshtv822)
    | MenhirState1977 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv823 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv824)
    | MenhirState1975 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv826)
    | MenhirState1974 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv828)
    | MenhirState1973 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv829 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv830)
    | MenhirState1972 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv831 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv832)
    | MenhirState1971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv833 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv834)
    | MenhirState1969 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv837 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv835 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1969 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce254 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv836)) : 'freshtv838)
    | MenhirState1965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv839 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv840)
    | MenhirState1962 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv843 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv841 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1962 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce119 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv842)) : 'freshtv844)
    | MenhirState1958 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv845 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_package_type)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv846)
    | MenhirState1955 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv847 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv848)
    | MenhirState1952 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv849 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv850)
    | MenhirState1950 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv851 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv852)
    | MenhirState1949 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv853 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv854)
    | MenhirState1948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv855 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv856)
    | MenhirState1947 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv857 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce117 _menhir_env (Obj.magic _menhir_stack) : 'freshtv858)
    | MenhirState1945 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv859 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv860)
    | MenhirState1944 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv861 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv862)
    | MenhirState1943 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv863 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv864)
    | MenhirState1939 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv865 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1939 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv866)
    | MenhirState1935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv867 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1935 : 'freshtv868)
    | MenhirState1932 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv869 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1932 : 'freshtv870)
    | MenhirState1929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv871 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv872)
    | MenhirState1924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv873 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1924 : 'freshtv874)
    | MenhirState1923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv875 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce502 _menhir_env (Obj.magic _menhir_stack) : 'freshtv876)
    | MenhirState1922 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv877 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv878)
    | MenhirState1921 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv879 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv880)
    | MenhirState1920 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv881 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce503 _menhir_env (Obj.magic _menhir_stack) : 'freshtv882)
    | MenhirState1919 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv883 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv884)
    | MenhirState1917 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv885 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv886)
    | MenhirState1912 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv887 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv888)
    | MenhirState1910 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv889 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce498 _menhir_env (Obj.magic _menhir_stack) : 'freshtv890)
    | MenhirState1909 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv891 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv892)
    | MenhirState1906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv893 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv894)
    | MenhirState1904 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv895 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv896)
    | MenhirState1902 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv897 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce538 _menhir_env (Obj.magic _menhir_stack) : 'freshtv898)
    | MenhirState1900 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv899 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv900)
    | MenhirState1898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv901 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv902)
    | MenhirState1896 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv903 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv904)
    | MenhirState1895 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv905 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv906)
    | MenhirState1894 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv907 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce61 _menhir_env (Obj.magic _menhir_stack) : 'freshtv908)
    | MenhirState1893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv909 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv910)
    | MenhirState1891 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv911 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 80444 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv912)
    | MenhirState1888 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv913 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv914)
    | MenhirState1887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv915 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv916)
    | MenhirState1886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv917 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv918)
    | MenhirState1885 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv919 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce296 _menhir_env (Obj.magic _menhir_stack) : 'freshtv920)
    | MenhirState1882 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv921 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1882 : 'freshtv922)
    | MenhirState1881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv923 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv924)
    | MenhirState1880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv925 * Lexing.position * _menhir_state * 'tv_many_class_declarations * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce307 _menhir_env (Obj.magic _menhir_stack) : 'freshtv926)
    | MenhirState1877 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv927 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1877 : 'freshtv928)
    | MenhirState1876 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv929 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv930)
    | MenhirState1875 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv931 * Lexing.position * _menhir_state * 'tv_many_class_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce308 _menhir_env (Obj.magic _menhir_stack) : 'freshtv932)
    | MenhirState1872 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv933 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonlocal_module_binding_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1872 : 'freshtv934)
    | MenhirState1871 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv935 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv936)
    | MenhirState1870 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv937 * Lexing.position * _menhir_state * 'tv_many_nonlocal_module_bindings * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce303 _menhir_env (Obj.magic _menhir_stack) : 'freshtv938)
    | MenhirState1867 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv939 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1867 : 'freshtv940)
    | MenhirState1864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv941 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        (_menhir_reduce641 _menhir_env (Obj.magic _menhir_stack) : 'freshtv942)
    | MenhirState1861 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv943 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv944)
    | MenhirState1859 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv945 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv946)
    | MenhirState1858 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv947 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80529 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        (_menhir_reduce811 _menhir_env (Obj.magic _menhir_stack) : 'freshtv948)
    | MenhirState1857 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv949 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80537 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        (_menhir_reduce383 _menhir_env (Obj.magic _menhir_stack) MenhirState1857 : 'freshtv950)
    | MenhirState1856 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv951 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80545 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        (_menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) MenhirState1856 : 'freshtv952)
    | MenhirState1855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv953 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80553 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState1855 : 'freshtv954)
    | MenhirState1854 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv955 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv956)
    | MenhirState1853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv957 * Lexing.position * _menhir_state * 'tv_many_type_declarations * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce299 _menhir_env (Obj.magic _menhir_stack) : 'freshtv958)
    | MenhirState1846 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv959 * Lexing.position * _menhir_state * 'tv_structure_item * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1846 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv960)
    | MenhirState1842 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv961 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1842 : 'freshtv962)
    | MenhirState1834 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv963 * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv964)
    | MenhirState1832 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv965 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce348 _menhir_env (Obj.magic _menhir_stack) : 'freshtv966)
    | MenhirState1831 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv967 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_non_arrowed_class_constructor_type) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv968)
    | MenhirState1827 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv969 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv970)
    | MenhirState1826 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv971 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv972)
    | MenhirState1825 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv973 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce347 _menhir_env (Obj.magic _menhir_stack) : 'freshtv974)
    | MenhirState1824 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv975 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv976)
    | MenhirState1823 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv977 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv978)
    | MenhirState1820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv979 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce54 _menhir_env (Obj.magic _menhir_stack) : 'freshtv980)
    | MenhirState1819 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv981 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_constructor_type) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv982)
    | MenhirState1816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv983 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80625 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv984)
    | MenhirState1815 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv985 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80634 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv986)
    | MenhirState1814 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv987 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80643 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv988)
    | MenhirState1809 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv989 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv990)
    | MenhirState1808 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv991 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv992)
    | MenhirState1807 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv993 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80662 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv994)
    | MenhirState1806 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv995 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80671 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv996)
    | MenhirState1805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv997 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80680 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv998)
    | MenhirState1801 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1000)
    | MenhirState1800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1001 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1002)
    | MenhirState1799 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1003 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1004)
    | MenhirState1795 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1005 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1006)
    | MenhirState1794 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1007 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1008)
    | MenhirState1790 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1009 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce355 _menhir_env (Obj.magic _menhir_stack) MenhirState1790 : 'freshtv1010)
    | MenhirState1788 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1011 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1012)
    | MenhirState1787 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1013 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1014)
    | MenhirState1786 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1015 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 80728 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce340 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1016)
    | MenhirState1781 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1017 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80736 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1018)
    | MenhirState1780 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1019 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80745 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1020)
    | MenhirState1779 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1021 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80754 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1022)
    | MenhirState1774 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1023 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1024)
    | MenhirState1773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1025 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1026)
    | MenhirState1772 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1027 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80773 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1028)
    | MenhirState1771 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1029 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80782 "reason_parser.ml"
        ) * Lexing.position)) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1030)
    | MenhirState1770 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1031 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 80791 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1032)
    | MenhirState1768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1033 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce7 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1034)
    | MenhirState1767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1035 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1036)
    | MenhirState1766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1037 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1038)
    | MenhirState1763 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1041 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1039 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1763 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce53 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1040)) : 'freshtv1042)
    | MenhirState1757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1043 * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce19 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1044)
    | MenhirState1753 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1045 * Lexing.position * _menhir_state * 'tv_class_field) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce695 _menhir_env (Obj.magic _menhir_stack) MenhirState1753 : 'freshtv1046)
    | MenhirState1749 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1047 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1749 : 'freshtv1048)
    | MenhirState1745 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1049 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1745 : 'freshtv1050)
    | MenhirState1744 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1051 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1052)
    | MenhirState1742 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1053 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_parent_binder) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1742 : 'freshtv1054)
    | MenhirState1739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1055 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce652 _menhir_env (Obj.magic _menhir_stack) MenhirState1739 : 'freshtv1056)
    | MenhirState1738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1057 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1058)
    | MenhirState1737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1059 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1060)
    | MenhirState1735 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1061 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1735 : 'freshtv1062)
    | MenhirState1734 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1063 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1064)
    | MenhirState1731 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1065 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_method_) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1731 : 'freshtv1066)
    | MenhirState1729 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1067 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce567 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1068)
    | MenhirState1728 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1069 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1070)
    | MenhirState1726 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((((((('freshtv1071 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce569 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1072)
    | MenhirState1725 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((((('freshtv1073 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1074)
    | MenhirState1723 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1075 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1076)
    | MenhirState1721 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1077 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1078)
    | MenhirState1720 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1079 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1080)
    | MenhirState1719 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1081 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce568 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1082)
    | MenhirState1718 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1083 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1084)
    | MenhirState1714 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1085 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1086)
    | MenhirState1713 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1087 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce408 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1088)
    | MenhirState1712 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1089 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1090)
    | MenhirState1711 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1091 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1092)
    | MenhirState1710 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1093 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1094)
    | MenhirState1709 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1095 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce403 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1096)
    | MenhirState1708 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1097 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1098)
    | MenhirState1707 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1099 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 80951 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1100)
    | MenhirState1704 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1101 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1102)
    | MenhirState1703 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1103 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1104)
    | MenhirState1702 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1105 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1106)
    | MenhirState1700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1107 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1108)
    | MenhirState1698 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1109 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1110)
    | MenhirState1695 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1111 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1112)
    | MenhirState1693 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1113 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1114)
    | MenhirState1692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1115 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1116)
    | MenhirState1691 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1117 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1118)
    | MenhirState1690 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1119 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1120)
    | MenhirState1689 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1121 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce843 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1122)
    | MenhirState1688 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1123 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1124)
    | MenhirState1686 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1125 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce842 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1126)
    | MenhirState1685 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1127 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1128)
    | MenhirState1684 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1129 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1130)
    | MenhirState1683 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1131 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1132)
    | MenhirState1680 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1133 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1134)
    | MenhirState1678 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1135 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1136)
    | MenhirState1676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1137 * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1138)
    | MenhirState1674 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1139 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1674 : 'freshtv1140)
    | MenhirState1671 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1141 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1142)
    | MenhirState1669 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1143 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1144)
    | MenhirState1668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1145 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1146)
    | MenhirState1667 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1147 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1148)
    | MenhirState1666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1149 * Lexing.position * _menhir_state * 'tv_class_self_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce695 _menhir_env (Obj.magic _menhir_stack) MenhirState1666 : 'freshtv1150)
    | MenhirState1664 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1151 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1152)
    | MenhirState1663 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1153 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1663 : 'freshtv1154)
    | MenhirState1662 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1155 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1156)
    | MenhirState1658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1157 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1158)
    | MenhirState1657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1159 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1160)
    | MenhirState1650 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1161 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce30 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1162)
    | MenhirState1648 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1163 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce13 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1164)
    | MenhirState1647 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1165 * Lexing.position * _menhir_state * 'tv_class_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce11 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1166)
    | MenhirState1643 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1167 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce16 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1168)
    | MenhirState1642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1169 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce15 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1170)
    | MenhirState1641 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1171 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1172)
    | MenhirState1640 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1173 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1174)
    | MenhirState1639 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1175 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1176)
    | MenhirState1637 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1177 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1178)
    | MenhirState1635 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1179 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1180)
    | MenhirState1634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1181 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_rec_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1182)
    | MenhirState1633 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1183 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1184)
    | MenhirState1632 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1185 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState1632 : 'freshtv1186)
    | MenhirState1631 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1187 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1188)
    | MenhirState1630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1189 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 81169 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1190)
    | MenhirState1627 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1191 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 81178 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1192)
    | MenhirState1626 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1193 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 81187 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1194)
    | MenhirState1623 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1195 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_type_declaration_details) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1623 : 'freshtv1196)
    | MenhirState1617 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1197 * Lexing.position) = Obj.magic _menhir_stack in
        (raise _eRR : 'freshtv1198)
    | MenhirState1615 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1199 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 81204 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce362 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1200)
    | MenhirState1610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1201 * Lexing.position * _menhir_state * 'tv_class_sig_fields) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce637 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1202)
    | MenhirState1609 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1203 * Lexing.position * _menhir_state * 'tv_class_self_type) * Lexing.position * _menhir_state * 'tv_class_sig_fields) = Obj.magic _menhir_stack in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState1609 : 'freshtv1204)
    | MenhirState1606 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1205 * Lexing.position * _menhir_state * 'tv_item_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1606 : 'freshtv1206)
    | MenhirState1603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1207 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constrain_field) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1603 : 'freshtv1208)
    | MenhirState1601 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1209 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1210)
    | MenhirState1599 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1211 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1212)
    | MenhirState1595 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1213 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1595 : 'freshtv1214)
    | MenhirState1594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1215 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_instance_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce40 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1216)
    | MenhirState1593 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1217 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce33 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1218)
    | MenhirState1592 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1219 * Lexing.position * _menhir_state * 'tv_clty_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce32 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1220)
    | MenhirState1588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1221 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1222)
    | MenhirState1587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1223 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1224)
    | MenhirState1586 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1225 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1226)
    | MenhirState1581 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1227 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1581 : 'freshtv1228)
    | MenhirState1579 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1229 * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1230)
    | MenhirState1576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1231 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1232)
    | MenhirState1575 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1233 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1234)
    | MenhirState1573 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1235 * Lexing.position * _menhir_state * Lexing.position) * 'tv_private_virtual_flags) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1236)
    | MenhirState1566 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1237 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1238)
    | MenhirState1563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1239 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_value_type) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1563 : 'freshtv1240)
    | MenhirState1561 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1241 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1242)
    | MenhirState1559 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1243 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_virtual_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1244)
    | MenhirState1558 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1245 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1246)
    | MenhirState1556 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1247 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1248)
    | MenhirState1554 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1249 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1250)
    | MenhirState1553 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1251 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1252)
    | MenhirState1552 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1253 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1254)
    | MenhirState1551 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1255 * Lexing.position * _menhir_state * 'tv_class_self_type) = Obj.magic _menhir_stack in
        (_menhir_reduce358 _menhir_env (Obj.magic _menhir_stack) MenhirState1551 : 'freshtv1256)
    | MenhirState1545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1257 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1258)
    | MenhirState1544 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1259 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce355 _menhir_env (Obj.magic _menhir_stack) MenhirState1544 : 'freshtv1260)
    | MenhirState1542 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1261 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 81350 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1262)
    | MenhirState1539 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1263 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 81359 "reason_parser.ml"
        ) * Lexing.position) * _menhir_state * 'tv_class_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1264)
    | MenhirState1538 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1265 * _menhir_state * 'tv_virtual_flag) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 81368 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1266)
    | MenhirState1536 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1267 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1268)
    | MenhirState1534 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1269 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1270)
    | MenhirState1532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1271 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1532 : 'freshtv1272)
    | MenhirState1530 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1273 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_extension_constructor_rebind) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1530 : 'freshtv1274)
    | MenhirState1528 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1275 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1528 : 'freshtv1276)
    | MenhirState1526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1277 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1526 : 'freshtv1278)
    | MenhirState1525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1279 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1280)
    | MenhirState1524 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1281 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1524 : 'freshtv1282)
    | MenhirState1522 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1283 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1522 : 'freshtv1284)
    | MenhirState1520 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1285 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1520 : 'freshtv1286)
    | MenhirState1519 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1287 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1288)
    | MenhirState1518 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1289 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1518 : 'freshtv1290)
    | MenhirState1516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1291 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1516 : 'freshtv1292)
    | MenhirState1514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1293 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1514 : 'freshtv1294)
    | MenhirState1513 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1295 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1296)
    | MenhirState1512 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1297 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1512 : 'freshtv1298)
    | MenhirState1509 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1299 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1509 : 'freshtv1300)
    | MenhirState1507 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1301 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1507 : 'freshtv1302)
    | MenhirState1506 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1303 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1304)
    | MenhirState1505 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1305 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1505 : 'freshtv1306)
    | MenhirState1503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1307 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81463 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1503 : 'freshtv1308)
    | MenhirState1501 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1309 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1310)
    | MenhirState1500 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1311 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce450 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1312)
    | MenhirState1498 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1313 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1314)
    | MenhirState1496 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1315 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1496 : 'freshtv1316)
    | MenhirState1494 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1317 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81489 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce335 _menhir_env (Obj.magic _menhir_stack) MenhirState1494 : 'freshtv1318)
    | MenhirState1493 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1319 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1320)
    | MenhirState1492 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1321 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1322)
    | MenhirState1491 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1323 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81507 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1324)
    | MenhirState1490 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1325 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81516 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState1490 : 'freshtv1326)
    | MenhirState1489 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1327 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1328)
    | MenhirState1487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1329 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1487 : 'freshtv1330)
    | MenhirState1485 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1331 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_primitive_declaration) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState1485 : 'freshtv1332)
    | MenhirState1483 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1333 * Lexing.position * _menhir_state * (
# 898 "reason_parser.mly"
       (string * string option)
# 81537 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce677 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1334)
    | MenhirState1482 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1335 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1336)
    | MenhirState1480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1337 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1338)
    | MenhirState1478 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1339 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1340)
    | MenhirState1477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1341 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1477 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1342)
    | MenhirState1476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1343 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1344)
    | MenhirState1475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1345 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1346)
    | MenhirState1474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1347 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1348)
    | MenhirState1472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1349 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1350)
    | MenhirState1471 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1351 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1352)
    | MenhirState1468 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1353 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_rec_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1354)
    | MenhirState1467 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1355 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81593 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1356)
    | MenhirState1466 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1357 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81602 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1358)
    | MenhirState1464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1359 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1360)
    | MenhirState1462 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1361 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 81616 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_declaration) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1362)
    | MenhirState1460 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1363 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 81625 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1364)
    | MenhirState1459 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1365 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 81634 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1366)
    | MenhirState1457 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1367 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1368)
    | MenhirState1445 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1369 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1370)
    | MenhirState1444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1371 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1372)
    | MenhirState1443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1373 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1374)
    | MenhirState1442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1375 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1376)
    | MenhirState1441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1377 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 81668 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1378)
    | MenhirState1438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1379 * Lexing.position * _menhir_state * 'tv_with_constraints) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1380)
    | MenhirState1436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1381 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1382)
    | MenhirState1435 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1383 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1384)
    | MenhirState1434 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1385 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1386)
    | MenhirState1433 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1387 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81697 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1388)
    | MenhirState1432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1389 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81706 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1390)
    | MenhirState1430 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1391 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1392)
    | MenhirState1429 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1393 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1394)
    | MenhirState1428 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1395 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1396)
    | MenhirState1427 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1397 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * _menhir_state * 'tv_with_type_binder) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1398)
    | MenhirState1425 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1399 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1400)
    | MenhirState1422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1401 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1402)
    | MenhirState1421 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1403 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1404)
    | MenhirState1420 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1405 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1406)
    | MenhirState1419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1407 * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1408)
    | MenhirState1418 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1409 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 81760 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1410)
    | MenhirState1414 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1411 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 81769 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1412)
    | MenhirState1411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1413 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 81778 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1414)
    | MenhirState1409 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1415 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1416)
    | MenhirState1408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1417 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState1408 : 'freshtv1418)
    | MenhirState1407 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1419 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1420)
    | MenhirState1406 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1421 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1422)
    | MenhirState1405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1423 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1424)
    | MenhirState1402 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1426)
    | MenhirState1401 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1427 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1428)
    | MenhirState1400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1429 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1430)
    | MenhirState1398 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1431 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1432)
    | MenhirState1397 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1433 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1434)
    | MenhirState1393 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1435 * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1436)
    | MenhirState1392 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1437 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) * Lexing.position * _menhir_state * 'tv_sig_extension_constructors) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1438)
    | MenhirState1390 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1440)
    | MenhirState1389 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1441 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1442)
    | MenhirState1387 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1443 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1444)
    | MenhirState1386 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1445 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1446)
    | MenhirState1384 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1447 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1448)
    | MenhirState1383 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1449 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1450)
    | MenhirState1380 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1451 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1452)
    | MenhirState1379 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1453 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1454)
    | MenhirState1377 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1455 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81886 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1456)
    | MenhirState1376 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1457 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81895 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1458)
    | MenhirState1375 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1459 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1460)
    | MenhirState1374 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1461 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1462)
    | MenhirState1372 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1463 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * _menhir_state * 'tv_potentially_long_ident_and_optional_type_parameters) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1464)
    | MenhirState1369 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1465 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_declaration_details) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1466)
    | MenhirState1367 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1467 * Lexing.position * _menhir_state * 'tv_type_strictly_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1468)
    | MenhirState1364 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1469 * Lexing.position * _menhir_state * 'tv_optional_type_parameter_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1470)
    | MenhirState1361 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1471 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1472)
    | MenhirState1359 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1473 * Lexing.position * _menhir_state * 'tv_constraints) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1474)
    | MenhirState1358 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1475 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 81944 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) * Lexing.position * _menhir_state * 'tv_constraints) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1476)
    | MenhirState1357 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1477 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 81953 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) * Lexing.position * _menhir_state * 'tv_type_kind) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1478)
    | MenhirState1352 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1479 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1480)
    | MenhirState1351 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1481 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1482)
    | MenhirState1347 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1483 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1484)
    | MenhirState1346 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1485 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1486)
    | MenhirState1345 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1487 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 81982 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1488)
    | MenhirState1344 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1489 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1490)
    | MenhirState1338 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1491 * _menhir_state * 'tv_label_declarations) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1492)
    | MenhirState1337 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1493 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) * _menhir_state * 'tv_label_declarations) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1494)
    | MenhirState1332 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1495 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 82006 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_poly_type) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1496)
    | MenhirState1330 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1497 * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1498)
    | MenhirState1328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1499 * Lexing.position * _menhir_state * 'tv_mutable_flag * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 82020 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1500)
    | MenhirState1324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1501 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_private_flag * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1502)
    | MenhirState1320 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1503 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1504)
    | MenhirState1319 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1505 * Lexing.position * _menhir_state * 'tv_constructor_declaration_no_leading_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1506)
    | MenhirState1316 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1507 * Lexing.position * _menhir_state * 'tv_constructor_declarations_leading_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1508)
    | MenhirState1313 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1509 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1510)
    | MenhirState1312 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1511 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1512)
    | MenhirState1310 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1513 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1514)
    | MenhirState1309 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1515 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1516)
    | MenhirState1307 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1517 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1518)
    | MenhirState1306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1519 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1520)
    | MenhirState1303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1521 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1522)
    | MenhirState1302 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1523 * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1524)
    | MenhirState1300 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1525 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 82089 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1526)
    | MenhirState1299 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1527 * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 82098 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1528)
    | MenhirState1296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1529 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1530)
    | MenhirState1295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1531 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1532)
    | MenhirState1293 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1533 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1534)
    | MenhirState1292 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1535 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1536)
    | MenhirState1290 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1537 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1538)
    | MenhirState1289 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1539 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1540)
    | MenhirState1288 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1541 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1542)
    | MenhirState1287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1543 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1544)
    | MenhirState1285 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1545 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1546)
    | MenhirState1284 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1547 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1548)
    | MenhirState1282 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1549 * Lexing.position * _menhir_state * 'tv_attribute) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1550)
    | MenhirState1280 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1551 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 82162 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_generalized_constructor_arguments) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1552)
    | MenhirState1278 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1553 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1554)
    | MenhirState1277 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1555 * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1556)
    | MenhirState1275 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1557 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1558)
    | MenhirState1272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1559 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 82186 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1560)
    | MenhirState1271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1561 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 82195 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1562)
    | MenhirState1269 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1563 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 82204 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1564)
    | MenhirState1268 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1565 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1566)
    | MenhirState1267 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1567 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 82218 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1568)
    | MenhirState1264 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1569 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1570)
    | MenhirState1260 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1571 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1572)
    | MenhirState1256 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1573 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1574)
    | MenhirState1254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1575 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 82242 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1576)
    | MenhirState1253 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1577 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1578)
    | MenhirState1252 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1579 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1580)
    | MenhirState1251 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1581 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState1251 : 'freshtv1582)
    | MenhirState1250 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1583 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1584)
    | MenhirState1249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1585 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 82270 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce453 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1586)
    | MenhirState1248 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1587 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1588)
    | MenhirState1247 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1589 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1590)
    | MenhirState1246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1591 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1592)
    | MenhirState1245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1593 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1594)
    | MenhirState1244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1595 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 82298 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1596)
    | MenhirState1239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1597 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_let_binding_body) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1598)
    | MenhirState1238 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1599 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1600)
    | MenhirState1237 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1601 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1602)
    | MenhirState1236 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1603 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1604)
    | MenhirState1235 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1605 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1606)
    | MenhirState1234 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1607 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1608)
    | MenhirState1232 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1609 * Lexing.position * _menhir_state * 'tv_simple_pattern_not_ident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1610)
    | MenhirState1227 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1611 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1612)
    | MenhirState1225 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1613 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1614)
    | MenhirState1224 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1615 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * 'tv_type_constraint) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1616)
    | MenhirState1221 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1617 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1618)
    | MenhirState1219 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1619 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_typevar_list * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1620)
    | MenhirState1217 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1621 * _menhir_state * 'tv_typevar_list * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1622)
    | MenhirState1215 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1623 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1624)
    | MenhirState1213 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1625 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1626)
    | MenhirState1211 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1627 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_lident_list)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1628)
    | MenhirState1208 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1629 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 82387 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1630)
    | MenhirState1207 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1631 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1632)
    | MenhirState1206 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1633 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1634)
    | MenhirState1205 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1635 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1636)
    | MenhirState1204 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1637 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1638)
    | MenhirState1199 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1639 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1640)
    | MenhirState1198 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1641 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1642)
    | MenhirState1197 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1643 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1644)
    | MenhirState1196 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1645 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1646)
    | MenhirState1195 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1647 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1648)
    | MenhirState1194 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1649 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1650)
    | MenhirState1193 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1651 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1652)
    | MenhirState1192 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1653 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 82451 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1654)
    | MenhirState1189 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1655 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1656)
    | MenhirState1188 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1657 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 82465 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1658)
    | MenhirState1185 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1659 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1660)
    | MenhirState1184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1661 * Lexing.position * _menhir_state * 'tv_val_ident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1662)
    | MenhirState1183 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1663 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1664)
    | MenhirState1181 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1665 * Lexing.position * _menhir_state * 'tv_item_attribute) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1666)
    | MenhirState1176 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1667 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1668)
    | MenhirState1175 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1669 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1670)
    | MenhirState1174 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1671 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1672)
    | MenhirState1171 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1673 * _menhir_state * 'tv_item_extension_sugar * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1674)
    | MenhirState1167 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1675 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1676)
    | MenhirState1166 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1677 * Lexing.position * _menhir_state * 'tv_let_bindings * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1678)
    | MenhirState1161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1679 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_post_item_attributes) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1680)
    | MenhirState1159 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1681 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1682)
    | MenhirState1158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1683 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1684)
    | MenhirState1157 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1685 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1686)
    | MenhirState1156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1687 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1688)
    | MenhirState1155 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1689 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1690)
    | MenhirState1154 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1691 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1692)
    | MenhirState1149 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1695 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1693 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1149 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce206 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1694)) : 'freshtv1696)
    | MenhirState1147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1697 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1698)
    | MenhirState1144 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1701 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1699 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1144 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce211 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1700)) : 'freshtv1702)
    | MenhirState1142 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1703 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1704)
    | MenhirState1139 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1705 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1706)
    | MenhirState1138 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1707 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1708)
    | MenhirState1137 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1709 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1710)
    | MenhirState1135 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1711 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce602 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1712)
    | MenhirState1134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1713 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1714)
    | MenhirState1131 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1715 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1716)
    | MenhirState1124 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1717 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce474 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1718)
    | MenhirState1123 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1719 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error1122 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv1720)
    | MenhirState1120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1721 * _menhir_state * 'tv_lbl_expr) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce478 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1722)
    | MenhirState1117 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1723 * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        _menhir_error1116 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv1724)
    | MenhirState1115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1725 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1726)
    | MenhirState1113 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1727 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1728)
    | MenhirState1109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1729 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1109 : 'freshtv1730)
    | MenhirState1108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1731 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1732)
    | MenhirState1103 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1733 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1734)
    | MenhirState1102 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1735 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1736)
    | MenhirState1099 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1737 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1738)
    | MenhirState1098 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1739 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1740)
    | MenhirState1092 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1741 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) = Obj.magic _menhir_stack in
        (_menhir_error1091 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1742)
    | MenhirState1089 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1743 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1744)
    | MenhirState1087 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1745 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1746)
    | MenhirState1086 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1747 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce419 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1748)
    | MenhirState1084 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1749 * _menhir_state * 'tv_expr_comma_seq) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce635 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1750)
    | MenhirState1083 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1751 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_seq) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState1083 : 'freshtv1752)
    | MenhirState1081 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1753 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1754)
    | MenhirState1078 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1757 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1755 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState1078 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce215 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv1756)) : 'freshtv1758)
    | MenhirState1076 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1759 * _menhir_state * 'tv_expr_comma_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1760)
    | MenhirState1074 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1761 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_expr_comma_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1762)
    | MenhirState1071 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1763 * Lexing.position * _menhir_state * 'tv_core_type * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1764)
    | MenhirState1068 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1765 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1766)
    | MenhirState1066 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1767 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1768)
    | MenhirState1065 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1769 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1770)
    | MenhirState1063 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1771 * _menhir_state * 'tv_expr_optional_constraint * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1772)
    | MenhirState1061 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1773 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce100 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1774)
    | MenhirState1060 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1775 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce102 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1776)
    | MenhirState1059 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1777 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error1058 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv1778)
    | MenhirState1057 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1779 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce105 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1780)
    | MenhirState1056 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1783 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1781 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce221 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1782)) : 'freshtv1784)
    | MenhirState1054 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1785 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1786)
    | MenhirState1053 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1787 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1788)
    | MenhirState1052 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1789 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce104 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1790)
    | MenhirState1051 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1793 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1791 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce219 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1792)) : 'freshtv1794)
    | MenhirState1048 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1795 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1049 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1048 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1796)
    | MenhirState1047 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1797 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1798)
    | MenhirState1046 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1799 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce109 _menhir_env (Obj.magic _menhir_stack) : 'freshtv1800)
    | MenhirState1045 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1801 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1802)
    | MenhirState1044 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1803 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1804)
    | MenhirState1043 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1805 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1806)
    | MenhirState1041 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1807 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1808)
    | MenhirState1038 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1809 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_leading_bar_match_cases) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1810)
    | MenhirState1037 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1811 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1812)
    | MenhirState1036 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1813 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1814)
    | MenhirState1035 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1815 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1816)
    | MenhirState1034 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1817 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1818)
    | MenhirState1033 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1819 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1820)
    | MenhirState1032 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1821 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1822)
    | MenhirState1027 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1823 * _menhir_state * 'tv_labeled_simple_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1824)
    | MenhirState1025 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1825 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1826)
    | MenhirState1024 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1827 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1828)
    | MenhirState1023 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1829 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1830)
    | MenhirState1022 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1831 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1832)
    | MenhirState1021 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1833 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1834)
    | MenhirState1020 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1835 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1836)
    | MenhirState1019 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1837 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1838)
    | MenhirState1018 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1839 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1840)
    | MenhirState1016 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1841 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1842)
    | MenhirState1015 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1843 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1844)
    | MenhirState1014 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1845 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1846)
    | MenhirState1013 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1847 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1848)
    | MenhirState1012 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1849 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1850)
    | MenhirState1011 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1851 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1852)
    | MenhirState1009 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1853 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1854)
    | MenhirState1008 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1855 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1856)
    | MenhirState1007 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1857 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1858)
    | MenhirState1006 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1859 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1860)
    | MenhirState1003 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1861 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1004 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1003 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1862)
    | MenhirState1002 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1863 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1864)
    | MenhirState1001 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1865 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1866)
    | MenhirState1000 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1867 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1868)
    | MenhirState999 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1869 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1870)
    | MenhirState998 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1871 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1872)
    | MenhirState997 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1873 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1874)
    | MenhirState996 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1875 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1876)
    | MenhirState995 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1877 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1878)
    | MenhirState994 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1879 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1880)
    | MenhirState993 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1881 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1882)
    | MenhirState992 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1883 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1884)
    | MenhirState991 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1885 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1886)
    | MenhirState990 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1887 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 832 "reason_parser.mly"
       (string)
# 83023 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1888)
    | MenhirState989 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1889 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 832 "reason_parser.mly"
       (string)
# 83032 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1890)
    | MenhirState988 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1891 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1892)
    | MenhirState987 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1893 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1894)
    | MenhirState986 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1895 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1896)
    | MenhirState985 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1897 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1898)
    | MenhirState984 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1899 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 833 "reason_parser.mly"
       (string)
# 83061 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1900)
    | MenhirState983 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1901 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 833 "reason_parser.mly"
       (string)
# 83070 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1902)
    | MenhirState982 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1903 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 834 "reason_parser.mly"
       (string)
# 83079 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1904)
    | MenhirState981 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1905 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 834 "reason_parser.mly"
       (string)
# 83088 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1906)
    | MenhirState980 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1907 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1908)
    | MenhirState979 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1909 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1910)
    | MenhirState978 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1911 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1912)
    | MenhirState977 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1913 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1914)
    | MenhirState976 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1915 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1916)
    | MenhirState975 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1917 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1918)
    | MenhirState974 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1919 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1920)
    | MenhirState973 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1921 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1922)
    | MenhirState972 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1923 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1924)
    | MenhirState971 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1925 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1926)
    | MenhirState970 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1927 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1928)
    | MenhirState969 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1929 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1930)
    | MenhirState968 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1931 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 835 "reason_parser.mly"
       (string)
# 83157 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1932)
    | MenhirState967 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1933 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 835 "reason_parser.mly"
       (string)
# 83166 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1934)
    | MenhirState966 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1935 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1936)
    | MenhirState965 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1937 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1938)
    | MenhirState964 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1939 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1940)
    | MenhirState963 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1941 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1942)
    | MenhirState961 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1943 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1944)
    | MenhirState960 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1945 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1946)
    | MenhirState955 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1947 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1948)
    | MenhirState953 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1949 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1950)
    | MenhirState951 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1951 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 836 "reason_parser.mly"
       (string)
# 83215 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1952)
    | MenhirState950 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1953 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 836 "reason_parser.mly"
       (string)
# 83224 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1954)
    | MenhirState949 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1955 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1956)
    | MenhirState948 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1957 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1958)
    | MenhirState947 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1959 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1960)
    | MenhirState946 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1961 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1962)
    | MenhirState940 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1963 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1964)
    | MenhirState938 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1965 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 83258 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1966)
    | MenhirState936 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1967 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1968)
    | MenhirState935 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1969 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 83272 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1970)
    | MenhirState933 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1971 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1972)
    | MenhirState931 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1973 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1974)
    | MenhirState930 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1975 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1976)
    | MenhirState929 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1977 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1978)
    | MenhirState926 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1979 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run927 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState926 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv1980)
    | MenhirState925 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv1981 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1982)
    | MenhirState924 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1983 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1984)
    | MenhirState923 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1985 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1986)
    | MenhirState922 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1987 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1988)
    | MenhirState921 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv1989 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1990)
    | MenhirState920 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv1991 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1992)
    | MenhirState914 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv1993 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1994)
    | MenhirState913 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv1995 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1996)
    | MenhirState912 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv1997 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv1998)
    | MenhirState911 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv1999 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2000)
    | MenhirState910 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2001 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2002)
    | MenhirState909 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2003 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2004)
    | MenhirState908 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2005 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2006)
    | MenhirState907 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2007 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2008)
    | MenhirState906 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2009 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 83375 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2010)
    | MenhirState903 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2011 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2012)
    | MenhirState902 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2013 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 83389 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2014)
    | MenhirState899 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2015 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2016)
    | MenhirState898 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2017 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2018)
    | MenhirState897 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2019 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2020)
    | MenhirState896 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2021 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2022)
    | MenhirState895 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2023 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2024)
    | MenhirState894 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2025 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2026)
    | MenhirState893 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2027 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2028)
    | MenhirState892 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2029 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce103 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2030)
    | MenhirState891 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2033 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2031 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce217 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2032)) : 'freshtv2034)
    | MenhirState888 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2035 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run889 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState888 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2036)
    | MenhirState887 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2037 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce95 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2038)
    | MenhirState886 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2039 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2040)
    | MenhirState885 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2041 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce94 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2042)
    | MenhirState884 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2043 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2044)
    | MenhirState883 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2045 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce93 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2046)
    | MenhirState882 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2047 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2048)
    | MenhirState881 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2049 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce92 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2050)
    | MenhirState880 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2051 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2052)
    | MenhirState879 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2053 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce90 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2054)
    | MenhirState878 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2055 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2056)
    | MenhirState877 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2057 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 832 "reason_parser.mly"
       (string)
# 83493 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce79 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2058)
    | MenhirState876 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2059 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 832 "reason_parser.mly"
       (string)
# 83501 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2060)
    | MenhirState875 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2061 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce89 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2062)
    | MenhirState874 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2063 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2064)
    | MenhirState873 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2065 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce99 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2066)
    | MenhirState872 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2067 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2068)
    | MenhirState871 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2069 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 833 "reason_parser.mly"
       (string)
# 83528 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce80 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2070)
    | MenhirState870 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2071 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 833 "reason_parser.mly"
       (string)
# 83536 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2072)
    | MenhirState869 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2073 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 834 "reason_parser.mly"
       (string)
# 83545 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce81 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2074)
    | MenhirState868 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2075 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 834 "reason_parser.mly"
       (string)
# 83553 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2076)
    | MenhirState867 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2077 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce98 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2078)
    | MenhirState866 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2079 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2080)
    | MenhirState865 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2081 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce86 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2082)
    | MenhirState864 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2083 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2084)
    | MenhirState863 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2085 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce87 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2086)
    | MenhirState862 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2087 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2088)
    | MenhirState861 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2089 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce91 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2090)
    | MenhirState860 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2091 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2092)
    | MenhirState859 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2093 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce84 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2094)
    | MenhirState858 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2095 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2096)
    | MenhirState857 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2097 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce85 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2098)
    | MenhirState856 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2099 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2100)
    | MenhirState855 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2101 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 835 "reason_parser.mly"
       (string)
# 83616 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce82 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2102)
    | MenhirState854 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2103 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 835 "reason_parser.mly"
       (string)
# 83624 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2104)
    | MenhirState853 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2105 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce97 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2106)
    | MenhirState852 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2107 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2108)
    | MenhirState851 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2109 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce96 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2110)
    | MenhirState850 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2111 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2112)
    | MenhirState848 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2113 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce101 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2114)
    | MenhirState847 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2115 * Lexing.position * _menhir_state * 'tv_additive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2116)
    | MenhirState842 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2117 * Lexing.position * _menhir_state * 'tv_simple_non_labeled_expr_list * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce268 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2118)
    | MenhirState840 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2119 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce203 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2120)
    | MenhirState838 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2121 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 836 "reason_parser.mly"
       (string)
# 83668 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce83 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2122)
    | MenhirState837 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2123 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 836 "reason_parser.mly"
       (string)
# 83676 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2124)
    | MenhirState836 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2125 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce88 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2126)
    | MenhirState835 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2127 * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2128)
    | MenhirState834 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2129 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce106 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2130)
    | MenhirState833 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2131 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2132)
    | MenhirState827 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2133 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_labeled_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce65 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2134)
    | MenhirState825 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2135 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 83707 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error821 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2136)
    | MenhirState823 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2137 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce495 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2138)
    | MenhirState822 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2139 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 83720 "reason_parser.ml"
        ) * Lexing.position)) = Obj.magic _menhir_stack in
        (_menhir_error821 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2140)
    | MenhirState820 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2141 * Lexing.position * _menhir_state * 'tv_less_aggressive_simple_expression * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce64 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2142)
    | MenhirState818 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2143 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce73 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2144)
    | MenhirState817 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2145 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce204 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2146)
    | MenhirState816 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2147 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2148)
    | MenhirState815 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2149 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2150)
    | MenhirState814 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2151 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce495 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2152)
    | MenhirState813 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2153 * Lexing.position * _menhir_state * 'tv_subtractive * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2154)
    | MenhirState812 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2155 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce107 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2156)
    | MenhirState811 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2157 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2158)
    | MenhirState805 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2159 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2160)
    | MenhirState801 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2163 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2161 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState801 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce249 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2162)) : 'freshtv2164)
    | MenhirState800 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2165 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2166)
    | MenhirState799 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2167 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2168)
    | MenhirState798 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2169 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2170)
    | MenhirState797 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2171 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2172)
    | MenhirState796 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2173 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2174)
    | MenhirState794 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2175 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2176)
    | MenhirState793 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2177 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2178)
    | MenhirState791 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2179 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2180)
    | MenhirState780 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2181 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2182)
    | MenhirState775 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2185 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2183 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState775 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce247 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2184)) : 'freshtv2186)
    | MenhirState773 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2187 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2188)
    | MenhirState772 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2189 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2190)
    | MenhirState771 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2191 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2192)
    | MenhirState770 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2193 * _menhir_state * 'tv_pattern_with_bar) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2194)
    | MenhirState768 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2195 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2196)
    | MenhirState767 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2197 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2198)
    | MenhirState766 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2199 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2200)
    | MenhirState765 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2201 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2202)
    | MenhirState764 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2203 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2204)
    | MenhirState763 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2205 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2206)
    | MenhirState762 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2207 * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2208)
    | MenhirState761 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2209 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_module_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2210)
    | MenhirState759 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2211 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2212)
    | MenhirState758 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2213 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_functor_args * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2214)
    | MenhirState757 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2215 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2216)
    | MenhirState756 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2217 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2218)
    | MenhirState755 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2219 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2220)
    | MenhirState754 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2221 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2222)
    | MenhirState753 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2223 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce374 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2224)
    | MenhirState749 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2225 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2226)
    | MenhirState748 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2227 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce77 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2228)
    | MenhirState747 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2229 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2230)
    | MenhirState746 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2231 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * 'tv_direction_flag) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2232)
    | MenhirState743 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2233 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2234)
    | MenhirState741 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2235 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2236)
    | MenhirState740 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2237 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2238)
    | MenhirState739 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2239 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2240)
    | MenhirState738 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2241 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2242)
    | MenhirState737 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2243 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2244)
    | MenhirState736 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2245 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2246)
    | MenhirState735 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2247 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 895 "reason_parser.mly"
       (string)
# 83986 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2248)
    | MenhirState734 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2249 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 895 "reason_parser.mly"
       (string)
# 83995 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2250)
    | MenhirState733 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2251 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2252)
    | MenhirState730 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2253 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2254)
    | MenhirState729 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2255 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 84014 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2256)
    | MenhirState722 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2259 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2257 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState722 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce289 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2258)) : 'freshtv2260)
    | MenhirState719 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2263 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2261 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState719 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce286 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2262)) : 'freshtv2264)
    | MenhirState714 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2267 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((((('freshtv2265 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState714 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce171 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2266)) : 'freshtv2268)
    | MenhirState713 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((((('freshtv2269 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2270)
    | MenhirState712 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2271 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2272)
    | MenhirState711 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2273 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2274)
    | MenhirState705 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2275 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState705 : 'freshtv2276)
    | MenhirState700 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2277 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState700 : 'freshtv2278)
    | MenhirState699 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2279 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce481 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2280)
    | MenhirState692 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2281 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2282)
    | MenhirState688 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2283 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce60 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2284)
    | MenhirState687 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2285 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2286)
    | MenhirState686 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2287 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce55 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2288)
    | MenhirState681 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2291 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2289 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState681 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce289 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2290)) : 'freshtv2292)
    | MenhirState678 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2295 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2293 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState678 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce286 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2294)) : 'freshtv2296)
    | MenhirState676 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2297 * _menhir_state * 'tv_pattern_optional_constraint) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2298)
    | MenhirState668 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2299 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2300)
    | MenhirState666 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2301 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2302)
    | MenhirState665 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2303 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2304)
    | MenhirState660 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2305 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2306)
    | MenhirState658 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2309 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2307 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState658 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce165 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2308)) : 'freshtv2310)
    | MenhirState657 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2311 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2312)
    | MenhirState655 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2315 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2313 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState655 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce169 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2314)) : 'freshtv2316)
    | MenhirState654 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2317 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2318)
    | MenhirState650 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2319 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2320)
    | MenhirState649 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2321 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2322)
    | MenhirState646 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2323 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2324)
    | MenhirState644 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2325 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2326)
    | MenhirState642 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2327 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2328)
    | MenhirState635 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2329 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2330)
    | MenhirState634 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2331 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2332)
    | MenhirState630 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2333 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2334)
    | MenhirState619 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2335 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2336)
    | MenhirState615 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2337 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2338)
    | MenhirState614 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2339 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce660 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2340)
    | MenhirState613 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2341 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce662 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2342)
    | MenhirState612 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2343 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce661 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2344)
    | MenhirState611 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2345 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2346)
    | MenhirState610 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2347 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2348)
    | MenhirState609 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2349 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2350)
    | MenhirState605 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2351 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list_extension * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState605 : 'freshtv2352)
    | MenhirState603 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2353 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2354)
    | MenhirState602 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2355 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce658 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2356)
    | MenhirState594 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2357 * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2358)
    | MenhirState593 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2359 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) * Lexing.position * _menhir_state * 'tv_pattern * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce659 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2360)
    | MenhirState592 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2361 * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2362)
    | MenhirState590 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2363 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_comma_list * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce636 _menhir_env (Obj.magic _menhir_stack) MenhirState590 : 'freshtv2364)
    | MenhirState588 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2365 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2366)
    | MenhirState587 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2367 * _menhir_state * 'tv_pattern_two_or_more_comma_list) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2368)
    | MenhirState579 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2369 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2370)
    | MenhirState577 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2373 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2371 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState577 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce165 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2372)) : 'freshtv2374)
    | MenhirState576 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2375 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce168 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2376)
    | MenhirState574 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2379 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2377 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState574 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce169 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos) : 'freshtv2378)) : 'freshtv2380)
    | MenhirState573 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2381 * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce656 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2382)
    | MenhirState564 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2383 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_pattern_list) = Obj.magic _menhir_stack in
        (_menhir_reduce166 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2384)
    | MenhirState563 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2385 * Lexing.position * _menhir_state * 'tv_constr_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce274 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2386)
    | MenhirState554 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2387 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2388)
    | MenhirState553 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2389 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce374 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2390)
    | MenhirState550 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2391 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce275 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2392)
    | MenhirState548 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2393 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_pattern_without_or * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce173 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2394)
    | MenhirState545 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2395 * Lexing.position * _menhir_state * 'tv_signed_constant * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _) = _menhir_stack in
        _menhir_error544 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2396)
    | MenhirState540 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2397 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2398)
    | MenhirState533 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2399 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2400)
    | MenhirState532 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2401 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2402)
    | MenhirState531 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2403 * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error530 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2404)
    | MenhirState526 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2405 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState526 : 'freshtv2406)
    | MenhirState525 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2407 * _menhir_state * 'tv_lbl_pattern) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce484 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2408)
    | MenhirState519 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2409 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2410)
    | MenhirState518 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2411 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2412)
    | MenhirState516 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2413 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2414)
    | MenhirState514 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2415 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2416)
    | MenhirState511 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2417 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2418)
    | MenhirState505 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2419 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 84407 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce453 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2420)
    | MenhirState504 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2421 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2422)
    | MenhirState503 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2423 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2424)
    | MenhirState499 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2425 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 84425 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2426)
    | MenhirState495 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2427 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2428)
    | MenhirState487 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2429 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2430)
    | MenhirState486 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2431 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 84444 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2432)
    | MenhirState484 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2433 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 84453 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2434)
    | MenhirState481 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2435 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2436)
    | MenhirState480 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2437 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 84467 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2438)
    | MenhirState477 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2439 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2440)
    | MenhirState476 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2441 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2442)
    | MenhirState475 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2443 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState475 : 'freshtv2444)
    | MenhirState474 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2445 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2446)
    | MenhirState473 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2447 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2448)
    | MenhirState472 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2449 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2450)
    | MenhirState471 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2451 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 895 "reason_parser.mly"
       (string)
# 84505 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce244 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2452)
    | MenhirState470 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2453 * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) * Lexing.position * _menhir_state * (
# 895 "reason_parser.mly"
       (string)
# 84513 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2454)
    | MenhirState469 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2455 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_simple_expr * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce236 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2456)
    | MenhirState465 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2457 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2458)
    | MenhirState464 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2459 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2460)
    | MenhirState457 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2461 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2462)
    | MenhirState456 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2463 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState456 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2464)
    | MenhirState455 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2465 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2466)
    | MenhirState454 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2467 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState454 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2468)
    | MenhirState453 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2469 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2470)
    | MenhirState452 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2471 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState452 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2472)
    | MenhirState451 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2473 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2474)
    | MenhirState449 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2475 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_functor_arg_name * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2476)
    | MenhirState444 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2477 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2478)
    | MenhirState443 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2479 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 84578 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2480)
    | MenhirState442 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2481 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2482)
    | MenhirState441 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2483 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2484)
    | MenhirState440 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2485 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState440 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2486)
    | MenhirState439 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2487 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2488)
    | MenhirState438 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2489 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2490)
    | MenhirState437 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2491 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2492)
    | MenhirState436 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2493 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2494)
    | MenhirState434 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2495 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2496)
    | MenhirState432 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2497 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState432 : 'freshtv2498)
    | MenhirState426 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2499 * _menhir_state * 'tv_field_expr_list) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        (_menhir_reduce635 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2500)
    | MenhirState425 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2501 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_field_expr_list) = Obj.magic _menhir_stack in
        (_menhir_reduce634 _menhir_env (Obj.magic _menhir_stack) MenhirState425 : 'freshtv2502)
    | MenhirState423 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2503 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2504)
    | MenhirState422 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2505 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 84643 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error421 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2506)
    | MenhirState420 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2507 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2508)
    | MenhirState419 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2509 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2510)
    | MenhirState418 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2511 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2512)
    | MenhirState417 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2513 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2514)
    | MenhirState416 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2515 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2516)
    | MenhirState415 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2517 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2518)
    | MenhirState413 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2519 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2520)
    | MenhirState412 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2521 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2522)
    | MenhirState411 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2523 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2524)
    | MenhirState408 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2525 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2526)
    | MenhirState405 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2527 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2528)
    | MenhirState404 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2529 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2530)
    | MenhirState403 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2531 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2532)
    | MenhirState401 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2533 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2534)
    | MenhirState400 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2535 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_ident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState400 : 'freshtv2536)
    | MenhirState399 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2537 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2538)
    | MenhirState397 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2539 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState397 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2540)
    | MenhirState396 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2541 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2542)
    | MenhirState395 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2543 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce673 _menhir_env (Obj.magic _menhir_stack) MenhirState395 : 'freshtv2544)
    | MenhirState394 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2545 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_override_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2546)
    | MenhirState392 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2547 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2548)
    | MenhirState389 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2549 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2550)
    | MenhirState337 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2551 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2552)
    | MenhirState334 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2553 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState334 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2554)
    | MenhirState333 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2555 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2556)
    | MenhirState332 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2557 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce57 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2558)
    | MenhirState329 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2559 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce134 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2560)
    | MenhirState328 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2561 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce141 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2562)
    | MenhirState327 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2563 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 84785 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2564)
    | MenhirState325 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2565 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 84794 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2566)
    | MenhirState324 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2569 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 84803 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2567 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 84810 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce823 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2568)) : 'freshtv2570)
    | MenhirState317 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2571 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        (_menhir_reduce135 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2572)
    | MenhirState313 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2573 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * 'tv_row_field) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2574)
    | MenhirState307 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2575 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2576)
    | MenhirState306 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2577 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2578)
    | MenhirState303 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2579 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2580)
    | MenhirState301 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2581 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2582)
    | MenhirState296 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2583 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) * _menhir_state * 'tv_name_tag_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2584)
    | MenhirState295 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2585 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) * _menhir_state * 'tv_row_field_list) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2586)
    | MenhirState292 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2587 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2588)
    | MenhirState291 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2589 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2590)
    | MenhirState288 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2591 * _menhir_state * 'tv_field) * Lexing.position * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2592)
    | MenhirState287 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2593 * _menhir_state * 'tv_field) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2594)
    | MenhirState284 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2595 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2596)
    | MenhirState277 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2597 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2598)
    | MenhirState273 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2599 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2600)
    | MenhirState272 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2601 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2602)
    | MenhirState271 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2603 * Lexing.position * _menhir_state * 'tv_core_type2 * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2604)
    | MenhirState270 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2605 * _menhir_state * 'tv_core_type_comma_list) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2606)
    | MenhirState263 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2607 * Lexing.position * _menhir_state * 'tv_non_arrowed_core_type * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2608)
    | MenhirState255 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2609 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) * Lexing.position * _menhir_state * 'tv_non_arrowed_simple_core_type_list) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2610)
    | MenhirState254 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2611 * Lexing.position * _menhir_state * 'tv_type_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2612)
    | MenhirState251 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2613 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 84922 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2614)
    | MenhirState250 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2615 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2616)
    | MenhirState249 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2617 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2618)
    | MenhirState246 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2619 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2620)
    | MenhirState245 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2621 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2622)
    | MenhirState244 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2623 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce139 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2624)
    | MenhirState243 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2625 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2626)
    | MenhirState242 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2627 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2628)
    | MenhirState240 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2629 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2630)
    | MenhirState239 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2631 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _) = _menhir_stack in
        _menhir_error238 _menhir_env (Obj.magic _menhir_stack)) : 'freshtv2632)
    | MenhirState234 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2633 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 84975 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce453 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2634)
    | MenhirState233 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2635 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 906 "reason_parser.mly"
       (string)
# 84983 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let (((_menhir_stack, _, _menhir_s, _), _, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2636)
    | MenhirState216 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2637 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2638)
    | MenhirState215 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2639 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2640)
    | MenhirState212 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2641 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2642)
    | MenhirState210 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2643 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2644)
    | MenhirState196 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2645 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2646)
    | MenhirState195 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2647 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2648)
    | MenhirState193 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2649 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2650)
    | MenhirState191 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2651 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2652)
    | MenhirState190 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2653 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState190 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2654)
    | MenhirState189 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2655 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2656)
    | MenhirState187 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2657 * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 85041 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2658)
    | MenhirState184 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2659 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2660)
    | MenhirState181 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2661 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2662)
    | MenhirState180 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2663 * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 85060 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2664)
    | MenhirState178 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2665 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2666)
    | MenhirState177 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2667 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2668)
    | MenhirState175 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2669 * _menhir_state) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState175 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2670)
    | MenhirState174 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2671 * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2672)
    | MenhirState173 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2673 * Lexing.position * _menhir_state * 'tv_name_tag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2674)
    | MenhirState170 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2675 * _menhir_state * 'tv_row_field_list) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2676)
    | MenhirState164 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2677 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2678)
    | MenhirState163 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2679 * Lexing.position * _menhir_state * Lexing.position) * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2680)
    | MenhirState162 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2681 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2682)
    | MenhirState161 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2683 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2684)
    | MenhirState159 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2685 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2686)
    | MenhirState158 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2687 * _menhir_state * Lexing.position) * _menhir_state * 'tv_opt_bar) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2688)
    | MenhirState156 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2689 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2690)
    | MenhirState154 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2691 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2692)
    | MenhirState153 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2693 * Lexing.position * _menhir_state * 'tv_label * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2694)
    | MenhirState147 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2695 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2696)
    | MenhirState143 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2697 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 85148 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2698)
    | MenhirState141 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2699 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_label_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2700)
    | MenhirState138 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2701 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2702)
    | MenhirState137 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2703 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2704)
    | MenhirState135 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2705 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2706)
    | MenhirState134 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2707 * Lexing.position * _menhir_state * 'tv_mty_longident * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2708)
    | MenhirState130 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2709 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85182 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2710)
    | MenhirState129 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2711 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2712)
    | MenhirState128 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2713 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2714)
    | MenhirState124 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2715 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2716)
    | MenhirState122 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2717 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2718)
    | MenhirState121 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2719 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_class_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2720)
    | MenhirState120 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2721 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2722)
    | MenhirState118 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2723 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85221 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2724)
    | MenhirState115 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2725 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2726)
    | MenhirState112 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2727 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2728)
    | MenhirState111 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2729 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2730)
    | MenhirState109 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2731 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2732)
    | MenhirState108 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2733 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85250 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce448 _menhir_env (Obj.magic _menhir_stack) MenhirState108 : 'freshtv2734)
    | MenhirState106 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2735 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2736)
    | MenhirState105 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2737 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 85263 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * 'tv_optional_type_parameters) = Obj.magic _menhir_stack in
        (_menhir_reduce812 _menhir_env (Obj.magic _menhir_stack) MenhirState105 : 'freshtv2738)
    | MenhirState102 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2739 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2740)
    | MenhirState98 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2741 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _), _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2742)
    | MenhirState92 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2743 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2744)
    | MenhirState90 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2745 * Lexing.position * _menhir_state * (
# 857 "reason_parser.mly"
       (string)
# 85286 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_reduce640 _menhir_env (Obj.magic _menhir_stack) MenhirState90 : 'freshtv2746)
    | MenhirState88 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2747 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2748)
    | MenhirState87 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2749 * Lexing.position * _menhir_state * 'tv_mod_ext2 * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
        ((let ((_menhir_stack, _, _menhir_s, _, _), _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2750)
    | MenhirState84 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (((('freshtv2751 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85304 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2752)
    | MenhirState83 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ((('freshtv2753 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85313 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2754)
    | MenhirState82 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2755 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85322 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2756)
    | MenhirState81 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2757 * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2758)
    | MenhirState79 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : (('freshtv2759 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85336 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_mod_ext_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2760)
    | MenhirState78 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2761 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85345 "reason_parser.ml"
        ) * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2762)
    | MenhirState77 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2763 * Lexing.position * _menhir_state * (
# 906 "reason_parser.mly"
       (string)
# 85354 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2764)
    | MenhirState76 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2765 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_nonrec_flag * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2766)
    | MenhirState74 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2767 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2768)
    | MenhirState73 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2769 * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * 'tv_attr_id) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState73 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2770)
    | MenhirState71 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2771 * Lexing.position * _menhir_state * 'tv_single_attr_id * Lexing.position)) = Obj.magic _menhir_stack in
        (_menhir_error70 _menhir_env (Obj.magic _menhir_stack) : 'freshtv2772)
    | MenhirState19 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2773 * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2774)
    | MenhirState17 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2775 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2776)
    | MenhirState15 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2777 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2778)
    | MenhirState10 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : ('freshtv2779 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) * _menhir_state) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _menhir_s) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2780)
    | MenhirState9 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2781 * Lexing.position * _menhir_state * 'tv_mod_longident * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2782)
    | MenhirState7 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2783 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2784)
    | MenhirState6 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2785 * Lexing.position * _menhir_state * (
# 882 "reason_parser.mly"
       (string)
# 85411 "reason_parser.ml"
        ) * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2786)
    | MenhirState2 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2787 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv2788)
    | MenhirState1 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2789 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2790)
    | MenhirState0 ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv2791) = Obj.magic _menhir_stack in
        (_menhir_run1 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p : 'freshtv2792)

and _menhir_run417 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState417 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState417

and _menhir_run3 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 906 "reason_parser.mly"
       (string)
# 85485 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce576 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run2583 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | NONREC ->
        _menhir_run75 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2583 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUAL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce604 _menhir_env (Obj.magic _menhir_stack) MenhirState2583

and _menhir_run419 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState419 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState419

and _menhir_run4 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce378 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run764 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState764 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState764

and _menhir_run5 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 898 "reason_parser.mly"
       (string * string option)
# 85617 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce369 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run2619 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2619 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2619

and _menhir_run6 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 882 "reason_parser.mly"
       (string)
# 85641 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState6 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState6

and _menhir_run335 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce322 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run336 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce321 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run337 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run387 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run386 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run385 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run384 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run383 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run382 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run381 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run380 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run379 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run378 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run377 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run376 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run375 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run374 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run373 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run372 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run371 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run370 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run369 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run368 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run367 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run366 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run365 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run364 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run363 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run362 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run361 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run360 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run359 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run358 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run357 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run356 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run355 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run354 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run353 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run352 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run351 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run350 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run349 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run348 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run347 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run346 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run345 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run344 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run343 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run342 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run341 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run340 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run339 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run338 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState337 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState337

and _menhir_run1400 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BANG ->
        _menhir_run393 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1400 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _ ->
        _menhir_reduce644 _menhir_env (Obj.magic _menhir_stack) MenhirState1400
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1400

and _menhir_run7 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run8 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState7 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState7 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState7

and _menhir_run14 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 871 "reason_parser.mly"
       (nativeint)
# 85851 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce373 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run2628 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv51 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | LIDENT _v ->
            _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2629 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UIDENT _v ->
            _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2629 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2629) : 'freshtv52)
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv53 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv54)

and _menhir_run406 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce796 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run407 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce795 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run771 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER ->
        _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AMPERSAND ->
        _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run805 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARBAR ->
        _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | COLONCOLON ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv49 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState771 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | RPAREN ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv45 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LPAREN ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv41 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | ASSERT ->
                    _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BACKQUOTE ->
                    _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | BANG ->
                    _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | CHAR _v ->
                    _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FALSE ->
                    _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FLOAT _v ->
                    _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FOR ->
                    _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | FUN ->
                    _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | IF ->
                    _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT _v ->
                    _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT32 _v ->
                    _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | INT64 _v ->
                    _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LAZY ->
                    _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACE ->
                    _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACELESS ->
                    _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKET ->
                    _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETBAR ->
                    _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LBRACKETPERCENT ->
                    _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LIDENT _v ->
                    _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | LPAREN ->
                    _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUS ->
                    _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | MINUSDOT ->
                    _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NATIVEINT _v ->
                    _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | NEW ->
                    _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUS ->
                    _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PLUSDOT ->
                    _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | PREFIXOP _v ->
                    _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | STRING _v ->
                    _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | SWITCH ->
                    _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRUE ->
                    _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | TRY ->
                    _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | UIDENT _v ->
                    _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | WHILE ->
                    _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState791 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState791) : 'freshtv42)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv43 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * Lexing.position) = Obj.magic _menhir_stack in
                ((let ((_menhir_stack, _, _menhir_s, _), _, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv44)) : 'freshtv46)
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv47 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv48)) : 'freshtv50)
    | COLONEQUAL ->
        _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | GREATER ->
        _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP0 _v ->
        _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP1 _v ->
        _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP2 _v ->
        _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP3 _v ->
        _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INFIXOP4 _v ->
        _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run788 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run787 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run772 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run410 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSEQ ->
        _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run187 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RPAREN ->
        _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STAR ->
        _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState771 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState771

and _menhir_run792 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 857 "reason_parser.mly"
       (string)
# 86130 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _v, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AMPERAMPER | AMPERSAND | AND | AS | ASSERT | BACKQUOTE | BANG | BAR | BARBAR | BARRBRACKET | BEGIN | CHAR _ | CLASS | COLON | COLONCOLON | COLONEQUAL | COLONGREATER | COMMA | COMMENT _ | CONSTRAINT | DO | DONE | DOT | DOTDOT | DOTDOTDOT | DOWNTO | ELSE | END | EOF | EOL | EQUALGREATER | EXCEPTION | EXPLICITLY_PASSED_OPTIONAL | EXTERNAL | FALSE | FLOAT _ | FOR | FUN | FUNCTION | FUNCTOR | GREATER | GREATERRBRACE | GREATERRBRACKET | IF | IN | INCLUDE | INFIXOP0 _ | INFIXOP1 _ | INFIXOP2 _ | INFIXOP3 _ | INFIXOP4 _ | INHERIT | INITIALIZER | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACELESS | LBRACKET | LBRACKETAT | LBRACKETATAT | LBRACKETATATAT | LBRACKETBAR | LBRACKETGREATER | LBRACKETLESS | LBRACKETPERCENT | LBRACKETPERCENTPERCENT | LESS | LESSDOTDOTGREATER | LESSGREATER | LESSMINUS | LET | LIDENT _ | LPAREN | MATCH | METHOD | MINUS | MINUSDOT | MINUSGREATER | MODULE | MUTABLE | NATIVEINT _ | NEW | NONREC | OBJECT | OF | OPEN | OPTIONAL_NO_DEFAULT | OR | PERCENT | PLUS | PLUSDOT | PLUSEQ | PREFIXOP _ | PRIVATE | QUESTION | QUOTE | RBRACE | RBRACKET | REC | RPAREN | SEMI | SEMISEMI | SHARP | SHARPOP _ | SIG | STAR | STRING _ | STRUCT | SWITCH | THEN | TILDE | TO | TRUE | TRY | TYPE | UIDENT _ | UNDERSCORE | VAL | VIRTUAL | WHEN | WHILE | WITH ->
        _menhir_reduce834 _menhir_env (Obj.magic _menhir_stack)
    | EQUAL ->
        _menhir_reduce460 _menhir_env (Obj.magic _menhir_stack)

and _menhir_run2635 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | MODULE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv39 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState2635 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | REC ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv37 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState2636 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | UIDENT _v ->
                _menhir_run2637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2639 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2639) : 'freshtv38)
        | UIDENT _v ->
            _menhir_run2637 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2636 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2636) : 'freshtv40)
    | REC ->
        _menhir_run435 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2635 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE | CHAR _ | EXCEPTION | FALSE | FLOAT _ | INT _ | INT32 _ | INT64 _ | LAZY | LBRACE | LBRACKET | LBRACKETBAR | LBRACKETPERCENT | LIDENT _ | LPAREN | MINUS | NATIVEINT _ | PLUS | SHARP | STRING _ | TRUE | UIDENT _ | UNDERSCORE ->
        _menhir_reduce686 _menhir_env (Obj.magic _menhir_stack) MenhirState2635
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2635

and _menhir_run1474 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState1474 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState1474

and _menhir_run19 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState19 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState19

and _menhir_run178 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BARRBRACKET ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv35 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState178 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce230 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv36)
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState178 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState178

and _menhir_run2032 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AND ->
        _menhir_run69 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | AS ->
        _menhir_run68 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run67 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BEGIN ->
        _menhir_run66 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run65 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT ->
        _menhir_run64 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DO ->
        _menhir_run63 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DONE ->
        _menhir_run62 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOWNTO ->
        _menhir_run61 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ELSE ->
        _menhir_run60 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | END ->
        _menhir_run59 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run58 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run57 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run56 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run55 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run54 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTION ->
        _menhir_run53 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUNCTOR ->
        _menhir_run52 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run51 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IN ->
        _menhir_run50 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run49 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INHERIT ->
        _menhir_run48 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INITIALIZER ->
        _menhir_run47 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run46 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run45 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run44 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | METHOD ->
        _menhir_run43 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run42 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MUTABLE ->
        _menhir_run41 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run40 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NONREC ->
        _menhir_run39 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OBJECT ->
        _menhir_run38 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run37 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OR ->
        _menhir_run36 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PRIVATE ->
        _menhir_run35 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | REC ->
        _menhir_run34 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SIG ->
        _menhir_run33 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRUCT ->
        _menhir_run32 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run31 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | THEN ->
        _menhir_run30 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TO ->
        _menhir_run29 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run28 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run27 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run26 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run25 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VAL ->
        _menhir_run24 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run23 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHEN ->
        _menhir_run22 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run21 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WITH ->
        _menhir_run20 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2032 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2032

and _menhir_run418 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run1099 _menhir_env (Obj.magic _menhir_stack) MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | RBRACKET ->
        _menhir_run604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState418 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState418

and _menhir_run420 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | GREATERRBRACE ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv33 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState420 in
        ((let _menhir_env = _menhir_discard _menhir_env in
        _menhir_reduce240 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s) : 'freshtv34)
    | LIDENT _v ->
        _menhir_run421 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState420 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState420

and _menhir_run475 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | AS ->
        _menhir_run1657 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | DOTDOTDOT ->
        _menhir_run1115 _menhir_env (Obj.magic _menhir_stack) MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run434 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run433 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState475 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CONSTRAINT | INHERIT | INITIALIZER | LBRACKETATATAT | LBRACKETPERCENTPERCENT | METHOD | RBRACE | VAL ->
        _menhir_reduce38 _menhir_env (Obj.magic _menhir_stack) MenhirState475
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState475

and _menhir_run793 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState793 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState793

and _menhir_run620 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 841 "reason_parser.mly"
       (int64)
# 86850 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce372 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run621 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 840 "reason_parser.mly"
       (int32)
# 86859 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce371 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run622 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 839 "reason_parser.mly"
       (int)
# 86868 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce367 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run2644 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | FUN ->
        _menhir_run757 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2644 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run452 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2644 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2644 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run412 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2644 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2644 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2644

and _menhir_run894 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState894 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState894

and _menhir_run898 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BAR ->
        _menhir_run767 _menhir_env (Obj.magic _menhir_stack) MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv31 * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
        let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
        let (_menhir_s : _menhir_state) = MenhirState898 in
        let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
        ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
        let _menhir_env = _menhir_discard _menhir_env in
        let _tok = _menhir_env._menhir_token in
        match _tok with
        | AMPERAMPER ->
            _menhir_run572 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | AMPERSAND ->
            _menhir_run571 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BACKQUOTE ->
            _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BANG ->
            _menhir_run570 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | BARBAR ->
            _menhir_run569 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | CHAR _v ->
            _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | COLONEQUAL ->
            _menhir_run568 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | EXCEPTION ->
            _menhir_run540 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FALSE ->
            _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | FLOAT _v ->
            _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | GREATER ->
            _menhir_run539 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP0 _v ->
            _menhir_run538 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP1 _v ->
            _menhir_run537 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP2 _v ->
            _menhir_run536 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP3 _v ->
            _menhir_run535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INFIXOP4 _v ->
            _menhir_run534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT _v ->
            _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT32 _v ->
            _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | INT64 _v ->
            _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LAZY ->
            _menhir_run532 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACE ->
            _menhir_run516 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKET ->
            _menhir_run514 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETBAR ->
            _menhir_run511 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LBRACKETPERCENT ->
            _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESS ->
            _menhir_run510 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSDOTDOTGREATER ->
            _menhir_run509 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LESSGREATER ->
            _menhir_run508 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LIDENT _v ->
            _menhir_run188 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | LPAREN ->
            _menhir_run495 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUS ->
            _menhir_run489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MINUSDOT ->
            _menhir_run488 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | MODULE ->
            _menhir_run230 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | NATIVEINT _v ->
            _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | OR ->
            _menhir_run229 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PERCENT ->
            _menhir_run228 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUS ->
            _menhir_run222 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSDOT ->
            _menhir_run221 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PLUSEQ ->
            _menhir_run220 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | PREFIXOP _v ->
            _menhir_run219 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | RPAREN ->
            _menhir_run218 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | SHARP ->
            _menhir_run212 _menhir_env (Obj.magic _menhir_stack) MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STAR ->
            _menhir_run185 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | STRING _v ->
            _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TRUE ->
            _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | TYPE ->
            let (_menhir_env : _menhir_env) = _menhir_env in
            let (_menhir_stack : ('freshtv29 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
            let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
            let (_menhir_s : _menhir_state) = MenhirState899 in
            let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
            ((let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
            let _menhir_env = _menhir_discard _menhir_env in
            let _tok = _menhir_env._menhir_token in
            match _tok with
            | LIDENT _v ->
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv25 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                let (_v : (
# 857 "reason_parser.mly"
       (string)
# 87095 "reason_parser.ml"
                )) = _v in
                let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                ((let _menhir_stack = (_menhir_stack, _endpos, _v, _startpos) in
                let _menhir_env = _menhir_discard _menhir_env in
                let _tok = _menhir_env._menhir_token in
                match _tok with
                | RPAREN ->
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv21 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 87107 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    let (_endpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_curr_p in
                    let (_startpos : Lexing.position) = _menhir_env._menhir_lexbuf.Lexing.lex_start_p in
                    ((let _menhir_stack = (_menhir_stack, _endpos, _startpos) in
                    let _menhir_env = _menhir_discard _menhir_env in
                    let _tok = _menhir_env._menhir_token in
                    match _tok with
                    | BACKQUOTE ->
                        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | CHAR _v ->
                        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | COLON ->
                        _menhir_run1022 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | EQUALGREATER ->
                        _menhir_run907 _menhir_env (Obj.magic _menhir_stack) MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FALSE ->
                        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | FLOAT _v ->
                        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT _v ->
                        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT32 _v ->
                        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | INT64 _v ->
                        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACE ->
                        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKET ->
                        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETBAR ->
                        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LBRACKETPERCENT ->
                        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LIDENT _v ->
                        _menhir_run485 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | LPAREN ->
                        _menhir_run903 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | MINUS ->
                        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | NATIVEINT _v ->
                        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | PLUS ->
                        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | SHARP ->
                        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | STRING _v ->
                        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | TRUE ->
                        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UIDENT _v ->
                        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | UNDERSCORE ->
                        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState902 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
                    | _ ->
                        assert (not _menhir_env._menhir_error);
                        _menhir_env._menhir_error <- true;
                        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState902) : 'freshtv22)
                | _ ->
                    assert (not _menhir_env._menhir_error);
                    _menhir_env._menhir_error <- true;
                    let (_menhir_env : _menhir_env) = _menhir_env in
                    let (_menhir_stack : ((('freshtv23 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * (
# 857 "reason_parser.mly"
       (string)
# 87172 "reason_parser.ml"
                    ) * Lexing.position) = Obj.magic _menhir_stack in
                    ((let ((_menhir_stack, _, _menhir_s, _), _, _, _) = _menhir_stack in
                    _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv24)) : 'freshtv26)
            | _ ->
                assert (not _menhir_env._menhir_error);
                _menhir_env._menhir_error <- true;
                let (_menhir_env : _menhir_env) = _menhir_env in
                let (_menhir_stack : (('freshtv27 * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) * Lexing.position * _menhir_state * Lexing.position) = Obj.magic _menhir_stack in
                ((let (_menhir_stack, _, _menhir_s, _) = _menhir_stack in
                _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) _menhir_s) : 'freshtv28)) : 'freshtv30)
        | UIDENT _v ->
            _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | UNDERSCORE ->
            _menhir_run211 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState899 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
        | _ ->
            assert (not _menhir_env._menhir_error);
            _menhir_env._menhir_error <- true;
            _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState899) : 'freshtv32)
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState898 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState898

and _menhir_run908 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState908 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState908

and _menhir_run623 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 821 "reason_parser.mly"
       (string)
# 87272 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce370 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run624 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce377 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _startpos

and _menhir_run2647 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2647 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2647 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2647

and _menhir_run2654 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | COLONCOLON ->
        _menhir_run2608 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run2604 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run2599 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run2595 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2654 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run2589 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2654 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2654

and _menhir_run2754 : _menhir_env -> 'ttv_tail -> _menhir_state -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s ->
    let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv19) = Obj.magic _menhir_stack in
    let (_menhir_s : _menhir_state) = _menhir_s in
    ((let _1 = () in
    let _v : 'tv__use_file = 
# 1131 "reason_parser.mly"
                                                   ( [] )
# 87328 "reason_parser.ml"
     in
    _menhir_goto__use_file _menhir_env _menhir_stack _menhir_s _v) : 'freshtv20)

and _menhir_run2660 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | TYPE ->
        _menhir_run2052 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | VIRTUAL ->
        _menhir_run1535 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2660 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _ ->
        _menhir_reduce847 _menhir_env (Obj.magic _menhir_stack) MenhirState2660
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2660

and _menhir_run625 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> (
# 800 "reason_parser.mly"
       (char)
# 87352 "reason_parser.ml"
) -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _v _startpos ->
    let _menhir_env = _menhir_discard _menhir_env in
    _menhir_reduce368 _menhir_env (Obj.magic _menhir_stack) _endpos _menhir_s _v _startpos

and _menhir_run730 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState730 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState730

and _menhir_run164 : _menhir_env -> 'ttv_tail -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LIDENT _v ->
        _menhir_run126 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState164 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run125 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState164 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState164

and _menhir_run920 : _menhir_env -> 'ttv_tail -> Lexing.position -> _menhir_state -> Lexing.position -> 'ttv_return =
  fun _menhir_env _menhir_stack _endpos _menhir_s _startpos ->
    let _menhir_stack = (_menhir_stack, _endpos, _menhir_s, _startpos) in
    let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run15 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState920 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState920

and _menhir_discard : _menhir_env -> _menhir_env =
  fun _menhir_env ->
    let lexer = _menhir_env._menhir_lexer in
    let lexbuf = _menhir_env._menhir_lexbuf in
    let _tok = lexer lexbuf in
    {
      _menhir_lexer = lexer;
      _menhir_lexbuf = lexbuf;
      _menhir_token = _tok;
      _menhir_error = false;
    }

and _menhir_init : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> _menhir_env =
  fun lexer lexbuf ->
    let _tok = Obj.magic () in
    {
      _menhir_lexer = lexer;
      _menhir_lexbuf = lexbuf;
      _menhir_token = _tok;
      _menhir_error = false;
    }

and implementation : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1059 "reason_parser.mly"
      (Parsetree.structure)
# 87504 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv17) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run811 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run465 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run464 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run463 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run1534 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run1489 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run1478 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run462 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run461 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run738 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run474 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run1476 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run460 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run459 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run458 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run457 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run432 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run423 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run416 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run455 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run415 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run189 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run453 _menhir_env (Obj.magic _menhir_stack) MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run441 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run414 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run408 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run398 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run183 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run181 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run392 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run180 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run179 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run177 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run176 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run17 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run74 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run16 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run2 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState0 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        _menhir_reduce787 _menhir_env (Obj.magic _menhir_stack) MenhirState0
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState0) : 'freshtv18))

and interface : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1061 "reason_parser.mly"
      (Parsetree.signature)
# 87611 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv15) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | CLASS ->
        _menhir_run2051 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run2048 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2039 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2036 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run1409 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run1404 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run1252 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2563 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        _menhir_reduce704 _menhir_env (Obj.magic _menhir_stack) MenhirState2563
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2563) : 'freshtv16))

and parse_core_type : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1067 "reason_parser.mly"
      (Parsetree.core_type)
# 87650 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv13) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | LBRACKET ->
        _menhir_run162 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETGREATER ->
        _menhir_run159 _menhir_env (Obj.magic _menhir_stack) MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETLESS ->
        _menhir_run156 _menhir_env (Obj.magic _menhir_stack) MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESS ->
        _menhir_run147 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSDOTDOTGREATER ->
        _menhir_run146 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LESSGREATER ->
        _menhir_run145 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run142 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run128 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | QUOTE ->
        _menhir_run124 _menhir_env (Obj.magic _menhir_stack) MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run120 _menhir_env (Obj.magic _menhir_stack) MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run77 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run119 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2570 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2570) : 'freshtv14))

and parse_expression : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1069 "reason_parser.mly"
      (Parsetree.expression)
# 87693 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv11) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2574 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2574) : 'freshtv12))

and parse_pattern : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1071 "reason_parser.mly"
      (Parsetree.pattern)
# 87776 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv9) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXCEPTION ->
        _menhir_run642 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run619 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run615 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run603 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run588 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run18 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run210 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run204 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run198 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run193 _menhir_env (Obj.magic _menhir_stack) MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UNDERSCORE ->
        _menhir_run192 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2578 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2578) : 'freshtv10))

and toplevel_phrase : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1063 "reason_parser.mly"
      (Parsetree.toplevel_phrase)
# 87839 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv7) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run2660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv5) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = MenhirState2582 in
        ((let (_menhir_env : _menhir_env) = _menhir_env in
        let (_menhir_stack : 'freshtv3) = Obj.magic _menhir_stack in
        let (_menhir_s : _menhir_state) = _menhir_s in
        ((let _1 = () in
        let _v : 'tv__toplevel_phrase = 
# 1125 "reason_parser.mly"
                                        ( raise End_of_file)
# 87869 "reason_parser.ml"
         in
        _menhir_goto__toplevel_phrase _menhir_env _menhir_stack _menhir_s _v) : 'freshtv4)) : 'freshtv6)
    | EXCEPTION ->
        _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2647 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2644 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run2635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run2628 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run2619 _menhir_env (Obj.magic _menhir_stack) MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2582 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2582) : 'freshtv8))

and use_file : (Lexing.lexbuf -> token) -> Lexing.lexbuf -> (
# 1065 "reason_parser.mly"
      (Parsetree.toplevel_phrase list)
# 87960 "reason_parser.ml"
) =
  fun lexer lexbuf ->
    let _menhir_env = _menhir_init lexer lexbuf in
    Obj.magic (let (_menhir_env : _menhir_env) = _menhir_env in
    let (_menhir_stack : 'freshtv1) = ((), _menhir_env._menhir_lexbuf.Lexing.lex_curr_p) in
    ((let _menhir_env = _menhir_discard _menhir_env in
    let _tok = _menhir_env._menhir_token in
    match _tok with
    | ASSERT ->
        _menhir_run920 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BACKQUOTE ->
        _menhir_run164 _menhir_env (Obj.magic _menhir_stack) MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | BANG ->
        _menhir_run730 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CHAR _v ->
        _menhir_run625 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | CLASS ->
        _menhir_run2660 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EOF ->
        _menhir_run2754 _menhir_env (Obj.magic _menhir_stack) MenhirState2753
    | EXCEPTION ->
        _menhir_run2654 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | EXTERNAL ->
        _menhir_run2647 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FALSE ->
        _menhir_run624 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FLOAT _v ->
        _menhir_run623 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FOR ->
        _menhir_run908 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | FUN ->
        _menhir_run898 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | IF ->
        _menhir_run894 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INCLUDE ->
        _menhir_run2644 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT _v ->
        _menhir_run622 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT32 _v ->
        _menhir_run621 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | INT64 _v ->
        _menhir_run620 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LAZY ->
        _menhir_run793 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACE ->
        _menhir_run475 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACELESS ->
        _menhir_run420 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKET ->
        _menhir_run418 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETATATAT ->
        _menhir_run2032 _menhir_env (Obj.magic _menhir_stack) MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETBAR ->
        _menhir_run178 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENT ->
        _menhir_run19 _menhir_env (Obj.magic _menhir_stack) MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LBRACKETPERCENTPERCENT ->
        _menhir_run1474 _menhir_env (Obj.magic _menhir_stack) MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LET ->
        _menhir_run2635 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LIDENT _v ->
        _menhir_run792 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | LPAREN ->
        _menhir_run771 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUS ->
        _menhir_run407 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MINUSDOT ->
        _menhir_run406 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | MODULE ->
        _menhir_run2628 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NATIVEINT _v ->
        _menhir_run14 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | NEW ->
        _menhir_run7 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | OPEN ->
        _menhir_run1400 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PERCENT ->
        _menhir_run337 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUS ->
        _menhir_run336 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PLUSDOT ->
        _menhir_run335 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | PREFIXOP _v ->
        _menhir_run6 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SHARP ->
        _menhir_run2619 _menhir_env (Obj.magic _menhir_stack) MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | STRING _v ->
        _menhir_run5 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | SWITCH ->
        _menhir_run764 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRUE ->
        _menhir_run4 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TRY ->
        _menhir_run419 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | TYPE ->
        _menhir_run2583 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | UIDENT _v ->
        _menhir_run3 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _v _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | WHILE ->
        _menhir_run417 _menhir_env (Obj.magic _menhir_stack) _menhir_env._menhir_lexbuf.Lexing.lex_curr_p MenhirState2753 _menhir_env._menhir_lexbuf.Lexing.lex_start_p
    | _ ->
        assert (not _menhir_env._menhir_error);
        _menhir_env._menhir_error <- true;
        _menhir_errorcase _menhir_env (Obj.magic _menhir_stack) MenhirState2753) : 'freshtv2))

# 4362 "reason_parser.mly"
  

# 88069 "reason_parser.ml"

# 220 "/Users/hzhang295/.opam/4.02.3+local-git-master/lib/menhir/standard.mly"
  


# 88075 "reason_parser.ml"

end
module 
Reason_lexer
= struct
#1 "reason_lexer.ml"
# 52 "reason_lexer.mll"
 
open Lexing
open Misc
open Reason_parser

type error =
  | Illegal_character of char
  | Illegal_escape of string
  | Unterminated_comment of Location.t
  | Unterminated_string
  | Unterminated_string_in_comment of Location.t * Location.t
  | Keyword_as_label of string
  | Literal_overflow of string
;;

exception Error of error * Location.t;;

(* The table of keywords *)

let keyword_table =
  create_hashtable 149 [
    "and", AND;
    "as", AS;
    "assert", ASSERT;
    "begin", BEGIN;
    "class", CLASS;
    "constraint", CONSTRAINT;
    "do", DO;
    "done", DONE;
    "downto", DOWNTO;
    "else", ELSE;
    "end", END;
    "exception", EXCEPTION;
    "external", EXTERNAL;
    "false", FALSE;
    "for", FOR;
    "fun", FUN;
    "function", FUNCTION;
    "functor", FUNCTOR;
    "if", IF;
    "in", IN;
    "include", INCLUDE;
    "inherit", INHERIT;
    "initializer", INITIALIZER;
    "lazy", LAZY;
    "let", LET;
    "switch", SWITCH;
    "match", MATCH; (* Including MATCH for source transforming compat *)
    "method", METHOD;
    "module", MODULE;
    "mutable", MUTABLE;
    "new", NEW;
    "nonrec", NONREC;
    "object", OBJECT;
    "of", OF;
    "open", OPEN;
    "or", OR;
(*  "parser", PARSER; *)
    "private", PRIVATE;
    "rec", REC;
    "sig", SIG;
    "struct", STRUCT;
    "then", THEN;
    "to", TO;
    "true", TRUE;
    "try", TRY;
    "type", TYPE;
    "val", VAL;
    "virtual", VIRTUAL;
    "when", WHEN;
    "while", WHILE;
    "with", WITH;

    "mod", INFIXOP3("mod");
    "land", INFIXOP3("land");
    "lor", INFIXOP3("lor");
    "lxor", INFIXOP3("lxor");
    "lsl", INFIXOP4("lsl");
    "lsr", INFIXOP4("lsr");
    "asr", INFIXOP4("asr")
]

(* To buffer string literals *)

let initial_string_buffer = Bytes.create 256
let string_buff = ref initial_string_buffer
let string_index = ref 0

let reset_string_buffer () =
  string_buff := initial_string_buffer;
  string_index := 0

let store_string_char c =
  if !string_index >= Bytes.length !string_buff then begin
    let new_buff = Bytes.create (Bytes.length (!string_buff) * 2) in
    Bytes.blit !string_buff 0 new_buff 0 (Bytes.length !string_buff);
    string_buff := new_buff
  end;
  Bytes.unsafe_set !string_buff !string_index c;
  incr string_index

let store_string s =
  for i = 0 to String.length s - 1 do
    store_string_char s.[i];
  done

let store_lexeme lexbuf =
  store_string (Lexing.lexeme lexbuf)

let get_stored_string () =
  let s = Bytes.sub_string !string_buff 0 !string_index in
  string_buff := initial_string_buffer;
  s

(* To store the position of the beginning of a string and comment *)
let string_start_loc = ref Location.none;;
let comment_start_loc = ref [];;
let in_comment () = !comment_start_loc <> [];;
let is_in_string = ref false
let in_string () = !is_in_string
let print_warnings = ref true

(* To translate escape sequences *)

let char_for_backslash = function
  | 'n' -> '\010'
  | 'r' -> '\013'
  | 'b' -> '\008'
  | 't' -> '\009'
  | c   -> c

let char_for_decimal_code lexbuf i =
  let c = 100 * (Char.code(Lexing.lexeme_char lexbuf i) - 48) +
           10 * (Char.code(Lexing.lexeme_char lexbuf (i+1)) - 48) +
                (Char.code(Lexing.lexeme_char lexbuf (i+2)) - 48) in
  if (c < 0 || c > 255) then
    if in_comment ()
    then 'x'
    else raise (Error(Illegal_escape (Lexing.lexeme lexbuf),
                      Location.curr lexbuf))
  else Char.chr c

let char_for_hexadecimal_code lexbuf i =
  let d1 = Char.code (Lexing.lexeme_char lexbuf i) in
  let val1 = if d1 >= 97 then d1 - 87
             else if d1 >= 65 then d1 - 55
             else d1 - 48
  in
  let d2 = Char.code (Lexing.lexeme_char lexbuf (i+1)) in
  let val2 = if d2 >= 97 then d2 - 87
             else if d2 >= 65 then d2 - 55
             else d2 - 48
  in
  Char.chr (val1 * 16 + val2)

(* To convert integer literals, allowing max_int + 1 (PR#4210) *)

let cvt_int_literal s =
  - int_of_string ("-" ^ s)
let cvt_int32_literal s =
  Int32.neg (Int32.of_string ("-" ^ String.sub s 0 (String.length s - 1)))
let cvt_int64_literal s =
  Int64.neg (Int64.of_string ("-" ^ String.sub s 0 (String.length s - 1)))
let cvt_nativeint_literal s =
  Nativeint.neg (Nativeint.of_string ("-" ^ String.sub s 0
                                                       (String.length s - 1)))

(* Remove underscores from float literals *)

let remove_underscores s =
  let l = String.length s in
  let b = Bytes.create l in
  let rec remove src dst =
    if src >= l then
      if dst >= l then s else Bytes.sub_string b 0 dst
    else
      match s.[src] with
        '_' -> remove (src + 1) dst
      |  c  -> Bytes.set b dst c; remove (src + 1) (dst + 1)
  in remove 0 0

(* Update the current location with file name and line number. *)

let update_loc lexbuf file line absolute chars =
  let pos = lexbuf.lex_curr_p in
  let new_file = match file with
                 | None -> pos.pos_fname
                 | Some s -> s
  in
  lexbuf.lex_curr_p <- { pos with
    pos_fname = new_file;
    pos_lnum = if absolute then line else pos.pos_lnum + line;
    pos_bol = pos.pos_cnum - chars;
  }
;;

let preprocessor = ref None

(* Warn about Latin-1 characters used in idents *)

let warn_latin1 lexbuf =
  Location.prerr_warning (Location.curr lexbuf)
    (Warnings.Deprecated "ISO-Latin1 characters in identifiers")
;;

(* Error report *)

open Format

let report_error ppf = function
  | Illegal_character c ->
      fprintf ppf "Illegal character (%s)" (Char.escaped c)
  | Illegal_escape s ->
      fprintf ppf "Illegal backslash escape in string or character (%s)" s
  | Unterminated_comment _ ->
      fprintf ppf "Comment not terminated"
  | Unterminated_string ->
      fprintf ppf "String literal not terminated"
  | Unterminated_string_in_comment (_, loc) ->
      fprintf ppf "This comment contains an unterminated string literal@.\
                   %aString literal begins here"
              Location.print_error loc
  | Keyword_as_label kwd ->
      fprintf ppf "`%s' is a keyword, it cannot be used as label name" kwd
  | Literal_overflow ty ->
      fprintf ppf "Integer literal exceeds the range of representable \
                   integers of type %s" ty

let () =
  Location.register_error_of_exn
    (function
      | Error (err, loc) ->
          Some (Location.error_of_printer loc report_error err)
      | _ ->
          None
    )

# 240 "reason_lexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\161\255\162\255\224\000\235\000\018\001\053\001\088\001\
    \191\255\123\001\158\001\198\255\091\000\195\001\068\000\083\000\
    \230\001\218\255\220\255\221\255\223\255\009\002\093\002\128\002\
    \163\002\008\001\193\002\238\255\014\003\047\003\120\003\072\004\
    \024\005\232\005\072\000\199\006\210\006\249\006\055\007\126\000\
    \254\255\001\000\046\002\022\008\033\008\072\008\107\008\142\008\
    \177\008\255\255\005\000\089\000\212\008\247\008\026\009\090\000\
    \061\009\097\000\096\009\098\000\131\009\099\000\100\000\101\000\
    \107\000\006\000\110\000\166\009\201\009\208\255\209\255\087\000\
    \248\255\239\255\240\255\241\255\116\000\180\002\114\000\126\000\
    \195\003\147\004\099\005\116\001\063\002\134\000\196\255\237\255\
    \004\006\248\009\117\000\121\000\007\000\236\255\235\255\231\255\
    \113\005\190\002\208\000\234\255\210\002\219\000\233\255\130\007\
    \220\000\232\255\216\000\229\255\228\255\222\009\250\003\017\010\
    \247\000\226\255\011\000\039\001\040\001\074\001\057\001\226\255\
    \012\000\055\010\090\010\125\010\218\000\212\255\206\255\163\010\
    \198\010\233\010\012\011\047\011\228\000\202\000\199\255\200\255\
    \201\255\211\000\185\255\187\255\193\255\082\011\190\255\117\011\
    \152\011\187\011\222\011\030\001\163\001\243\255\244\255\013\000\
    \245\255\012\002\032\006\253\255\229\000\237\000\255\255\254\255\
    \252\255\220\011\055\012\244\000\251\000\017\000\251\255\250\255\
    \249\255\089\012\024\003\252\000\248\255\218\003\253\000\247\255\
    \128\012\005\001\246\255\018\000\075\001\247\255\248\255\249\255\
    \035\001\199\012\255\255\250\255\233\012\176\004\253\255\037\001\
    \055\001\091\001\136\005\252\255\016\013\251\255\109\001\122\001\
    \252\255\024\013\254\255\255\255\110\001\117\001\253\255\053\013\
    \022\001\025\001\049\001\085\001\033\001\087\001\040\001\019\000\
    \255\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\089\000\086\000\085\000\076\000\071\000\
    \255\255\063\000\060\000\255\255\053\000\050\000\048\000\041\000\
    \079\000\255\255\255\255\255\255\255\255\030\000\040\000\036\000\
    \091\000\033\000\058\000\255\255\012\000\012\000\011\000\010\000\
    \009\000\008\000\044\000\052\000\005\000\004\000\003\000\002\000\
    \255\255\094\000\094\000\092\000\091\000\090\000\087\000\255\255\
    \255\255\255\255\255\255\255\255\083\000\074\000\075\000\255\255\
    \084\000\255\255\090\000\255\255\088\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\039\000\006\000\255\255\255\255\045\000\
    \255\255\255\255\255\255\255\255\255\255\013\000\255\255\255\255\
    \255\255\012\000\012\000\012\000\013\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \024\000\024\000\024\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\025\000\255\255\255\255\255\255\255\255\028\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\031\000\080\000\038\000\042\000\255\255\255\255\085\000\
    \081\000\051\000\085\000\082\000\067\000\066\000\255\255\255\255\
    \255\255\069\000\255\255\255\255\255\255\061\000\255\255\072\000\
    \073\000\078\000\077\000\255\255\255\255\255\255\255\255\012\000\
    \255\255\012\000\012\000\255\255\012\000\012\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \008\000\008\000\255\255\255\255\005\000\005\000\255\255\001\000\
    \005\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\003\000\255\255\255\255\255\255\
    \002\000\255\255\255\255\001\000\255\255\255\255\255\255\255\255\
    \255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\000\000\000\000\255\255\255\255\255\255\
    \255\255\090\000\255\255\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \255\255\095\000\255\255\255\255\255\255\000\000\000\000\000\000\
    \255\255\255\255\255\255\000\000\255\255\255\255\000\000\255\255\
    \255\255\000\000\255\255\000\000\000\000\115\000\255\255\255\255\
    \255\255\000\000\255\255\115\000\116\000\115\000\118\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\149\000\000\000\000\000\255\255\
    \000\000\163\000\255\255\000\000\255\255\255\255\000\000\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\000\000\000\000\
    \000\000\255\255\255\255\255\255\000\000\255\255\255\255\000\000\
    \255\255\255\255\000\000\255\255\181\000\000\000\000\000\000\000\
    \255\255\187\000\000\000\000\000\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\000\000\255\255\000\000\255\255\200\000\
    \000\000\255\255\000\000\000\000\255\255\255\255\000\000\255\255\
    \255\255\255\255\210\000\213\000\255\255\213\000\255\255\255\255\
    \000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\039\000\040\000\040\000\039\000\041\000\065\000\049\000\
    \040\000\091\000\050\000\065\000\092\000\113\000\119\000\150\000\
    \114\000\120\000\179\000\164\000\150\000\216\000\165\000\179\000\
    \039\000\007\000\027\000\022\000\005\000\003\000\021\000\025\000\
    \019\000\018\000\023\000\006\000\017\000\016\000\015\000\024\000\
    \029\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\034\000\014\000\013\000\035\000\009\000\036\000\
    \004\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\012\000\042\000\011\000\004\000\038\000\
    \020\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\026\000\010\000\008\000\037\000\126\000\
    \132\000\124\000\071\000\052\000\056\000\070\000\069\000\039\000\
    \052\000\056\000\039\000\046\000\060\000\060\000\058\000\044\000\
    \046\000\058\000\060\000\058\000\044\000\043\000\072\000\135\000\
    \005\000\134\000\043\000\133\000\094\000\005\000\039\000\085\000\
    \093\000\085\000\083\000\083\000\084\000\084\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\082\000\084\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\136\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\099\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \002\000\043\000\102\000\105\000\043\000\043\000\043\000\107\000\
    \125\000\139\000\137\000\043\000\004\000\043\000\043\000\004\000\
    \004\000\004\000\091\000\138\000\159\000\092\000\004\000\158\000\
    \004\000\004\000\043\000\168\000\043\000\043\000\043\000\043\000\
    \043\000\111\000\167\000\172\000\175\000\004\000\111\000\004\000\
    \004\000\004\000\004\000\004\000\178\000\183\000\191\000\255\255\
    \198\000\113\000\113\000\005\000\114\000\114\000\005\000\005\000\
    \005\000\209\000\210\000\211\000\064\000\005\000\043\000\005\000\
    \005\000\191\000\214\000\119\000\193\000\191\000\120\000\147\000\
    \004\000\004\000\118\000\215\000\005\000\004\000\005\000\005\000\
    \005\000\005\000\005\000\117\000\113\000\183\000\046\000\114\000\
    \184\000\046\000\046\000\046\000\043\000\000\000\043\000\212\000\
    \046\000\212\000\046\000\146\000\089\000\191\000\000\000\004\000\
    \193\000\004\000\117\000\000\000\116\000\186\000\066\000\046\000\
    \005\000\046\000\145\000\046\000\046\000\046\000\000\000\183\000\
    \203\000\052\000\198\000\205\000\052\000\052\000\052\000\203\000\
    \000\000\000\000\205\000\052\000\203\000\052\000\052\000\204\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \005\000\057\000\052\000\046\000\052\000\143\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\005\000\000\000\000\000\005\000\
    \005\000\005\000\000\000\000\000\083\000\083\000\005\000\185\000\
    \005\000\005\000\000\000\000\000\000\000\150\000\000\000\000\000\
    \151\000\046\000\000\000\046\000\051\000\005\000\052\000\005\000\
    \005\000\005\000\005\000\005\000\000\000\000\000\000\000\005\000\
    \074\000\000\000\005\000\005\000\005\000\155\000\000\000\000\000\
    \000\000\005\000\153\000\005\000\005\000\156\000\000\000\000\000\
    \000\000\000\000\157\000\083\000\052\000\000\000\052\000\066\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\005\000\000\000\
    \075\000\000\000\073\000\000\000\005\000\000\000\000\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \129\000\127\000\000\000\000\000\000\000\000\000\201\000\005\000\
    \142\000\005\000\066\000\140\000\005\000\005\000\000\000\005\000\
    \005\000\128\000\005\000\005\000\000\000\000\000\000\000\046\000\
    \255\255\000\000\046\000\046\000\046\000\000\000\000\000\000\000\
    \000\000\046\000\000\000\046\000\122\000\000\000\164\000\000\000\
    \000\000\165\000\141\000\000\000\005\000\000\000\154\000\066\000\
    \046\000\005\000\046\000\046\000\123\000\046\000\046\000\255\255\
    \255\255\000\000\005\000\000\000\000\000\005\000\005\000\121\000\
    \000\000\255\255\000\000\166\000\005\000\000\000\005\000\005\000\
    \049\000\255\255\000\000\050\000\000\000\000\000\000\000\005\000\
    \000\000\005\000\057\000\005\000\046\000\005\000\005\000\005\000\
    \005\000\005\000\255\255\182\000\000\000\000\000\000\000\048\000\
    \000\000\000\000\005\000\043\000\005\000\255\255\000\000\255\255\
    \045\000\046\000\000\000\046\000\000\000\044\000\000\000\000\000\
    \000\000\000\000\046\000\000\000\046\000\066\000\110\000\005\000\
    \162\000\000\000\005\000\005\000\005\000\047\000\004\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\000\000\202\000\000\000\000\000\110\000\111\000\000\000\
    \111\000\111\000\111\000\111\000\000\000\005\000\000\000\005\000\
    \111\000\000\000\111\000\111\000\004\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\111\000\
    \000\000\111\000\111\000\111\000\111\000\111\000\084\000\000\000\
    \000\000\058\000\000\000\152\000\058\000\058\000\058\000\000\000\
    \000\000\000\000\005\000\058\000\047\000\058\000\058\000\108\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\112\000\058\000\111\000\058\000\058\000\058\000\058\000\
    \058\000\000\000\000\000\000\000\044\000\000\000\000\000\044\000\
    \044\000\044\000\000\000\000\000\000\000\106\000\044\000\000\000\
    \044\000\044\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\111\000\000\000\111\000\059\000\044\000\058\000\044\000\
    \044\000\044\000\044\000\044\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\100\000\100\000\
    \100\000\100\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \000\000\076\000\000\000\000\000\058\000\086\000\058\000\063\000\
    \000\000\044\000\101\000\101\000\101\000\101\000\101\000\101\000\
    \101\000\101\000\101\000\101\000\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\077\000\000\000\000\000\000\000\000\000\
    \000\000\076\000\000\000\000\000\000\000\000\000\000\000\044\000\
    \088\000\044\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\077\000\087\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \173\000\173\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\000\000\076\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\074\000\000\000\000\000\077\000\000\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\000\000\000\000\000\000\000\000\028\000\000\000\000\000\
    \000\000\078\000\000\000\076\000\076\000\000\000\000\000\000\000\
    \000\000\000\000\075\000\074\000\073\000\000\000\079\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\080\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\028\000\000\000\
    \000\000\078\000\000\000\000\000\076\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\075\000\000\000\073\000\079\000\030\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\080\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\000\000\000\000\000\000\000\000\030\000\
    \000\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\110\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\174\000\174\000\174\000\174\000\174\000\174\000\
    \174\000\174\000\174\000\174\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\110\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\000\000\000\000\000\000\000\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\000\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\031\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\000\000\000\000\000\000\000\000\031\000\
    \000\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\000\000\000\000\000\000\000\000\000\000\074\000\
    \194\000\194\000\194\000\194\000\194\000\194\000\194\000\194\000\
    \194\000\194\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\081\000\000\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\000\000\000\000\000\000\000\000\000\000\075\000\
    \000\000\073\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\000\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\032\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\000\000\000\000\000\000\000\000\032\000\
    \000\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\000\000\000\000\000\000\000\000\074\000\
    \000\000\000\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \195\000\195\000\195\000\195\000\195\000\195\000\195\000\195\000\
    \195\000\195\000\082\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\075\000\
    \000\000\073\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\000\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\033\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\000\000\000\000\000\000\000\000\033\000\
    \000\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\088\000\000\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\161\000\
    \087\000\161\000\161\000\161\000\161\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\000\000\160\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\000\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\000\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \005\000\000\000\000\000\005\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\005\000\056\000\005\000\005\000\056\000\056\000\
    \056\000\000\000\000\000\000\000\000\000\056\000\000\000\056\000\
    \056\000\005\000\000\000\005\000\005\000\067\000\068\000\005\000\
    \000\000\000\000\000\000\000\000\056\000\000\000\056\000\056\000\
    \056\000\056\000\056\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\056\000\000\000\000\000\056\000\056\000\056\000\
    \000\000\000\000\000\000\066\000\056\000\005\000\056\000\056\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\055\000\000\000\
    \056\000\000\000\000\000\056\000\000\000\056\000\056\000\056\000\
    \056\000\056\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\005\000\000\000\005\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\056\000\000\000\
    \056\000\000\000\000\000\000\000\000\000\055\000\000\000\056\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\033\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\000\000\000\000\000\000\000\000\056\000\000\000\056\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\000\000\000\000\000\000\000\000\033\000\000\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\000\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\000\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\043\000\
    \000\000\000\000\043\000\043\000\043\000\000\000\000\000\000\000\
    \000\000\043\000\044\000\043\000\043\000\044\000\044\000\044\000\
    \000\000\000\000\000\000\000\000\044\000\000\000\044\000\044\000\
    \043\000\000\000\043\000\043\000\043\000\043\000\043\000\000\000\
    \000\000\000\000\000\000\044\000\000\000\044\000\044\000\044\000\
    \044\000\044\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\058\000\000\000\000\000\058\000\058\000\058\000\000\000\
    \000\000\000\000\064\000\058\000\043\000\058\000\058\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\063\000\000\000\044\000\
    \000\000\000\000\058\000\000\000\058\000\058\000\058\000\058\000\
    \058\000\000\000\000\000\000\000\046\000\000\000\000\000\046\000\
    \046\000\046\000\043\000\000\000\043\000\000\000\046\000\000\000\
    \046\000\046\000\000\000\000\000\000\000\044\000\000\000\044\000\
    \000\000\000\000\000\000\000\000\059\000\046\000\058\000\046\000\
    \046\000\046\000\046\000\046\000\000\000\000\000\000\000\056\000\
    \000\000\000\000\056\000\056\000\056\000\000\000\000\000\000\000\
    \000\000\056\000\000\000\056\000\056\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\058\000\000\000\058\000\057\000\
    \056\000\046\000\056\000\056\000\056\000\056\000\056\000\000\000\
    \000\000\000\000\052\000\000\000\000\000\052\000\052\000\052\000\
    \000\000\000\000\000\000\000\000\052\000\000\000\052\000\052\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\046\000\
    \000\000\046\000\055\000\052\000\056\000\052\000\053\000\052\000\
    \052\000\052\000\000\000\000\000\000\000\052\000\000\000\000\000\
    \052\000\052\000\052\000\000\000\000\000\000\000\000\000\052\000\
    \000\000\052\000\052\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\056\000\000\000\056\000\051\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\000\000\000\000\000\000\
    \052\000\000\000\000\000\052\000\052\000\052\000\000\000\000\000\
    \000\000\000\000\052\000\000\000\052\000\052\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\052\000\000\000\052\000\
    \051\000\052\000\052\000\052\000\054\000\052\000\052\000\052\000\
    \000\000\000\000\000\000\052\000\000\000\000\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\000\000\052\000\000\000\052\000\
    \052\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \052\000\000\000\052\000\051\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\000\000\000\000\000\000\056\000\000\000\
    \000\000\056\000\056\000\056\000\000\000\000\000\000\000\000\000\
    \056\000\000\000\056\000\056\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\052\000\000\000\052\000\051\000\056\000\
    \052\000\056\000\056\000\056\000\056\000\056\000\000\000\000\000\
    \000\000\058\000\000\000\000\000\058\000\058\000\058\000\000\000\
    \000\000\000\000\000\000\058\000\000\000\058\000\058\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\052\000\000\000\
    \052\000\055\000\058\000\056\000\058\000\058\000\058\000\058\000\
    \058\000\000\000\000\000\000\000\060\000\000\000\000\000\060\000\
    \060\000\060\000\000\000\000\000\000\000\000\000\060\000\000\000\
    \060\000\060\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\056\000\000\000\056\000\062\000\060\000\058\000\060\000\
    \060\000\060\000\060\000\060\000\000\000\000\000\000\000\005\000\
    \000\000\000\000\005\000\005\000\005\000\000\000\000\000\000\000\
    \000\000\005\000\000\000\005\000\005\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\058\000\000\000\058\000\061\000\
    \005\000\060\000\005\000\005\000\005\000\005\000\005\000\117\000\
    \113\000\000\000\005\000\114\000\000\000\005\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\005\000\000\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\117\000\060\000\
    \116\000\060\000\066\000\005\000\005\000\005\000\005\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\000\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \098\000\000\000\098\000\000\000\000\000\000\000\000\000\098\000\
    \000\000\000\000\005\000\000\000\005\000\066\000\000\000\005\000\
    \097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\
    \097\000\097\000\111\000\000\000\111\000\111\000\111\000\111\000\
    \000\000\000\000\000\000\000\000\111\000\000\000\111\000\111\000\
    \000\000\000\000\000\000\000\000\000\000\005\000\000\000\005\000\
    \000\000\000\000\000\000\111\000\000\000\111\000\111\000\111\000\
    \111\000\111\000\000\000\000\000\098\000\000\000\000\000\000\000\
    \005\000\000\000\098\000\005\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\005\000\005\000\098\000\000\000\
    \000\000\000\000\098\000\000\000\098\000\112\000\000\000\111\000\
    \096\000\005\000\000\000\005\000\005\000\005\000\005\000\005\000\
    \000\000\000\000\000\000\046\000\000\000\000\000\046\000\046\000\
    \046\000\000\000\000\000\000\000\000\000\046\000\000\000\046\000\
    \046\000\000\000\000\000\000\000\000\000\111\000\000\000\111\000\
    \000\000\000\000\000\000\066\000\046\000\005\000\046\000\046\000\
    \046\000\046\000\046\000\000\000\000\000\000\000\046\000\000\000\
    \000\000\046\000\046\000\046\000\000\000\000\000\000\000\000\000\
    \046\000\000\000\046\000\046\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\005\000\000\000\005\000\057\000\046\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\005\000\000\000\000\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \005\000\130\000\000\000\000\000\000\000\000\000\046\000\000\000\
    \046\000\057\000\000\000\046\000\000\000\005\000\255\255\005\000\
    \005\000\005\000\005\000\005\000\000\000\000\000\000\000\005\000\
    \000\000\000\000\005\000\005\000\005\000\000\000\000\000\000\000\
    \000\000\005\000\000\000\005\000\005\000\000\000\000\000\000\000\
    \255\255\046\000\000\000\046\000\000\000\000\000\000\000\066\000\
    \005\000\005\000\005\000\005\000\005\000\005\000\005\000\000\000\
    \000\000\000\000\005\000\000\000\000\000\005\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\005\000\000\000\005\000\005\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\005\000\
    \000\000\005\000\066\000\005\000\005\000\005\000\005\000\005\000\
    \005\000\005\000\000\000\000\000\000\000\005\000\000\000\000\000\
    \005\000\005\000\005\000\000\000\000\000\000\000\000\000\005\000\
    \000\000\005\000\005\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\005\000\066\000\005\000\005\000\
    \005\000\005\000\131\000\005\000\005\000\000\000\000\000\000\000\
    \005\000\000\000\000\000\005\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\005\000\000\000\005\000\005\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\005\000\000\000\005\000\
    \066\000\005\000\005\000\005\000\005\000\005\000\005\000\005\000\
    \000\000\000\000\000\000\005\000\000\000\000\000\005\000\005\000\
    \005\000\000\000\000\000\000\000\000\000\005\000\000\000\005\000\
    \005\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \005\000\000\000\005\000\066\000\005\000\005\000\005\000\005\000\
    \005\000\005\000\005\000\000\000\000\000\000\000\052\000\000\000\
    \000\000\052\000\052\000\052\000\000\000\000\000\000\000\000\000\
    \052\000\000\000\052\000\052\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\005\000\000\000\005\000\066\000\052\000\
    \005\000\052\000\144\000\052\000\052\000\052\000\000\000\000\000\
    \000\000\052\000\000\000\000\000\052\000\052\000\052\000\000\000\
    \000\000\000\000\000\000\052\000\000\000\052\000\052\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\005\000\000\000\
    \005\000\051\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\000\000\000\000\000\000\046\000\000\000\000\000\046\000\
    \046\000\046\000\000\000\000\000\000\000\000\000\046\000\000\000\
    \046\000\046\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\052\000\000\000\052\000\051\000\046\000\052\000\046\000\
    \046\000\046\000\046\000\046\000\000\000\000\000\000\000\046\000\
    \000\000\000\000\046\000\046\000\046\000\000\000\000\000\000\000\
    \000\000\046\000\000\000\046\000\046\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\052\000\000\000\052\000\057\000\
    \046\000\046\000\046\000\046\000\046\000\046\000\046\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\046\000\
    \000\000\046\000\057\000\161\000\046\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\171\000\
    \160\000\171\000\046\000\000\000\046\000\000\000\171\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\
    \170\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\176\000\171\000\000\000\000\000\000\000\000\000\
    \000\000\171\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \000\000\000\000\000\000\000\000\000\000\171\000\000\000\000\000\
    \000\000\171\000\000\000\171\000\000\000\000\000\000\000\169\000\
    \177\000\177\000\177\000\177\000\177\000\177\000\177\000\177\000\
    \177\000\177\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \000\000\177\000\177\000\177\000\177\000\177\000\177\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\191\000\000\000\000\000\192\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\177\000\177\000\177\000\177\000\177\000\177\000\190\000\
    \000\000\190\000\000\000\000\000\000\000\000\000\190\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\189\000\
    \189\000\189\000\189\000\189\000\189\000\189\000\189\000\189\000\
    \189\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\196\000\196\000\196\000\196\000\196\000\196\000\196\000\
    \196\000\196\000\196\000\190\000\000\000\000\000\000\000\000\000\
    \000\000\190\000\196\000\196\000\196\000\196\000\196\000\196\000\
    \000\000\000\000\000\000\000\000\000\000\190\000\000\000\000\000\
    \000\000\190\000\000\000\190\000\000\000\000\000\000\000\188\000\
    \197\000\197\000\197\000\197\000\197\000\197\000\197\000\197\000\
    \197\000\197\000\196\000\196\000\196\000\196\000\196\000\196\000\
    \000\000\197\000\197\000\197\000\197\000\197\000\197\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\197\000\197\000\197\000\197\000\197\000\197\000\207\000\
    \000\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \207\000\207\000\207\000\000\000\207\000\206\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \000\000\000\000\206\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\041\000\000\000\000\000\041\000\050\000\
    \065\000\092\000\050\000\065\000\092\000\114\000\120\000\151\000\
    \114\000\120\000\151\000\165\000\179\000\215\000\165\000\179\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\014\000\
    \012\000\015\000\034\000\051\000\055\000\034\000\034\000\039\000\
    \051\000\055\000\039\000\057\000\059\000\061\000\062\000\063\000\
    \057\000\059\000\061\000\062\000\063\000\064\000\071\000\012\000\
    \066\000\012\000\064\000\012\000\090\000\066\000\039\000\076\000\
    \091\000\076\000\078\000\078\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\079\000\079\000\
    \079\000\079\000\079\000\079\000\079\000\079\000\085\000\085\000\
    \085\000\085\000\085\000\085\000\085\000\085\000\085\000\085\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\012\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\098\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\101\000\104\000\003\000\003\000\003\000\106\000\
    \124\000\132\000\133\000\003\000\004\000\003\000\003\000\004\000\
    \004\000\004\000\025\000\137\000\156\000\025\000\004\000\157\000\
    \004\000\004\000\003\000\163\000\003\000\003\000\003\000\003\000\
    \003\000\112\000\164\000\171\000\174\000\004\000\112\000\004\000\
    \004\000\004\000\004\000\004\000\177\000\184\000\191\000\025\000\
    \184\000\115\000\116\000\005\000\115\000\116\000\005\000\005\000\
    \005\000\208\000\209\000\210\000\003\000\005\000\003\000\005\000\
    \005\000\192\000\212\000\118\000\192\000\191\000\118\000\004\000\
    \147\000\004\000\116\000\214\000\005\000\147\000\005\000\005\000\
    \005\000\005\000\005\000\117\000\117\000\180\000\006\000\117\000\
    \180\000\006\000\006\000\006\000\003\000\255\255\003\000\211\000\
    \006\000\213\000\006\000\006\000\025\000\193\000\255\255\004\000\
    \193\000\004\000\117\000\255\255\117\000\180\000\005\000\006\000\
    \005\000\006\000\006\000\006\000\006\000\006\000\255\255\198\000\
    \204\000\007\000\198\000\204\000\007\000\007\000\007\000\205\000\
    \255\255\255\255\205\000\007\000\199\000\007\000\007\000\199\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\005\000\255\255\
    \005\000\006\000\007\000\006\000\007\000\007\000\007\000\007\000\
    \007\000\255\255\255\255\255\255\009\000\255\255\255\255\009\000\
    \009\000\009\000\255\255\255\255\083\000\083\000\009\000\180\000\
    \009\000\009\000\255\255\255\255\255\255\148\000\255\255\255\255\
    \148\000\006\000\255\255\006\000\007\000\009\000\007\000\009\000\
    \009\000\009\000\009\000\009\000\255\255\255\255\255\255\010\000\
    \083\000\255\255\010\000\010\000\010\000\148\000\255\255\255\255\
    \255\255\010\000\148\000\010\000\010\000\148\000\255\255\255\255\
    \255\255\255\255\148\000\083\000\007\000\255\255\007\000\009\000\
    \010\000\009\000\010\000\010\000\010\000\010\000\010\000\255\255\
    \083\000\255\255\083\000\255\255\013\000\255\255\255\255\013\000\
    \013\000\013\000\255\255\255\255\255\255\255\255\013\000\255\255\
    \013\000\013\000\255\255\255\255\255\255\255\255\199\000\009\000\
    \009\000\009\000\010\000\010\000\010\000\013\000\255\255\013\000\
    \013\000\013\000\013\000\013\000\255\255\255\255\255\255\016\000\
    \025\000\255\255\016\000\016\000\016\000\255\255\255\255\255\255\
    \255\255\016\000\255\255\016\000\016\000\255\255\153\000\255\255\
    \255\255\153\000\010\000\255\255\010\000\255\255\148\000\013\000\
    \016\000\013\000\016\000\016\000\016\000\016\000\016\000\115\000\
    \116\000\255\255\021\000\255\255\255\255\021\000\021\000\021\000\
    \255\255\210\000\255\255\153\000\021\000\255\255\021\000\021\000\
    \042\000\118\000\255\255\042\000\255\255\255\255\255\255\013\000\
    \255\255\013\000\016\000\021\000\016\000\021\000\021\000\021\000\
    \021\000\021\000\117\000\180\000\255\255\255\255\255\255\042\000\
    \255\255\255\255\042\000\042\000\042\000\211\000\255\255\213\000\
    \042\000\042\000\255\255\042\000\255\255\042\000\255\255\255\255\
    \255\255\255\255\016\000\255\255\016\000\021\000\022\000\021\000\
    \153\000\255\255\042\000\042\000\042\000\042\000\042\000\084\000\
    \084\000\084\000\084\000\084\000\084\000\084\000\084\000\084\000\
    \084\000\255\255\199\000\255\255\255\255\022\000\022\000\255\255\
    \022\000\022\000\022\000\022\000\255\255\021\000\255\255\021\000\
    \022\000\255\255\022\000\022\000\042\000\022\000\022\000\022\000\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \255\255\022\000\022\000\022\000\022\000\022\000\084\000\255\255\
    \255\255\023\000\255\255\148\000\023\000\023\000\023\000\255\255\
    \255\255\255\255\042\000\023\000\042\000\023\000\023\000\023\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\022\000\023\000\022\000\023\000\023\000\023\000\023\000\
    \023\000\255\255\255\255\255\255\024\000\255\255\255\255\024\000\
    \024\000\024\000\255\255\255\255\255\255\024\000\024\000\255\255\
    \024\000\024\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\022\000\255\255\022\000\023\000\024\000\023\000\024\000\
    \024\000\024\000\024\000\024\000\077\000\077\000\077\000\077\000\
    \077\000\077\000\077\000\077\000\077\000\077\000\097\000\097\000\
    \097\000\097\000\097\000\097\000\097\000\097\000\097\000\097\000\
    \255\255\077\000\255\255\255\255\023\000\026\000\023\000\024\000\
    \255\255\024\000\100\000\100\000\100\000\100\000\100\000\100\000\
    \100\000\100\000\100\000\100\000\153\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\077\000\255\255\255\255\255\255\255\255\
    \255\255\077\000\255\255\255\255\255\255\255\255\255\255\024\000\
    \026\000\024\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\026\000\026\000\026\000\026\000\
    \026\000\026\000\026\000\026\000\028\000\026\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\
    \170\000\170\000\255\255\028\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\028\000\255\255\255\255\029\000\255\255\029\000\
    \029\000\029\000\029\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\255\255\255\255\255\255\255\255\028\000\255\255\255\255\
    \255\255\029\000\255\255\028\000\029\000\255\255\255\255\255\255\
    \255\255\255\255\028\000\029\000\028\000\255\255\029\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\029\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\029\000\255\255\
    \255\255\029\000\255\255\255\255\029\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\029\000\255\255\029\000\029\000\030\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\029\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\255\255\255\255\255\255\255\255\030\000\
    \255\255\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\080\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\110\000\080\000\080\000\080\000\080\000\
    \080\000\080\000\173\000\173\000\173\000\173\000\173\000\173\000\
    \173\000\173\000\173\000\173\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\110\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\080\000\080\000\080\000\080\000\
    \080\000\080\000\110\000\110\000\110\000\110\000\110\000\110\000\
    \110\000\110\000\110\000\110\000\255\255\255\255\255\255\255\255\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\255\255\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\031\000\
    \030\000\030\000\030\000\030\000\030\000\030\000\030\000\030\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\255\255\255\255\255\255\255\255\031\000\
    \255\255\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\081\000\081\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\081\000\081\000\081\000\081\000\
    \081\000\081\000\255\255\255\255\255\255\255\255\255\255\081\000\
    \189\000\189\000\189\000\189\000\189\000\189\000\189\000\189\000\
    \189\000\189\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\081\000\255\255\081\000\081\000\081\000\081\000\
    \081\000\081\000\255\255\255\255\255\255\255\255\255\255\081\000\
    \255\255\081\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\255\255\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\032\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\255\255\255\255\255\255\255\255\032\000\
    \255\255\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\096\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \096\000\096\000\096\000\255\255\255\255\255\255\255\255\082\000\
    \255\255\255\255\096\000\096\000\096\000\096\000\096\000\096\000\
    \194\000\194\000\194\000\194\000\194\000\194\000\194\000\194\000\
    \194\000\194\000\082\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\082\000\
    \255\255\082\000\096\000\096\000\096\000\096\000\096\000\096\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\255\255\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\033\000\
    \032\000\032\000\032\000\032\000\032\000\032\000\032\000\032\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\255\255\255\255\255\255\255\255\033\000\
    \255\255\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\088\000\255\255\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\088\000\
    \088\000\088\000\088\000\088\000\088\000\088\000\088\000\154\000\
    \088\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\154\000\154\000\154\000\154\000\154\000\
    \154\000\154\000\154\000\255\255\154\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\255\255\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\255\255\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \035\000\255\255\255\255\035\000\035\000\035\000\255\255\255\255\
    \255\255\255\255\035\000\036\000\035\000\035\000\036\000\036\000\
    \036\000\255\255\255\255\255\255\255\255\036\000\255\255\036\000\
    \036\000\035\000\255\255\035\000\035\000\035\000\035\000\035\000\
    \255\255\255\255\255\255\255\255\036\000\255\255\036\000\036\000\
    \036\000\036\000\036\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\037\000\255\255\255\255\037\000\037\000\037\000\
    \255\255\255\255\255\255\035\000\037\000\035\000\037\000\037\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\036\000\255\255\
    \036\000\255\255\255\255\037\000\255\255\037\000\037\000\037\000\
    \037\000\037\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\035\000\255\255\035\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\036\000\255\255\
    \036\000\255\255\255\255\255\255\255\255\037\000\255\255\037\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\038\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\255\255\255\255\255\255\255\255\037\000\255\255\037\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\255\255\255\255\255\255\255\255\038\000\255\255\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\103\000\103\000\103\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\103\000\103\000\103\000\103\000\103\000\
    \103\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\103\000\103\000\103\000\103\000\103\000\
    \103\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\255\255\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\255\255\038\000\
    \038\000\038\000\038\000\038\000\038\000\038\000\038\000\043\000\
    \255\255\255\255\043\000\043\000\043\000\255\255\255\255\255\255\
    \255\255\043\000\044\000\043\000\043\000\044\000\044\000\044\000\
    \255\255\255\255\255\255\255\255\044\000\255\255\044\000\044\000\
    \043\000\255\255\043\000\043\000\043\000\043\000\043\000\255\255\
    \255\255\255\255\255\255\044\000\255\255\044\000\044\000\044\000\
    \044\000\044\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\045\000\255\255\255\255\045\000\045\000\045\000\255\255\
    \255\255\255\255\043\000\045\000\043\000\045\000\045\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\044\000\255\255\044\000\
    \255\255\255\255\045\000\255\255\045\000\045\000\045\000\045\000\
    \045\000\255\255\255\255\255\255\046\000\255\255\255\255\046\000\
    \046\000\046\000\043\000\255\255\043\000\255\255\046\000\255\255\
    \046\000\046\000\255\255\255\255\255\255\044\000\255\255\044\000\
    \255\255\255\255\255\255\255\255\045\000\046\000\045\000\046\000\
    \046\000\046\000\046\000\046\000\255\255\255\255\255\255\047\000\
    \255\255\255\255\047\000\047\000\047\000\255\255\255\255\255\255\
    \255\255\047\000\255\255\047\000\047\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\045\000\255\255\045\000\046\000\
    \047\000\046\000\047\000\047\000\047\000\047\000\047\000\255\255\
    \255\255\255\255\048\000\255\255\255\255\048\000\048\000\048\000\
    \255\255\255\255\255\255\255\255\048\000\255\255\048\000\048\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\046\000\
    \255\255\046\000\047\000\048\000\047\000\048\000\048\000\048\000\
    \048\000\048\000\255\255\255\255\255\255\052\000\255\255\255\255\
    \052\000\052\000\052\000\255\255\255\255\255\255\255\255\052\000\
    \255\255\052\000\052\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\047\000\255\255\047\000\048\000\052\000\048\000\
    \052\000\052\000\052\000\052\000\052\000\255\255\255\255\255\255\
    \053\000\255\255\255\255\053\000\053\000\053\000\255\255\255\255\
    \255\255\255\255\053\000\255\255\053\000\053\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\048\000\255\255\048\000\
    \052\000\053\000\052\000\053\000\053\000\053\000\053\000\053\000\
    \255\255\255\255\255\255\054\000\255\255\255\255\054\000\054\000\
    \054\000\255\255\255\255\255\255\255\255\054\000\255\255\054\000\
    \054\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \052\000\255\255\052\000\053\000\054\000\053\000\054\000\054\000\
    \054\000\054\000\054\000\255\255\255\255\255\255\056\000\255\255\
    \255\255\056\000\056\000\056\000\255\255\255\255\255\255\255\255\
    \056\000\255\255\056\000\056\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\053\000\255\255\053\000\054\000\056\000\
    \054\000\056\000\056\000\056\000\056\000\056\000\255\255\255\255\
    \255\255\058\000\255\255\255\255\058\000\058\000\058\000\255\255\
    \255\255\255\255\255\255\058\000\255\255\058\000\058\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\054\000\255\255\
    \054\000\056\000\058\000\056\000\058\000\058\000\058\000\058\000\
    \058\000\255\255\255\255\255\255\060\000\255\255\255\255\060\000\
    \060\000\060\000\255\255\255\255\255\255\255\255\060\000\255\255\
    \060\000\060\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\056\000\255\255\056\000\058\000\060\000\058\000\060\000\
    \060\000\060\000\060\000\060\000\255\255\255\255\255\255\067\000\
    \255\255\255\255\067\000\067\000\067\000\255\255\255\255\255\255\
    \255\255\067\000\255\255\067\000\067\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\058\000\255\255\058\000\060\000\
    \067\000\060\000\067\000\067\000\067\000\067\000\067\000\109\000\
    \109\000\255\255\068\000\109\000\255\255\068\000\068\000\068\000\
    \255\255\255\255\255\255\255\255\068\000\255\255\068\000\068\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\109\000\060\000\
    \109\000\060\000\067\000\068\000\067\000\068\000\068\000\068\000\
    \068\000\068\000\255\255\255\255\255\255\255\255\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \089\000\255\255\089\000\255\255\255\255\255\255\255\255\089\000\
    \255\255\255\255\067\000\255\255\067\000\068\000\255\255\068\000\
    \089\000\089\000\089\000\089\000\089\000\089\000\089\000\089\000\
    \089\000\089\000\111\000\255\255\111\000\111\000\111\000\111\000\
    \255\255\255\255\255\255\255\255\111\000\255\255\111\000\111\000\
    \255\255\255\255\255\255\255\255\255\255\068\000\255\255\068\000\
    \255\255\255\255\255\255\111\000\255\255\111\000\111\000\111\000\
    \111\000\111\000\255\255\255\255\089\000\255\255\255\255\255\255\
    \121\000\255\255\089\000\121\000\121\000\121\000\255\255\255\255\
    \255\255\255\255\121\000\255\255\121\000\121\000\089\000\255\255\
    \255\255\255\255\089\000\255\255\089\000\111\000\255\255\111\000\
    \089\000\121\000\255\255\121\000\121\000\121\000\121\000\121\000\
    \255\255\255\255\255\255\122\000\255\255\255\255\122\000\122\000\
    \122\000\255\255\255\255\255\255\255\255\122\000\255\255\122\000\
    \122\000\255\255\255\255\255\255\255\255\111\000\255\255\111\000\
    \255\255\255\255\255\255\121\000\122\000\121\000\122\000\122\000\
    \122\000\122\000\122\000\255\255\255\255\255\255\123\000\255\255\
    \255\255\123\000\123\000\123\000\255\255\255\255\255\255\255\255\
    \123\000\255\255\123\000\123\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\121\000\255\255\121\000\122\000\123\000\
    \122\000\123\000\123\000\123\000\123\000\123\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\127\000\255\255\255\255\127\000\
    \127\000\127\000\255\255\255\255\255\255\255\255\127\000\255\255\
    \127\000\127\000\255\255\255\255\255\255\255\255\122\000\255\255\
    \122\000\123\000\255\255\123\000\255\255\127\000\109\000\127\000\
    \127\000\127\000\127\000\127\000\255\255\255\255\255\255\128\000\
    \255\255\255\255\128\000\128\000\128\000\255\255\255\255\255\255\
    \255\255\128\000\255\255\128\000\128\000\255\255\255\255\255\255\
    \089\000\123\000\255\255\123\000\255\255\255\255\255\255\127\000\
    \128\000\127\000\128\000\128\000\128\000\128\000\128\000\255\255\
    \255\255\255\255\129\000\255\255\255\255\129\000\129\000\129\000\
    \255\255\255\255\255\255\255\255\129\000\255\255\129\000\129\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\127\000\
    \255\255\127\000\128\000\129\000\128\000\129\000\129\000\129\000\
    \129\000\129\000\255\255\255\255\255\255\130\000\255\255\255\255\
    \130\000\130\000\130\000\255\255\255\255\255\255\255\255\130\000\
    \255\255\130\000\130\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\128\000\255\255\128\000\129\000\130\000\129\000\
    \130\000\130\000\130\000\130\000\130\000\255\255\255\255\255\255\
    \131\000\255\255\255\255\131\000\131\000\131\000\255\255\255\255\
    \255\255\255\255\131\000\255\255\131\000\131\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\129\000\255\255\129\000\
    \130\000\131\000\130\000\131\000\131\000\131\000\131\000\131\000\
    \255\255\255\255\255\255\141\000\255\255\255\255\141\000\141\000\
    \141\000\255\255\255\255\255\255\255\255\141\000\255\255\141\000\
    \141\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \130\000\255\255\130\000\131\000\141\000\131\000\141\000\141\000\
    \141\000\141\000\141\000\255\255\255\255\255\255\143\000\255\255\
    \255\255\143\000\143\000\143\000\255\255\255\255\255\255\255\255\
    \143\000\255\255\143\000\143\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\131\000\255\255\131\000\141\000\143\000\
    \141\000\143\000\143\000\143\000\143\000\143\000\255\255\255\255\
    \255\255\144\000\255\255\255\255\144\000\144\000\144\000\255\255\
    \255\255\255\255\255\255\144\000\255\255\144\000\144\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\141\000\255\255\
    \141\000\143\000\144\000\143\000\144\000\144\000\144\000\144\000\
    \144\000\255\255\255\255\255\255\145\000\255\255\255\255\145\000\
    \145\000\145\000\255\255\255\255\255\255\255\255\145\000\255\255\
    \145\000\145\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\143\000\255\255\143\000\144\000\145\000\144\000\145\000\
    \145\000\145\000\145\000\145\000\255\255\255\255\255\255\146\000\
    \255\255\255\255\146\000\146\000\146\000\255\255\255\255\255\255\
    \255\255\146\000\255\255\146\000\146\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\144\000\255\255\144\000\145\000\
    \146\000\145\000\146\000\146\000\146\000\146\000\146\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\145\000\
    \255\255\145\000\146\000\161\000\146\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\161\000\
    \161\000\161\000\161\000\161\000\161\000\161\000\161\000\162\000\
    \161\000\162\000\146\000\255\255\146\000\255\255\162\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\162\000\
    \162\000\162\000\162\000\162\000\162\000\162\000\162\000\162\000\
    \162\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\169\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \169\000\169\000\169\000\162\000\255\255\255\255\255\255\255\255\
    \255\255\162\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \255\255\255\255\255\255\255\255\255\255\162\000\255\255\255\255\
    \255\255\162\000\255\255\162\000\255\255\255\255\255\255\162\000\
    \176\000\176\000\176\000\176\000\176\000\176\000\176\000\176\000\
    \176\000\176\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \255\255\176\000\176\000\176\000\176\000\176\000\176\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\185\000\255\255\255\255\185\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\176\000\176\000\176\000\176\000\176\000\176\000\185\000\
    \255\255\185\000\255\255\255\255\255\255\255\255\185\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\185\000\
    \185\000\185\000\185\000\185\000\185\000\185\000\185\000\185\000\
    \185\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\188\000\188\000\188\000\188\000\188\000\188\000\188\000\
    \188\000\188\000\188\000\185\000\255\255\255\255\255\255\255\255\
    \255\255\185\000\188\000\188\000\188\000\188\000\188\000\188\000\
    \255\255\255\255\255\255\255\255\255\255\185\000\255\255\255\255\
    \255\255\185\000\255\255\185\000\255\255\255\255\255\255\185\000\
    \196\000\196\000\196\000\196\000\196\000\196\000\196\000\196\000\
    \196\000\196\000\188\000\188\000\188\000\188\000\188\000\188\000\
    \255\255\196\000\196\000\196\000\196\000\196\000\196\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\196\000\196\000\196\000\196\000\196\000\196\000\201\000\
    \255\255\201\000\201\000\201\000\201\000\201\000\201\000\201\000\
    \201\000\201\000\201\000\201\000\201\000\201\000\201\000\201\000\
    \201\000\201\000\201\000\201\000\201\000\201\000\201\000\201\000\
    \201\000\201\000\201\000\255\255\207\000\201\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \207\000\207\000\207\000\207\000\207\000\207\000\207\000\207\000\
    \255\255\255\255\207\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\185\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\010\000\036\000\000\000\
    \000\000\012\000\000\000\000\000\000\000\002\000\000\000\027\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \002\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\039\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\019\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\036\000\036\000\000\000\036\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\000\000\000\000\001\000\022\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\007\000\001\000\000\000\000\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\001\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\022\000\116\000\185\000\192\000\116\000\193\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \022\000\255\255\116\000\000\000\117\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\109\000\110\000\255\255\255\255\
    \022\000\022\000\022\000\022\000\022\000\022\000\022\000\022\000\
    \022\000\022\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\110\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\110\000\110\000\110\000\110\000\
    \110\000\110\000\110\000\110\000\110\000\110\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \116\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\004\255\255\005\255\255\007\255\006\255\255\003\255\000\004\
    \001\005\255\007\255\255\006\255\007\255\255\000\004\001\005\003\
    \006\002\007\255\001\255\255\000\001\255";
}

let rec token lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 8 (-1) ;   __ocaml_lex_token_rec lexbuf 0
and __ocaml_lex_token_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 325 "reason_lexer.mll"
                 (
      match !preprocessor with
      | None ->
        raise (Error(Illegal_character (Lexing.lexeme_char lexbuf 0),
                     Location.curr lexbuf))
      | Some _ ->
        update_loc lexbuf None 1 false 0;
        token lexbuf )
# 1424 "reason_lexer.ml"

  | 1 ->
# 334 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 0;
        match !preprocessor with
        | None -> token lexbuf
        | Some _ -> EOL
      )
# 1433 "reason_lexer.ml"

  | 2 ->
# 340 "reason_lexer.mll"
      ( token lexbuf )
# 1438 "reason_lexer.ml"

  | 3 ->
# 342 "reason_lexer.mll"
      ( UNDERSCORE )
# 1443 "reason_lexer.ml"

  | 4 ->
# 344 "reason_lexer.mll"
      ( TILDE )
# 1448 "reason_lexer.ml"

  | 5 ->
# 346 "reason_lexer.mll"
      ( QUESTION )
# 1453 "reason_lexer.ml"

  | 6 ->
# 349 "reason_lexer.mll"
      ( OPTIONAL_NO_DEFAULT )
# 1458 "reason_lexer.ml"

  | 7 ->
# 351 "reason_lexer.mll"
      ( EXPLICITLY_PASSED_OPTIONAL )
# 1463 "reason_lexer.ml"

  | 8 ->
# 353 "reason_lexer.mll"
      ( let s = Lexing.lexeme lexbuf in
        try Hashtbl.find keyword_table s
        with Not_found -> LIDENT s )
# 1470 "reason_lexer.ml"

  | 9 ->
# 357 "reason_lexer.mll"
      ( warn_latin1 lexbuf; LIDENT (Lexing.lexeme lexbuf) )
# 1475 "reason_lexer.ml"

  | 10 ->
# 359 "reason_lexer.mll"
      ( UIDENT(Lexing.lexeme lexbuf) )
# 1480 "reason_lexer.ml"

  | 11 ->
# 361 "reason_lexer.mll"
      ( warn_latin1 lexbuf; UIDENT(Lexing.lexeme lexbuf) )
# 1485 "reason_lexer.ml"

  | 12 ->
# 363 "reason_lexer.mll"
      ( try
          INT (cvt_int_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "int", Location.curr lexbuf))
      )
# 1494 "reason_lexer.ml"

  | 13 ->
# 369 "reason_lexer.mll"
      ( FLOAT (remove_underscores(Lexing.lexeme lexbuf)) )
# 1499 "reason_lexer.ml"

  | 14 ->
# 371 "reason_lexer.mll"
      ( try
          INT32 (cvt_int32_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "int32", Location.curr lexbuf)) )
# 1507 "reason_lexer.ml"

  | 15 ->
# 376 "reason_lexer.mll"
      ( try
          INT64 (cvt_int64_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "int64", Location.curr lexbuf)) )
# 1515 "reason_lexer.ml"

  | 16 ->
# 381 "reason_lexer.mll"
      ( try
          NATIVEINT (cvt_nativeint_literal (Lexing.lexeme lexbuf))
        with Failure _ ->
          raise (Error(Literal_overflow "nativeint", Location.curr lexbuf)) )
# 1523 "reason_lexer.ml"

  | 17 ->
# 386 "reason_lexer.mll"
      ( reset_string_buffer();
        is_in_string := true;
        let string_start = lexbuf.lex_start_p in
        string_start_loc := Location.curr lexbuf;
        string lexbuf;
        is_in_string := false;
        lexbuf.lex_start_p <- string_start;
        STRING (get_stored_string(), None) )
# 1535 "reason_lexer.ml"

  | 18 ->
# 395 "reason_lexer.mll"
      ( reset_string_buffer();
        let delim = Lexing.lexeme lexbuf in
        let delim = String.sub delim 1 (String.length delim - 2) in
        is_in_string := true;
        let string_start = lexbuf.lex_start_p in
        string_start_loc := Location.curr lexbuf;
        quoted_string delim lexbuf;
        is_in_string := false;
        lexbuf.lex_start_p <- string_start;
        STRING (get_stored_string(), Some delim) )
# 1549 "reason_lexer.ml"

  | 19 ->
# 406 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 1;
        CHAR (Lexing.lexeme_char lexbuf 1) )
# 1555 "reason_lexer.ml"

  | 20 ->
# 409 "reason_lexer.mll"
      ( CHAR(Lexing.lexeme_char lexbuf 1) )
# 1560 "reason_lexer.ml"

  | 21 ->
# 411 "reason_lexer.mll"
      ( CHAR(char_for_backslash (Lexing.lexeme_char lexbuf 2)) )
# 1565 "reason_lexer.ml"

  | 22 ->
# 413 "reason_lexer.mll"
      ( CHAR(char_for_decimal_code lexbuf 2) )
# 1570 "reason_lexer.ml"

  | 23 ->
# 415 "reason_lexer.mll"
      ( CHAR(char_for_hexadecimal_code lexbuf 3) )
# 1575 "reason_lexer.ml"

  | 24 ->
# 417 "reason_lexer.mll"
      ( let l = Lexing.lexeme lexbuf in
        let esc = String.sub l 1 (String.length l - 1) in
        raise (Error(Illegal_escape esc, Location.curr lexbuf))
      )
# 1583 "reason_lexer.ml"

  | 25 ->
# 422 "reason_lexer.mll"
      ( let start_loc = Location.curr lexbuf  in
        comment_start_loc := [start_loc];
        reset_string_buffer ();
        let end_loc = comment lexbuf in
        let s = get_stored_string () in
        reset_string_buffer ();
        COMMENT (s, { start_loc with
                      Location.loc_end = end_loc.Location.loc_end })
      )
# 1596 "reason_lexer.ml"

  | 26 ->
# 432 "reason_lexer.mll"
      ( let loc = Location.curr lexbuf  in
        if !print_warnings then
          Location.prerr_warning loc Warnings.Comment_start;
        comment_start_loc := [loc];
        reset_string_buffer ();
        let end_loc = comment lexbuf in
        let s = get_stored_string () in
        reset_string_buffer ();
        COMMENT (s, { loc with Location.loc_end = end_loc.Location.loc_end })
      )
# 1610 "reason_lexer.ml"

  | 27 ->
# 443 "reason_lexer.mll"
      ( let loc = Location.curr lexbuf in
        Location.prerr_warning loc Warnings.Comment_not_end;
        lexbuf.Lexing.lex_curr_pos <- lexbuf.Lexing.lex_curr_pos - 1;
        let curpos = lexbuf.lex_curr_p in
        lexbuf.lex_curr_p <- { curpos with pos_cnum = curpos.pos_cnum - 1 };
        STAR
      )
# 1621 "reason_lexer.ml"

  | 28 ->
# 451 "reason_lexer.mll"
      ( SHARPOP(Lexing.lexeme lexbuf) )
# 1626 "reason_lexer.ml"

  | 29 ->
let
# 452 "reason_lexer.mll"
                                   num
# 1632 "reason_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 453 "reason_lexer.mll"
                                           name
# 1637 "reason_lexer.ml"
= Lexing.sub_lexeme_opt lexbuf lexbuf.Lexing.lex_mem.(3) lexbuf.Lexing.lex_mem.(2) in
# 455 "reason_lexer.mll"
      ( update_loc lexbuf name (int_of_string num) true 0;
        token lexbuf
      )
# 1643 "reason_lexer.ml"

  | 30 ->
# 458 "reason_lexer.mll"
         ( AMPERSAND )
# 1648 "reason_lexer.ml"

  | 31 ->
# 459 "reason_lexer.mll"
         ( AMPERAMPER )
# 1653 "reason_lexer.ml"

  | 32 ->
# 460 "reason_lexer.mll"
         ( BACKQUOTE )
# 1658 "reason_lexer.ml"

  | 33 ->
# 461 "reason_lexer.mll"
         ( QUOTE )
# 1663 "reason_lexer.ml"

  | 34 ->
# 462 "reason_lexer.mll"
         ( LPAREN )
# 1668 "reason_lexer.ml"

  | 35 ->
# 463 "reason_lexer.mll"
         ( RPAREN )
# 1673 "reason_lexer.ml"

  | 36 ->
# 464 "reason_lexer.mll"
         ( STAR )
# 1678 "reason_lexer.ml"

  | 37 ->
# 465 "reason_lexer.mll"
         ( COMMA )
# 1683 "reason_lexer.ml"

  | 38 ->
# 466 "reason_lexer.mll"
         ( MINUSGREATER )
# 1688 "reason_lexer.ml"

  | 39 ->
# 467 "reason_lexer.mll"
         ( EQUALGREATER )
# 1693 "reason_lexer.ml"

  | 40 ->
# 468 "reason_lexer.mll"
         ( SHARP )
# 1698 "reason_lexer.ml"

  | 41 ->
# 469 "reason_lexer.mll"
         ( DOT )
# 1703 "reason_lexer.ml"

  | 42 ->
# 470 "reason_lexer.mll"
         ( DOTDOT )
# 1708 "reason_lexer.ml"

  | 43 ->
# 471 "reason_lexer.mll"
         ( DOTDOTDOT )
# 1713 "reason_lexer.ml"

  | 44 ->
# 472 "reason_lexer.mll"
         ( COLON )
# 1718 "reason_lexer.ml"

  | 45 ->
# 473 "reason_lexer.mll"
         ( COLONCOLON )
# 1723 "reason_lexer.ml"

  | 46 ->
# 474 "reason_lexer.mll"
         ( COLONEQUAL )
# 1728 "reason_lexer.ml"

  | 47 ->
# 475 "reason_lexer.mll"
         ( COLONGREATER )
# 1733 "reason_lexer.ml"

  | 48 ->
# 476 "reason_lexer.mll"
         ( SEMI )
# 1738 "reason_lexer.ml"

  | 49 ->
# 477 "reason_lexer.mll"
         ( SEMISEMI )
# 1743 "reason_lexer.ml"

  | 50 ->
# 478 "reason_lexer.mll"
         ( LESS )
# 1748 "reason_lexer.ml"

  | 51 ->
# 479 "reason_lexer.mll"
         ( LESSMINUS )
# 1753 "reason_lexer.ml"

  | 52 ->
# 480 "reason_lexer.mll"
         ( EQUAL )
# 1758 "reason_lexer.ml"

  | 53 ->
# 481 "reason_lexer.mll"
         ( LBRACKET )
# 1763 "reason_lexer.ml"

  | 54 ->
# 482 "reason_lexer.mll"
         ( LBRACKETBAR )
# 1768 "reason_lexer.ml"

  | 55 ->
# 483 "reason_lexer.mll"
         ( LBRACKETLESS )
# 1773 "reason_lexer.ml"

  | 56 ->
# 484 "reason_lexer.mll"
         ( LBRACKETGREATER )
# 1778 "reason_lexer.ml"

  | 57 ->
# 485 "reason_lexer.mll"
         ( RBRACKET )
# 1783 "reason_lexer.ml"

  | 58 ->
# 486 "reason_lexer.mll"
         ( LBRACE )
# 1788 "reason_lexer.ml"

  | 59 ->
# 487 "reason_lexer.mll"
         ( LBRACELESS )
# 1793 "reason_lexer.ml"

  | 60 ->
# 488 "reason_lexer.mll"
         ( BAR )
# 1798 "reason_lexer.ml"

  | 61 ->
# 489 "reason_lexer.mll"
         ( BARBAR )
# 1803 "reason_lexer.ml"

  | 62 ->
# 490 "reason_lexer.mll"
         ( BARRBRACKET )
# 1808 "reason_lexer.ml"

  | 63 ->
# 491 "reason_lexer.mll"
         ( GREATER )
# 1813 "reason_lexer.ml"

  | 64 ->
# 497 "reason_lexer.mll"
         ( RBRACE )
# 1818 "reason_lexer.ml"

  | 65 ->
# 498 "reason_lexer.mll"
         ( GREATERRBRACE )
# 1823 "reason_lexer.ml"

  | 66 ->
# 499 "reason_lexer.mll"
         ( LBRACKETAT )
# 1828 "reason_lexer.ml"

  | 67 ->
# 500 "reason_lexer.mll"
         ( LBRACKETPERCENT )
# 1833 "reason_lexer.ml"

  | 68 ->
# 501 "reason_lexer.mll"
          ( LBRACKETPERCENTPERCENT )
# 1838 "reason_lexer.ml"

  | 69 ->
# 502 "reason_lexer.mll"
          ( LBRACKETATAT )
# 1843 "reason_lexer.ml"

  | 70 ->
# 503 "reason_lexer.mll"
           ( LBRACKETATATAT )
# 1848 "reason_lexer.ml"

  | 71 ->
# 504 "reason_lexer.mll"
         ( BANG )
# 1853 "reason_lexer.ml"

  | 72 ->
# 505 "reason_lexer.mll"
         ( INFIXOP0 "!=" )
# 1858 "reason_lexer.ml"

  | 73 ->
# 506 "reason_lexer.mll"
          ( INFIXOP0 "!==" )
# 1863 "reason_lexer.ml"

  | 74 ->
# 507 "reason_lexer.mll"
           ( INFIXOP0 "!=" )
# 1868 "reason_lexer.ml"

  | 75 ->
# 508 "reason_lexer.mll"
            ( INFIXOP0 "!==" )
# 1873 "reason_lexer.ml"

  | 76 ->
# 509 "reason_lexer.mll"
         ( PLUS )
# 1878 "reason_lexer.ml"

  | 77 ->
# 510 "reason_lexer.mll"
         ( PLUSDOT )
# 1883 "reason_lexer.ml"

  | 78 ->
# 511 "reason_lexer.mll"
         ( PLUSEQ )
# 1888 "reason_lexer.ml"

  | 79 ->
# 512 "reason_lexer.mll"
         ( MINUS )
# 1893 "reason_lexer.ml"

  | 80 ->
# 513 "reason_lexer.mll"
         ( MINUSDOT )
# 1898 "reason_lexer.ml"

  | 81 ->
# 514 "reason_lexer.mll"
         ( LESSGREATER )
# 1903 "reason_lexer.ml"

  | 82 ->
# 515 "reason_lexer.mll"
           ( LESSDOTDOTGREATER )
# 1908 "reason_lexer.ml"

  | 83 ->
# 517 "reason_lexer.mll"
            ( PREFIXOP(Lexing.lexeme lexbuf) )
# 1913 "reason_lexer.ml"

  | 84 ->
# 519 "reason_lexer.mll"
            ( PREFIXOP(Lexing.lexeme lexbuf) )
# 1918 "reason_lexer.ml"

  | 85 ->
# 521 "reason_lexer.mll"
            ( INFIXOP0(Lexing.lexeme lexbuf) )
# 1923 "reason_lexer.ml"

  | 86 ->
# 523 "reason_lexer.mll"
            ( INFIXOP1(Lexing.lexeme lexbuf) )
# 1928 "reason_lexer.ml"

  | 87 ->
# 525 "reason_lexer.mll"
            ( INFIXOP2(Lexing.lexeme lexbuf) )
# 1933 "reason_lexer.ml"

  | 88 ->
# 536 "reason_lexer.mll"
            ( INFIXOP4(Lexing.lexeme lexbuf))
# 1938 "reason_lexer.ml"

  | 89 ->
# 537 "reason_lexer.mll"
            ( PERCENT )
# 1943 "reason_lexer.ml"

  | 90 ->
# 539 "reason_lexer.mll"
            ( INFIXOP3(Lexing.lexeme lexbuf) )
# 1948 "reason_lexer.ml"

  | 91 ->
# 541 "reason_lexer.mll"
            ( INFIXOP3(Lexing.lexeme lexbuf))
# 1953 "reason_lexer.ml"

  | 92 ->
# 543 "reason_lexer.mll"
            ( INFIXOP3(Lexing.lexeme lexbuf) )
# 1958 "reason_lexer.ml"

  | 93 ->
# 544 "reason_lexer.mll"
        ( EOF )
# 1963 "reason_lexer.ml"

  | 94 ->
# 546 "reason_lexer.mll"
      ( raise (Error(Illegal_character (Lexing.lexeme_char lexbuf 0),
                     Location.curr lexbuf))
      )
# 1970 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_token_rec lexbuf __ocaml_lex_state

and comment lexbuf =
    __ocaml_lex_comment_rec lexbuf 148
and __ocaml_lex_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 552 "reason_lexer.mll"
      ( comment_start_loc := (Location.curr lexbuf) :: !comment_start_loc;
        store_lexeme lexbuf;
        comment lexbuf;
      )
# 1985 "reason_lexer.ml"

  | 1 ->
# 557 "reason_lexer.mll"
      ( match !comment_start_loc with
        | [] -> assert false
        | [_] -> comment_start_loc := []; Location.curr lexbuf
        | _ :: l -> comment_start_loc := l;
                  store_lexeme lexbuf;
                  comment lexbuf;
       )
# 1996 "reason_lexer.ml"

  | 2 ->
# 565 "reason_lexer.mll"
      (
        string_start_loc := Location.curr lexbuf;
        store_string_char '"';
        is_in_string := true;
        begin try string lexbuf
        with Error (Unterminated_string, str_start) ->
          match !comment_start_loc with
          | [] -> assert false
          | loc :: _ ->
            let start = List.hd (List.rev !comment_start_loc) in
            comment_start_loc := [];
            raise (Error (Unterminated_string_in_comment (start, str_start),
                          loc))
        end;
        is_in_string := false;
        store_string_char '"';
        comment lexbuf )
# 2017 "reason_lexer.ml"

  | 3 ->
# 583 "reason_lexer.mll"
      (
        let delim = Lexing.lexeme lexbuf in
        let delim = String.sub delim 1 (String.length delim - 2) in
        string_start_loc := Location.curr lexbuf;
        store_lexeme lexbuf;
        is_in_string := true;
        begin try quoted_string delim lexbuf
        with Error (Unterminated_string, str_start) ->
          match !comment_start_loc with
          | [] -> assert false
          | loc :: _ ->
            let start = List.hd (List.rev !comment_start_loc) in
            comment_start_loc := [];
            raise (Error (Unterminated_string_in_comment (start, str_start),
                          loc))
        end;
        is_in_string := false;
        store_string_char '|';
        store_string delim;
        store_string_char '}';
        comment lexbuf )
# 2042 "reason_lexer.ml"

  | 4 ->
# 606 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2047 "reason_lexer.ml"

  | 5 ->
# 608 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 1;
        store_lexeme lexbuf;
        comment lexbuf
      )
# 2055 "reason_lexer.ml"

  | 6 ->
# 613 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2060 "reason_lexer.ml"

  | 7 ->
# 615 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2065 "reason_lexer.ml"

  | 8 ->
# 617 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2070 "reason_lexer.ml"

  | 9 ->
# 619 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2075 "reason_lexer.ml"

  | 10 ->
# 621 "reason_lexer.mll"
      ( match !comment_start_loc with
        | [] -> assert false
        | loc :: _ ->
          let start = List.hd (List.rev !comment_start_loc) in
          comment_start_loc := [];
          raise (Error (Unterminated_comment start, loc))
      )
# 2086 "reason_lexer.ml"

  | 11 ->
# 629 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 0;
        store_lexeme lexbuf;
        comment lexbuf
      )
# 2094 "reason_lexer.ml"

  | 12 ->
# 634 "reason_lexer.mll"
      ( store_lexeme lexbuf; comment lexbuf )
# 2099 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec lexbuf __ocaml_lex_state

and string lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1) ;   __ocaml_lex_string_rec lexbuf 180
and __ocaml_lex_string_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 638 "reason_lexer.mll"
      ( () )
# 2111 "reason_lexer.ml"

  | 1 ->
let
# 639 "reason_lexer.mll"
                                  space
# 2117 "reason_lexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 640 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false (String.length space);
        if in_comment () then store_lexeme lexbuf;
        string lexbuf
      )
# 2124 "reason_lexer.ml"

  | 2 ->
# 645 "reason_lexer.mll"
      ( if in_comment () then store_lexeme lexbuf
        else store_string_char(char_for_backslash(Lexing.lexeme_char lexbuf 1));
        string lexbuf )
# 2131 "reason_lexer.ml"

  | 3 ->
# 649 "reason_lexer.mll"
      ( if in_comment () then store_lexeme lexbuf
        else store_string_char(char_for_decimal_code lexbuf 1);
        string lexbuf )
# 2138 "reason_lexer.ml"

  | 4 ->
# 653 "reason_lexer.mll"
      ( if in_comment () then store_lexeme lexbuf
        else store_string_char(char_for_hexadecimal_code lexbuf 2);
        string lexbuf )
# 2145 "reason_lexer.ml"

  | 5 ->
# 657 "reason_lexer.mll"
      ( if in_comment ()
        then string lexbuf
        else begin
(*  Should be an error, but we are very lax.
          raise (Error (Illegal_escape (Lexing.lexeme lexbuf),
                        Location.curr lexbuf))
*)
          let loc = Location.curr lexbuf in
          Location.prerr_warning loc Warnings.Illegal_backslash;
          store_string_char (Lexing.lexeme_char lexbuf 0);
          store_string_char (Lexing.lexeme_char lexbuf 1);
          string lexbuf
        end
      )
# 2163 "reason_lexer.ml"

  | 6 ->
# 672 "reason_lexer.mll"
      ( if not (in_comment ()) then
          Location.prerr_warning (Location.curr lexbuf) Warnings.Eol_in_string;
        update_loc lexbuf None 1 false 0;
        store_lexeme lexbuf;
        string lexbuf
      )
# 2173 "reason_lexer.ml"

  | 7 ->
# 679 "reason_lexer.mll"
      ( is_in_string := false;
        raise (Error (Unterminated_string, !string_start_loc)) )
# 2179 "reason_lexer.ml"

  | 8 ->
# 682 "reason_lexer.mll"
      ( store_string_char(Lexing.lexeme_char lexbuf 0);
        string lexbuf )
# 2185 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_string_rec lexbuf __ocaml_lex_state

and quoted_string delim lexbuf =
    __ocaml_lex_quoted_string_rec delim lexbuf 199
and __ocaml_lex_quoted_string_rec delim lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 687 "reason_lexer.mll"
      ( update_loc lexbuf None 1 false 0;
        store_lexeme lexbuf;
        quoted_string delim lexbuf
      )
# 2200 "reason_lexer.ml"

  | 1 ->
# 692 "reason_lexer.mll"
      ( is_in_string := false;
        raise (Error (Unterminated_string, !string_start_loc)) )
# 2206 "reason_lexer.ml"

  | 2 ->
# 695 "reason_lexer.mll"
      (
        let edelim = Lexing.lexeme lexbuf in
        let edelim = String.sub edelim 1 (String.length edelim - 2) in
        if delim = edelim then ()
        else (store_lexeme lexbuf; quoted_string delim lexbuf)
      )
# 2216 "reason_lexer.ml"

  | 3 ->
# 702 "reason_lexer.mll"
      ( store_string_char(Lexing.lexeme_char lexbuf 0);
        quoted_string delim lexbuf )
# 2222 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_quoted_string_rec delim lexbuf __ocaml_lex_state

and skip_sharp_bang lexbuf =
    __ocaml_lex_skip_sharp_bang_rec lexbuf 208
and __ocaml_lex_skip_sharp_bang_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 707 "reason_lexer.mll"
       ( update_loc lexbuf None 3 false 0 )
# 2234 "reason_lexer.ml"

  | 1 ->
# 709 "reason_lexer.mll"
       ( update_loc lexbuf None 1 false 0 )
# 2239 "reason_lexer.ml"

  | 2 ->
# 710 "reason_lexer.mll"
       ( () )
# 2244 "reason_lexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_skip_sharp_bang_rec lexbuf __ocaml_lex_state

;;

# 712 "reason_lexer.mll"
 

  let token_with_comments lexbuf =
    match !preprocessor with
    | None -> token lexbuf
    | Some (_init, preprocess) -> preprocess token lexbuf

  let last_comments = ref []
  let rec token lexbuf =
    match token_with_comments lexbuf with
        COMMENT (s, comment_loc) ->
          last_comments := (s, comment_loc) :: !last_comments;
          token lexbuf
      | tok -> tok
  let comments () = List.rev !last_comments

  let init () =
    is_in_string := false;
    last_comments := [];
    comment_start_loc := [];
    match !preprocessor with
    | None -> ()
    | Some (init, _preprocess) -> init ()

  let set_preprocessor init preprocess =
    preprocessor := Some (init, preprocess)


# 2280 "reason_lexer.ml"

end
module Reason_entry : sig 
#1 "reason_entry.mli"
(* Copyright (C) 2015-2016 Bloomberg Finance L.P.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * In addition to the permissions granted to you by the LGPL, you may combine
 * or link a "work that uses the Library" with a publicly distributed version
 * of this file to produce a combined library or application, then distribute
 * that combined work under the terms of your choosing, with no requirement
 * to comply with the obligations normally placed on you by section 4 of the
 * LGPL version 3 (or the corresponding section of a later version of the LGPL
 * should you choose to use a later version).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. *)



end = struct
#1 "reason_entry.ml"
(* Copyright (C) 2015-2016 Bloomberg Finance L.P.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * In addition to the permissions granted to you by the LGPL, you may combine
 * or link a "work that uses the Library" with a publicly distributed version
 * of this file to produce a combined library or application, then distribute
 * that combined work under the terms of your choosing, with no requirement
 * to comply with the obligations normally placed on you by section 4 of the
 * LGPL version 3 (or the corresponding section of a later version of the LGPL
 * should you choose to use a later version).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. *)

let implementation lexbuf =
  Reason_lexer.init ();
  Reason_parser.implementation Reason_lexer.token lexbuf


let () = Jsoo_main.make_compiler "reason" implementation

end
